# Server

_Note that during development, a plugin called Luanalysis (a fork of emmylua) is being used to provide static analysis 
of the code for type-safety_

## Script

As this mod utilises an extension, a small script is used to load this. Additionally, this script will also create the 
necessary materials definitions for all maps installed. This happens on game load which will not require a game restart
for the textures to be picked up.

## Explosions Extension
The main interface between this mod and BeamNG is the `extensions/explosionsNG.lua` extension file. This is responsible
for updating handlers (see below) with events from BeamNG as well as providing endpoints to be accessed by the UI
component of this mod.

## Handlers
There are a handful of handlers in the `/handlers` directory. These are responsible for creating, keeping and updating
arrays of active objects. These handlers provide an interface to be interacted with by the main
[extension](#Explosions-Extension), and also by other miscellaneous classes in a central place.

### World Objects Handler
This handler is a relatively small but unique and crucial in its purpose.

One problem faced when creating objects "on-the-fly" is that it is computationally expensive to do so. A noticeable 
freeze can be seen when creating multiple explosives, as they can each contain objects for sounds, lights, particles and
models. However, it is computationally _inexpensive_ to keep these cached in game (but hidden from view until needed 
again).

This handler is responsible for hiding the creation of actual in-game assets and instead provide an interface for an
explosive to request assets and to then release it when it is finished. When a map is loaded, an initial set of in-game 
assets will be created (and crucially locked to prevent the player from deleting them in the world editor). As other 
places in the code need these assets, this handler will loan them out. In the event that this handler does not have an 
appropriate asset, this handler will create a new one to loan out. Upon it being given back to the handler after it has 
been used, it will not be deleted, but instead stored in the cache for future usage.

It should be noted that at present, SFX emitters and particle effects will __NOT__ be cached in any way; this handler 
will always create a new object when requested and always delete the object when released. This is because BeamNG (or 
the Torque 3D engine) does not work properly when changing tracks or emitters, and makes re-use impossible.

At present, the cache does not ever reduce in size.

## Stores
Additionally, there is a store in the `/stores` directory. These are responsible for storing information related to
objects and handling the creation and retrieval of any object. This includes `.json` files for any necessary data. Each
store has a corresponding 'facade' to provide an interface for retrieving this information in an additionally type-safe
way. This is functionally the same as a 'barrel' file for providing an interface into the store. Files inside the store
should __NOT__ be accessed directly, but should instead make use of this facade.

## Explosives
For each explosive, there will be two associated files: the main explosive class and a stats class. These have been
separated as they are separate entities (though still related), such that you do not need an explosive's instance to get
the stats.

Each explosive will extend off of the `Explosive` base class. Some explosives are simple, however others may be
compositions of multiple different explosives in one. One of the inherited methods provides a mapping between a given
preset of an explosive and its associated options. These govern the variance between two presets of the same explosive.

Each explosive stats class will extend off of the `ExplosiveStats` base class. See [stat mapping](#stat-mapping) for
details on the stats themselves.

It should be noted that all explosives are inside the `/explosives` directory. There is multiple subdirectories, as 
detailed below: 
- `/base` - these are the base classes used for all other explosives and stats
- `/composed` - these are explosives which shouldn't be used as standalone (i.e. they should be used as part of another
  explosive via composition/aggregation). These explosives may or may not have an associated stats class, dependent on 
  how much variance they allow for. Additionally, these explosives will also typically only have one preset option (the
  "default" option) as it is expected that the explosive they are part of will define a set of options specific to  that 
  class
- `/contained` - these are the main explosives which map directly to the UI. These are handled and maintained within the
  [explosives store](#stores) and are served up to the client as DTOs
- `/containers` - these are groups of explosives that act as an explosive themselves, but manage a set of their own
  explosives inside of them, whilst maintaining the `Explosive` class interface

### Explosives, Sub-Explosives and Super-Explosives
Terminology wise, any explosive can be considered to be a super-explosive or a sub-explosive, dependent on context. This 
is particularly helpful when describing the stats of particular explosives.

**Explosive** - this is the general term and refers to any class/entity which extends the base `Explosive` abstract 
class

**Super-Explosive** - these explosives create and manage multiple explosives during their lifetime as distinct
entities. For example, an explosive which fires out a set of other explosives would be considered a super-explosive, in
relation to other explosives. Note that there can be many layers of super-explosives - i.e. a super-explosive can itself 
contain other super-explosives.

**Sub-Explosive** - these are explosives that are contained within a super-explosive

Super-explosives and sub-explosives are relative terms and only apply when considering two explosive together. See the 
following diagram for clarity.

![img.png](../explosive_terminology.png)

_Note how `ExplosiveB` can be considered either a sub-explosive or super-explosive, dependent on the other explosive it
is being compared to_

### Stat Mapping
An explosive's stats are algorithmically calculated info each explosive, preset pairing. Each explosive's stats class 
will return stats for that specific explosive; note that the values are absolute values, and the base class provides a
way to get a normalised set of stats from `[0, 1]`.

It should be noted that the normalisation of these absolute stats to `[0, 1]` _isn't necessarily_ linear, but instead 
sometimes logarithmic.

It should also be noted that the below descriptions/definitions are the general case and, due to the variance of 
explosives, any explosive may define their stat(s) differently.

#### Energy
The energy is based on the maximal force a vehicle will be inflicted with. Specifically, this is the **maximum** amount
of force any node in the vehicle can undergo as a **direct** result from the explosive (i.e. the force applied to nodes
during a shockwave effect _do_ count, but the resultant crash into a wall as a result of this force _does not_ count).
Explosive's will typically add up the maximum number of explosions/effects they produce to get an absolute value.

The actual value will depend on whether the explosive is a 
[super-explosive](#explosives-sub-explosives-and-super-explosives) or not, and will require substantially more energy
if it is. The reason for this is that an explosive with a single powerful effect only needs to hit a vehicle once to
destroy it, but an explosive that has two effects which are half as powerful will almost certainly not do the same level 
of damage.

#### Range
The range for each explosive is determined by the **maximum** range it can possibly attack. For example, `The Anomaly`
may attack up to 400m with its EMP attack. Although its main shockwave attack can only reach 200m, the maximum value is
taken and so its absolute range stat will be `400`.

**Gravitational Effects** - some explosives may perform gravitational effects on vehicles. The gravitational effect is 
technically infinite - only at an infinite distance away would a vehicle experience no force exerted on it. In practice,
this is not a helpful metric and so an analytical approach to this has been taken. The following is a definition, 
although arbitrary, that provides sensible/comparable values:

![img.png](../gravitational_range_stat_formula.png)

The `range constant` is defined as the smallest force required to overcome the frictional resistance of tires, on 
asphalt, from a Cherrier Vivace (Hillclimb Configuration). This value works out at around a value of `~488 Newtons` 
(assuming the mass of the vehicle is what is stated at `1420kg` and is used in the calculation - _value correct as of 
time of writing_).

#### Accuracy
The accuracy of an explosive is considered by one of two different ways:

**Average Speed** - what is the average speed needed for a vehicle which starts off at the centre of the explosive (from 
the perspective of the explosive's effect), in order to avoid being affected? The value for the normalisation of this 
stat when it returns `1` is set in a static field; see the definition of the `ExplosiveStats` class.

**Coverage** - what is the average ratio of the combined sub-explosives areas compared to the total area of the 
super-explosive? Unlike other stats, this does not consider the maximal area, but rather the average area as this is 
more meaningful in practice. Also note that this metric will only be used for a 
[super-explosive](#explosives-sub-explosives-and-super-explosives). The following formula describes how this average
coverage is calculated:

![img.png](../accuracy_stat_coverage_formula.png)

There is a special case of the **Coverage** accuracy for explosives that have gravitational effects. This fall under
this category, but will instead use a constant value. The constant is calculated as the ratio of the gravitational range
for a mass that produces `1 kilonewton` of force, compared to the [gravitational range](#range):

![img.png](../gravitational_accuracy_constant.png).

Again, the definition is mostly arbitrary but performs well in practise.


#### Attack Rate
The attack rate of an explosive is the minimal delay between two attacks from an explosive. This stat varies from 
explosive to explosive due to its nature, however, in general this stat is split into two main cases:

**Single Attack** - a single attack explosive is one which creates one, and only one, effect during its lifetime. In 
this case, all explosive that fall into this group will be given the same attack rate value, which is set inside a 
static field; see the definition of the `ExplosiveStats` class.

**Multi-Attack** - if the explosive instead creates multiple effects during its lifetime (in general, this explosive
will be a [super-explosive](#explosives-sub-explosives-and-super-explosives)), then the attack rate is the minimal
delay between two attacks/effects. In cases where this is not controlled, (i.e. explosives are thrown outwards and so 
the delay between two explosives detonating is highly dependent on the objects/geography of the map), this may be 
calculated differently as the ratio of total explosive lifetime compared to the number of explosives.

### Rank/Tier
Below is the list of each rank and their associated tier

| Tier [1, 4] | Rank [1, 10] |
|:------------|-------------:|
| 1           |      1, 2, 3 |
| 2           |      4, 5, 6 |
| 3           |      7, 8, 9 |
| 4           |           10 |

It should be noted that the tier for each explosive is hardcoded (due to various factors such as needing the icons to
match the tier colour), but the general rule is the tier of the most powerful preset of an explosive is taken as
the explosive's tier as a whole. For example, an explosive may have a preset-rank pairing of
`["small" -> 2, "default" -> 3, "large" -> 4]`. In this case, the "large" preset has the highest rank and so it is
chosen. This corresponds to tier 2, and so the explosive is considered a tier 2 explosive.

It should be noted that presets in general try to stay within one tier, and the above example is an outlier for 
demonstration purposes.

## Classes and Helpers
These are the main components that vary from the explosives themselves, various mathematical functions and others.
Helper functions are placed inside the `/helpers` directory and are self-contained, providing useful functions.
Unrelated classes are placed in the `/misc` directory, with other classes grouped in related directories. Classes can
access copies of object arrays in the [handlers](#handlers) and call methods on those objects.

## Animation State Machines
Some classes manage an internal state, designated by a constrained string type, and have animations attached to them,
based on what state an instance is currently in. These classes will make use of the `AnimationStateMachine` class which
provides this functionality for managing state and animations. Each class will be paired with an animations file which
will contain the animation definition(s), and expose a single function to create new instances of this state machine for
any given instance.

Note that the animations class will have full access to all methods and fields inside the instance being animated, 
including those marked private. In OOP, animations would be defined as inner classes, however this concept does not
really exist in Lua and so encapsulation is intentionally violated.

## beamng.def.lua
This file contains type definitions for BeamNG classes, methods and functions, to allow for type checking to work with
IDEs.

## interfaces.def.lua
This file contains interfaces (defined as classes) which class implementations can use their base class. Unfortunately,
this has limitations in that only one interface can be implemented per class (as they are defined as classes as 
interfaces do not exist), and they must be defined on a single line so no method documentation can be given to them.

## Class Structure
Static methods will be defined first, followed by abstract methods, then any overridden or implemented methods.
Constructors will be defined as the last method.

## Additional Code Documentation
This section contains some additional documentation and diagrams to explain some parts of the
classes/functions/interactions in this mod which may be beneficial to understand what certain things do.

### Anomaly Class
See the following image for a diagram of the lifecycle of the `Anomaly` class (in `/misc/anomaly.lua`)

![Anomaly State Diagram](../anomaly_state_diagram.png)