local classes = require("lua/explosionsNG/helpers/classes")
local Timer = require("lua/explosionsNG/timers/Timer")

--- @alias AnimationTimeContext { elapsed: number, elapsedRatio: number, timeLeft: number, timeLeftRatio: number }
--- @alias AnimationFn<T> fun(entity: T, dt: number, context?: any): void @A function that should perform an action on the entity, based on the current dt time elapsed since the last update
--- @alias AnimationTimeFn<T> fun(entity: T, dt: number, timeContext: AnimationTimeContext, context?: any): void @Am `AnimationFn` with an additional context around the timer

--- A class that interacts with a particular entity on every update cycle. This class can handle both finite and
--- infinite animations, with the former performing an emission of the finish event.
---
--- Each animation may make use of a `context` field that can contain anything to allow parametrisation/customization of
--- an animation. This is difficult to type correctly, and so no type information is constrained.
--- @class Animation<T>
--- @field entity T The entity being animated on
--- @field animationFn AnimationFn<T> | AnimationTimeFn<T> A function that animates the entity at every update cycle
--- @field finishFn (fun(): void) | nil A function to call when the animation finishes
--- @field context any A set of context parameters the animation can use
--- @field timer Timer | nil A timer that when expired, calls the `finishFn`
local Animation = classes.newClass()

--- Updates the next cycle of animation. If this animation is timed, then the animation will only update i the timer has
--- not completed.
--- @param dt number The time since the last update
function Animation:update(dt)
  if not self.timer then
    local castedAnimationFn = --[[--- @type AnimationFn<T> ]] self.animationFn
    castedAnimationFn(self.entity, dt)
    return
  end

  self:animateWithTimer(dt, --[[--- @not nil ]] self.timer)
end

--- Updates the next cycle of animation with the given timer. If the timer has not completed, this method will update
--- the entity with the animation function before updating the timer itself.
--- @param dt number The time since the last update
--- @param timer Timer The timer to use and update
--- @private
function Animation:animateWithTimer(dt, timer)
  if timer:isComplete() then return end

  --- @type AnimationTimeContext
  local timeContext = {
    elapsed = timer:elapsedTime(),
    elapsedRatio = timer:elapsedRatio(),
    timeLeft = timer:timeLeft(),
    timeLeftRatio = timer:timeLeftRatio()
  }

  local castedAnimationFn = --[[--- @type AnimationTimeFn<T> ]] self.animationFn
  castedAnimationFn(self.entity, dt, timeContext, self.context)

  timer:update(dt)
end

--- __Constructor__
--- @generic S
--- @param entity S The entity to animate on
--- @param animationFn AnimationFn<S> | AnimationTimeFn<S> The function to animate the entity at every update cycle. An `AnimationTimeFb` should be given instead if the following `period` is given
--- @param context? any Any context parameters to use
--- @param finishFn? fun(): void An optional function to call when the animation finishes. If the `period` is given, this should be given too.
--- @param period? number An optional period to specify how long the animation lasts for. If not given, the animation is indefinite.
--- @return Animation<S> The constructed instance
function Animation:new(entity, animationFn, context, finishFn, period)
  local self = --[[--- @type Animation<S> ]] classes.setmetatable({}, Animation)
  self.entity = entity
  self.animationFn = animationFn
  self.context = context
  self.finishFn = finishFn
  self.timer = (period ~= nil and finishFn ~= nil) and Timer:new(period, --[[--- @not nil ]] finishFn, true) or nil
  return self
end

return Animation
