local classes = require("lua/explosionsNG/helpers/classes")
local Animation = require("lua/explosionsNG/animations/Animation")

--- A state machine like object that maintains a set of animations, with a current animation being updatable. Animations
--- can be switched between and allows for noop animations.
---
--- __Generics__
--- - `T` is the unique state for each animation
--- - `S` is the type of entity that the animations will act upon
--- @class AnimationStateMachine<T, S>
--- @field private entity S The entity that is being animated
--- @field private currentState T The current state
--- @field private currentAnimation Animation<S> The current animation, matching the current state
--- @field private animationProducers { [T]: (fun(context?: any): Animation<S> | nil) } The list of animation producer functions, for each state
local AnimationStateMachine = classes.newClass()

--- If there is a current animation, this method updates the animation
--- @param dt number The time since the last update
function AnimationStateMachine:update(dt)
  if not self.currentAnimation then return end
  self.currentAnimation:update(dt)
end

--- Switches from the current animation to the request animation, given its unique state
--- @param state T The new animation to change to
--- @param context? any Any context parameters to use for the animation
function AnimationStateMachine:switch(state, context)
  local newAnimation = self.animationProducers[state](context)
  self.currentAnimation = newAnimation
  self.currentState = state
end

--- Creates a new animation with the given parameters and adds it to the stored list with the corresponding state
--- @param state T The state to add the animation under
--- @param animationFn AnimationFn<S> | AnimationTimeFn<S> The function to animate the entity at every update cycle. See `Animation` for details on which type to use when
--- @param finishFn? fun(): void An optional function to call when the animation finishes
--- @param period? number An optional period to specify how long the animation lasts for. If not given, the animation is indefinite
function AnimationStateMachine:add(state, animationFn, finishFn, period)
  self.animationProducers[state] = function(context)
    return Animation:new(self.entity, animationFn, context, finishFn, period)
  end
end

--- Creates a new no-operation animation and adds it to the stored list with the corresponding state. When the animation
--- is changed to, nothing will be called and it will be considered an infinite animation.
--- @param state T The state to add the no-operation animation under
function AnimationStateMachine:addNoop(state)
  self.animationProducers[state] = function() return nil end
end

--- @return T The current state
function AnimationStateMachine:getState()
  return self.currentState
end

--- __Constructor__
--- @generic U, V
--- @param entity V The entity that is being animated
--- @return AnimationStateMachine<U, V> The constructed instance
function AnimationStateMachine:new(entity)
  local self = --[[--- @type AnimationStateMachine<U, V> ]] classes.setmetatable({}, AnimationStateMachine)
  self.entity = entity
  self.currentState = nil
  self.currentAnimation = nil
  self.animationProducers = --[[--- @type any ]] {} -- Luanalysis really does not like this generic table type :/
  return self
end

return AnimationStateMachine
