local classes = require("lua/explosionsNG/helpers/classes")

-- TODO Also not used. Only transformPoint method is used by BoundingBox

--- @class RectangularArea
--- @field private centre LuaVec3 The centre point of the rectangle
--- @field private rotation LuaQuat The rotation of the rectangle
--- @field private width number The width of the rectangle
--- @field private height number The height of the rectangle
--- @field private depth number The depth of the rectangle
local RectangularArea = classes.newClass()

--- __Static__
---
--- Transforms the given `point` by rotating the point around the `rotation` and translating using the `centre`.
--- If `toRelative` is true, this method assumes the `point` is in the absolute coordinate system and transforms it into
--- a point that is in the relative coordinate system with respect to this rectangular area as an axis-aligned box.
--- @param centre LuaVec3 The centre point of the rectangular area
--- @param rotation LuaQuat The rotation around the centre point that the rectangular area is angled
--- @param point LuaVec3 The point to transform
--- @param toRelative? boolean Whether to transform the point into the relative coordinate system. Defaults to false
--- @return LuaVec3 The transform point
function RectangularArea:transformPoint(centre, rotation, point, toRelative)
  if toRelative then return (point - centre):rotated(rotation:inversed()) end
  return point:rotated(rotation) + centre
end

--- __Static__
---
--- Returns whether the given point is inside the extents of the rectangular area, along a particular given axis
--- @param point LuaVec3 The point to check
--- @param minimumExtent LuaVec3 The minimum extent of the rectangular area
--- @param maximumExtent LuaVec3 The maximum extent of the rectangular area
--- @param axis Axis The axis to consider
--- @return boolean Whether the point is inside the extents of the rectangular area, along the given axis
--- @private
function RectangularArea:isPointWithinAxis(point, minimumExtent, maximumExtent, axis)
  local minimum = minimumExtent[axis]
  local maximum = maximumExtent[axis]
  return point >= minimum and point <= maximum
end

--- Returns whether this rectangle area contains the given position or not
--- @param position LuaVec3 The position to check
--- @return boolean Whether this rectangle contains the given position or not
function RectangularArea:contains(position)
  -- Convert position so that it is in the relative coordinate system of both
  local relativePosition = RectangularArea:transformPoint(self.centre, self.rotation, position, true)
  local minimumExtent, maximumExtent = self:getRelativeExtents()
  return (
      RectangularArea:isPointWithinAxis(relativePosition, minimumExtent, maximumExtent, "x") and
      RectangularArea:isPointWithinAxis(relativePosition, minimumExtent, maximumExtent, "y") and
      RectangularArea:isPointWithinAxis(relativePosition, minimumExtent, maximumExtent, "z")
  )
end

--- @return LuaVec3, LuaVec3 The minimum and maximum extents of the area in the relative coordinates system with respect to the area's centre
--- @private
function RectangularArea:getRelativeExtents()
  local maximum = vec3(self.width, self.depth, self.height) / 2
  local minimum = -maximum
  return minimum, maximum
end

--- @return LuaVec3 The centre point of the rectangle area
function RectangularArea:getCentre()
  return vec3(self.centre)
end

--- __Constructor__
--- @param centre LuaVec3 The centre point of the rectangle
--- @param width number The width of the rectangle. Note that the width refers to the +X axis, and its independent of the rotation.
--- @param height number The height of the rectangle. Note that the width refers to the +Z axis, and its independent of the rotation.
--- @param depth number The height of the rectangle. Note that the width refers to the +Y axis, and its independent of the rotation.
--- @param rotation? LuaQuat The rotation of the rectangle. Defaults to no rotation.
--- @return RectangularArea The constructed instance
function RectangularArea:new(centre, width, height, depth, rotation)
  local self = classes.setmetatable({}, RectangularArea)
  self.centre = vec3(centre)
  self.width = width
  self.height = height
  self.depth = depth
  self.rotation = rotation or quat()
  return self
end

return RectangularArea
