--- Type definitions for various BeamNG and Torque3D classes, methods and functions.
---
--- Only those used by this mod are added.
---
--- Note that some types defined here (and used within this mod) are prefixed with `Lua` or `TS`. This is because type
--- checking plugins do not allow a type to have a constructor with the same name.
--- I.e. vec3 -> LuaVec3; QuatF -> TSQuatF

--- @alias void nil

------------------------------------------------------------------------------------------------------------------------
--- `LuaVec3`
------------------------------------------------------------------------------------------------------------------------

--- @alias Vector {x: number, y: number, z: number}

--- @class LuaVec3 : Vector
LuaVec3 = {}

--- @param position? userdata | Vector
--- @return LuaVec3
--- @overload fun(x: number, y: number, z: number): LuaVec3
function vec3(position) end

--- @return number
function LuaVec3:length() end

--- @param LuaQuat LuaQuat
--- @return LuaVec3
function LuaVec3:rotated(LuaQuat) end

--- @return number, number, number
function LuaVec3:xyz() end

--- @param x LuaVec3
--- @param y LuaVec3
--- @return LuaVec3
function LuaVec3:__add(x, y) end

--- @param x LuaVec3
--- @param y LuaVec3
--- @return LuaVec3
function LuaVec3:__sub(x, y) end

--- @param x LuaVec3
--- @param y LuaVec3
--- @return LuaVec3
--- @overload fun(x: LuaVec3, y: number): LuaVec3
function LuaVec3:__div(x, y) end

--- @param x LuaVec3
--- @param y LuaVec3
--- @return LuaVec3
--- @overload fun(x: LuaVec3, y: number): LuaVec3
function LuaVec3:__mul(x, y) end

--- @return LuaVec3
function LuaVec3:__unm() end

--- @return { number, number, number }
function LuaVec3:toTable() end

--- @return Vector
function LuaVec3:toDict() end

--- @param point LuaVec3
--- @return number
function LuaVec3:squaredDistance(point) end

--- @param point LuaVec3
--- @return number
function LuaVec3:distance(point) end

--- @return LuaVec3
function LuaVec3:normalized() end

--- @param vector LuaVec3
--- @return LuaQuat
function LuaVec3:getRotationTo(vector) end

--- @return LuaVec3
function LuaVec3:z0() end

------------------------------------------------------------------------------------------------------------------------
--- `LuaQuat`
------------------------------------------------------------------------------------------------------------------------

--- @class LuaQuat
--- @field x number
--- @field y number
--- @field z number
--- @field w number
LuaQuat = {}

--- @param quat? userdata | LuaQuat
--- @return LuaQuat
--- @overload fun(x: number, y: number, z: number, w: number): LuaQuat
function quat(quat) end

--- @return LuaQuat
function LuaQuat:inversed() end

--- @return { x: number, y: number, z: number, w: number }
function LuaQuat:toTorqueQuat() end

--- @return LuaQuat
function LuaQuat:normalized() end

--- @param x number
--- @param y number
--- @param z number
--- @return LuaQuat
function LuaQuat:setFromEuler(x, y, z) end

--- @return LuaVec3
function LuaQuat:toEulerYXZ() end

------------------------------------------------------------------------------------------------------------------------
--- `scenetree`
------------------------------------------------------------------------------------------------------------------------

--- @generic T : SceneObject
--- @param name string
--- @return T
function findObject(name) end

--- @param className string
--- @return SceneObject[]
function findClassObjects(className) end

scenetree = {
  findObject = findObject,
  findClassObjects = findClassObjects,
  MissionGroup = --[[--- @type SceneTreeFolder ]] {}
}

------------------------------------------------------------------------------------------------------------------------
--- `SceneTreeFolder` (type name arbitrarily made up - represents a folder object in the scenetree)
------------------------------------------------------------------------------------------------------------------------

--- @class SceneTreeFolder
SceneTreeFolder = {}

---@param field "locked"
---@param section number
---@param value string | number
function SceneTreeFolder:setField(field, section, value) end

--- @param object SceneTreeFolder | SceneObject
function SceneTreeFolder:addObject(object) end

--- @param name string
function SceneTreeFolder:registerObject(name) end

------------------------------------------------------------------------------------------------------------------------
--- `TorqueScript`
------------------------------------------------------------------------------------------------------------------------

--- @param cmd string
function eval(cmd) end

TorqueScript = {
  eval = eval
}

------------------------------------------------------------------------------------------------------------------------
--- `extensions`
------------------------------------------------------------------------------------------------------------------------

--- @type table<string, any>
extensions = {}

------------------------------------------------------------------------------------------------------------------------
--- `bullettime`
------------------------------------------------------------------------------------------------------------------------

--- @return number
function get() end

bullettime = {
  get = get
}

------------------------------------------------------------------------------------------------------------------------
--- `SceneObject`
------------------------------------------------------------------------------------------------------------------------

--- @alias SceneObjectField "position" | "scale" | "rotationEuler" | "hidden" | "locked"
--- @alias TSStaticField "dynamic" | "instanceColor" | "shapeName"
--- @alias ParticleEmitterField "velocity" | "emitter" | "dataBlock" | "active"
--- @alias PointLightField "radius" | "brightness" | "color" | "animationType" | "castShadows"
--- @alias SFXEmitterField "isLooping" | "filename" | "track" | "referenceDistance" | "maxDistance"
--- @alias SunSkyField "skyBrightness" | "brightness"

--- @param type string
--- @return SceneObject
function createObject(type) end

--- @class SceneObject
--- @field public name string
--- @field public shapeName string
--- @field private pos userdata
SceneObject = {}

--- @return userdata
function SceneObject:getPosition() end

--- @param cmd string
function SceneObject:queueLuaCommand(cmd) end

function SceneObject:delete() end

---@param field SceneObjectField | TSStaticField | ParticleEmitterField | PointLightField | SFXEmitterField | SunSkyField
---@param section number
---@param value string | number
function SceneObject:setField(field, section, value) end

--- @param name string
function SceneObject:registerObject(name) end

--- @return number
function SceneObject:getID() end

function SceneObject:postApply() end

------------------------------------------------------------------------------------------------------------------------
--- `VehicleSceneObject` (type name arbitrarily made up - this is what is returned from a scenetree/be object query for
--- a vehicle)
------------------------------------------------------------------------------------------------------------------------

--- @class VehicleSceneObject : SceneObject
VehicleSceneObject = {}

--- @param nodeId number
--- @return LuaVec3
function VehicleSceneObject:getNodePosition(nodeId) end

--- @return number
function VehicleSceneObject:getRefNodeId() end

--- @return number
function VehicleSceneObject:getNodeCount() end

--- @return LuaVec3
function VehicleSceneObject:getDirectionVector() end

--- @return LuaVec3
function VehicleSceneObject:getDirectionVectorUp() end

--- @return LuaVec3
function VehicleSceneObject:getVelocity() end

------------------------------------------------------------------------------------------------------------------------
--- `SunSkySceneObject` (type name arbitrarily made up - this is what is returned from a scenetree/be object query for
--- a SunSky object)
------------------------------------------------------------------------------------------------------------------------

--- @class SunSkySceneObject : SceneObject
--- @field public skyBrightness number
--- @field public brightness number
SunSkySceneObject = {}

------------------------------------------------------------------------------------------------------------------------
--- `be`
------------------------------------------------------------------------------------------------------------------------

--- @class be
be = {}

--- @generic T : SceneObject
--- @param id number
--- @return T
function be:getObjectByID(id) end

--- @param playerId number
--- @return number
function be:getPlayerVehicleID(playerId) end

------------------------------------------------------------------------------------------------------------------------
--- `Engine`
------------------------------------------------------------------------------------------------------------------------

--- @alias PlayOnceOptions { unique: boolean | nil, position: LuaVec3 | nil }

--- @param channel "AudioGui"
--- @param sound string
--- @param options? PlayOnceOptions | nil
function playOnce(channel, sound, options) end

Engine = {
  Audio = {
    playOnce = playOnce
  }
}

------------------------------------------------------------------------------------------------------------------------
--- `FS`
------------------------------------------------------------------------------------------------------------------------

--- @class FS
FS = {}

--- @param copyFromFileName string
--- @param copyToFileName string
function FS:copyFile(copyFromFileName, copyToFileName) end

--- @param fileName string
--- @return boolean
function FS:fileExists(fileName) end

------------------------------------------------------------------------------------------------------------------------
--- `core_environment`
------------------------------------------------------------------------------------------------------------------------

--- @alias TimeOfDay { time: number }

--- @return TimeOfDay
function getTimeOfDay() end

--- @param time TimeOfDay
function setTimeOfDay(time) end

--- @return number
function getGravity() end

core_environment = {
  getTimeOfDay = getTimeOfDay,
  setTimeOfDay = setTimeOfDay,
  getGravity = getGravity
}

------------------------------------------------------------------------------------------------------------------------
--- `core_camera`
------------------------------------------------------------------------------------------------------------------------

--- @return LuaVec3
function getPosition() end

core_camera = {
  getPosition = getPosition
}

------------------------------------------------------------------------------------------------------------------------
--- `core_terrain`
------------------------------------------------------------------------------------------------------------------------

--- @param point LuaVec3
--- @return number | nil
function getTerrainHeight(point) end

core_terrain = {
  getTerrainHeight = getTerrainHeight
}

------------------------------------------------------------------------------------------------------------------------
--- `core_levels`
------------------------------------------------------------------------------------------------------------------------

--- @return string[]
function getLevelPaths() end

--- @param levelPath string
--- @return string
function getLevelName(levelPath) end

core_levels = {
  getLevelPaths = getLevelPaths,
  getLevelName = getLevelName
}

------------------------------------------------------------------------------------------------------------------------
--- `core_camera`
------------------------------------------------------------------------------------------------------------------------

--- @param name string
--- @param data? any
function trigger(name, data) end

guihooks = {
  trigger = trigger
}

------------------------------------------------------------------------------------------------------------------------
--- `simTimeAuthority`
------------------------------------------------------------------------------------------------------------------------

--- @return boolean
function getPause() end

simTimeAuthority = {
  getPause = getPause
}

------------------------------------------------------------------------------------------------------------------------
--- `Misc Functions`
------------------------------------------------------------------------------------------------------------------------

--- @alias File table<any, any>

function dump(...) end

--- @return string
function dumps(...) end

--- @param level "A" | "D" | "E" | "I" | "W"
--- @param tag string
--- @param message string
function log(level, tag, message) end

--- @param vec1 LuaVec3
--- @param vec2? LuaVec3
--- @return LuaQuat
function quatFromDir(vec1, vec2) end

--- @param filename string
--- @return File
function readFile(filename) end

--- @param file File
--- @return table<any, any>
function jsonDecode(file) end

--- @param position LuaVec3
--- @param direction LuaVec3
--- @param searchDistance number
--- @return number
function castRayStatic(position, direction, searchDistance) end

--- @param h number
--- @param s number
--- @param v number
--- @return number, number, number
function HSVtoRGB(h, s, v) end

--- @param r number
--- @param g number
--- @param b number
--- @return number, number, number
function RGBtoHSV(r, g, b) end

--- @param extension string
--- @param mode string
function setExtensionUnloadMode(extension, mode) end

--- @return string
function getMissionFilename() end