local classes = require("lua/explosionsNG/helpers/classes")
local Timer = require("lua/explosionsNG/timers/Timer")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias ExplosiveState "ready" | "active" | "finished"
--- @alias AnyExplosive Explosive<any, any>

--- __Abstract__
---
--- The superclass that all explosive subclasses will inherit from.
---
--- The constructor takes in a parameter called `presetMapper`. This is used when the configuration specifies a
--- preset. Subclasses calling this method should implement the `getPresetOptions` and pass this method as this
--- parameter (the actual function can be a wrapper, but should make use of/call this method). In normal OOP, this would
--- be called within this constructor, however, as Lua does not support OOP (only mimicry), this method when called as
--- `self:getPresetOptions` will refer to this class as the parameter `self`, not the subclass. Therefore, it cannot be
--- used. See `getPresetOptions` for more details on that method.
---
--- __Generics__
--- - `T` is the type for the preset of the explosive
--- - `S` is the type of the options of the explosive
--- @class Explosive<T, S>
--- @field private state ExplosiveState The current state of the explosive
--- @field private fuse number The delay between the explosive being activated and it detonating
--- @field private fuseTimer Timer The timer of which the explosive is now active, but has not started exploding.
--- @field protected options S The options for the explosive
--- @field protected position LuaVec3 The in-game position of the explosive
--- @field protected onFinish Array<fun(): void> A set of functions to call when this explosive's state is set to `"Finished"`
local Explosive = classes.newClass()

--- Optional method that can be overridden in subclasses.
--- Called when the explosive is detonated.
--- @protected
function Explosive:onDetonation() end

--- Optional method that can be overridden in subclasses.
--- Called when the explosive is finished.
--- @protected
function Explosive:onDetonationFinished() end

--- __Abstract__
---
--- Whether the detonation of this explosive has now finished.
--- Note that this should be determined independently of the `state` of this explosive
--- @return boolean
--- @protected
function Explosive:isDetonationFinished()
  error("isDetonationFinished is not implemented")
  return false
end

--- __Abstract__
--- @param dt number The time since the last frame
--- @protected
function Explosive:updateDetonation(dt)
  error("updateDetonation is not implemented")
end

--- __Abstract__
---
--- Returns a given explosive options object with values for the given `preset`.
---
--- An implementation of this method __must__ remain static in nature (i.e. not access any instance fields or methods)
--- in order for this method to be called correctly in the constructor. This method is defined as an instance method
--- purely for the generic types to apply.
---
--- This means it is therefore safe to statically access this method.
--- @param preset T The preset of the explosive to get the options of
--- @return S The options of the explosive
function Explosive:getPresetOptions(preset)
  error("getPresetOptions is not implemented")
  return nil
end

--- __Abstract__
---
--- Destroys the explosive, removing it and all its effects.
--- @protected
function Explosive:destroy()
  error("destroy is not implemented")
  return nil
end

--- Sets the `state` to `"active"`.
--- Calls `onDetonation`.
--- @private
function Explosive:detonate()
  self.state = "active"
  self:onDetonation()
end

--- Activates the explosive, begin its fuse. If the fuse is `0`, this explosive will detonate on the next `update` call.
function Explosive:activate()
  self.state = "fused"
  self.fuseTimer = Timer:new(self.fuse, function() self:detonate() end, true)
end

--- Destroys the explosive and cleans up the instance.
--- This method will then call the `postExplosion` method.
function Explosive:deactivate()
  self:destroy()
  self:postExplosion()
end

--- Sets `state` to `"finished"`.
--- This method will then call the `onDetonationFinished` method, before calling all callbacks in the `onFinish` array.
--- @private
function Explosive:postExplosion()
  self.state = "finished"
  self:onDetonationFinished()
  self.onFinish:forEach(function(callback) callback() end)
end

--- Sets the fuse length of this explosive
--- @param fuse number The new fuse length to set
function Explosive:setFuse(fuse)
  self.fuse = fuse
end

--- @return number The fuse length of the explosive
function Explosive:getFuse()
  return self.fuse
end

--- If the call to `isActive` returns false, this method does nothing.
--- Otherwise, this method calls `updateDetonation` and `updateDetonation`.
--- If the call to `isDetonationFinished` returns true, this method calls `postExplosion`.
--- @param dt number The time since the last frame
function Explosive:update(dt)
  if not self:isActive() then return end

  if self.state == "fused" then
    self.fuseTimer:update(dt)
    return
  end

  self:updateDetonation(dt)

  if self:isDetonationFinished() then self:postExplosion() end
end

--- @return boolean Whether the `state` is `"active"` or `"fused"`
--- @private
function Explosive:isActive()
  return self.state == "active" or self.state == "fused"
end

--- Adds a given `callback` function to the `"finished"` event. Whenever this explosive's `state` changes to
---`"Finished"`, this callback function will be called.
--- @param callback fun(): void The callback function
function Explosive:registerOnFinishedHandler(callback)
  self.onFinish:add(callback)
end

--- __Constructor__
---
--- __Generics__
--- - `U` is the type for the preset of the explosive
--- - `V` is the type of the options of the explosive
--- @generic U, V
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: U | nil, options: V | nil } The configuration of the explosive. Either properties can be given; if both are given then the `preset` is ignored
--- @param presetMapper fun(self: any, preset: U): V A mapping function that produces an options object from a preset. `self` is unused
--- @return Explosive<U, V> The constructed instance
--- @protected
function Explosive:new(position, configuration, presetMapper)
  local self = --[[--- @type Explosive<U, V> ]] setmetatable({}, Explosive)
  self.options = configuration.options and configuration.options or presetMapper(self, configuration.preset)
  self.position = vec3(position)
  self.state = "ready"
  self.fuse = 0
  self.fuseTimer = nil
  self.onFinish = Array:new({})

  if not self.options then
    log("E", "explosivesNG", "No options have been provided with the following configuration: "..dumps(configuration))
  end

  return self
end

return Explosive
