local classes = require("lua/explosionsNG/helpers/classes")
local maths = require("lua/explosionsNG/helpers/maths")
local constants = require("lua/explosionsNG/constants")

--- @alias PresetStats { energy: number, range: number, accuracy: number, attackRate: number }
--- @alias AccuracyMetric "averageSpeed" | "coverage"
--- @alias EnergyMetric "single" | "multiple"

--- __Abstract__
---
--- The base stats class that all other explosive stats will inherit from. This class contains a set of abstract
--- methods, one for each stat metric.
---
--- For the accuracy stat, a subclass must pass in which metric it is using to determine how to quantify and normalise
--- this value (note that the constructor defaults this value):
---  - `"averageSpeed"` - The accuracy value implementation will be taken to be the average speed at which a vehicle
---                       must drive at to avoid the explosive's effect(s). This value will be normalised linearly, with
---                       it normalising to `1` when this value equals and exceeds `MAX_AVERAGE_SPEED_REQUIREMENT`
--- - `"coverage"` - The accuracy value implementation will be taken to be a normalised value which should correspond to
---                  the ratio of total effect(s) coverage (in area) against the total maximal coverage the explosive
---                  covers
--- @class ExplosiveStats<T>
--- @field protected explosiveOptions T The explosive options to get the stats of
--- @field private accuracyMetric AccuracyMetric The metric to use when normalising an accuracy value from a subclass
--- @field private energyMetric EnergyMetric The metric to use when normalising an energy value from a subclass
--- @field protected MAX_SINGLE_ENERGY_REQUIREMENT number __Static__ The absolute energy value required needed to return a normalised value of `1` for this stat, if the explosive produces a single energy effect
--- @field protected MAX_MULTI_ENERGY_REQUIREMENT number __Static__ The absolute energy value required needed to return a normalised value of `1` for this stat, if the explosive produces multiple energy effects
--- @field protected MAX_RANGE_REQUIREMENT number __Static__ The absolute range value required needed to return a normalised value of `1` for this stat
--- @field protected MAX_AVERAGE_SPEED_REQUIREMENT number __Static__ The absolute speed value (in ms^-1) required needed to return a normalised value of `1` for the accuracy stat when in the `"averageSpeed"`metric
--- @field protected SINGLE_ATTACK_RATE_VALUE number __Static__ The value to use for the absolute attack rate stat for an explosive who only attacks once
local ExplosiveStats = classes.newClass()
ExplosiveStats.MAX_SINGLE_ENERGY_REQUIREMENT = 625000
ExplosiveStats.MAX_MULTI_ENERGY_REQUIREMENT = 1750000
ExplosiveStats.MAX_RANGE_REQUIREMENT = 1000
ExplosiveStats.MAX_AVERAGE_SPEED_REQUIREMENT = 90
ExplosiveStats.SINGLE_ATTACK_RATE_VALUE = 10
ExplosiveStats.GRAVITATIONAL_ACCURACY_VALUE = 0.7

--- __Static__
---
--- Returns the gravitational range state for an object with a given mass.
--- More specifically, this returns the range (in m) for which the gravitational force is equal to a particular constant.
--- For details on this, see the relevant stats section in the `README.md`.
--- @param mass number The mass of the object that is the source of the gravitational effect
--- @return number The gravitational range
--- @protected
function ExplosiveStats:getGravitationalRange(mass)
  local VEHICLE_MASS = 1420
  local FORCE_CONSTANT = 488
  return math.sqrt((constants.GRAVITATIONAL * mass * VEHICLE_MASS) / FORCE_CONSTANT)
end

--- __Abstract__
--- @return number The absolute energy stat from the given options of the explosive
--- @protected
function ExplosiveStats:getAbsoluteEnergyStat()
  error("getEnergyStat is not implemented")
  return nil
end

--- __Abstract__
--- @return number The absolute range stat from the given options of the explosive
--- @protected
function ExplosiveStats:getAbsoluteRangeStat()
  error("getRange is not implemented")
  return nil
end

--- __Abstract__
--- @return number The absolute accuracy stat from the given options of the explosive
--- @protected
function ExplosiveStats:getAbsoluteAccuracyStat()
  error("getAccuracyStat is not implemented")
  return nil
end

--- __Abstract__
--- @return number The absolute accuracy stat from the given options of the explosive
--- @protected
function ExplosiveStats:getAbsoluteAttackRateStat()
  error("getAttackRateStat is not implemented")
  return nil
end

--- Normalises the given absolute accuracy value into a value between [0, 1]
--- @param value number The absolute accuracy
--- @return number The normalised accuracy
--- @private
function ExplosiveStats:normaliseAccuracy(value)
  if self.accuracyMetric == "averageSpeed" then
    return maths.constrain(value / ExplosiveStats.MAX_AVERAGE_SPEED_REQUIREMENT, 0, 1)
  end

  return value -- area coverage should already be in a range of [0, 1]
end

--- @return number The absolute energy value required needed to return a normalised value of 1 for this stat, based on the `EnergyMetric` being used
--- @private
function ExplosiveStats:getEnergyRequirement()
  return self.energyMetric == "single"
      and ExplosiveStats.MAX_SINGLE_ENERGY_REQUIREMENT
      or ExplosiveStats.MAX_MULTI_ENERGY_REQUIREMENT
end

--- @return PresetStats The combined absolute stats
function ExplosiveStats:getAbsoluteStats()
  return {
    energy = self:getAbsoluteEnergyStat(),
    range = self:getAbsoluteRangeStat(),
    accuracy = self:getAbsoluteAccuracyStat(),
    attackRate = self:getAbsoluteAttackRateStat()
  }
end

--- @return PresetStats The stats with each stat being between [0, 1]
function ExplosiveStats:getNormalisedStats()
  local absoluteStats = self:getAbsoluteStats()

  --- A wrapper to the diminishingReturns function that constrains the value to [0, 1]
  local clampedDiminishingReturns = function(value, clampToMax)
    local diminished = maths.diminishingReturns(value, clampToMax)
    return maths.constrain(diminished, 0, 1)
  end

  return {
    energy = clampedDiminishingReturns(absoluteStats.energy, self:getEnergyRequirement()),
    range = clampedDiminishingReturns(absoluteStats.range, ExplosiveStats.MAX_RANGE_REQUIREMENT),
    accuracy = self:normaliseAccuracy(absoluteStats.accuracy),
    attackRate = clampedDiminishingReturns(1 / absoluteStats.attackRate, 1)
  }
end

--- __Constructor__
--- @generic S
--- @param explosiveOptions S The explosive options to get the stats of
--- @param accuracyMetric? AccuracyMetric The metric to use when normalising an accuracy value from a subclass. Defaults to `"averageSpeed"`
--- @param energyMetric? EnergyMetric The metric to use when normalising an energy value from a subclass. Defaults to `"single"`
--- @return ExplosiveStats<S> The constructed instance
function ExplosiveStats:new(explosiveOptions, accuracyMetric, energyMetric)
  local self = --[[--- @type ExplosiveStats<S> ]] setmetatable({}, ExplosiveStats)
  self.explosiveOptions = explosiveOptions
  self.accuracyMetric = accuracyMetric or "averageSpeed"
  self.energyMetric = energyMetric or "single"

  if not self.explosiveOptions then
    log("E", "explosivesNG", "No options have been provided")
  end

  return self
end

return ExplosiveStats
