local classes = require("lua/explosionsNG/helpers/classes")
local maths = require("lua/explosionsNG/helpers/maths")
local ExplosiveStats = require("lua/explosionsNG/explosives/base/ExplosiveStats")
local ProjectileVolcano = require("lua/explosionsNG/projectiles/groups/ProjectileVolcano")

--- Shouldn't need to declare `S` as a generic for this class, but it has to for luanalysis to pick up the correct type for `U` :/

--- __Abstract__
---
--- An extension on the base `ExplosiveStats` class for explosives that contain other explosives as part of their
--- effects, as fires them out at a random trajectory. Terminology wise, the overall explosive will be referred to as
--- the "super-explosive", while the contained explosives it produces will be referred to as "sub-explosive(s)".
--- @class SuperExplosiveStats<T, S, U : ExplosiveStats<S>> : ExplosiveStats<T>
--- @field protected subExplosiveStats U The sub-explosive stats that this super-explosive produces
local SuperExplosiveStats = classes.newClass(ExplosiveStats)

--- __Static__
---
--- Returns the maximum range a projectile can be fired from a `ProjectileVolcano`
--- @param ejectionMultiplier number The multiplier of the velocity of the projectile being ejected at
--- @return number The maximum range a projectile can be fired from a `ProjectileVolcano`
--- @protected
function SuperExplosiveStats:getMaximumRangeOfProjectileVolcano(ejectionMultiplier)
  return ProjectileVolcano:getMaximumRange() * ejectionMultiplier
end

--- Returns the ratio of the expected range coverage the combined sub-explosives will cover in total, compared to the
--- super-explosive's maximum range coverage. This should be used as a helper for explosive stats that created
--- sub-explosives, that they themselves use a shockwave (or spherical ranged attack) as their detonation.
---
--- See `getAreaCoverageRatio` for details on the formula; see the README.md for a visual of the formula, specific to
--- this use case.
--- @return number The ratio of the average range coverage the sub-explosive will cover, compared to the super-explosive's range coverage
--- @protected
function SuperExplosiveStats:getSubShockwaveCoverageRatio()
  local numberOfSubExplosives = self:getNumberOfSubExplosives()
  local superExplosiveRange = self:getAbsoluteRangeStat()
  local subExplosiveRange = self.subExplosiveStats:getAbsoluteRangeStat()
  return maths.getAreaCoverageRatio(numberOfSubExplosives, superExplosiveRange, subExplosiveRange)
end

--- __Abstract__
--- @return number The number of sub-explosives that this super-explosive creates
--- @protected
function SuperExplosiveStats:getNumberOfSubExplosives()
  error("getNumberOfSubExplosives is not implemented")
  return nil
end

--- __Abstract__
--- @return number The maximum range of the super-explosive itself. In general, this should be how far a sub-explosive can be placed by the super-explosive
--- @protected
function SuperExplosiveStats:getSuperExplosiveRange()
  error("getSuperExplosiveRange is not implemented")
  return nil
end

--- @return number The absolute energy stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function SuperExplosiveStats:getAbsoluteRangeStat()
  -- The maximum range will be a sub-explosive that sits at the edge of the super-explosive's range, plus it's own range
  return self:getSuperExplosiveRange() + self.subExplosiveStats:getAbsoluteRangeStat()
end

--- @return number The absolute energy stat
--- @see ExplosiveStats#getAbsoluteEnergyStat
function SuperExplosiveStats:getAbsoluteEnergyStat()
  return self:getNumberOfSubExplosives() * self.subExplosiveStats:getAbsoluteEnergyStat()
end

--- @return number The absolute accuracy stat
--- @see ExplosiveStats#getAbsoluteAccuracyStat
function SuperExplosiveStats:getAbsoluteAccuracyStat()
  local subExplosiveNormalisedStats = self.subExplosiveStats:getNormalisedStats()
  local coverageAccuracy = self:getSubShockwaveCoverageRatio()
  -- Average the two
  return (coverageAccuracy + subExplosiveNormalisedStats.accuracy) / 2
end

--- __Constructor__
--- @generic V, W, X : ExplosiveStats<W>
--- @param explosiveOptions V The explosive options to get the stats of
--- @param subExplosiveStats X The sub-explosive stats that this super-explosive produces
--- @return SuperExplosiveStats<V, W, X> The constructed instance
function SuperExplosiveStats:new(explosiveOptions, subExplosiveStats)
  local self = classes.setmetatable(ExplosiveStats:new(explosiveOptions, "coverage", "multiple"), SuperExplosiveStats)
  self.subExplosiveStats = subExplosiveStats
  return --[[--- @type SuperExplosiveStats<V, W, X> ]] self
end

return SuperExplosiveStats
