local classes = require("lua/explosionsNG/helpers/classes")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local ShockwaveExplosive = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosive")
local Mine = require("lua/explosionsNG/worldObjects/Mine")

--- Properties:
---  - `mineRadius`       - The radius of the range of the danger zone of the mine
---  - `mineLife`         - The life (in seconds) of the mine before it naturally activates itself
---  - `shockwaveOptions` - The options to apply to the shockwave that is emitted when the mine detonates
--- @alias MineExplosiveOptions { mineRadius: number | nil, mineLife: number | nil, shockwaveExplosiveOptions: ShockwaveExplosiveOptions | nil }
--- @alias MineExplosivePreset "default"

--- @class MineExplosive : Explosive<MineExplosivePreset, MineExplosiveOptions>
--- @field mine Mine The mine object
--- @field shockwave ShockwaveExplosive The shockwave that will be created when the mine detonates
--- @field MARKER_ROTATION LuaVec3 __Static__ The rotation of the danger zone marker to apply to any new marker
local MineExplosive = classes.newClass(Explosive)
MineExplosive.MARKER_ROTATION = vec3(0, math.pi / 2, 0)

--- __Static__
---
--- Creates a new in-game marker object to represent a mine's danger area
--- @param position LuaVec3 The position to place the mine at
--- @param options MineExplosiveOptions The options of the explosive to use
--- @return SceneObject The created in-game object
function MineExplosive:createMineObject(position, options)
  return worldObjectsHandler.requestTSStatic({
    modelName = "sphere_danger_marker",
    position = position,
    dynamic = false,
    instanceColour = "0 0 0 0",
    rotationEuler = MineExplosive.MARKER_ROTATION,
    scale = options.mineRadius
  })
end

--- __Static__
---
--- Maps a list of `MineExplosiveOptions` to a list of `MineOptions`
--- @param options MineExplosiveOptions The options to map
--- @return MineOptions The mapped options
function MineExplosive:getMineOptions(options)
  return {
    range = options.mineRadius,
    life = options.mineLife
  }
end

--- __Static__
---
--- Creates and returns a new `Mine`, placing it at the given `position` with the given `options`. On detonation, the
--- mine will call the given `detonationFn` function when it detonates.
--- @param position LuaVec3 The position to place the mine at
--- @param detonationFn fun(): void A callback function to call when the mine detonates
--- @param options MineExplosiveOptions The list of options to use
--- @return Mine The created mine
function MineExplosive:createMine(position, detonationFn, options)
  local object = MineExplosive:createMineObject(position, options)
  local mineOptions = MineExplosive:getMineOptions(options)
  return Mine:new(position, object, detonationFn, mineOptions)
end

--- @param preset MineExplosivePreset The preset of the explosive to get the options of
--- @return MineExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function MineExplosive:getPresetOptions(preset)
  return {
    mineRadius = 25,
    mineLife = 15,
    shockwaveExplosiveOptions = {
      forceMultiplier = 3,
      maxRadius = 50,
      time = 1,
      shockwaveOptions = {
        audioTrack = "generic_explosion"
      }
    }
  }
end

--- @see Explosive#onDetonation
function MineExplosive:onDetonation()
  local mine = MineExplosive:createMine(self.position, function() self:onMineDetonation() end, self.options)
  mine:arm()
  self.mine = mine
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function MineExplosive:isDetonationFinished()
  return self.shockwave and self.shockwave:isDetonationFinished()
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function MineExplosive:updateDetonation(dt)
  if self.shockwave then
    self.shockwave:update(dt)
    -- If the shockwave is defined, the mine need not be updated as it will have finished
    return
  end

  self.mine:update(dt)
end

--- Creates a new shockwave and activates it
--- @private
function MineExplosive:onMineDetonation()
  local shockwave = ShockwaveExplosive:new(self.position, { options = self.options.shockwaveExplosiveOptions })
  shockwave:activate()
  self.shockwave = shockwave
end

--- @see Explosive#destroy
function MineExplosive:destroy()
  if self.mine then
    self.mine:destroy()
  end

  if self.shockwave then
    self.shockwave:deactivate()
  end
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: MineExplosivePreset | nil, options: MineExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return MineExplosive The constructed instance
function MineExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, MineExplosive.getPresetOptions), MineExplosive)
  self.mine = nil
  self.shockwave = nil
  return self
end

return MineExplosive
