local classes = require("lua/explosionsNG/helpers/classes")
local vehicles = require("lua/explosionsNG/handlers/vehicles")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Anomaly = require("lua/explosionsNG/worldObjects/anomalies/Anomaly")
local AttackingAnomaly = require("lua/explosionsNG/worldObjects/anomalies/attacks/AttackingAnomaly")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")

--- @alias AnomalyExplosivePreset "default" | "large"

--- @class AnomalyExplosive : Explosive<AnomalyExplosivePreset, AnomalyOptions>
--- @field private anomaly Anomaly The underlying anomaly entity. This should only be used for destroying the anomaly.
--- @field private attackingAnomaly AttackingAnomaly The anomaly entity, wrapped in an attacking state
--- @field private empTimer Timer The timer for the EMP. Upon finishing, the EMP is then activated
--- @field private isQueueingForceAttack boolean Whether a `discharge` state is being queued up for the anomaly
--- @field private attackRange number The range at which the anomaly will be eligible to discharge itself, should a vehicle fall inside this range
--- @field private EMP_TIMER number __Static__ The time that should elapse before each EMP shockwave begins
--- @field private EMP_TIME number __Static__ The amount of time the EMP shockwave lasts for
--- @field private ATTACK_RANGE_MULTIPLIER number __Static__ The multiplier of the static radius of the anomaly that corresponds to its attack range. See `attackRange` for details.
--- @field public DISCHARGE_TIME number __Static__ The amount of time the discharging shockwaves last for
local AnomalyExplosive = classes.newClass(Explosive)
AnomalyExplosive.EMP_TIMER = 12
AnomalyExplosive.EMP_TIME = 2
AnomalyExplosive.ATTACK_RANGE_MULTIPLIER = 5
AnomalyExplosive.DISCHARGE_TIME = 1.5

--- __Static__
--- @return AnomalyExplosivePreset[] The list of presets for this explosive
function AnomalyExplosive:getPresets()
  return { "default", "large" }
end

--- __Static__
---
--- Returns the maximum number of discharges an anomaly with the given amount of life can perform
--- @param life number The anomaly life (in seconds)
--- @return number The average number of discharges an anomaly with the given amount of life will perform
function AnomalyExplosive:getMaximumNumberOfDischarges(life)
  local totalNumberOfFluctuations = life * math.floor(Anomaly.FLUCTUATION_PERIOD / AnomalyExplosive.EMP_TIMER)
  return (life - totalNumberOfFluctuations) / Anomaly.DISCHARGING_PERIOD
end

--- __Static__
---
--- Returns the bounding area that makes up this explosive. In particular, this will be the maximum bounds of the
--- anomaly object itself.
--- @param preset AnomalyExplosivePreset The preset of the explosive to get the bounding area of
--- @return number The area of the objects that make up this explosive
function AnomalyExplosive:getBoundingArea(preset)
  local options = AnomalyExplosive:getPresetOptions(preset)
  return Anomaly:getMaxRadiusOf(options.scale)
end

--- __Static__
---
--- Returns the maximum range the EMP pulse can reach for the given `scale`
--- @param scale number The scale of the anomaly
--- @return number The maximum range the EMP pulse can reach
--- @private
function AnomalyExplosive:getEmpRange(scale)
  return scale * 200
end

--- __Static__
---
--- Returns the maximum range the discharge shockwave can reach for the given `scale`
--- @param scale number The scale of the anomaly
--- @return number The maximum range the discharge shockwave can reach
function AnomalyExplosive:getDischargeRange(scale)
  return scale * 100
end

--- @param preset AnomalyExplosivePreset The preset of the explosive to get the options of
--- @return AnomalyOptions The options
--- @see Explosive#getPresetOptions
function AnomalyExplosive:getPresetOptions(preset)
  if preset == "default" then return { scale = 1.5, numberOfLayers = 10, life = 30 } end
  return { scale = 3, numberOfLayers = 14, life = 45 }
end

--- Opens the anomaly and begins the EMP timer
--- @see Explosive#onDetonation
function AnomalyExplosive:onDetonation()
  self.attackingAnomaly:open()
  self.empTimer:start()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function AnomalyExplosive:isDetonationFinished()
  return self.attackingAnomaly:isFullyClosed()
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function AnomalyExplosive:updateDetonation(dt)
  self:tryAndAttack()
  self.attackingAnomaly:update(dt)
  self.empTimer:update(dt)
end

--- If this explosive is not currently queueing up a force attack, and the anomaly can attack, then this method does so
--- @private
function AnomalyExplosive:tryAndAttack()
  if self.isQueueingForceAttack then return end

  if self:canAnomalyAttack() then
    self:queueForceAttack()
    self.isQueueingForceAttack = true
  end
end

--- Checks whether the anomaly has closed and if so, this method returns `false`.
--- Otherwise, this methods checks whether at least one vehicle is within the attack zone of the `anomaly`.
--- @return boolean Whether a vehicle is within range of the `anomaly` and it has not closed
--- @private
function AnomalyExplosive:canAnomalyAttack()
  if self.attackingAnomaly:isFullyClosed() then return false end
  return vehicles.isVehicleInRange(self.position, self.attackRange)
end

--- Queues up an EMP attack by the anomaly
--- @private
function AnomalyExplosive:queueEmpAttack()
  self.attackingAnomaly:attack({
    animation = "fluctuating",
    attack = "emp",
    maxRadius = AnomalyExplosive:getEmpRange(self.options.scale),
    time = AnomalyExplosive.EMP_TIME
  })
end

--- Queues up a force attack by the anomaly
--- @private
function AnomalyExplosive:queueForceAttack()
  self.attackingAnomaly:attack({
    animation = "discharging",
    attack = "force",
    maxRadius = AnomalyExplosive:getDischargeRange(self.options.scale),
    time = AnomalyExplosive.DISCHARGE_TIME,
    onAttackFn = function() self.isQueueingForceAttack = false end
  })
end

--- Creates a new `AttackingAnomaly` instance at the position with the options set in this instance
function AnomalyExplosive:initAnomaly()
  local anomaly = Anomaly:new(self.position, self.options)
  self.anomaly = anomaly
  self.attackingAnomaly = AttackingAnomaly:new(anomaly)
  self.attackRange = anomaly:getStaticRadius() * AnomalyExplosive.ATTACK_RANGE_MULTIPLIER
end

--- @see Explosive#destroy
function AnomalyExplosive:destroy()
  self.attackingAnomaly:destroy()
  self.anomaly:destroy()
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: AnomalyExplosivePreset | nil, options: AnomalyOptions | nil } The configuration of the explosive. Either properties can be given; if both are given then the `preset` is ignored
--- @return AnomalyExplosive The constructed instance
function AnomalyExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, AnomalyExplosive.getPresetOptions), AnomalyExplosive)
  self.empTimer = IntervalTimer:new(AnomalyExplosive.EMP_TIMER, function() self:queueEmpAttack() end)
  self.isQueueingForceAttack = false
  self.attackRange = nil
  self:initAnomaly()
  return self
end

return AnomalyExplosive
