local classes = require("lua/explosionsNG/helpers/classes")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Timer = require("lua/explosionsNG/timers/Timer")
local Star = require("lua/explosionsNG/worldObjects/celestials/star/Star")
local BlackHole = require("lua/explosionsNG/worldObjects/celestials/blackHole/BlackHole")
local VehiclePlanet = require("lua/explosionsNG/vehicles/VehiclePlanet")
local Vehicle = require("lua/explosionsNG/vehicles/Vehicle")
local Pulsar = require("lua/explosionsNG/worldObjects/Pulsar")
local effectFactory = require("lua/explosionsNG/vehicles/effects/factory")

--- Properties:
---  - `starRadius` - the radius (in m) of the star(s)
---  - `blackHoleLifeTime` - the life time of the black hole
---  - `blackHoleMass` - the mass (in kg) of the black hole
--- @alias BlackHoleExplosiveOptions { starRadius: number, blackHoleLifeTime: number, blackHoleMass: number }
--- @alias BlackHoleExplosivePreset "default" | "supermassive"

--- @class BlackHoleExplosive : Explosive<BlackHoleExplosivePreset, BlackHoleExplosiveOptions>
--- @field private star Star The star object during the star phase
--- @field private blackHole BlackHole The black hole object during the black hole phase
--- @field private pulsar Pulsar The pulsar that additionally pulls vehicles towards the black hole centre
--- @field private vehiclePlanet VehiclePlanet The vehicle planet that will be applied to all vehicles
--- @field private blackHoleLifeTimer Timer The black hole life timer. On completion, the black hole will be evaporated.
--- @field private vehiclesEventHandler VehicleEventFn The event handler for a vehicle event on the vehicles' handler
--- @field private STAR_COLOUR string __Static__ The colour of the star
--- @field private VEHICLE_PLANET_RADIUS number __Static__ The radius of the vehicle planet at which the gravitational effect tapers off
--- @field public PULSAR_INTERVAL number __Static__ The interval (in s) of time between two pulses within a pulsar
--- @field public PULSAR_SHOCKWAVE_TIME number __Static__ The time (in s) each pulsar shockwave lasts for
--- @field public PULSAR_FORCE_MULTIPLIER number __Static__ The force multiplier of each pulsar shockwave
local BlackHoleExplosive = classes.newClass(Explosive)
BlackHoleExplosive.STAR_COLOUR = "1 0 0"
BlackHoleExplosive.VEHICLE_PLANET_RADIUS = 25
BlackHoleExplosive.PULSAR_INTERVAL = 3
BlackHoleExplosive.PULSAR_SHOCKWAVE_TIME = 2.5
BlackHoleExplosive.PULSAR_FORCE_MULTIPLIER = 10

--- __Static__
--- @return BlackHoleExplosivePreset[] The list of presets for this explosive
function BlackHoleExplosive:getPresets()
  return { "default", "supermassive" }
end

--- __Static__
---
--- Returns the bounding box that makes up this explosive. In particular, this will be the bounding sphere of the star.
--- @param preset BlackHoleExplosivePreset The preset of the explosive to get the bounding area of
--- @return BoundingArea The area of the objects that make up this explosive
function BlackHoleExplosive:getBoundingArea(preset)
  local options = BlackHoleExplosive:getPresetOptions(preset)
  return options.starRadius
end

--- __Static__
---
--- Returns the radius of a black hole from the radius of the star it is born from
--- @param starRadius number The star radius
--- @return number The radius of the black hole, from the star radius
--- @private
function BlackHoleExplosive:starRadiusToBlackHoleRadius(starRadius)
  return starRadius * 4
end

--- __Static__
---
--- Returns the radius of a pulsar from the radius of the star
--- @param starRadius number The star radius
--- @return number The radius of a pulsar, from the star radius
function BlackHoleExplosive:starRadiusToPulsarRadius(starRadius)
  return starRadius * 10
end

--- __Static__
---
--- Returns a `VehiclePlanet` to apply to vehicles that performs a gravitational affect on them
--- @param position LuaVec3 The position of the black hole
--- @param mass number The mass (in kg) of the black hole
--- @return VehiclePlanet The created object
--- @private
function BlackHoleExplosive:createVehiclePlanet(position, mass)
  return VehiclePlanet:new(position, BlackHoleExplosive.VEHICLE_PLANET_RADIUS, mass)
end

--- __Static__
---
--- Returns the number of shockwaves a pulsar will produce for the given options
--- @param options BlackHoleExplosiveOptions The options for the explosive
--- @return number The number of shockwaves for a pulsar
function BlackHoleExplosive:getNumberOfPulsarShockwaves(options)
  return options.blackHoleLifeTime / BlackHoleExplosive.PULSAR_INTERVAL
end

--- __Static__
---
--- Returns a new `Pulsar` that is placed at the given position and is constructed with options, derived from the given
--- options.
--- @param position LuaVec3 The position to place the pulsar at
--- @param options BlackHoleExplosiveOptions The options for the explosive
--- @return Pulsar The created pulsar
--- @private
function BlackHoleExplosive:createPulsar(position, options)
  local numberOfShockwaves = BlackHoleExplosive:getNumberOfPulsarShockwaves(options)
  local maxRadius = BlackHoleExplosive:starRadiusToPulsarRadius(options.starRadius)

  return Pulsar:new(position, {
    time = BlackHoleExplosive.PULSAR_SHOCKWAVE_TIME,
    interval = BlackHoleExplosive.PULSAR_INTERVAL,
    maxRadius = maxRadius,
    numberOfShockwaves = numberOfShockwaves,
    affectFn = function(id, event)
      return effectFactory.createForceEffect(id, event.position, maxRadius, BlackHoleExplosive.PULSAR_FORCE_MULTIPLIER, true)
    end
  })
end

--- @param preset BlackHoleExplosivePreset The preset of the explosive to get the options of
--- @return BlackHoleExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function BlackHoleExplosive:getPresetOptions(preset)
  if preset == "default" then
    return { starRadius = 30, blackHoleLifeTime = 30, blackHoleMass = 3e15 }
  end

  if preset == "supermassive" then
    return { starRadius = 50, blackHoleLifeTime = 60, blackHoleMass = 1e16 }
  end
end

--- @see Explosive#onDetonation
function BlackHoleExplosive:onDetonation()
  self.star = self:createSingleStar()
end

--- @see Explosive#onDetonationFinished
function BlackHoleExplosive:onDetonationFinished()
  self:unregisterEventsOnHandler()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function BlackHoleExplosive:isDetonationFinished()
  local isStarDead = self.star and self.star:isDead()
  local isBlackHoleDead = self.blackHole and self.blackHole:isDead()
  return isStarDead and isBlackHoleDead
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function BlackHoleExplosive:updateDetonation(dt)
  self.blackHoleLifeTimer:update(dt)
  if self.star then self.star:update(dt) end
  if self.blackHole then self.blackHole:update(dt) end
  if self.pulsar then self.pulsar:update(dt) end
end

--- Creates a new star from the options and position of this instance
--- @return Star The created star
--- @private
function BlackHoleExplosive:createSingleStar()
  local star = Star:new(self.position, self.options.starRadius, BlackHoleExplosive.STAR_COLOUR)

  local events = star:getEvents()
  events:registerEventHandler("formed", function() star:collapse() end)
  events:registerEventHandler("compressed", function() self:onStarDeath() end)

  return star
end

--- Creates a new black hole from the options and position of this instance
--- @return BlackHole The created black hole
--- @private
function BlackHoleExplosive:createBlackHole()
  local radius = BlackHoleExplosive:starRadiusToBlackHoleRadius(self.options.starRadius)
  local blackHole = BlackHole:new(self.position, radius)

  local events = blackHole:getEvents()
  events:registerEventHandler("formed", function()
    self.blackHoleLifeTimer:start()
    self.pulsar:begin()
  end)
  events:registerEventHandler("evaporated", function() self:updatePlanetOnVehicles(false) end)

  return blackHole
end

--- Called when the star dies.
--- Creates a new black hole and applies the `vehiclePlanet` to all current vehicles.
--- This method will also set up the events on the vehicles' handler.
--- @private
function BlackHoleExplosive:onStarDeath()
  self.blackHole = self:createBlackHole()
  self.pulsar = BlackHoleExplosive:createPulsar(self.position, self.options)
  self:registerEventsOnHandler()
  self:updatePlanetOnVehicles(true)
end

--- @see Explosive#destroy
function BlackHoleExplosive:destroy()
  if self.star then
    self.star:destroy()
  end

  if self.blackHole then
    self.blackHole:destroy()
  end

  if self.pulsar then
    self.pulsar:destroy()
  end

  self:updatePlanetOnVehicles(false)
  self:unregisterEventsOnHandler()
end

--- Adds or removes the `vehiclePlanet` from this instance to the vehicles on the vehicles' handler
--- @param add boolean Whether to add the planet to the vehicle. If `false`, it will instead remove the planet.
--- @private
function BlackHoleExplosive:updatePlanetOnVehicles(add)
  local method = add and Vehicle.addPlanet or Vehicle.removePlanet
  vehiclesHandler.getVehicles():forEach(function(vehicle) method(vehicle, self.vehiclePlanet) end)
end

--- Registers various event handlers to the vehicles' handler to add the `vehiclePlanet` as necessary
--- @private
function BlackHoleExplosive:registerEventsOnHandler()
  vehiclesHandler.registerHandler("spawned", self.vehiclesEventHandler)
  vehiclesHandler.registerHandler("resetted", self.vehiclesEventHandler)
end

--- Unregisters various event handlers to the vehicles' handler
--- @private
function BlackHoleExplosive:unregisterEventsOnHandler()
  vehiclesHandler.unregisterHandler("spawned", self.vehiclesEventHandler)
  vehiclesHandler.unregisterHandler("resetted", self.vehiclesEventHandler)
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: BlackHoleExplosivePreset | nil, options: BlackHoleExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return BlackHoleExplosive The constructed instance
function BlackHoleExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, BlackHoleExplosive.getPresetOptions), BlackHoleExplosive)
  self.star = nil
  self.blackHole = nil
  self.pulsar = nil
  self.blackHoleLifeTimer = Timer:new(self.options.blackHoleLifeTime, function() self.blackHole:evaporate() end)
  self.vehiclePlanet = BlackHoleExplosive:createVehiclePlanet(self.position, self.options.blackHoleMass)
  self.vehiclesEventHandler = function(vehicle) vehicle:addPlanet(self.vehiclePlanet) end
  return self
end

return BlackHoleExplosive
