local classes = require("lua/explosionsNG/helpers/classes")
local ExplosiveStats = require("lua/explosionsNG/explosives/base/ExplosiveStats")
local BlackHoleExplosive = require("lua/explosionsNG/explosives/contained/blackHole/BlackHoleExplosive")
local ShockwaveExplosiveStats = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosiveStats")

--- @class BlackHoleExplosiveStats : ExplosiveStats<BlackHoleExplosiveOptions>
--- @field private pulsarShockwaveStats ShockwaveExplosiveStats The stats for each pulsar explosive
local BlackHoleExplosiveStats = classes.newClass(ExplosiveStats)

--- @return number The absolute energy stat
--- @see ExplosiveStats#getAbsoluteEnergyStat
function BlackHoleExplosiveStats:getAbsoluteEnergyStat()
  local numberOfShockwaves = BlackHoleExplosive:getNumberOfPulsarShockwaves(self.explosiveOptions)
  return self.pulsarShockwaveStats:getAbsoluteEnergyStat() * numberOfShockwaves
end

--- @return number The absolute range stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function BlackHoleExplosiveStats:getAbsoluteRangeStat()
  local gravitationalRange = ExplosiveStats:getGravitationalRange(self.explosiveOptions.blackHoleMass)
  local pulsarRange = self.pulsarShockwaveStats:getAbsoluteRangeStat()
  return math.max(gravitationalRange, pulsarRange)
end

--- @return number The absolute accuracy stat
--- @see ExplosiveStats#getAbsoluteAccuracyStat
function BlackHoleExplosiveStats:getAbsoluteAccuracyStat()
  local pulsarAccuracy = self.pulsarShockwaveStats:getAbsoluteAccuracyStat()
  local gravitationalAccuracy = ExplosiveStats.GRAVITATIONAL_ACCURACY_VALUE * ExplosiveStats.MAX_AVERAGE_SPEED_REQUIREMENT
  return math.max(pulsarAccuracy, gravitationalAccuracy)
end

--- @return number The absolute attack rate stat
--- @see ExplosiveStats#getAbsoluteAttackRateStat
function BlackHoleExplosiveStats:getAbsoluteAttackRateStat()
  return BlackHoleExplosive.PULSAR_SHOCKWAVE_TIME
end

--- __Constructor__
--- @param explosiveOptions BlackHoleExplosiveOptions The explosive options to get the stats of
--- @return BlackHoleExplosiveStats The constructed instance
function BlackHoleExplosiveStats:new(explosiveOptions)
  local self = classes.setmetatable(ExplosiveStats:new(explosiveOptions, "averageSpeed", "multiple"), BlackHoleExplosiveStats)
  self.pulsarShockwaveStats = ShockwaveExplosiveStats:new({
    time = BlackHoleExplosive.PULSAR_SHOCKWAVE_TIME,
    forceMultiplier = BlackHoleExplosive.PULSAR_FORCE_MULTIPLIER,
    maxRadius = BlackHoleExplosive:starRadiusToPulsarRadius(explosiveOptions.starRadius)
  })
  return self
end

return BlackHoleExplosiveStats
