local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local ShockwaveExplosive = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosive")
local ProjectileVolcano = require("lua/explosionsNG/projectiles/groups/ProjectileVolcano")

--- Properties:
---  - `numberOfPieces` - the number of small explosives this cluster will explode into
---  - `ejectionMultiplier` - the multiplier of the velocity when projectiles are thrown out of this explosive
---  - `recursiveFactor` - the recursion of other clusters inside this one. The total explosives being thrown out will
---                        be `numberOfPieces ^ (recursiveFactor + 1)`. Any non-zero integer is permitted. For example,
---                        if set to 0, this explosive has no recursion; if set to 2, then this explosive will split two
---                        additional times
---  - `shockwaveExplosiveOptions` - the list of options to use for shockwave explosions
--- @alias ClusterExplosiveOptions { numberOfPieces: number, ejectionMultiplier: number, recursiveFactor: number, shockwaveExplosiveOptions: ShockwaveExplosiveOptions | nil }
--- @alias ClusterExplosivePreset "default" | "recursive" | "compact" | "area"

--- @class ClusterExplosive : Explosive<ClusterExplosivePreset, ClusterExplosiveOptions>
--- @field volcano ProjectileVolcano The volcano of explosives that are thrown out
--- @field PROJECTILE_INTERVAL number __Static__ The time between firing two explosives in succession from the `volcano`
--- @field RECURSIVE_EJECTION_SCALE_FACTORY number __Static__ The scale factor to apply to new explosives for each recursive iteration
local ClusterExplosive = classes.newClass(Explosive)
ClusterExplosive.PROJECTILE_INTERVAL = 0.3
ClusterExplosive.RECURSIVE_EJECTION_SCALE_FACTORY = 1.5

--- __Static__
--- @return ClusterExplosivePreset[] The list of presets for this explosive
function ClusterExplosive:getPresets()
  return { "default", "recursive", "compact", "area" }
end

--- __Static__
---
--- Creates and returns a payload explosive that is set to detonate at the given `position`. If the recursive factor of
--- the given `options` is `0`, then this method will return a `ShockwaveExplosive`. Else, this method will return
--- another `ClusterExplosive` with the same list of options to those given, except with its recursive factor reduced by
--- `1`.
--- @param position LuaVec3 The position at which the payload should be detonated at
--- @param options ClusterExplosiveOptions The list of options from the explosive that is carrying the payload
--- @return AnyExplosive The payload
function ClusterExplosive:getPayload(position, options)
  if options.recursiveFactor == 0 then
    return --[[--- @type AnyExplosive ]] ShockwaveExplosive:new(position, {
      options = options.shockwaveExplosiveOptions
    })
  end

  return --[[--- @type AnyExplosive ]] ClusterExplosive:new(position, {
    options = tables.merge(options, --[[--- @type ClusterExplosiveOptions ]] {
      ejectionMultiplier = options.ejectionMultiplier * ClusterExplosive.RECURSIVE_EJECTION_SCALE_FACTORY,
      recursiveFactor = options.recursiveFactor - 1,
    })
  })
end

--- __Static__
---
--- Fills in any key not in the given `options` with default values
--- @param options ClusterExplosiveOptions The list of predefined options
--- @return ProjectileVolcanoOptions The complete set of options with defaults filled in where empty values are given
function ClusterExplosive:getVolcanoOptions(options)
  return {
    ejectionMultiplier = options.ejectionMultiplier,
    numberOfProjectiles = options.numberOfPieces,
    interval = ClusterExplosive.PROJECTILE_INTERVAL
  }
end

--- __Static__
---
--- Creates a new `ProjectileVolcano` that is placed at the given `position`, with a producer function (see
--- `getPayload`) and a set of options that are derived from the given `options` (see `getVolcanoOptions`)
--- @param position LuaVec3 The position to set the volcano to
--- @param options ClusterExplosiveOptions The list of options of the explosive to use
--- @return ProjectileVolcano The created projectile volcano
--- @see ClusterExplosive#getPayload
function ClusterExplosive:createVolcano(position, options)
  local producer = function(projectilePosition) return ClusterExplosive:getPayload(projectilePosition, options) end
  local volcanoOptions = ClusterExplosive:getVolcanoOptions(options)
  return ProjectileVolcano:new(position, producer, volcanoOptions)
end

--- @param preset ClusterExplosivePreset The preset of the explosive to get the options of
--- @return ClusterExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function ClusterExplosive:getPresetOptions(preset)
  --- @type ShockwaveExplosiveOptions
  local baseShockwaveExplosiveOptions = { time = 1.5, maxRadius = 100, forceMultiplier = 3 }

  if preset == "default" then
    return {
      numberOfPieces = 6,
      ejectionMultiplier = 1,
      recursiveFactor = 0,
      shockwaveExplosiveOptions = baseShockwaveExplosiveOptions
    }
  end

  if preset == "compact" then
    return {
      numberOfPieces = 6,
      ejectionMultiplier = 0.5,
      recursiveFactor = 0,
      shockwaveExplosiveOptions = tables.merge(baseShockwaveExplosiveOptions, --[[--- @type ShockwaveExplosiveOptions ]] { forceMultiplier = 5 })
    }
  end

  if preset == "area" then
    return {
      numberOfPieces = 8,
      ejectionMultiplier = 2,
      recursiveFactor = 0,
      shockwaveExplosiveOptions = tables.merge(baseShockwaveExplosiveOptions, --[[--- @type ShockwaveExplosiveOptions ]] { forceMultiplier = 3 })
    }
  end

  if preset == "recursive" then
    return {
      numberOfPieces = 2,
      ejectionMultiplier = 0.25,
      recursiveFactor = 3,
      shockwaveExplosiveOptions = baseShockwaveExplosiveOptions
    }
  end
end

--- @see Explosive#onDetonation
function ClusterExplosive:onDetonation()
  self.volcano:erupt()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function ClusterExplosive:isDetonationFinished()
  return self.volcano:isFinished()
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function ClusterExplosive:updateDetonation(dt)
  self.volcano:update(dt)
end

--- @see Explosive#destroy
function ClusterExplosive:destroy()
  self.volcano:destroy()
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: ClusterExplosivePreset | nil, options: ClusterExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return ClusterExplosive The constructed instance
function ClusterExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, ClusterExplosive.getPresetOptions), ClusterExplosive)
  self.volcano = ClusterExplosive:createVolcano(position, self.options)
  return self
end

return ClusterExplosive
