local classes = require("lua/explosionsNG/helpers/classes")
local SuperExplosiveStats = require("lua/explosionsNG/explosives/base/SuperExplosiveStats")
local ProjectileVolcano = require("lua/explosionsNG/projectiles/groups/ProjectileVolcano")
local ShockwaveExplosiveStats = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosiveStats")

--- @class ClusterExplosiveStats : SuperExplosiveStats<ClusterExplosiveOptions, ShockwaveExplosiveOptions, ShockwaveExplosiveStats>
--- @field shockwaveStats ShockwaveExplosiveStats The list of options for the shockwave payload
local ClusterExplosiveStats = classes.newClass(SuperExplosiveStats)

--- @return number The number of sub-explosives that this super explosive creates
--- @see SuperExplosiveStats#getNumberOfSubExplosives
function ClusterExplosiveStats:getNumberOfSubExplosives()
  return self.explosiveOptions.numberOfPieces ^ self:getNumberOfRecursions()
end

--- @return number The maximum range of the super-explosive itself. In general, this should be how far a sub-explosive can be placed by the super-explosive
--- @see SuperExplosiveStats#getSuperExplosiveRange
function ClusterExplosiveStats:getSuperExplosiveRange()
  -- The average range will be equivalent to the maximum range a projectile can travel, multiplied by the number of
  -- recursions (i.e. the number of "hops" the last projectile's position will be at)
  local maxProjectileRangePerHop = SuperExplosiveStats:getMaximumRangeOfProjectileVolcano(self.explosiveOptions.ejectionMultiplier)
  return maxProjectileRangePerHop * self:getNumberOfRecursions()
end

--- @return number The absolute attack rate stat
--- @see ExplosiveStats#getAbsoluteAttackRateStat
function ClusterExplosiveStats:getAbsoluteAttackRateStat()
  local lifeTime = ProjectileVolcano:getTotalProjectileAirTime() * self:getNumberOfRecursions()
  return lifeTime / self:getNumberOfSubExplosives()
end

--- @return number The number of recursions the options allow for. This will be an integer >= 1
function ClusterExplosiveStats:getNumberOfRecursions()
  return self.explosiveOptions.recursiveFactor + 1
end

--- __Constructor__
--- @param explosiveOptions ClusterExplosiveOptions The explosive options to get the stats of
--- @return ClusterExplosiveStats The constructed instance
function ClusterExplosiveStats:new(explosiveOptions)
  local shockwaveExplosiveOptions = explosiveOptions.shockwaveExplosiveOptions

  if not shockwaveExplosiveOptions then
    error("Shockwave explosive options cannot be nil!")
    return nil
  end

  local shockwaveStats = ShockwaveExplosiveStats:new(--[[--- @not nil ]] shockwaveExplosiveOptions)
  local self = classes.setmetatable(SuperExplosiveStats:new(explosiveOptions, shockwaveStats), ClusterExplosiveStats)
  return self
end

return ClusterExplosiveStats
