local classes = require("lua/explosionsNG/helpers/classes")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Convergence = require("lua/explosionsNG/worldObjects/anomalies/convergence/Convergence")

--- @alias ConvergenceExplosivePreset "easy" | "normal" | "hard"

--- @class ConvergenceExplosive : Explosive<ConvergenceExplosivePreset, ConvergenceOptions>
--- @field private convergence Convergence The convergence object
--- @field private finished boolean Whether the convergence has finished
local ConvergenceExplosive = classes.newClass(Explosive)

--- __Static__
--- @return ConvergenceExplosivePreset[] The list of presets for this explosive
function ConvergenceExplosive:getPresets()
  return { "easy", "normal", "hard" }
end

--- __Static__
---
--- Returns the bounding area that makes up this explosive
--- @param preset ConvergenceExplosivePreset The preset of the explosive to get the bounding area of
--- @return SphericalBound[] The area of the objects that make up this explosive
function ConvergenceExplosive:getBoundingArea(preset)
  return Convergence:getBoundingArea()
end

--- @param preset ConvergenceExplosivePreset The preset of the explosive to get the options of
--- @return ConvergenceOptions The options of the explosive
--- @see Explosive#getPresetOptions
function ConvergenceExplosive:getPresetOptions(preset)
  if preset == "easy" then
    return {
      phaseOptions = {
        timeMultiplier = 1.4
      }
    }
  end

  if preset == "normal" then
    return {
      phaseOptions = {
        timeMultiplier = 1.2
      }
    }
  end

  if preset == "hard" then
    return {
      phaseOptions = {
       timeMultiplier = 1
      }
    }
  end
end

--- @see Explosive#onDetonation
function ConvergenceExplosive:onDetonation()
  self.convergence:open()
  guihooks.trigger("engBossBarUpdate", self.convergence:getCurrentProgress())
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function ConvergenceExplosive:isDetonationFinished()
  return self.finished
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function ConvergenceExplosive:updateDetonation(dt)
  self.convergence:update(dt)
end

--- @return ConvergencePhaseProgress[] Te current progress through the convergence
function ConvergenceExplosive:getCurrentProgress()
  return self.convergence:getCurrentProgress()
end

--- @return boolean Whether the player has failed the current phase
function ConvergenceExplosive:hasFailedCurrentPhase()
  return self.convergence:hasFailedCurrentPhase()
end

--- Creates a new convergence object and registers its events to this instance
--- @param position LuaVec3 The position to place the convergence object
--- @param options ConvergenceOptions The list of options to use
--- @return Convergence The created convergence
--- @private
function ConvergenceExplosive:createConvergence(position, options)
  local convergence = Convergence:new(position, options)

  local events = convergence:getEvents()
  events:registerEventHandler("phaseUpdated", function() self:onPhaseUpdate(convergence) end)
  events:registerEventHandler("completed", function() self:onCompleted() end)
  events:registerEventHandler("playerAffected", function(wasPlayerAffected) self:onPlayerAffected(wasPlayerAffected) end)

  return convergence
end

--- Called when the convergence emits a phase update.
--- Notifies the UI with the current progress of the given convergence instance.
--- @param convergence Convergence The convergence instance
--- @private
function ConvergenceExplosive:onPhaseUpdate(convergence)
  local data = convergence:getCurrentProgress()
  guihooks.trigger("engBossBarUpdate", data)
end

--- Called when the convergence entity is defeated/finishes.
--- Notifies the UI and sets the internal `finished` flag to `true`.
--- @private
function ConvergenceExplosive:onCompleted()
  -- Send nothing to let the UI know it has finished
  guihooks.trigger("engBossBarUpdate")
  self.finished = true
end

--- Called when the player vehicle is affected (or not) during a phase/subphase.
--- Notifies the UI with the given flag.
--- @param wasPlayerAffected boolean Whether the player vehicle was affected
--- @private
function ConvergenceExplosive:onPlayerAffected(wasPlayerAffected)
  guihooks.trigger("engBossBarFailed", wasPlayerAffected)
end

--- @see Explosive#destroy
function ConvergenceExplosive:destroy()
  self.convergence:destroy()
  self:onCompleted()
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: ConvergenceExplosivePreset | nil, options: ConvergenceOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return ConvergenceExplosive The constructed instance
function ConvergenceExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, ConvergenceExplosive.getPresetOptions), ConvergenceExplosive)
  self.convergence = self:createConvergence(position, self.options)
  self.finished = false
  return self
end

return ConvergenceExplosive
