local classes = require("lua/explosionsNG/helpers/classes")
local effectFactory = require("lua/explosionsNG/vehicles/effects/factory")
local Array = require("lua/explosionsNG/misc/Array")
local Orb = require("lua/explosionsNG/worldObjects/Orb")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local CircularPath = require("lua/explosionsNG/pathing/CircularPath")

--- @alias OrbType "zero" | "light" | "heavy" | "negative"

--- Properties:
---  - `numberOfOrbs` - the number of orbs to generate
---  - `orbOptions` - the list of options to use for the orbs. Note that the `colour` will be overwritten.
--- @alias GravityOrbsExplosiveOptions { numberOfOrbs: number | nil, orbOptions: OrbOptions | nil }
--- @alias GravityOrbsExplosivePreset "default" | "dense" | "energetic" | "lingering"

--- @class GravityOrbsExplosive : Explosive<GravityOrbsExplosivePreset, GravityOrbsExplosiveOptions>
--- @field private orbs Array<Orb> The list of orbs
--- @field private EFFECT_DURATION number __Static__
--- @field private TYPES Array<OrbType> __Static__
local GravityOrbsExplosive = classes.newClass(Explosive)
GravityOrbsExplosive.EFFECT_DURATION = 3
GravityOrbsExplosive.TYPES = Array:new({ "zero", "light", "heavy", "negative" })

--- __Static__
--- @return GravityOrbsExplosivePreset[] The list of presets for this explosive
function GravityOrbsExplosive:getPresets()
  return { "default", "dense", "energetic", "lingering" }
end

--- __Static__
--- @return OrbType The random orb type
--- @private
function GravityOrbsExplosive:getRandomType()
  local random = math.random(1, GravityOrbsExplosive.TYPES:length())
  return GravityOrbsExplosive.TYPES:nth(random)
end

--- __Static__
---
--- Returns the RGB colour of the corresponding given `OrbType`
--- @param type OrbType The type of orb to get the colour of
--- @return string The RGB colour
--- @private
function GravityOrbsExplosive:getOrbColour(type)
  if type == "zero"  then return "1 1 1" end
  if type == "light" then return "1 1 0" end
  if type == "heavy" then return "1 0 0" end
  return "0 0.2 1"
end

--- __Static__
---
--- Returns the gravity (in m/s^-2) of the corresponding given `OtbType`
--- @param type OrbType The type of orb to get the gravity
--- @return number The gravity
--- @private
function GravityOrbsExplosive:getGravity(type)
  if type == "zero"  then return 0 end
  if type == "light" then return -4 end
  if type == "heavy" then return -30 end
  return 1
end

--- __Static__
---
--- Returns a merge list of options from the given list of options, and the orb type.
--- Specifically, the returned options will be a list of options from the given options, but with the `colour` set to
--- the colour corresponding with the orb type.
--- @param options OrbOptions The list of orb options
--- @param type OrbType The type of orb
--- @return OrbOptions The merged options
--- @private
function GravityOrbsExplosive:getOptions(options, type)
  return {
    colour = GravityOrbsExplosive:getOrbColour(type),
    speed = options.speed,
    life = options.life,
    radius = options.radius
  }
end

--- __Static__
---
--- Creates and returns a new list of orbs with a gravitational effect to them
--- @param options GravityOrbsExplosiveOptions The list of options of the explosive
--- @param position LuaVec3 The position of the explosive itself
--- @return Array<Orb> The generated list of orbs
--- @private
function GravityOrbsExplosive:createOrbs(options, position)
  local numberOfOrbs = options.numberOfOrbs
  local ringPath = CircularPath:new(position, 1)

  return Array:from(numberOfOrbs, function(index)
    local rotation = (index / numberOfOrbs) * 2 * math.pi
    local rotatedPosition = ringPath:getPositionAtOffsetRotation(rotation)

    local direction = position - rotatedPosition
    local orbPosition = position + direction -- Add one second of movement

    local orbType = GravityOrbsExplosive:getRandomType()
    local orbOptions = GravityOrbsExplosive:getOptions(options.orbOptions or {}, orbType)
    local gravity = GravityOrbsExplosive:getGravity(orbType)
    --- @type VehicleEffectProducer
    local affectFn = function(id) return effectFactory.createLocalGravityEffect(id, gravity, GravityOrbsExplosive.EFFECT_DURATION) end

    return Orb:new(orbPosition, direction, affectFn, orbOptions)
  end)
end

--- @param preset GravityOrbsExplosivePreset The preset of the explosive to get the options of
--- @return GravityOrbsExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function GravityOrbsExplosive:getPresetOptions(preset)
  if preset == "default" then
    return {
      numberOfOrbs = 7,
      orbOptions = {
        speed = 3,
        life = 25,
        radius = 7.5
      }
    }
  end

  if preset == "dense" then
    return {
      numberOfOrbs = 12,
      orbOptions = {
        speed = 3,
        life = 25,
        radius = 5
      }
    }
  end

  if preset == "energetic" then
    return {
      numberOfOrbs = 7,
      orbOptions = {
        speed = 6,
        life = 35,
        radius = 7.5
      }
    }
  end

  if preset == "lingering" then
    return {
      numberOfOrbs = 7,
      orbOptions = {
        speed = 3,
        life = 45,
        radius = 7.5
      }
    }
  end
end

--- @see Explosive#onDetonation
function GravityOrbsExplosive:onDetonation()
  self.orbs = GravityOrbsExplosive:createOrbs(self.options, self.position)
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function GravityOrbsExplosive:isDetonationFinished()
  return self.orbs and self.orbs:all(function(orb) return orb:hasFinished() end)
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function GravityOrbsExplosive:updateDetonation(dt)
  if self.orbs then
    self.orbs:forEach(function(orb) orb:update(dt) end)
  end
end

--- @see Explosive#destroy
function GravityOrbsExplosive:destroy()
  self.orbs:forEach(function(orb) orb:destroy() end)
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: GravityOrbsExplosivePreset | nil, options: GravityOrbsExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return GravityOrbsExplosive The constructed instance
function GravityOrbsExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, GravityOrbsExplosive.getPresetOptions), GravityOrbsExplosive)
  self.orbs = nil
  return self
end

return GravityOrbsExplosive
