local classes = require("lua/explosionsNG/helpers/classes")
local maths = require("lua/explosionsNG/helpers/maths")
local ExplosiveStats = require("lua/explosionsNG/explosives/base/ExplosiveStats")
local GravityOrbsExplosive = require("lua/explosionsNG/explosives/contained/gravityOrbs/GravityOrbsExplosive")

--- @class GravityOrbsExplosiveStats : ExplosiveStats<GravityOrbsExplosiveOptions>
local GravityOrbsExplosiveStats = classes.newClass(ExplosiveStats)

--- @return number The absolute energy stat
--- @see ExplosiveStats#getAbsoluteEnergyStat
function GravityOrbsExplosiveStats:getAbsoluteEnergyStat()
  return 0 -- Does not do direct damage
end

--- @return number The absolute range stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function GravityOrbsExplosiveStats:getAbsoluteRangeStat()
  local orbOptions = --[[--- @type OrbOptions ]] self.explosiveOptions.orbOptions
  return orbOptions.speed * (orbOptions.life + 1) -- +1 as the orbs get one second of "free" movement when created
end

--- @return number The absolute accuracy stat
--- @see ExplosiveStats#getAbsoluteAccuracyStat
function GravityOrbsExplosiveStats:getAbsoluteAccuracyStat()
  local mainRange = self:getAbsoluteRangeStat()
  local orbOptions = --[[--- @type OrbOptions ]] self.explosiveOptions.orbOptions

  local mainArea = math.pi * mainRange ^ 2
  -- Calculate the rough area that the circle projection of the orb will "cut" into the main area.
  -- This is approximated as the semi circle of the orb, plus the rectangle that stretches from the centre of the area
  -- to the edge of the area.
  -- This is approximate as it counts the area outside of the range by assuming the area path the orb leaves is
  -- rectangular, not rounded.
  local orbArea = ((math.pi * orbOptions.radius ^ 2) / 2) + mainRange * orbOptions.radius

  return maths.getAreaCoverageRatio(self.explosiveOptions.numberOfOrbs, mainArea, orbArea)
end

--- @return number The absolute attack rate stat
--- @see ExplosiveStats#getAbsoluteAttackRateStat
function GravityOrbsExplosiveStats:getAbsoluteAttackRateStat()
  return GravityOrbsExplosive.EFFECT_DURATION
end

--- __Constructor__
--- @param explosiveOptions GravityOrbsExplosiveOptions The explosive options to get the stats of
--- @return GravityOrbsExplosiveStats The constructed instance
function GravityOrbsExplosiveStats:new(explosiveOptions)
  local self = classes.setmetatable(ExplosiveStats:new(explosiveOptions, "coverage"), GravityOrbsExplosiveStats)
  return self
end

return GravityOrbsExplosiveStats
