local classes = require("lua/explosionsNG/helpers/classes")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local MineExplosive = require("lua/explosionsNG/explosives/composed/mine/MineExplosive")
local ProjectileVolcano = require("lua/explosionsNG/projectiles/groups/ProjectileVolcano")

--- Properties:
---   - `numberOfPieces`     - the number of mines that will be created by the explosive
---   - `ejectionMultiplier` - the multiplier of the velocity when projectiles are thrown out of this explosive
---   - `mineOptions`        - a list of `MineExplosiveOptions` to use when creating a new mine with this explosive
--- @alias LandMineExplosiveOptions { numberOfMines: number, ejectionMultiplier: number, mineOptions: MineExplosiveOptions }
--- @alias LandMineExplosivePreset "default" | "wide" | "narrow"

--- @class LandMineExplosive : Explosive<LandMineExplosivePreset, LandMineExplosiveOptions>
--- @field volcano ProjectileVolcano The volcano of mines that are thrown out
--- @field PROJECTILE_INTERVAL number __Static__ The time between firing two projectiles in succession
local LandMineExplosive = classes.newClass(Explosive)
LandMineExplosive.PROJECTILE_INTERVAL = 0.75

--- __Static__
--- @return LandMineExplosivePreset[] The list of presets for this explosive
function LandMineExplosive:getPresets()
  return { "default", "wide", "narrow" }
end

--- __Static__
---
--- Creates and returns a payload explosive that is set to detonate at the given `position`
--- @param position LuaVec3 The position at which the payload should be detonated at
--- @param options MineExplosiveOptions The list of options from the explosive that is carrying the payload
--- @return AnyExplosive The payload
function LandMineExplosive:getPayload(position, options)
  return --[[--- @type AnyExplosive ]] MineExplosive:new(position, { options = options })
end

--- __Static__
---
--- Returns a list of `ProjectileVolcanoOptions` from a list of given `LandMineExplosiveOptions`. The `interval` is set
--- from the `PROJECTILE_INTERVAL` static value.
--- @param options LandMineExplosiveOptions The options to use
--- @return ProjectileVolcanoOptions The mapped options
function LandMineExplosive:getVolcanoOptions(options)
  return {
    ejectionMultiplier = options.ejectionMultiplier,
    numberOfProjectiles = options.numberOfMines,
    interval = LandMineExplosive.PROJECTILE_INTERVAL
  }
end

--- __Static__
---
--- Creates a new `ProjectileVolcano` that is placed at the given `position`, with a producer function (see
--- `getPayload`) and a set of options that are derived from the given `options` (see `getVolcanoOptions`)
--- @param position LuaVec3 The position to set the volcano to
--- @param options LandMineExplosiveOptions The list of options of the explosive to use
--- @return ProjectileVolcano The created projectile volcano
--- @see ClusterExplosive#getPayload
function LandMineExplosive:createVolcano(position, options)
  local volcanoOptions = LandMineExplosive:getVolcanoOptions(options)
  local producer = function(projectilePosition) return LandMineExplosive:getPayload(projectilePosition, options.mineOptions) end
  return ProjectileVolcano:new(position, producer, volcanoOptions)
end

--- @param preset LandMineExplosivePreset The preset of the explosive to get the options of
--- @return LandMineExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function LandMineExplosive:getPresetOptions(preset)
  local shockwaveOptions = --[[--- @type ShockwaveOptions ]] { audioTrack = "generic_explosion" }

  if preset == "default" then
    return {
      ejectionMultiplier = 1,
      numberOfMines = 6,
      mineOptions = {
        mineLife = 20,
        mineRadius = 30,
        shockwaveExplosiveOptions = {
          time = 1,
          maxRadius = 30,
          forceMultiplier = 6,
          shockwaveOptions = shockwaveOptions
        }
      }
    }
  end

  if preset == "narrow" then
    return {
      ejectionMultiplier = 0.75,
      numberOfMines = 6,
      mineOptions = {
        mineLife = 15,
        mineRadius = 20,
        shockwaveExplosiveOptions = {
          time = 0.5,
          maxRadius = 20,
          forceMultiplier = 7,
          shockwaveOptions = shockwaveOptions
        }
      }
    }
  end

  if preset == "wide" then
    return {
      ejectionMultiplier = 2.5,
      numberOfMines = 9,
      mineOptions = {
        mineLife = 30,
        mineRadius = 80,
        shockwaveExplosiveOptions = {
          time = 2,
          maxRadius = 80,
          forceMultiplier = 7,
          shockwaveOptions = shockwaveOptions
        }
      }
    }
  end
end

--- @see Explosive#onDetonation
function LandMineExplosive:onDetonation()
  self.volcano:erupt()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function LandMineExplosive:isDetonationFinished()
  return self.volcano:isFinished()
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function LandMineExplosive:updateDetonation(dt)
  self.volcano:update(dt)
end

--- @see Explosive#destroy
function LandMineExplosive:destroy()
  self.volcano:destroy()
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: LandMineExplosivePreset | nil, options: LandMineExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return LandMineExplosive The constructed instance
function LandMineExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, LandMineExplosive.getPresetOptions), LandMineExplosive)
  self.volcano = LandMineExplosive:createVolcano(position, self.options)
  return self
end

return LandMineExplosive
