local classes = require("lua/explosionsNG/helpers/classes")
local collisions = require("lua/explosionsNG/helpers/collisions")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local LightningExplosive = require("lua/explosionsNG/explosives/composed/lightning/LightningExplosive")
local Array = require("lua/explosionsNG/misc/Array")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")
local Timer = require("lua/explosionsNG/timers/Timer")

--- Properties:
---   - `attackDelay` - The time (in seconds) between two lightning bolts
---   - `life` - The life time (in seconds) of the explosive
---   - `radius` - The maximum distance from the centre position that a lightning bolt could be placed
--- @alias LightningStormExplosiveOptions { attackDelay: number, life: number, radius: number }
--- @alias LightningStormExplosivePreset "default" | "focused" | "persistent"

--- @class LightningStormExplosive : Explosive<LightningStormExplosivePreset, LightningStormExplosiveOptions>
--- @field private boltTimer IntervalTimer The interval which will create a new lightning bolt on every cycle
--- @field private lifeTimer Timer The time left for this explosive
--- @field private lightningBolts Array<LightningExplosive> The list of active lightning bolts
--- @field public LIGHTING_PRESET LightningExplosivePreset The preset of the lighting to use for the lightning bolts
local LightningStormExplosive = classes.newClass(Explosive)
LightningStormExplosive.LIGHTING_PRESET = "default"

--- __Static__
--- @return LightningStormExplosivePreset[] The list of presets for this explosive
function LightningStormExplosive:getPresets()
  return { "default", "focused", "persistent" }
end

--- @param preset LightningStormExplosivePreset The preset of the explosive to get the options of
--- @return LightningStormExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function LightningStormExplosive:getPresetOptions(preset)
  if preset ==    "default" then return { attackDelay = 3  , life = 25, radius = 25 } end
  if preset ==   "focused" then return { attackDelay = 1.5, life = 25, radius = 40 } end
  if preset == "persistent" then return { attackDelay = 3  , life = 40, radius = 60 } end
end

--- @see Explosive#onDetonation
function LightningStormExplosive:onDetonation()
  self:createNewLightningBolt()
  self.lifeTimer:start()
  self.boltTimer:start()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function LightningStormExplosive:isDetonationFinished()
  return self.lifeTimer:isComplete() and self.lightningBolts:all(LightningExplosive.isDetonationFinished)
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function LightningStormExplosive:updateDetonation(dt)
  self.boltTimer:update(dt)
  self.lifeTimer:update(dt)
  self.lightningBolts:forEach(function(bolt) bolt:update(dt) end)
end

--- @see Explosive#destroy
function LightningStormExplosive:destroy()
  self.lightningBolts:forEach(function(bolt) bolt:deactivate() end)
  self.lifeTimer:complete()
  self.boltTimer:stop()
end

--- Creates a new `LightningExplosive` that is within range of this explosive. This method will then activate it and add
--- it to the internal array.
--- @private
function LightningStormExplosive:createNewLightningBolt()
  local boltPosition = collisions.getRandomPointInRange(self.position, self.options.radius)
  local lightning = LightningExplosive:new(boltPosition, { preset = LightningStormExplosive.LIGHTING_PRESET })
  lightning:activate()
  self.lightningBolts:add(lightning)
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: LightningExplosivePreset | nil, options: LightningStormExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return LightningStormExplosive The constructed instance
function LightningStormExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, LightningStormExplosive.getPresetOptions), LightningStormExplosive)
  self.boltTimer = IntervalTimer:new(self.options.attackDelay, function() self:createNewLightningBolt() end)
  self.lifeTimer = Timer:new(self.options.life, function() self.boltTimer:stop() end)
  self.lightningBolts = Array:new(--[[--- @type LightningExplosive[] ]] {})
  return self
end

return LightningStormExplosive
