local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local effectFactory = require("lua/explosionsNG/vehicles/effects/factory")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Pulsar = require("lua/explosionsNG/worldObjects/Pulsar")
local Laser = require("lua/explosionsNG/worldObjects/laser/Laser")

--- Properties:
---  - `forceMultiplier` - the multiplier of force to apply to each shockwave
---  - `pulsarOptions` - a list of options to use for the pulsar object
---  - `laserOptions` - a list of options to use for the laser object. Note that some options will be overwritten.
--- @alias OrbitalStrikeExplosiveOptions { forceMultiplier: number, pulsarOptions: PulsarOptions, laserOptions: LaserOptions | nil }
--- @alias OrbitalStrikeExplosivePreset "default" | "focused" | "unfocused" | "imploded"

--- @class OrbitalStrikeExplosive : Explosive<OrbitalStrikeExplosivePreset, OrbitalStrikeExplosiveOptions>
--- @field private pulsar Pulsar The pulsar object created the shockwaves
--- @field private laser Laser The laser object providing visual effects
--- @field private finished boolean Whether the explosive has finished
--- @field private LASER_RADIUS_MULTIPLIER number The multiplier to apply when converting the maximum radius of the pulsar's shockwave, to the radius of the laser it should be associated with
local OrbitalStrikeExplosive = classes.newClass(Explosive)
OrbitalStrikeExplosive.LASER_RADIUS_MULTIPLIER = 0.2

--- __Static__
--- @return OrbitalStrikeExplosivePreset[] The list of presets for this explosive
function OrbitalStrikeExplosive:getPresets()
  return { "default", "focused", "unfocused", "imploded" }
end

--- __Static__
---
--- Returns a list of options to use for a `Pulsar` object, given its corresponding explosive options.
--- @param options OrbitalStrikeExplosiveOptions The list of options for the explosive
--- @return PulsarOptions The pulsar options to use for the corresponding explosive
--- @private
function OrbitalStrikeExplosive:applyPulsarOptionDefaults(options)
  local pulsarOptions = options.pulsarOptions
  --- @type ShockwaveOptions
  local shockwaveOptions = pulsarOptions.shockwaveOptions or {}

  return tables.merge(pulsarOptions, {
    affectFn = function(id, event)
      return effectFactory.createForceEffect(id, event.position, pulsarOptions.maxRadius, options.forceMultiplier, shockwaveOptions.implode)
    end
  })
end

--- __Static__
---
--- Returns a list of options to use for a `Laser` object, given its corresponding explosive options.
--- @param options OrbitalStrikeExplosiveOptions The list of options for the explosive
--- @return PulsarOptions The laser options to use for the corresponding explosive
--- @private
function OrbitalStrikeExplosive:applyLaserOptionOverwrites(options)
  return tables.merge(options.laserOptions or {}, {
    pulseTime = options.pulsarOptions.interval,
    radius = options.pulsarOptions.maxRadius * OrbitalStrikeExplosive.LASER_RADIUS_MULTIPLIER
  })
end

--- @param preset OrbitalStrikeExplosivePreset The preset of the explosive to get the options of
--- @return OrbitalStrikeExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function OrbitalStrikeExplosive:getPresetOptions(preset)
  if preset == "default" then
    return {
      forceMultiplier = 5,
      pulsarOptions = {
        numberOfShockwaves = 5,
        interval = 2,
        maxRadius = 50,
        time = 2
      }
    }
  end

  if preset == "unfocused" then
    return {
      forceMultiplier = 5,
      pulsarOptions = {
        numberOfShockwaves = 8,
        interval = 3,
        maxRadius = 125,
        time = 4
      }
    }
  end

  if preset == "focused" then
    return {
      forceMultiplier = 5,
      pulsarOptions = {
        numberOfShockwaves = 10,
        interval = 1,
        maxRadius = 50,
        time = 1.5
      }
    }
  end

  if preset == "imploded" then
    return {
      forceMultiplier = 5,
      pulsarOptions = {
        numberOfShockwaves = 5,
        interval = 2,
        maxRadius = 50,
        time = 2,
        shockwaveOptions = {
          implode = true
        }
      }
    }
  end
end

--- @see Explosive#onDetonation
function OrbitalStrikeExplosive:onDetonation()
  local pulsarOptions = OrbitalStrikeExplosive:applyPulsarOptionDefaults(self.options)
  self.pulsar = Pulsar:new(self.position, pulsarOptions)

  --- @type OrbitalStrikeExplosiveOptions
  local options = tables.merge(self.options, { pulsarOptions = pulsarOptions }) -- Add the updated values
  local laserOptions = OrbitalStrikeExplosive:applyLaserOptionOverwrites(options)

  local laser = Laser:new(self.position, laserOptions)
  local events = laser:getEvents()
  events:registerEventHandler("opened", function() self.pulsar:begin() end)
  events:registerEventHandler("closed", function() self.finished = true end)

  self.laser = laser
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function OrbitalStrikeExplosive:isDetonationFinished()
  return self.finished
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function OrbitalStrikeExplosive:updateDetonation(dt)
  if self.pulsar then
    self.pulsar:update(dt)
    self:checkAndCloseLaser()
  end

  if self.laser then
    self.laser:update(dt)
  end
end

--- Checks whether the `pulsar` has finished and if so, closes the laser and deletes the `pulsar` instance
--- @private
function OrbitalStrikeExplosive:checkAndCloseLaser()
  if self.pulsar:isFinished() then
    self.laser:close()
    self.pulsar = nil
  end
end

--- @see Explosive#destroy
function OrbitalStrikeExplosive:destroy()
  self.pulsar:destroy()
  self.laser:destroy()
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: OrbitalStrikeExplosivePreset | nil, options: OrbitalStrikeExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return OrbitalStrikeExplosive The constructed instance
function OrbitalStrikeExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, OrbitalStrikeExplosive.getPresetOptions), OrbitalStrikeExplosive)
  self.pulsar = nil
  self.laser = nil
  self.finished = false
  return self
end

return OrbitalStrikeExplosive
