local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local VehicleMultiEffect = require("lua/explosionsNG/vehicles/effects/VehicleMultiEffect")
local effectFactory = require("lua/explosionsNG/vehicles/effects/factory")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Shockwave = require("lua/explosionsNG/worldObjects/Shockwave")

--- @alias ShockwaveVehicleAffectedFn fun(vehicle: Vehicle): void
--- @alias ShockwaveAdditionalEffectFn fun(id: string, position: LuaVec3, event: VehicleEffectEvent): Array<VehicleEffect>
--- Properties:
---  - `time` - the time (in s) the shockwave exists for
---  - `maxRadius` - the maximum radius (in m) the shockwave can become
---  - `forceMultiplier` - the multiplier of maximum force the shockwave can inflict
---  - `additionalEffectsFn` - a provider function to apply additional effects to vehicles, in addition to the force effect
---  - `onVehicleAffected` a function that will be called when a vehicle is affected by this shockwave
---  - `shockwaveOptions` - a list of options to use for the shockwave
--- @alias ShockwaveExplosiveOptions { time: number, maxRadius: number, forceMultiplier: number, additionalEffectsFn: ShockwaveAdditionalEffectFn | nil, onVehicleAffected: ShockwaveVehicleAffectedFn | nil, shockwaveOptions: ShockwaveOptions | nil }
--- @alias ShockwaveExplosivePreset "default" | "implode" | "range" | "compact"

--- @class ShockwaveExplosive : Explosive<ShockwaveExplosivePreset, ShockwaveExplosiveOptions>
--- @field private shockwave Shockwave The shockwave object that this explosive wraps
local ShockwaveExplosive = classes.newClass(Explosive)

--- __Static__
--- @return ShockwaveExplosivePreset[] The list of presets for this explosive
function ShockwaveExplosive:getPresets()
  return { "default", "implode", "range", "compact" }
end

--- __Static__
---
--- Returns a new `VehicularEffect` for a given shockwave to use
--- @param id string The id of the shockwave
--- @param event VehicleEffectEvent The shockwave effect event
--- @param position LuaVec3 The position of the shockwave's centre
--- @param options ShockwaveExplosiveOptions The list of options of the shockwave
--- @return VehicularEffect The vehicle effect(s)
--- @private
function ShockwaveExplosive:getEffect(id, event, options, position)
  local implode = options.shockwaveOptions and (--[[--- @not nil ]] options.shockwaveOptions).implode
  local forceEffect = effectFactory.createForceEffect(id, event.position, options.maxRadius, options.forceMultiplier, implode)

  local additionalEffectsFn = options.additionalEffectsFn
  if not additionalEffectsFn then return forceEffect end

  local effects = (--[[--- @not nil ]] additionalEffectsFn)(id, position, event)
  effects:add(forceEffect)
  return VehicleMultiEffect:new(id, effects)
end

--- __Static__
---
--- Creates and returns a shockwave object with the given `position` and `options`
--- @param position LuaVec3 The position to set the shockwave to
--- @param options ShockwaveExplosiveOptions The list of options use
--- @return Shockwave The shockwave created
--- @private
function ShockwaveExplosive:createShockwave(position, options)
  --- @type VehicleEffectProducer
  local affectFn = function(id, event)
    local onVehicleAffected = options.onVehicleAffected
    if onVehicleAffected then onVehicleAffected(event.vehicle) end
    return ShockwaveExplosive:getEffect(id, event, options, position)
  end

  return Shockwave:new(position, options.maxRadius, options.time, affectFn, options.shockwaveOptions)
end

--- @param preset ShockwaveExplosivePreset The preset of the explosive to get the options of
--- @return ShockwaveExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function ShockwaveExplosive:getPresetOptions(preset)
  --- @type ShockwaveOptions
  local baseShockwaveOptions = { audioTrack = "generic_explosion" }

  if preset == "default" then
    return { time = 1.5, maxRadius = 50, forceMultiplier = 1, shockwaveOptions = baseShockwaveOptions }
  end

  if preset == "implode" then
    return { time = 1.5, maxRadius = 50, forceMultiplier = 1, shockwaveOptions = tables.merge(baseShockwaveOptions, { implode = true }) }
  end

  if preset == "range" then
    return { time = 2, maxRadius = 100, forceMultiplier = 1, shockwaveOptions = baseShockwaveOptions }
  end

  if preset == "compact" then
    return { time = 1, maxRadius = 25, forceMultiplier = 2, shockwaveOptions = baseShockwaveOptions }
  end
end

--- Activates the shockwave
--- @see Explosive#onDetonation
function ShockwaveExplosive:onDetonation()
  self.shockwave:activate()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function ShockwaveExplosive:isDetonationFinished()
  return self.shockwave:isFinished()
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function ShockwaveExplosive:updateDetonation(dt)
  self.shockwave:update(dt)
end

--- @see Explosive#destroy
function ShockwaveExplosive:destroy()
  self.shockwave:destroy()
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: ShockwaveExplosivePreset | nil, options: ShockwaveExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return ShockwaveExplosive The constructed instance
function ShockwaveExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, ShockwaveExplosive.getPresetOptions), ShockwaveExplosive)
  self.shockwave = ShockwaveExplosive:createShockwave(position, self.options)
  return self
end

return ShockwaveExplosive
