local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local Array = require("lua/explosionsNG/misc/Array")
local Vehicle = require("lua/explosionsNG/vehicles/Vehicle")
local VehicleEffect = require("lua/explosionsNG/vehicles/effects/VehicleEffect")
local VehicleMultiEffect = require("lua/explosionsNG/vehicles/effects/VehicleMultiEffect")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Star = require("lua/explosionsNG/worldObjects/celestials/star/Star")
local Shockwave = require("lua/explosionsNG/worldObjects/Shockwave")
local FiniteLight = require("lua/explosionsNG/worldObjects/environment/FiniteLight")
local BinaryStarSystem = require("lua/explosionsNG/worldObjects/celestials/BinaryStarSystem")
local ShockwaveExplosive = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosive")

--- Properties:
---  - `starRadius` - the radius (in m) of the star(s) created
---  - `isBinary`- whether to create a binary star system (`true`), or just a single star (`false`)
---  - `shockwaveOptions` - the list of options to use for the explosive shockwave. Note that the `maxRadius` will be
---                         used as the max radius for the clouds as well
--- @alias SupernovaExplosiveOptions { starRadius: number, isBinary: boolean, shockwaveExplosiveOptions: ShockwaveExplosiveOptions }
--- @alias SupernovaExplosivePreset "typeII" | "typeIa"

--- @class SupernovaExplosive : Explosive<SupernovaExplosivePreset, SupernovaExplosiveOptions>
--- @field private celestial Celestial<any> The celestial used during the non-exploding part of the explosive
--- @field private explosiveShockwave ShockwaveExplosive The shockwave explosion
--- @field private clouds Shockwave The stellar clouds
--- @field private light FiniteLight The stellar clouds' light
--- @field private MAX_VEHICLE_EFFECT_COOLDOWN number __Static__ The maximum delay of a vehicle being effected by the supernova
--- @field private MAX_LIGHT_BRIGHTNESS number __Static__ The maximum brightness of the light, after the supernova has occurred
--- @field private CLOUD_MAX_ALPHA number __Static__ The maximum alpha to use on the supernova clouds
--- @field private CLOUD_LIFETIME number __Static__ The time the supernova clouds exist for
--- @field private TIME_FOR_CLOUDS_TO_REACH_FULL_RANGE number __Static__ The time it takes for the supernova clouds to reach the full range of the instance
--- @field private STAR_RADIUS_TO_SYSTEM_MULTIPLIER number __Static__ The radius of the star system, compared to the size of the stars it contains
--- @field private SHOCKWAVE_EXPLOSIVE_COLOUR string __Static__ The colour of the shockwave explosive
--- @field private SINGLE_STAR_COLOUR string __Static__ The colour of the single star
--- @field private SYSTEM_STAR_COLOUR string __Static__ The colour of each star in the star system
local SupernovaExplosive = classes.newClass(Explosive)
SupernovaExplosive.MAX_VEHICLE_EFFECT_COOLDOWN = 0.5
SupernovaExplosive.MAX_LIGHT_BRIGHTNESS = 10
SupernovaExplosive.CLOUD_MAX_ALPHA = 0.6
SupernovaExplosive.CLOUD_LIFETIME = 30
SupernovaExplosive.TIME_FOR_CLOUDS_TO_REACH_FULL_RANGE = 5
SupernovaExplosive.STAR_RADIUS_TO_SYSTEM_MULTIPLIER = 5
SupernovaExplosive.SHOCKWAVE_EXPLOSIVE_COLOUR = "1 0 0"
SupernovaExplosive.SINGLE_STAR_COLOUR = "0 0.25 1"
SupernovaExplosive.SYSTEM_STAR_COLOUR = "1 1 1"

--- __Static__
--- @return SupernovaExplosivePreset[] The list of presets for this explosive
function SupernovaExplosive:getPresets()
  return { "typeII", "typeIa" }
end

--- __Static__
---
--- Returns the bounding box that makes up this explosive. In particular, this will be bounds of the star object
--- if it is a single star object going supernova, else it will be the bounds of the binary star system.
--- @param preset SupernovaExplosivePreset The preset of the explosive to get the bounding area of
--- @return BoundingArea The area of the objects that make up this explosive
function SupernovaExplosive:getBoundingArea(preset)
  local options = SupernovaExplosive:getPresetOptions(preset)
  local starRadius = options.starRadius

  if not options.isBinary then return starRadius end

  local startOrbitRadius = starRadius * SupernovaExplosive.STAR_RADIUS_TO_SYSTEM_MULTIPLIER
  -- Set offset to opposite sides
  return {
    { radius = starRadius, offset = vec3( startOrbitRadius, 0, 0) },
    { radius = starRadius, offset = vec3(-startOrbitRadius, 0, 0) },
  }
end

--- __Static__
--- @return ShockwaveOptions Returns the list of base options to use for the explosive shockwave
function SupernovaExplosive:getBaseShockwaveOptions()
  return {
    modelName = "supernova",
    maxAlpha = SupernovaExplosive.CLOUD_MAX_ALPHA,
    expansionFn = function(time)
      local ratio = SupernovaExplosive.TIME_FOR_CLOUDS_TO_REACH_FULL_RANGE / SupernovaExplosive.CLOUD_LIFETIME
      return 1 - math.exp(-time * ratio * SupernovaExplosive.CLOUD_LIFETIME)
    end
  }
end

--- __Static__
---
--- Creates and returns a new shockwave to use for the stellar clouds
--- @param position LuaVec3 The centre position of the clouds
--- @param maxRadius number The maximum radius the clouds can reach
--- @return Shockwave The created shockwave clouds
function SupernovaExplosive:createCloudShockwave(position, maxRadius)
  --- @type VehicleEffectProducer
  local affectFn = function(id, event)
    return SupernovaExplosive:createVehicleAffectFn(id, event.position, position, maxRadius)
  end

  local options = SupernovaExplosive:getBaseShockwaveOptions()
  return Shockwave:new(position, maxRadius, SupernovaExplosive.CLOUD_LIFETIME, affectFn, options)
end

--- __Static__
---
--- Creates and returns a new `VehicleMultiEffect` that applies two effects to a vehicle:
---  - A fire effect that ignites a node on the vehicle
---  - A dissolve effect that breaks a break group on the vehicle
--- Both effects are a timed effect, with the period of the effect being proportional to how close the affect position
--- and the supernova position are to each other.
--- @param id string The id of the explosive who created this effect
--- @param affectPosition LuaVec3 The position at which a vehicle is being affected
--- @param supernovaPosition LuaVec3 The position of the supernova explosive
--- @param supernovaMaxRadius number The maximum range of the supernova explosive
--- @return VehicleMultiEffect The create vehicle effect
function SupernovaExplosive:createVehicleAffectFn(id, affectPosition, supernovaPosition, supernovaMaxRadius)
  local distance = supernovaPosition:distance(affectPosition)
  local normDistance = distance / supernovaMaxRadius
  local cooldownPeriod = normDistance * SupernovaExplosive.MAX_VEHICLE_EFFECT_COOLDOWN;

  local fireEffect = VehicleEffect:newSingleTimedEffect(id, Vehicle.igniteRandomNode, cooldownPeriod)
  local dissolveEffect = VehicleEffect:newSingleTimedEffect(id, Vehicle.breakRandomBreakGroup, cooldownPeriod)
  return VehicleMultiEffect:new(id, Array:new({ fireEffect, dissolveEffect }))
end

--- __Static__
---
--- Returns a list of `ShockwaveExplosiveOptions`. These option will contain all of the shockwave options from the
--- given options, except with the colour and audio track overridden to specific values.
--- Notes this does not effect the given list of options.
--- @param options SupernovaExplosiveOptions The supernova options
--- @return ShockwaveExplosiveOptions The overridden options for the shockwave explosive
function SupernovaExplosive:overrideShockwaveOptions(options)
  local explosiveOptions = tables.copy(options.shockwaveExplosiveOptions)

  local shockwaveOptions = --[[--- @type ShockwaveOptions ]] explosiveOptions.shockwaveOptions or {}
  shockwaveOptions.colour = SupernovaExplosive.SHOCKWAVE_EXPLOSIVE_COLOUR
  shockwaveOptions.audioTrack = "supernova_explosion"
  explosiveOptions.shockwaveOptions = shockwaveOptions

  return explosiveOptions
end

--- @param preset SupernovaExplosivePreset The preset of the explosive to get the options of
--- @return SupernovaExplosiveOptions The options of the explosive
--- @see Explosive#getPresetOptions
function SupernovaExplosive:getPresetOptions(preset)
  if preset == "typeII" then
    return {
      starRadius = 25,
      isBinary = false,
      shockwaveExplosiveOptions = {
        forceMultiplier = 25,
        maxRadius = 1000,
        time = 2
      }
    }
  end

  return {
    starRadius = 15,
    isBinary = true,
    shockwaveExplosiveOptions = {
      forceMultiplier = 25,
      maxRadius = 1500,
      time = 3
    }
  }
end

--- @see Explosive#onDetonation
function SupernovaExplosive:onDetonation()
  self.celestial = self.options.isBinary and self:createBinarySystem() or self:createSingleStar()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function SupernovaExplosive:isDetonationFinished()
  local isLightExpired = self.light and self.light:isFinished()
  local isCelestialDead = self.celestial and self.celestial:isDead()
  local areCloudsFinished = self.clouds and self.clouds:isFinished()
  local isExplosiveShockwaveFinished = self.explosiveShockwave and self.explosiveShockwave:isDetonationFinished()
  return isLightExpired and isCelestialDead and areCloudsFinished and isExplosiveShockwaveFinished
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function SupernovaExplosive:updateDetonation(dt)
  if self.celestial then self.celestial:update(dt) end
  if self.light then self.light:update(dt) end
  if self.clouds then self.clouds:update(dt) end
  if self.explosiveShockwave then self.explosiveShockwave:update(dt) end
end

--- Creates a new binary star system from the options and position of this instance
--- @return BinaryStarSystem The created star system
--- @private
function SupernovaExplosive:createBinarySystem()
  local starRadius = self.options.starRadius
  local systemRadius = starRadius * SupernovaExplosive.STAR_RADIUS_TO_SYSTEM_MULTIPLIER

  local starSystem = BinaryStarSystem:new(self.position, systemRadius, starRadius, SupernovaExplosive.SYSTEM_STAR_COLOUR)
  local events = starSystem:getEvents()
  events:registerEventHandler("formed", function() starSystem:collapse() end)
  events:registerEventHandler("collided", function() self:onCelestialDeath() end)
  return starSystem
end

--- Creates a new star from the options and position of this instance
--- @return Star The created star
--- @private
function SupernovaExplosive:createSingleStar()
  local star = Star:new(self.position, self.options.starRadius, SupernovaExplosive.SINGLE_STAR_COLOUR)
  local events = star:getEvents()
  events:registerEventHandler("formed", function() star:collapse() end)
  events:registerEventHandler("compressed", function() self:onCelestialDeath() end)
  return star
end

--- Creates the main explosive shockwave and the cloud shockwaves, and begins the activation of both
--- @private
function SupernovaExplosive:onCelestialDeath()
  local shockwaveExplosiveOptions = SupernovaExplosive:overrideShockwaveOptions(self.options)

  self.explosiveShockwave = ShockwaveExplosive:new(self.position, { options = shockwaveExplosiveOptions })
  self.explosiveShockwave:activate()

  self.clouds = SupernovaExplosive:createCloudShockwave(self.position, shockwaveExplosiveOptions.maxRadius)
  self.clouds:activate()

  self.light = FiniteLight:new(SupernovaExplosive.CLOUD_LIFETIME, {
    position = self.position,
    radius = shockwaveExplosiveOptions.maxRadius,
    brightness = SupernovaExplosive.MAX_LIGHT_BRIGHTNESS
  })
end

--- @see Explosive#destroy
function SupernovaExplosive:destroy()
  if self.celestial then
    self.celestial:destroy()
  end

  if self.explosiveShockwave then
    self.explosiveShockwave:deactivate()
  end

  if self.clouds then
    self.clouds:destroy()
  end

  if self.light then
    self.light:destroy()
  end
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: SupernovaExplosivePreset | nil, options: SupernovaExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return SupernovaExplosive The constructed instance
function SupernovaExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, SupernovaExplosive.getPresetOptions), SupernovaExplosive)
  self.celestial = nil
  self.explosiveShockwave = nil
  self.clouds = nil
  self.light = nil
  return self
end

return SupernovaExplosive
