local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local collisions = require("lua/explosionsNG/helpers/collisions")
local positioning = require("lua/explosionsNG/helpers/positioning")
local audio = require("lua/explosionsNG/helpers/audio")
local Array = require("lua/explosionsNG/misc/Array")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Target = require("lua/explosionsNG/worldObjects/Target")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")
local CircularPath = require("lua/explosionsNG/pathing/CircularPath")

--- @alias ExplosiveAreaSoundEffectKey "firstTarget" | "firstDetonation"
--- Properties (based on keys from `ExplosiveAreaSoundEffectKey`):
---  - `firstTarget` - A sound to play when the first target is placed
---  - `firstDetonation` - A sound to play when the first explosive is detonated
--- @alias ExplosiveAreaSoundEffectGenericOptions<T> { firstTarget: T, firstDetonation: T }
--- @alias ExplosiveAreaSoundEffectOptions ExplosiveAreaSoundEffectGenericOptions<string | nil>
--- @alias ExplosiveAreaPlacement "followPlayer" | "ring"

--- Properties:
---  - `areaRadius` - The circular radius of the area covered. This option should not be used with the `"followPlayer"` `placement` option - either or should be given.
---  - `explosiveRadius` - The radius of effect of an individual explosive
---  - `numberOfExplosives` - The number of explosives in the area
---  - `interval` - The delay between two targets appearing
---  - `targetOptions` - The list of options to use for each target
---  - `placement` - The placement pattern to use when placing explosives. If this is set to `"followPlayer"`, then this should be given instead of `areaRadius`, and will explicitly overwrite that property. If this is not given, it is expected that the `areaRadius` will be given.
---  - `soundEffects` - A list of sound effects to play
--- @alias ExplosiveAreaOptions { areaRadius: number | nil, explosiveRadius: number | nil, numberOfExplosives: number | nil, interval: number | nil, targetOptions: TargetOptions | nil, placement: ExplosiveAreaPlacement | nil, soundEffects: ExplosiveAreaSoundEffectOptions | nil }
--- @alias ExplosiveAreaPreset nil @There are no set presets for this explosive

--- @class ExplosiveArea<T : AnyExplosive> : Explosive<ExplosiveAreaPreset, ExplosiveAreaOptions>
--- @field private explosiveProducer fun(position: LuaVec3): T A function that producers a new explosive, placed at the given position
--- @field private explosives Array<T> The list of active explosives
--- @field private targets Array<Target> The list of active targets
--- @field private interval IntervalTimer An interval which will create a new target when it completes. This will be `nil` if there is no delay between two targets being placed.
--- @field private hasPlayedAudio ExplosiveAreaSoundEffectGenericOptions<boolean> Whether audio has been played for each possible event/key
--- @field private awaitingPlayerVehicle boolean Whether this pattern is waiting for the player to be in a vehicle
--- @field private ringPath CircularPath | nil A circular path to generate new ring area patterns. This is lazy instantiated and will not be used if the options do not specify the `"ring"` placement.
--- @field private isDestroyed boolean Whether the instance has been destroyed. This will only be `true` with external influence to the instance.
--- @field private PLAYER_POSITIONAL_ACCURACY number __Static__ How accurate to the player the positioning of new explosives will be
--- @field private PREDICTION_TIME number __Static__ How far in advanced the prediction system will be. In other words, it is how far in advance the prediction system will predict whether the player will be
local ExplosiveArea = classes.newClass(Explosive)
ExplosiveArea.PLAYER_POSITIONAL_ACCURACY = 50
ExplosiveArea.PREDICTION_TIME = 2

--- @return ExplosiveAreaOptions The options of the explosive
function ExplosiveArea:getPresetOptions()
  error("There are no set presets for this explosive")
  return {}
end

--- Returns a set of options to use for a target that will be placed at the given position
--- @param targetPosition LuaVec3 The position of the target these options are for
--- @return TargetOptions The options
function ExplosiveArea:getTargetOptions(targetPosition)
  local options = tables.copy(self.options.targetOptions or {})
  options.showMarker = options.showMarker or false
  options.onFinish = function() self:onTargetFinish(targetPosition) end
  return options
end

--- @see Explosive#onDetonation
function ExplosiveArea:onDetonation()
  if self.interval then
    self.interval:start()
    return
  end

  self:checkPlayerAvailability(function(player)
    -- If there is no interval, create all targets in one go
    Array:iterate(self.options.numberOfExplosives, function() self:placeTarget(player) end)
  end)
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function ExplosiveArea:isDetonationFinished()
  local hasAllExplosives = self.explosives:length() == self.options.numberOfExplosives
  local allExplosivesFinished = self.explosives:all(function(explosive) return explosive:isDetonationFinished() end)
  return self.isDestroyed or hasAllExplosives and allExplosivesFinished
end

--- Updates the targets and the explosive
--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function ExplosiveArea:updateDetonation(dt)
  if self.awaitingPlayerVehicle then return end

  self.targets:forEach(function(target) target:update(dt) end)
  self.explosives:forEach(function(explosive) explosive:update(dt) end)

  -- Update interval while there are still targets to be placed
  if self.interval and self.targets:length() < self.options.numberOfExplosives then
    self.interval:update(dt)
  end
end

--- Called when a target is finished displaying.
--- This method creates a new explosive and activates it. This method will also attempt to play an audio sound for the
--- first activated explosive, if it is the first explosive to be activated.
--- @param targetPosition LuaVec3 The target's position
--- @private
function ExplosiveArea:onTargetFinish(targetPosition)
  local explosive = self.explosiveProducer(targetPosition)
  explosive:activate()
  self.explosives:add(explosive)
  
  self:tryAndPlayAudio("firstDetonation", targetPosition)
end

--- Returns a new position for a target. If the options in this instance specify to follow the player, then the returned
--- position will return a point within a random range of the player. If not, the returned position will be a point
--- within a random range of the centre position of this explosive.
--- @param player Vehicle The player vehicle
--- @return LuaVec3 The new target position
--- @private
function ExplosiveArea:getNewTargetPosition(player)
  local placement = self.options.placement

  if not placement then
    return collisions.getRandomPointInRange(self.position, self.options.areaRadius)
  end

  if placement == "followPlayer" then
    return positioning.predictPosition(player, ExplosiveArea.PREDICTION_TIME, ExplosiveArea.PLAYER_POSITIONAL_ACCURACY)
  end

  return self:nextRingPath()
end

--- Returns the next position from the internal `ringPath`, before updating the instance again for the next call.
--- This method will also create this path, should it not exist beforehand.
--- @return LuaVec3 The next ring position
--- @private
function ExplosiveArea:nextRingPath()
  -- Lazy instantiated
  if not self.ringPath then
    self.ringPath = CircularPath:new(self.position, self.options.areaRadius)
  end

  local castedPath = --[[--- @not nil ]] self.ringPath
  local position = castedPath:getCurrentPosition()

  -- Update for next explosive
  local rotationalIncrement = (2 * math.pi) / self.options.numberOfExplosives
  castedPath:addRotation(rotationalIncrement)

  return position
end

--- Creates and new target and places it at a random position. See `getNewTargetPosition` for details on the position.
--- This method will also attempt to play an audio sound for the first target, if it is the first target to be placed.
--- @param player Vehicle The player vehicle
--- @private
function ExplosiveArea:placeTarget(player)
  local targetPosition = self:getNewTargetPosition(player)

  local options = self:getTargetOptions(targetPosition)
  local target = Target:new(targetPosition, self.options.explosiveRadius, options)
  self.targets:add(target)

  self:tryAndPlayAudio("firstTarget", targetPosition)
end

--- Checks whether a sound effect was set within the options that matches the given key, and that the audio has not yet
--- been played before. If this is the case, then this method will play the sound effect and set the relevant flag.
--- @param key ExplosiveAreaSoundEffectKey The sound effect key to consider
--- @param position LuaVec3 The position to use as the sound effects source
--- @private
function ExplosiveArea:tryAndPlayAudio(key, position)
  if self.hasPlayedAudio[key] or not self.options.soundEffects then return end

  local soundToPlay = (--[[--- @not nil ]] self.options.soundEffects)[key]
  if not soundToPlay then return end

  audio.playOnce(soundToPlay, { position = position })
  self.hasPlayedAudio[key] = true
end

--- The interval timer has elapsed and emitted.
--- Places a new target at a random position.
--- @private
function ExplosiveArea:onIntervalEnded()
  self:checkPlayerAvailability(function(player) self:placeTarget(player) end)
end

--- Checks whether the player is in a vehicle on the vehicles' handler. If it is, this method will immediately call the
--- given callback function. If not, this method will wait for the player vehicle to be present. before later calling
--- this callback with the newly spawned player vehicle.
--- @param callback VehicleEventFn The callback function
--- @private
function ExplosiveArea:checkPlayerAvailability(callback)
  local player = vehiclesHandler.getPlayerVehicle()

  if not player then
    self:waitForPlayer(callback)
    return
  end

  callback(player)
end

--- Registers an internal callback with the vehicles' handler, which will fire when the player vehicle is spawned. This
--- callback will then call the given callback function with the player vehicle.
--- @param callback VehicleEventFn The callback function
--- @private
function ExplosiveArea:waitForPlayer(callback)
  self.awaitingPlayerVehicle = true

  vehiclesHandler.registerHandler("playerSpawned", function(playerVehicle)
    callback(playerVehicle)
    self.awaitingPlayerVehicle = false
  end)
end

--- @see Explosive#destroy
function ExplosiveArea:destroy()
  self.targets:forEach(function(target) target:destroy() end)
  self.explosives:forEach(function(explosive) explosive:deactivate() end)
  self.interval:stop()
  self.isDestroyed = true
end

--- __Constructor__
--- @generic S : AnyExplosive
--- @param position LuaVec3 The centre of the area
--- @param explosiveProducer fun(position: LuaVec3): S A function that producers a new explosive, placed at the given position
--- @param configuration { preset: ExplosiveAreaPreset | nil, options: ExplosiveAreaOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return ExplosiveArea<S> The constructed instance
function ExplosiveArea:new(position, explosiveProducer, configuration)
  local self = --[[--- @type ExplosiveArea<S> ]] classes.setmetatable(
      --[[--- @type ExplosiveArea<S> ]] Explosive:new(position, configuration, ExplosiveArea.getPresetOptions),
      ExplosiveArea
  )
  self.explosiveProducer = explosiveProducer
  self.explosives = Array:new({})
  self.targets = Array:new({})
  self.interval = self.options.interval and IntervalTimer:new(self.options.interval, function() self:onIntervalEnded() end)
  self.hasPlayedAudio = { firstTarget = false, firstDetonation = false }
  self.awaitingPlayerVehicle = false
  self.ringPath = nil
  self.isDestroyed = false
  return self
end

return ExplosiveArea
