local bounds = require("lua/explosionsNG/helpers/bounds")
local positioning = require("lua/explosionsNG/helpers/positioning")
local strings = require("lua/explosionsNG/helpers/strings")
local Array = require("lua/explosionsNG/misc/Array")
local explosivesHandler = require("lua/explosionsNG/handlers/explosives")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local explosivesFacade = require("lua/explosionsNG/stores/explosives/facade")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")

--- @alias ChaosModeConfigurationDto { interval: number, maximumConcurrentExplosives: number, allowedExplosives: string[], placementRandomness: number }
--- @alias ChaosModeConfiguration { interval: number, maximumConcurrentExplosives: number, allowedExplosives: Array<string>, placementRandomness: number }
--- @alias ChaosModeConfigurationWithtimer { interval: number, maximumConcurrentExplosives: number, allowedExplosives: string[], placementRandomness: number, timer: number }

--- An unique id for this mode
local MODE_ID = "chaos"

--- A timer that upon every completion cycle will create a new random explosive
--- @type IntervalTimer
local timer = nil
--- The next available id that can be used for creating a new explosive
local nextExplosiveId = 0
--- The current configuration settings of this mode
--- @type ChaosModeConfiguration
local configuration = nil

--- @return boolean Whether the mode is active or not. if `true`, the `timer` will be defined.
local function isActive()
  return timer ~= nil
end

--- @return number The number of generated explosives from this mode
local function getNumberOfGeneratedExplosives()
  return explosivesHandler.getIds()
      :filter(function(explosive) return strings.startsWith(explosive.uiId, MODE_ID) end)
      :length()
end

--- Called when an explosive finished.
---
--- This function will check to see if the concurrency limit is now greater than the current number of generated
--- explosives. If it is, this function will start the timer.
---
--- Note that if this mode is not active, then this function will do nothing
local function onExplosiveFinished()
  if not isActive() then return end

  -- Start the timer if we have dipped below the concurrency limit
  if getNumberOfGeneratedExplosives() < configuration.maximumConcurrentExplosives then
    timer:start()
  end
end

--- Creates a random explosive and adds it to the explosives handler.
--- This function will also notify the UI of the added explosive.
local function addRandomExplosive()
  local vehicle = vehiclesHandler.getPlayerVehicle()

  if not vehicle then
    log("I", "explosivesNG", "Player is not currently in a vehicle. This explosive will be skipped.")
    return
  end

  local id = MODE_ID..nextExplosiveId
  nextExplosiveId = nextExplosiveId + 1

  local explosiveId = explosivesFacade.getExplosives()
      :filter(function(explosive) return configuration.allowedExplosives:find(explosive) ~= nil end)
      :randomElement()

  local presetIds = explosivesFacade.getPresetsForExplosive(explosiveId)
  local presetId = presetIds:randomElement()

  --log("I", "explosivesNG", "Chaos mode: spawning "..explosiveId.." with "..presetId.." preset")

  -- Find a random position, and adjust by the bounding area of explosives
  local basePosition = positioning.getRandomPointAroundVehicle(vehicle, configuration.placementRandomness)
  local boundingRange = explosivesFacade.getBoundingArea(explosiveId, presetId)
  local elevation = bounds.getElevationOffset(boundingRange)
  local position = basePosition + elevation

  local explosive = explosivesFacade.createNewExplosive(explosiveId, presetId, position)
  local identifier = explosivesHandler.addExplosive(id, explosiveId, presetId, explosive)
  -- Note that order here matters. The callbacks on any explosive is guaranteed to be sequential.
  -- The explosivesHandler will register its own event handler to remove the explosive from its own state. We want this
  -- event handler to trigger afterwards, so that this function gets the correct number of active explosives.
  explosive:registerOnFinishedHandler(onExplosiveFinished)

  guihooks.trigger("engExplosiveAdded", identifier)
end

--- Called when the timer interval elapses.
--- This function will create a new explosive, and will check to see if the concurrency limit has been reached. If it
--- has, this function will additionally stop the timer.
local function onIntervalFinished()
  addRandomExplosive()

  -- Stop the timer to prevent anymore explosives being generated
  if getNumberOfGeneratedExplosives() >= configuration.maximumConcurrentExplosives then
    timer:stop()
  end
end

--- Updates the mode. If the mode is not active, then this function will do nothing.
--- @param dt number The time since the last frame
local function update(dt)
  if not isActive() then return end
  timer:update(dt)

  guihooks.trigger("engChaosModeTimer", timer:elapsedRatio())
end

--- Activate this mode and starts the explosive timer
local function start()
  timer = IntervalTimer:new(configuration.interval, onIntervalFinished, true)
end

--- Stops this mod and destroys the explosive timer.
--- Additionally, this function notifies the UI of the timer being destroyed.
local function stop()
  timer = nil
  guihooks.trigger("engChaosModeTimer")
end

--- Sets the configuration for this mode
--- @param configDto ChaosModeConfigurationDto The configuration to set
local function setConfigurationByDto(configDto)
  configuration = {
    interval = configDto.interval,
    maximumConcurrentExplosives = configDto.maximumConcurrentExplosives,
    placementRandomness = configDto.placementRandomness,
    allowedExplosives = Array:new(configDto.allowedExplosives)
  }
end

--- @return ChaosModeConfigurationWithtimer | nil The current configuration. If the mode is not active, then `nil` is returned.
local function getActiveConfiguration()
  if not isActive() then return nil end

  return {
    interval = configuration.interval,
    maximumConcurrentExplosives = configuration.maximumConcurrentExplosives,
    placementRandomness = configuration.placementRandomness,
    allowedExplosives = configuration.allowedExplosives:toPrimitive(),
    timer = timer:elapsedRatio(),
  }
end

return {
  isActive = isActive,
  update = update,
  start = start,
  stop = stop,
  setConfigurationByDto = setConfigurationByDto,
  getActiveConfigurationDto = getActiveConfiguration
}