local Array = require("lua/explosionsNG/misc/Array")

--- @alias ExplosiveIdentifier { uiId: string, explosiveId: string, presetId: string, fuse: number }
--- @alias ExplosiveWithIdentifier { identifier: ExplosiveIdentifier, explosive: AnyExplosive }
--- @alias BossData { progress: ConvergencePhaseProgress[], failed: boolean }

local explosives = Array:new(--[[--- @type ExplosiveWithIdentifier[] ]] {})

--- Removes the given `explosive` from the `explosives` Array
--- @param explosive AnyExplosive The explosive to remove
local function removeExplosive(explosive)
  local foundExplosive = explosives:findBy(function(explosiveWithId) return explosiveWithId.explosive == explosive end)

  if not foundExplosive then
    log("W", "explosivesNG", "Could not remove explosive as it doesn't exist")
    return
  end

  explosives:remove(foundExplosive)
  guihooks.trigger("engExplosiveFinished", foundExplosive.identifier.uiId)
end

--- Registers a handler to the finish event of the explosive and activates the explosive.
--- This function will also adds the explosive to this handlers internal array.
--- @param uiId string The UI id to store with the explosive
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @param explosive AnyExplosive The explosive to add
--- @return ExplosiveIdentifier The identifier of the explosive that was added
local function addExplosive(uiId, explosiveId, presetId, explosive)
  local identifier = { uiId = uiId, explosiveId = explosiveId, presetId = presetId, fuse = explosive:getFuse() }
  explosives:add({ identifier = identifier, explosive = explosive })
  explosive:registerOnFinishedHandler(function() removeExplosive(explosive) end)
  explosive:activate()
  return identifier
end

--- Updates each explosive in `explosives`
--- @param dt number The time since the last frame
local function update(dt)
  explosives:forEach(function(explosiveWithId) explosiveWithId.explosive:update(dt) end)
end

--- @return Array<ExplosiveIdentifier> The list of all ids for each explosive being handled by this handler
local function getIds()
  return explosives:map(function(explosiveWithId) return explosiveWithId.identifier end)
end

--- @return BossData The current progress through the "The Convergence" boss battle. `nil` if no boss is currently active.
local function getBossData()
  local convergence = explosives:findBy(function(explosiveWithId) return explosiveWithId.identifier.explosiveId == "convergence" end)
  if not convergence then return nil end

  local castedConvergence = --[[--- @type ConvergenceExplosive ]] convergence.explosive
  return { progress = castedConvergence:getCurrentProgress(), failed = castedConvergence:hasFailedCurrentPhase() }
end

--- Clears all explosives in the `explosives` array.
---
--- Unlike the `destroyAll` function, this function will not perform any additional actions on existing explosives
--- currently stored at the time of the function being called.
---
--- This function should be used when the explosives contained within this handler is out of date (for example: a new
--- level was loaded).
local function clear()
  explosives:clear()
end

--- Deletes all explosives before clearing the array.
--- Unlike the `clear` function, this function will ensure all explosives have been properly closed down.
local function deactivateAll()
  explosives:forEach(function(explosiveWithId) explosiveWithId.explosive:deactivate() end)
  --- Note that this function does not clear the array - the explosives have been destroyed and so they are responsible
  --- for them to correct call the "onFinished" event handler. If that method has been correctly defined, the callback
  --- will run which will filter them out; in essence acting as the clear. If we were to call `clear` here, we would be
  --- trying to clear the array twice.
end

--- Deactivates the explosive that matches the given uiId
--- @param explosiveUiId string The explosive uiId to destroy
local function deactivateExplosive(explosiveUiId)
  local explosive = explosives:findBy(function(explosiveWithId) return explosiveWithId.identifier.uiId == explosiveUiId end)

  if not explosive then
    log("W", "explosivesNG", "No explosive could be found with the uiId "..tostring(explosiveUiId))
    return
  end

  explosive.explosive:deactivate()
end

return {
  addExplosive = addExplosive,
  update = update,
  getIds = getIds,
  getBossData = getBossData,
  clear = clear,
  deactivateAll = deactivateAll,
  deactivateExplosive = deactivateExplosive
}