local Vehicle = require("lua/explosionsNG/vehicles/Vehicle")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias VehicleEvent "playerSpawned" |  "spawned" | "resetted"
--- @alias VehicleEventFn fun(player: Vehicle): void

local vehicles = Array:new(--[[--- @type Vehicle[] ]] {})

--- @type table<VehicleEvent, Array<VehicleEventFn>>
local eventCallbacks = {
  playerSpawned = Array:new(),
  spawned = Array:new(),
  resetted = Array:new()
}

--- A flag of whether the player is currently not in a vehicle
local isPlayerNotInAVehicle = true

--- Returns a predicate that matches the given vehicle `id`
--- @param id number The vehicle id that the predicate should match
--- @return fun(vehicle: Vehicle): boolean The predicate
local function vehicleIdPredicate(id)
  return function(vehicle) return vehicle:getId() == id end
end

--- Calls all registered callbacks for the given event, passing the given vehicle to them.
--- @param event VehicleEvent The event
--- @param vehicle Vehicle The vehicle triggering the event
--- @param clear? boolean Whether the clear all callbacks, once called
local function emitVehicleEvent(event, vehicle, clear)
  local callbacks = eventCallbacks[event]
  callbacks:forEach(function(callback) callback(vehicle) end)
  if clear then callbacks:clear() end
end

--- Creates a new `Vehicle` object using the game object found using the given `id`.
--- If the given `id` does not yet have  a `Vehicle` associated with it, this function adds it to the internal array.
--- Otherwise, this function will overwrite the existing object. This function will then return the vehicle.
--- @param id number The id of the vehicle to add
--- @return Vehicle The created vehicle
local function addVehicle(id)
  local predicate = vehicleIdPredicate(id)
  local newVehicle = Vehicle:new(id, --[[--- @type VehicleSceneObject ]] be:getObjectByID(id))
  local existingVehicle = vehicles:findBy(predicate)

  if existingVehicle then
    vehicles:replaceBy(newVehicle, predicate)
    log("I", "explosivesNG", "Replaced existing Vehicle instance for "..tostring(id))
    return newVehicle
  end

  vehicles:add(newVehicle)
  return newVehicle
end

--- Checks and returns whether at least one vehicle is within the given range of the given position
--- @param position LuaVec3 The centre of the range to check
--- @param range number The extent of the range (as a radius)
--- @return boolean whether at least one vehicle is within the given range of the given position
local function isVehicleInRange(position, range)
  return vehicles:some(function(vehicle) return vehicle:isWithin(position, range) end)
end

--- @return Array<Vehicle> A copy of the array of all vehicles
local function getVehicles()
  return vehicles:clone()
end

--- Returns the vehicle with the given id
--- @param id number The id of the vehicle to find
--- @return Vehicle The found vehicle. `nil` if no vehicle exists
local function getVehicle(id)
  return vehicles:findBy(vehicleIdPredicate(id))
end

--- Finds and returns the vehicle the player is in
--- @return Vehicle The found player vehicle. `nil` if the player is not in any vehicle
local function getPlayerVehicle()
  return vehicles:findBy(function(vehicle) return vehicle:isPlayer() end)
end

--- It should be noted that this function will check specifically whether there is a vehicle that this handler knows
--- about that reports it is a player vehicle; it is independent of the `isPlayerNotInAVehicle` flag.
--- @return boolean Whether the player is in a vehicle or not
local function isPlayerInAVehicle()
  return getPlayerVehicle() ~= nil
end

--- Stores the given callback function to this handler. When the event is triggered in this handler, this callback will
--- run.
---
--- __Important:__ it should be noted that for a `"playerSpawned"` event. once the callback is triggered, it will be
--- cleared from this handler. If a caller wishes to maintain a link to this event, after the callback is triggered they
--- should re-register the callback.
--- @param event VehicleEvent The event to register to
--- @param callback VehicleEventFn The handler callback
local function registerHandler(event, callback)
  eventCallbacks[event]:add(callback)
end

--- Removes the given callback function from this handler
--- @param event VehicleEvent The event to register to
--- @param callback VehicleEventFn The handler callback
local function unregisterHandler(event, callback)
  eventCallbacks[event]:remove(callback)
end

--- Checks whether a player vehicle exists. If it does, this function will then call all callbacks that have been
--- registered to this handler before clearing them and setting the `isPlayerNotInAVehicle` flag to `false`.
local function checkPlayerNowSpawned()
  local playerVehicle = getPlayerVehicle()

  if playerVehicle then
    emitVehicleEvent("playerSpawned", playerVehicle, true)
    isPlayerNotInAVehicle = false
  end
end

--- Updates every vehicle in this handler
--- @param dt number The time since the last frame
local function update(dt)
  vehicles:forEach(function(vehicle) vehicle:update(dt) end)
end

--- A vehicle has been removed.
--- Removes the given vehicle from the vehicles array, given the vehicle's id.
--- @param id number The id of the vehicle to remove
local function onVehicleDestroyed(id)
  vehicles = vehicles:removeIf(vehicleIdPredicate(id))
  isPlayerNotInAVehicle = not isPlayerInAVehicle()
end

--- The player vehicle has been changed.
--- Notifies the UI as to whether the player is currently in a vehicle or not.
local function onVehicleSwitched()
  if isPlayerNotInAVehicle then
    checkPlayerNowSpawned()
  end

  guihooks.trigger("engIsPlayerInAVehicle", isPlayerInAVehicle())
end

--- A vehicle has been reset.
--- Resets the vehicle with the given in-game object id
--- This function will then emit the `"resetted"` event to all registered callbacks.
--- @param id number The id of the vehicle
local function onVehicleResetted(id)
  local vehicle = getVehicle(id)

  if not vehicle then
    log("W", "explosivesNG", "A vehicle has been reset that does not exist on this handler.")
    return
  end

  vehicle:reset()
  emitVehicleEvent("resetted", vehicle)
end

--- A vehicle has been spawned.
--- Creates a new vehicle to this handler with the in-game object that matches the given id.
--- This function will then emit the `"spawned"` event to all registered callbacks.
--- @param id number The id of the vehicle
local function onVehicleSpawned(id)
  local vehicle = addVehicle(id)
  emitVehicleEvent("spawned", vehicle)
end

--- Removes any vehicles that no longer exist in-game.
--- This should be called if a vehicle is removed and the id is not known, i.e. loading a new level.
local function clear()
  vehicles = vehicles:removeIf(Vehicle.isRemovedFromLevel)
end

return {
  isVehicleInRange = isVehicleInRange,
  getVehicles = getVehicles,
  getPlayerVehicle = getPlayerVehicle,
  isPlayerInAVehicle = isPlayerInAVehicle,
  registerHandler = registerHandler,
  unregisterHandler = unregisterHandler,
  update = update,
  onVehicleDestroyed = onVehicleDestroyed,
  onVehicleSwitched = onVehicleSwitched,
  onVehicleResetted = onVehicleResetted,
  onVehicleSpawned = onVehicleSpawned,
  clear = clear
}
