local Array = require("lua/explosionsNG/misc/Array")
local simObjects = require("lua/explosionsNG/handlers/worldObjects/simObjects")

-- A predicate that will return true for any SceneObject
--- @type fun(element: SceneObject): true
local FIRST_AVAILABLE = function() return true end

local SIM_GROUP_NAME = "ExplosivesNG"
--- @type SceneTreeFolder
local SIM_GROUP = nil

local tsStatics = Array:new(--[[--- @type SceneObject[] ]] {})
local particleEmitters = Array:new(--[[--- @type SceneObject[] ]] {})
local pointLights = Array:new(--[[--- @type SceneObject[] ]] {})
local soundEmitters = Array:new(--[[--- @type SceneObject[] ]] {})

--- Checks to see whether a suitable object exists in the `existingObjects` by running the `searchBy` predicate against
--- each element. If a suitable element is found, then that object will be unhidden and it fields set by the list of
--- given `options`, before being returned. If no suitable object could be found, then this function will create one
--- using the `provider` function with the same `options` and return it.
--- @generic T
--- @param existingObjects Array<SceneObject> A list of existing objects to look for an existing object
--- @param searchBy fun(element: SceneObject): boolean A predicate that should return `true` for an existing object that can be returned
--- @param provider fun(options: T): SceneObject A provider function to create a new object
--- @param fieldSetter fun(object: SceneObject, options: T): void A function that should accept an object and apply the given options to it
--- @param options T The options to apply to the returned object
--- @return SceneObject The returned object
local function requestObject(existingObjects, searchBy, provider, fieldSetter, options)
  local found = existingObjects:findBy(searchBy)

  if found then
    existingObjects:remove(found)
    found:setField("hidden", 0, "0")
    fieldSetter(found, options)
    found:postApply()
    return found
  end

  local created = provider(options)
  SIM_GROUP:addObject(created)
  return created
end

--- Sets the `object` to be hidden from view and sets its fields to the given list of options. This function will then
--- add the object to the `objectsList`.
--- @generic T
--- @param objectsList Array<SceneObject> The list of objects to add the given object to
--- @param object SceneObject The object to add the list of objects
--- @param fieldSetter fun(object: SceneObject, options: T): void A function that should accept an object and apply the given options to it
--- @param resetOptions T The options to apply to the returned object
local function releaseObject(objectsList, object, fieldSetter, resetOptions)
  object:setField("hidden", 0, "1")
  fieldSetter(object, resetOptions)
  objectsList:add(object)
end

--- Returns a TSStatic object. This object will either:
---  - Be an existing object; if the `modelName` in the `options` matches an existing object that this handler has
---    available
---  - Be a new object; otherwise
--- The object will have the given `options` applied to it before being returned.
--- @param options TSStaticOptions The options to apply to the object upon returning it
--- @return SceneObject A TSStatic object
local function requestTSStatic(options)
  --- @type fun(element: SceneObject): boolean
  local searchByModelName = function(element)
    local shapeName = element.shapeName
    -- TorqueScript is very inconsistent with whether or not the shape name will have a prefixed "/" on it :/
    -- Strip it to avoid any issues with string equality below.
    if shapeName:sub(1, 1) == "/" then
      shapeName = shapeName:sub(2)
    end

    return shapeName == simObjects.modelToPath(options.modelName)
  end

  return requestObject(tsStatics, searchByModelName, simObjects.createTSStatic, simObjects.setTSStaticOptions, options)
end

--- Adds the TSStatic object to this handler.
--- __Important: Callers should NOT interact with this object again after passing it to this handler.__
--- @param object SceneObject The object to give to this handler
local function releaseTSStatic(object)
  -- Persist modelName as changing that is computationally expensive
  releaseObject(tsStatics, object, simObjects.setTSStaticOptions, { modelName = object.shapeName })
end

--- Returns a ParticleEmitter object. This object will either:
---  - Be an existing object; if any ParticleEmitter exists
---  - Be a new object; otherwise
--- The object will have the given `options` applied to it before being returned.
--- @param options ParticleEmitterOptions The options to apply to the object upon returning it
--- @return SceneObject A ParticleEmitter object
local function requestParticleEmitter(options)
  return requestObject(particleEmitters, FIRST_AVAILABLE, simObjects.createParticleEmitter, simObjects.setParticleEmitterOptions, options)
end

--- Adds the ParticleEmitter object to this handler.
--- __Important: Callers should NOT interact with this object again after passing it to this handler.__
--- @param object SceneObject The object to give to this handler
local function releaseParticleEmitter(object)
  releaseObject(particleEmitters, object, simObjects.setParticleEmitterOptions, { emitter = "" })
end

--- Returns a PointLight object. This object will either:
---  - Be an existing object; if any PointLight exists
---  - Be a new object; otherwise
--- The object will have the given `options` applied to it before being returned.
--- @param options PointLightOptions The options to apply to the object upon returning it
--- @return SceneObject A PointLight object
local function requestPointLight(options)
  return requestObject(pointLights, FIRST_AVAILABLE, simObjects.createPointLight, simObjects.setPointLightOptions, options)
end

--- Adds the PointLight object to this handler.
--- __Important: Callers should NOT interact with this object again after passing it to this handler.__
--- @param object SceneObject The object to give to this handler
local function releasePointLight(object)
  releaseObject(pointLights, object, simObjects.setPointLightOptions, --[[--- @type PointLightOptions ]] {})
end

--- Returns a new SFXEmitter object
--- The object will have the given `options` applied to it before being returned.
--- @param options SFXEmitterOptions The options to apply to the object upon returning it
--- @return SceneObject A SFXEmitter object
local function requestSoundEmitter(options)
  --requestObject(soundEmitters, FIRST_AVAILABLE, simObjects.createSoundEmitter, simObjects.setSoundEmitterOptions, options)
  local created = simObjects.createSoundEmitter(options)
  SIM_GROUP:addObject(created)
  return created
end

--- Adds the SFXEmitter object to this handler.
--- __Important: Callers should NOT interact with this object again after passing it to this handler.__
--- @param object SceneObject The object to give to this handler
local function releaseSoundEmitter(object)
  --releaseObject(soundEmitters, object, simObjects.setSoundEmitterOptions, --[[--- @type SFXEmitterOptions ]] {})
  object:delete()
end

--- Creates an initial list of all four types of objects (TSStatic, PointLight, ParticleEmitter, SFXEmitter).
--- This function will create TSStatics with multiple different model names as well.
--- @param group SceneTreeFolder The group to add the sim objects to
local function createInitialSimObjects(group)
  --- Adds the object to the `group` and hides it
  --- @param object SceneObject The object
  local onObjectCreation = function(object)
    object:setField("hidden", 0, "1")
    group:addObject(object)
  end

  tsStatics = Array:flatten(--[[--- @type Array<Array<SceneObject>> ]] Array:new({
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "projectile_sphere"                     }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "cylinder_marker"                       }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "cylinder_electric",    dynamic = false }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "sphere_marker",        dynamic = false }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "sphere_danger_marker", dynamic = false }) end),
    Array:from(50, function() return simObjects.createTSStatic({ modelName = "sphere"                                }) end),
    Array:from(5 , function() return simObjects.createTSStatic({ modelName = "sphere_opaque"                         }) end),
    Array:from(40, function() return simObjects.createTSStatic({ modelName = "anomaly_layer"                         }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "box"                                   }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "lightning1"                            }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "lightning2"                            }) end),
    Array:from(10, function() return simObjects.createTSStatic({ modelName = "lightning3"                            }) end),
    Array:from(5 , function() return simObjects.createTSStatic({ modelName = "supernova"                             }) end),
    Array:from(5 , function() return simObjects.createTSStatic({ modelName = "accretion_disc"                        }) end),
  }))
  pointLights = Array:from(30, function() return simObjects.createPointLight({}) end)
  --soundEmitters    = Array:from(10, function() return simObjects.createSoundEmitter({})                  end)
  --particleEmitters = Array:from(10, function() return simObjects.createParticleEmitter({ emitter = "" }) end)

  tsStatics:forEach(onObjectCreation)
  pointLights:forEach(onObjectCreation)
  --soundEmitters:forEach(onObjectCreation)
  --particleEmitters:forEach(onObjectCreation)
end

--- Creates the SimGroup in-game object
local function initialise()
  if SIM_GROUP and scenetree.findObject(SIM_GROUP_NAME) then
    log("W", "explosivesNG", "SimGroup already exists. Aborting the creation of initial objects...")
    return
  end

  if tsStatics:length() > 0 or pointLights:length() > 0 or soundEmitters:length() > 0 or particleEmitters:length() > 0 then
    log("W", "explosivesNG", "Attempting to create initial objects yet existing objects are stored. Proceeding and ignoring existing objects...")
  end

  local group = --[[--- @type SceneTreeFolder ]] createObject("SimGroup")
  group:setField("locked", 0, "1")
  group:registerObject(SIM_GROUP_NAME)
  createInitialSimObjects(group)
  scenetree.MissionGroup:addObject(group)

  SIM_GROUP = group
end

return {
  requestTSStatic = requestTSStatic,
  releaseTSStatic = releaseTSStatic,
  requestParticleEmitter = requestParticleEmitter,
  releaseParticleEmitter = releaseParticleEmitter,
  requestPointLight = requestPointLight,
  releasePointLight = releasePointLight,
  requestSoundEmitter = requestSoundEmitter,
  releaseSoundEmitter = releaseSoundEmitter,
  initialise = initialise
}