local vectors = require("lua/explosionsNG/helpers/vectors")

local ROOT_ASSET_FOLDER = "art/explosionsNG"
local MODEL_FOLDER = ROOT_ASSET_FOLDER
local SOUNDS_FOLDER = ROOT_ASSET_FOLDER.."/sounds"

--- @alias TSStaticOptions { modelName: string, dynamic: boolean | nil, position: Vector | nil, scale: LuaVec3 | number | nil, rotationEuler: LuaVec3 | nil, instanceColour: string | nil }
--- @alias ParticleEmitterOptions { emitter: string, position: Vector | nil, scale: LuaVec3 | number | nil, rotationEuler: LuaVec3 | nil, velocity: number | nil, dataBlock: string | nil }
--- @alias PointLightOptions { position: Vector | nil, brightness: number | nil, radius: number | nil, colour: string | nil, animationType: string | nil, applyRadiusToPosition: boolean | nil }
--- @alias SFXEmitterOptions { track: string | nil, position: Vector | nil, referenceDistance: number | nil, maxDistance: number | nil, isLooping: boolean | nil, useInternalSound: boolean | nil }

--- Returns a string in the format of `rootFolder/asset.extension`
--- @param asset string The asset
--- @param rootFolder string The root folder
--- @param extension string The extension type
--- @return string A string in the format of `rootFolder/asset.extension`
local function assetToPath(asset, rootFolder, extension)
  return rootFolder.."/"..asset.."."..extension
end

--- Converts a model name into its associated absolute filename
--- @param model string The model name
--- @return string The absolute path of the model
local function modelToPath(model)
  return assetToPath(model, MODEL_FOLDER, "dae")
end

--- Converts a sound/track name into its associated absolute filename
--- @param sound string The sound/track name
--- @return string The absolute path of the sound/track
local function soundToPath(sound)
  return assetToPath(sound, SOUNDS_FOLDER, "ogg")
end

--- Adds a proportional elevation to the given position, based on the radius of the point light.
--- This should be used for objects that are directly in/on an object (no light is emitted as it's considered obscured)
--- so that the light will illuminate the surroundings, but also so that the light will extend to the actual radius
--- desired (if you place a PointLight just above the ground with a radius of X, the actual light emitted will extend to
--- significantly less than X).
--- @param position Vector The initial position of the light
--- @param radius number The radius of the light
--- @return LuaVec3 The adjusted position
local function fixPointLightPosition(position, radius)
  return vec3(position) + vec3(0, 0, radius / 4)
end

--- Formats the given field in one of the following ways:
---  - If the `field` is a `LuaVec3` or `Vector`, this function returns the vector formatted as `"x y z"`
---  - If the `field` is a string, this function returns this string unaffected
---  - If the `field` is a number, this function returns a vector string formatted as `"x x x"`, where `x` is the value
---    of `field`
--- @param field number | string | LuaVec3 | Vector The field to format
--- @return string The formatted field
local function fieldNumberToString3d(field)
  if type(field) == "cdata" then return vectors.toString(--[[--- @type LuaVec3 | Vector ]] field) end
  if type(field) == "string" then return --[[--- @type string ]] field end

  local casted = --[[--- @type number ]] field
  return vectors.toString(vec3(casted, casted, casted))
end

--- Fills in any key not in the given `options` with default values
--- @param options TSStaticOptions List of defined options
--- @return TSStaticOptions The `options` with any undefined key being set to the default value
local function fillTSStaticOptions(options)
  return {
    modelName = options.modelName,
    position = options.position or vec3(),
    scale = options.scale or 1,
    rotationEuler = options.rotationEuler or vec3(),
    instanceColour = options.instanceColour or "1 1 1 1",
    dynamic = options.dynamic == nil and true or options.dynamic
  }
end

--- Fills in any key not in the given `options` with default values
--- @param options ParticleEmitterOptions List of defined options
--- @return ParticleEmitterOptions The `options` with any undefined key being set to the default value
local function fillParticleEmitterOptions(options)
  return {
    emitter = options.emitter,
    position = options.position or vec3(),
    scale = options.scale or 1,
    rotationEuler = options.rotationEuler or vec3(),
    velocity = options.velocity or 3,
    dataBlock = options.dataBlock or "lightExampleEmitterNodeData1"
  }
end

--- Fills in any key not in the given `options` with default values
--- @param options PointLightOptions List of defined options
--- @return PointLightOptions The `options` with any undefined key being set to the default value
local function fillPointLightOptions(options)
  return {
    position = options.position or vec3(),
    radius = options.radius or 1,
    colour = options.colour or "1 1 1 1",
    brightness = options.brightness or 1,
    animationType = options.animationType or "",
    applyRadiusToPosition = options.applyRadiusToPosition or false
  }
end

--- Fills in any key not in the given `options` with default values
--- @param options SFXEmitterOptions List of defined options
--- @return SFXEmitterOptions The `options` with any undefined key being set to the default value
local function fillSFXEmitterOptions(options)
  local maxDistance = options.maxDistance or 100
  return {
    track = options.track,
    maxDistance = maxDistance,
    position = options.position or vec3(),
    referenceDistance = options.referenceDistance or maxDistance / 5,
    isLooping = options.isLooping or false,
    useInternalSound = options.useInternalSound or false,
  }
end

--- Modifies the given object's fields with the given options, with defaults being used for options not given.
--- Note that the `modelName` option can either be a full path (`folder/model.dae`), or just the model name (`model`)
--- @param object SceneObject The TSStatic to modify
--- @param options TSStaticOptions A list of fields to set to the object. Any blank fields will be set to their default values
local function setTSStaticOptions(object, options)
  local allOptions = fillTSStaticOptions(options)

  if not allOptions.modelName:find(ROOT_ASSET_FOLDER) then
    allOptions.modelName = modelToPath(options.modelName)
  end

  object:setField("dynamic", 0, allOptions.dynamic and "1" or "0")
  object:setField("shapeName", 0, allOptions.modelName)
  object:setField("position", 0, fieldNumberToString3d(allOptions.position))
  object:setField("rotationEuler", 0, fieldNumberToString3d(allOptions.rotationEuler))
  object:setField("scale", 0, fieldNumberToString3d(allOptions.scale))
  object:setField("instanceColor", 0, allOptions.instanceColour)
end

--- Modifies the given particle emitter's fields with the given options, with defaults being used for options not given.
--- @param object SceneObject The particle emitter to modify
--- @param options ParticleEmitterOptions A list of fields to set to the particle emitter. Any blank fields will be set to their default values
local function setParticleEmitterOptions(object, options)
  local allOptions = fillParticleEmitterOptions(options)
  object:setField("emitter", 0, allOptions.emitter)
  object:setField("dataBlock", 0, allOptions.dataBlock)
  object:setField("velocity", 0, allOptions.velocity)
  object:setField("position", 0, fieldNumberToString3d(allOptions.position))
  object:setField("rotationEuler", 0, fieldNumberToString3d(allOptions.rotationEuler))
  object:setField("scale", 0, fieldNumberToString3d(allOptions.scale))
end

--- Modifies the given point light's fields with the given options, with defaults being used for options not given.
--- @param object SceneObject The particle emitter to modify
--- @param options PointLightOptions A list of fields to set to the point light. Any blank fields will be set to their default values
local function setPointLightOptions(object, options)
  local allOptions = fillPointLightOptions(options)
  local position = options.applyRadiusToPosition
      and fixPointLightPosition(--[[--- @not nil ]] allOptions.position, allOptions.radius)
      or allOptions.position

  object:setField("radius", 0, allOptions.radius)
  object:setField("color", 0, allOptions.colour)
  object:setField("brightness", 0, allOptions.brightness)
  object:setField("animationType", 0, allOptions.animationType)
  object:setField("position", 0, fieldNumberToString3d(position))
end

--- Modifies the given sound emitter's fields with the given track and options, with defaults being used for options not
--- given.
--- @param object SceneObject The particle emitter to modify
--- @param options SFXEmitterOptions A list of fields to set to the SFX emitter. Any blank fields will be set to their default values
local function setSoundEmitterOptions(object, options)
  local allOptions = fillSFXEmitterOptions(options)
  object:setField("isLooping", 0, allOptions.isLooping and "1" or "0")
  object:setField("referenceDistance", 0, allOptions.referenceDistance)
  object:setField("maxDistance", 0, allOptions.maxDistance)
  object:setField("position", 0, fieldNumberToString3d(allOptions.position))

  if allOptions.track then
    object:setField("filename", 0, soundToPath(allOptions.track))
  end
end

--local ID = 0

--- Registers a SceneObject and sets it to be locked. This will prevent deletion in-game and is important as if a player
--- were to delete an object in the editor while a class was using it, it will almost certainly crash the whole game.
--- @param object SceneObject The object to register and lock
local function registerObject(object)
  object:setField("locked", 0, "1")
  --object:registerObject("eng_object_"..ID)
  object:registerObject("")
  --ID = ID + 1
end

--- Creates and returns a TSStatic object
--- @param options TSStaticOptions A list of fields to set to the object when creating it. Any blank fields will be set to their default values.
--- @return SceneObject The created object
local function createTSStatic(options)
  local object = createObject("TSStatic")
  setTSStaticOptions(object, options)
  registerObject(object)
  return object
end

--- Creates and returns a ParticleEmitterNode object
--- @param options ParticleEmitterOptions A list of fields to set to the object when creating it. Any blank fields will be set to their default values.
--- @return SceneObject The created object
local function createParticleEmitter(options)
  local object = createObject("ParticleEmitterNode")
  setParticleEmitterOptions(object, options)
  registerObject(object)
  return object
end

--- Creates and returns a PointLight object
--- @param options PointLightOptions A list of fields to set to the object when creating it. Any blank fields will be set to their default values.
--- @return SceneObject The created object
local function createPointLight(options)
  local object = createObject("PointLight")
  setPointLightOptions(object, options)
  registerObject(object)
  return object
end

--- Creates and returns a SFXEmitter object
--- @param options SFXEmitterOptions A list of fields to set to the object when creating it. Any blank fields will be set to their default values.
--- @return SceneObject
local function createSoundEmitter(options)
  local object = createObject("SFXEmitter")
  setSoundEmitterOptions(object, options)
  registerObject(object)
  return object
end

return {
  modelToPath = modelToPath,
  soundToPath = soundToPath,
  fillTSStaticOptions = fillTSStaticOptions,
  fillParticleEmitterOptions = fillParticleEmitterOptions,
  fillSFXEmitterOptions = fillSFXEmitterOptions,
  fillPointLightOptions = fillPointLightOptions,
  setTSStaticOptions = setTSStaticOptions,
  setParticleEmitterOptions = setParticleEmitterOptions,
  setPointLightOptions = setPointLightOptions,
  setSoundEmitterOptions = setSoundEmitterOptions,
  createTSStatic = createTSStatic,
  createParticleEmitter = createParticleEmitter,
  createPointLight = createPointLight,
  createSoundEmitter = createSoundEmitter
}