local maths = require("lua/explosionsNG/helpers/maths")
local Array = require("lua/explosionsNG/misc/Array")

local POSITIVE_Z = vec3( 0,  0,  1)
local NEGATIVE_Z = vec3( 0,  0, -1)
local POSITIVE_X = vec3( 1,  0,  0)
local NEGATIVE_X = vec3(-1,  0,  0)
local POSITIVE_Y = vec3( 0,  1,  0)
local NEGATIVE_Y = vec3( 0, -1,  0)

local DIRECTIONS = Array:new({ POSITIVE_X, NEGATIVE_X, POSITIVE_Y, NEGATIVE_Y, POSITIVE_Z, NEGATIVE_Z })

--- The list of GroundPlane objects in the map. This should really only ever contain one and only the first will be used.
--- @type SceneObject[]
local groundPlanes = {}

--- Checks and returns whether the position is colliding/inside with the level terrain. Note that if the terrain does
--- not exist, which may be the case for some maps, this method will always return `false`.
--- @param position LuaVec3 The position to check
--- @return boolean Whether the position is colliding with the level terrain
local function isCollidingWithTerrain(position)
  local terrainHeight = core_terrain.getTerrainHeight(position)

  local isBelowTerrainHeight = terrainHeight and position.z <= terrainHeight
  -- If a GroundPlane exists, check the origin Z value
  local isBelowGroundPlane = groundPlanes and groundPlanes[1] and position.z <= 0

  return isBelowTerrainHeight or isBelowGroundPlane
end

--- Checks and returns whether the position is colliding/inside with a TSStatic object. This method will return `true`
--- if there is a static object within the tolerance.
--- @param position LuaVec3 The position to check
--- @param direction LuaVec3 The direction to check for collision
--- @param tolerance number The tolerance (in m) that is required for a collision to be considered
--- @param sourceSize? number The size of the position's source. Anywhere within this region a collision is not considered.
--- @return boolean Whether the position is colliding with a TSStatic object
local function isCollidingWithStaticObject(position, direction, tolerance, sourceSize)
  local staticCollision = castRayStatic(position, direction, sourceSize or 1)
  return staticCollision <= tolerance
end

--- Returns the elevation from either the position of the highest static object at the calculated xy position, or if
--- there is no static object, simply the ground position. Should neither or these exist, the z component will default
--- to `0`.
--- @param x number The x coordinate
--- @param y number The y coordinate
--- @return number The elevation of the first collision point at those coordinates
local function findElevationOfCollision(x, y)
  local ALTITUDE_TEST = 10000

  local consideringPosition = vec3(x, y, ALTITUDE_TEST)
  local terrainZ = core_terrain.getTerrainHeight(consideringPosition)
  -- Look down to the terrain and search as far as the terrain surface
  local distanceToStatic = castRayStatic(consideringPosition, NEGATIVE_Z, ALTITUDE_TEST)
  return distanceToStatic and ALTITUDE_TEST - distanceToStatic or terrainZ
end

--- Returns a random point within the circle that is centred around the given `centre` with the given `radius`. The z
--- component is set from either the position of the highest static object at the calculated xy position, or if there is
--- no static object, simply the ground position. Should neither or these exist, the z component will default to `0`.
--- @param centre LuaVec3 The centre position of the sphere of effect
--- @param radius number The radius of effect
--- @return LuaVec3 The random point
local function getRandomPointInRange(centre, radius)
  local x = centre.x + maths.randomSignedFloat(radius)
  local y = centre.y + maths.randomSignedFloat(radius)
  local z = findElevationOfCollision(x, y)
  return vec3(x, y, z)
end

--- Returns the axis at which the given position is colliding with either the terrain or a static object. `nil` will be
--- returned if there is no collision.
--- Note that in the very rare event that a collision is occurring on more than one axis, the order will follow the
--- definition of the directions; see `DIRECTIONS`.
--- @param position LuaVec3 The position to check
--- @param tolerance number The tolerance (in m) that is required for a collision to be considered
--- @param sourceSize? number The size of the position's source. Anywhere within this region a collision is not considered.
--- @return LuaVec3 The direction in which a collision is detected. `nil` if no collision
local function findDirectionOfCollision(position, tolerance, sourceSize)
  if isCollidingWithTerrain(position) then return NEGATIVE_Z end

  for _, direction in ipairs(DIRECTIONS:toPrimitive()) do
    if isCollidingWithStaticObject(position, direction, tolerance, sourceSize) then
      return direction
    end
  end
end

--- Resets the internal variables that are required for collision functions to be accurate.
--- Should be called when a new level is loaded.
local function onMapLoaded()
  groundPlanes = scenetree.findClassObjects("GroundPlane")
end

return {
  isCollidingWithTerrain = isCollidingWithTerrain,
  isCollidingWithStaticObject = isCollidingWithStaticObject,
  findElevationOfCollision = findElevationOfCollision,
  getRandomPointInRange = getRandomPointInRange,
  findDirectionOfCollision = findDirectionOfCollision,
  onMapLoaded = onMapLoaded
}