--- Rotates a point about the Z axis.
--- Note that the returned vector will have the same Z coordinate as the given `centre`.
--- @param centre LuaVec3 The centre to rotate around
--- @param radius number The radius to rotate from
--- @param angle number The angle to rotate around by
--- @return LuaVec3 The rotation about the Z axis
local function rotate2d(centre, radius, angle)
  local cos = radius * math.cos(angle)
  local sin = radius * math.sin(angle)
  return centre + vec3(cos, sin, 0)
end

--- Constrains a value between the range of `[min, max]`
--- @param value number The value to constrain
--- @param min number The minimum bound
--- @param max number The maximum bound
--- @return number The constrained value
local function constrain(value, min, max)
  return math.max(min, math.min(value, max))
end

--- @return number A random float in the range `[-1, 1]`
local function randomSigned()
  return 1 - 2 * math.random()
end

--- Returns a random float in the range `[-max, max]`
--- @param max number The maximal value than can be returned in both the positive and negative direction
--- @return number A random float in the range `[-max, max]`
local function randomSignedFloat(max)
  return randomSigned() * max
end

--- Returns the decimal part of the given value
--- @param value number The value
--- @return number The decimal part of the given value
local function decimal(value)
  return value - math.floor(value)
end

--- Returns the ratio of the expected range/area coverage the combined parts will cover in total, compared to the
--- maximum range/area of the overall part coverage.
---
--- This is defined as a geometric series with the following formula:
---  - Let: N = number of parts (N > 1), Rs = range/area of a single parts, Re = range/area of the overall container part
---  - coverage(N) = Rs/Re + Σ{i=2, N}( (1 - coverage(i-1)) * Rs/Re )
--- @return number The ratio of the average range/area coverage the sub-explosive will cover, compared to the super-explosive's range/area coverage
local function getAreaCoverageRatio(numberOfParts, mainRange, partRange)
  local rangeRatio = partRange / mainRange

  local coverage = rangeRatio
  for _ = 2, numberOfParts do
    coverage = coverage + (1 - coverage) * rangeRatio
  end

  return coverage
end

------------------------------------------------------------------------------------------------------------------------
--- These are useful functions, particularly for complex animations (such as The Anomaly).
--- See https://www.desmos.com/calculator/q7p2d7j0nw for a visual look at the functions.
------------------------------------------------------------------------------------------------------------------------

--- @param x number The input number
--- @return number The output with range `[0, 1]`. Cyclical and repeats every unit interval
local function hill(x)
  return (math.sin((2 * math.pi * x) - (math.pi / 2)) + 1) / 2
end

--- @param x number A number in the domain of `[0, 1]`
--- @return number The output with range of `[0, 1]`
local function smoothRampUp(x)
  return hill(x / 2)
end

--- @param x number A number in the domain of `[0, 1]`
--- @return number The output with range of `[0, 1]`
local function smoothRampDown(x)
  return 1 - smoothRampUp(x)
end

--- @param x number A number in the domain of `[0, 1]`
--- @param landingStart? number The point at which the ramp has reached max value. Defaults to `0.5`
--- @return number The output with range `[0, 1]`
local function landingRampUp(x, landingStart)
  local landing = landingStart or 0.5
  return x < landing and x / landing or 1
end

--- @param x number A number in the domain of `[0, 1]`
--- @param landingEnd? number The point at which the ramp begins to slope down. Defaults to `0.5`
--- @return number The output with range `[0, 1]`
local function landingRampDown(x, landingEnd)
  return landingRampUp(1 - x, landingEnd)
end

--- Note that this function is undefined (and will subsequently throw an error) if the `rampTime` is greater than half
--- of `totalTime`.
--- @param x number The input number
--- @param rampTime number The value that `x` must be equal to in order to have climbed to the plateau
--- @param totalTime number The domain of `x` which constituents one cycle of the formula
--- @return number The output with range of `[0, totalTime]`
local function smoothPlateau(x, rampTime, totalTime)
  if rampTime * 2 > totalTime then
    error("rampTime must be less than half of totalTime")
    return nil
  end

  -- By changing the domain from `[-inf, inf]` to `[0, totalTime]`, for which the function is defined for, this is
  -- allows the function to repeat
  x = x % totalTime

  if x <= rampTime             then return smoothRampUp(x / rampTime) end
  if x <= totalTime - rampTime then return 1 end
  return smoothRampDown(1 + (x - totalTime) / rampTime)
end

--- Returns a value which, as `value` gets bigger, the returned value increases by smaller and smaller amounts.
--- `returnMax` can be set to tune at what value `value` must be in order to return a value of `1`.
--- @param value number The input number
--- @param returnMax? number The value at which `value` will return `1`. Defaults to `1`.
--- @return number The output with range of `[0, infinity]`
local function diminishingReturns(value, returnMax)
  returnMax = returnMax or 1
  local movePeak = 9 / returnMax -- log10(9 + 1) = 1
  return math.log(movePeak * value + 1, 10)
end

return {
  rotate2d = rotate2d,
  constrain = constrain,
  randomSignedFloat = randomSignedFloat,
  decimalPart = decimal,
  getAreaCoverageRatio = getAreaCoverageRatio,
  repeatingHill = hill,
  smoothRampUp = smoothRampUp,
  smoothRampDown = smoothRampDown,
  landingRampUp = landingRampUp,
  landingRampDown = landingRampDown,
  smoothPlateau = smoothPlateau,
  diminishingReturns = diminishingReturns
}