local maths = require("lua/explosionsNG/helpers/maths")

--- Formats the vector into a string
--- @param vector Vector The vector to format
--- @param commas? boolean Optional parameter to use commas to space coordinates as "x, y, z". Defaults to false.
--- @return string The formatted vector in the format of "x y z"
local function toString(vector, commas)
  if commas then return string.format("%f, %f, %f", vector.x, vector.y, vector.z) end
  return string.format("%f %f %f", vector.x, vector.y, vector.z)
end

--- Returns a vector whose axises are a random value in the range of [-range, range], for the given range value
--- @param range number The maximum range on each axis in either the position or negative direction
--- @return LuaVec3 The random vector
local function random(range)
  return vec3(maths.randomSignedFloat(range), maths.randomSignedFloat(range), maths.randomSignedFloat(range))
end

--- Defaults any axis inside the given vector if its axis value is `0`
--- @param vector LuaVec3 The vector to default axis in
--- @param defaultValue number The value to replace `0` values
--- @return LuaVec3 The defaulted vector
local function defaultIfZero(vector, defaultValue)
  return vec3(
      vector.x == 0 and defaultValue or vector.x,
      vector.y == 0 and defaultValue or vector.y,
      vector.z == 0 and defaultValue or vector.z
  )
end

--- Multiplies each axis of the vectors together, for example:
---   `v1 = (1, 2, 3), v2 = (2, 4, 6)`;
---   `crossMultiply(v1, v2) = (1 * 2, 2 * 4, 3 * 6) = (2, 8, 18)`
--- This is commutative so the order does not matter (`crossMultiply(v1, v2) === crossMultiply(v2, v1)`)
--- @param vector1 LuaVec3 The first vector
--- @param vector2 LuaVec3 The second vector
--- @return LuaVec3 The resultant vector
local function crossMultiply(vector1, vector2)
  return vec3(vector1.x * vector2.x, vector1.y * vector2.y, vector1.z * vector2.z)
end

return {
  toString = toString,
  random = random,
  defaultIfZero = defaultIfZero,
  crossMultiply = crossMultiply
}