local classes = require("lua/explosionsNG/helpers/classes")

--- An array class that provides convenient methods for common tasks such as filtering, mapping and finding values.
---
--- Some methods that do not return anything specific will instead return the same array to allow for chaining methods
--- together.
---
--- It should be noted that a lot of functions internally will use the built-in ipairs iterator (i.e. they function as
--- wrappers). One reason is to take advantage of the handling of concurrent modification of an array by multiple
--- threads that comes with that function.
---
--- It should also be noted that some methods are in-place, and some methods return a new array and don't affect the
--- current array. Check the relevant method documentation for this information.
--- @class Array<T>
--- @field protected elements T[] The list of elements
--- @field protected len number The (cached) length of the array
local Array = classes.newClass()

--- __Static__
---
--- Creates and returns an array with the given `length` by providing each index to the `producer` function and using
--- the return values as elements
--- @generic S
--- @param length number The length of the array
--- @param producer fun(index: number): S The producer function that takes in an optional index and returns an associated element
--- @return Array<S> The array of produced values
function Array:from(length, producer)
  local array = Array:new(--[[--- @type S[] ]] {})
  for i = 1, length do array:add(producer(i)) end
  return array
end

--- __Static__
---
--- Performs an iteration of the given length, calling the callback function on each iteration.
--- __Important: to avoid unexpected behaviour, the array should not be altered, and should be treated as read only__
--- @generic S
--- @param length number The length of the array
--- @param callbackFn fun(index: number, array: Array<S>): void The function to call on each iteration
function Array:iterate(length, callbackFn)
  for i = 1, length do
    callbackFn(i, self)
  end
end

--- __Static__
---
--- Flattens an array by removing one layer of nesting from an array of arrays.
--- @generic S
--- @param nestedArray Array<Array<S>> The nested array to flatten
--- @return Array<S> The flattened nested array
function Array:flatten(nestedArray)
  local array = Array:new(--[[--- @type S[] ]] {})

  nestedArray:forEach(function(innerArray)
    innerArray:forEach(function(element) array:add(element) end)
  end)

  return array
end

--- Validates whether the given `index` is between [1, `self.len`].
--- This method throws an error if this is not the case.
--- @param index number The index to validate
--- @private
function Array:validateIndex(index)
  if index < 1 then error(string.format("Index (%d) must be > 0", index)) end
  if index > self.len then error(string.format("Index (%d) must be less than the array length (%d)", index, self.len)) end
end

--- Returns the element at the given index
--- @param index number The index
--- @return T The element
function Array:nth(index)
  self:validateIndex(index)
  return self.elements[index]
end

--- @return number The number of elements in the array
function Array:length()
  return self.len
end

--- @return boolean Whether there is at least one element in this array
function Array:isPopulated()
  return self.len >= 1
end

--- @return boolean Whether this array is empty
function Array:isEmpty()
  return self.len == 0
end

--- Removes and returns the first element in the array.
--- If the `takeLast` parameter is given, the last element is taken instead
--- @param takeLast? boolean Optional parameter of whether to take the last element instead of the first
--- @return T The taken item
function Array:pop(takeLast)
  if self:isEmpty() then
    error("Cannot take any item as the array is empty")
    return nil
  end

  local element = takeLast and self.elements[self.len] or self.elements[1]
  self:remove(element)
  return element
end

--- @return T A randomly selected item from this array
function Array:randomElement()
  if self:isEmpty() then
    error("Cannot take any item as the array is empty")
    return nil
  end

  return self.elements[math.random(self.len)]
end

--- For each element in the array, this method calls `callbackFn` with them, their index and the array itself.
--- __Important: to avoid unexpected behaviour, the array should not be altered, and should be treated as read only__
--- @param callbackFn fun(element: T, index: number, array: Array<T>): void The function to call on each element
function Array:forEach(callbackFn)
  for i, v in ipairs(self.elements) do
    callbackFn(v, i, self)
  end
end

--- Collapses all elements in the array to a single value using the given reducer function
--- @generic S
--- @param reducer fun(accumulator: S, current: T, index: number): S
--- @param initialValue S The starting value
--- @return S The reduced value of the array
function Array:reduce(reducer, initialValue)
  local accumulator = initialValue
  self:forEach(function(element, index) accumulator = reducer(accumulator, element, index) end)
  return accumulator
end

--- Finds the element that "wins" against all over elements according to the comparator function.
--- This function compares elements pairwise and only compares the current best element and the next element in the
--- array.
--- For example, this method should be used for finding the maximum or minimum value of the array, the longest string in
--- the array, etc.
--- Note that the first value for the `previous` parameter in the `comparator` function will be the first element in the
--- array.
--- @param comparator fun(previous: T, current: T, index: number): T The comparator function that returns the new "winning" element from the current best and the next element
--- @return T The "winning" element. This will be `nil` if the array is empty.
function Array:findByComparison(comparator)
  if self:length() == 0 then return nil end

  local currentBest = self:nth(1)
  self:forEach(function(element, index)
    if index == 1 then return end -- Skip first element
    currentBest = comparator(currentBest, element, index)
  end)
  return currentBest
end

--- For each element in the array, this method calls `callbackFn` with them and their index in the array and puts the
--- result into a new array, which is returned.
--- @generic S
--- @param callbackFn fun(element: T, index: number): S The function to map each element
--- @return Array<S> The array with each mapped element
function Array:map(callbackFn)
  local mappedArray = --[[--- @type Array<S> ]] Array:new()
  self:forEach(function(element, index) mappedArray:add(callbackFn(element, index)) end)
  return mappedArray
end

--- Filters the array by removing all elements that return `false` for the given `predicate`
--- @param predicate fun(element: T, index: number): boolean A predicate tha should return `true` for the given element to stay in the array
--- @return Array<T> The new array
function Array:filter(predicate)
  --- @type T[]
  local newElements = {}

  self:forEach(function(element, i)
    if predicate(element, i) then
      table.insert(newElements, element)
    end
  end)

  return Array:new(newElements)
end

--- Removes a given `element` from this array. The array will have the same elements if the given `element` does not
--- exist in the array.
--- Note that by default this uses referential equality. Provide the `equals` comparator for custom equality checking.
--- @param element T The element to remove
--- @param equals? fun(value1: T, value2: T): boolean An optional comparator to determine whether two elements are equal
--- @return self This array
function Array:remove(element, equals)
  local newTable = {}
  local newLength = 0
  local comparator = equals or function (value1, value2) return value1 == value2 end

  self:forEach(function(el)
    if not comparator(el, element) then
      table.insert(newTable, el)
      newLength = newLength + 1
    end
  end)

  self.elements = newTable
  self.len = newLength

  return self
end

--- Removes all elements in this array which return `true` for the given `predicate`.
--- This method acts as an inverse filter.
--- @param predicate fun(element: T): boolean A predicate that should return `true` for a given element that should be removed
--- @return Array<T> The resultant array
function Array:removeIf(predicate)
  return self:filter(function(element) return not predicate(element) end)
end

--- Adds a given `element` to this array. This element will be positioned at the end of the array,
--- @param element T The element to add
--- @return self This array
function Array:add(element)
  table.insert(self.elements, element)
  self.len = self.len + 1
  return self
end

--- Replaces the element with the new `element` at the given `index`
--- @param index number
--- @param element T
--- @return self This array
function Array:replace(index, element)
  self:validateIndex(index)
  self.elements[index] = element
  return self
end

--- Finds the first element that equals the given `element`.
--- This is done by referential equality.
--- @param element T The element to search for
--- @return T The found element, `nil` if no element found
function Array:find(element)
  return self:findBy(function(e) return e == element end)
end

--- Finds the first element that passes the given `predicate` function
--- @param predicate fun(element: T): boolean A predicate that should return `true` for the given element that should be returned
--- @return T The found element, `nil` if no element found
function Array:findBy(predicate)
  for _, v in ipairs(self.elements) do
    if predicate(v) then return v end
  end
end

--- Finds the index of the first element that passes the given `predicate` function
--- @param predicate fun(element: T): boolean A predicate that should return `true` for the given element that should be returned
--- @return number The found element's index, `-1` if no element found
function Array:findIndex(predicate)
  for i, v in ipairs(self.elements) do
    if predicate(v) then return i end
  end

  return -1
end

--- Returns a new array with only the first `n` elements from this array
--- @param n number The number of elements that can be kept in the array
--- @return Array<T> The new array
function Array:first(n)
  return self:filter(function(_, i) return i <= n end)
end

--- Returns whether, when called with `predicate`, at least one element in this array returns `true`
--- @param predicate fun(element: T): boolean A predicate that should return `true` for an element which passes the test
--- @return boolean Whether at least one element passes the test `predicate`
function Array:some(predicate)
  for _, v in ipairs(self.elements) do
    if predicate(v) then return true end
  end

  return false
end

--- Returns whether, when called with `predicate`, all elements in this array return `true`
--- @param predicate fun(element: T): boolean A predicate that should return `true` for an element which passes the test
--- @return boolean Whether all elements pass the test `predicate`
function Array:all(predicate)
  for _, v in ipairs(self.elements) do
    if not predicate(v) then return false end
  end

  return true
end

--- Replaces the first element found that returns `true` for the given `predicate` with the `replacement` element.
--- If no element is replaced, this method does nothing.
--- @param replacement T The replacement element
--- @param predicate fun(element: T): boolean A predicate that should return `true` for the element to replace
--- @return self This array
function Array:replaceBy(replacement, predicate)
  for i, v in ipairs(self.elements) do
    if predicate(v) then
      self.elements[i] = replacement
      return self
    end
  end

  return self
end

--- Removes all elements from this array
function Array:clear()
  self.elements = {}
  self.len = 0
end

--- @return Array<T> A shallow copy of the array
function Array:clone()
  return self:map(function(element) return element end)
end

--- @return T[] The primitive array from this array object
function Array:toPrimitive()
  local array = {}
  for i = 1, self.len do table.insert(array, self:nth(i)) end
  return array
end

--- @return string The string representation of this array
function Array:toString()
  local str = "{"
  for i = 1, self.len do
    str = str..self.elements[i]..","
  end
  -- Remove trailing ','
  return str:sub(1, -2).."}"
end

--- __Constructor__
--- @generic S
--- @param elements? S[]
--- @return Array<S> The constructed instance
function Array:new(elements)
  local self = --[[--- @type Array<S> ]] setmetatable({}, Array)
  self.elements = elements or {}
  self.len = #self.elements
  return self
end

return Array
