local classes = require("lua/explosionsNG/helpers/classes")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local Array = require("lua/explosionsNG/misc/Array")
local RectangularArea = require("lua/explosionsNG/areas/RectangularArea")

--- @alias Axis "x" | "y" | "z"

--- A bounding box that can be expanded to contain a new point and can be rotated around any arbitrary axis.
--- The actual bounds are stored as being axis-aligned (i.e. the rotation is not baked into the bounds).
--- Note that the constructor is private and instances should be created using the `fromPoints` factory method.
--- @class BoundingBox
--- @field protected min LuaVec3 The minimum coordinates of the bounding box
--- @field protected max LuaVec3 The maximum coordinates of the bounding box
--- @field protected centre LuaVec3 The centre of the bounding box
--- @field protected rotation LuaQuat The rotation of the bounding box
local BoundingBox = classes.newClass()

--- __Static__
---
--- Creates a new `BoundingBox` and sets its `centre` and `position` to those given. This method iterates through the
--- given `points` ands add them all to the bounding box.
--- This method then returns the created `BoundingBox`.
--- The given points __should__ be rotated about the `centre` with the given `rotation` quaternion.
--- @param points Array<LuaVec3> The points to add to the bounding box
--- @param centre LuaVec3 The centre point of the bounding box
--- @param rotation LuaQuat The rotation, relative to the `centre`, that the `points` are rotated about
--- @return BoundingBox The resulting bounding box
function BoundingBox:fromPoints(points, centre, rotation)
  local box = BoundingBox:new()
  box.centre = vec3(centre)
  box.rotation = quat(rotation)

  points
    :map(function(point) return box:transformPoint(point, true) end)
    :forEach(function(point) box:expand(point) end)

  return box
end

--- Transforms the given `point` by rotating the point around the `rotation` and translating using the `centre`.
--- If `toRelative` is true, this method assumes the `point` is in the absolute coordinate system and transforms it into
--- a point that is in the relative coordinate system with respect to this box as an axis-aligned box.
--- @param point LuaVec3 The point to transform
--- @param toRelative? boolean Whether to transform the point into the relative coordinate system. Defaults to false
--- @return LuaVec3 The transform point
--- @private
function BoundingBox:transformPoint(point, toRelative)
  return RectangularArea:transformPoint(self.centre, self.rotation, point, toRelative)
end

--- Updates the `min` or `max` boundary along the given `axis` it the given `value` is lesser or greater respectively
--- @param value number The value to consider
--- @param axis Axis The axis along which to compare the `value` to
--- @private
function BoundingBox:updateAxisBounds(value, axis)
  if value < self.min[axis] then self.min[axis] = value end
  if value > self.max[axis] then self.max[axis] = value end
end

--- For each axis of the `point`, this method calls `updateAxisBounds`
--- @param point LuaVec3 The point to update the bounds with
function BoundingBox:expand(point)
  self:updateAxisBounds(point.x, "x")
  self:updateAxisBounds(point.y, "y")
  self:updateAxisBounds(point.z, "z")
end

--- Returns the closest axis value along this bounding box (in the relative coordinate system) with respect to a
--- sphere's centre position.
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @param axis Axis The axis to find the closest value to
--- @return number The closest value along the `axis` to the sphere `centre`
--- @private
function BoundingBox:findClosestAxisToSphere(centre, axis)
  local min = self.min[axis]
  local max = self.max[axis]
  local sphere = centre[axis]

  if sphere < min then return min end
  if max < sphere then return max end
  return sphere -- Closest point to a sphere that is not the min or max will always be that of the sphere's value
end

--- Returns the closest point on this bounding box (in the absolute coordinate system) with respect to a sphere's centre
--- position.
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @return LuaVec3 The closest point to the sphere `centre`
function BoundingBox:findClosestPointToSphere(centre)
  -- Convert sphere so that it is in the relative coordinate system of both
  local relativeCentre = self:transformPoint(centre, true)
  local closestPoint = vec3(
      self:findClosestAxisToSphere(relativeCentre, 'x'),
      self:findClosestAxisToSphere(relativeCentre, 'y'),
      self:findClosestAxisToSphere(relativeCentre, 'z')
  )
  -- Convert point back to the absolute coordinate system
  return self:transformPoint(closestPoint)
end

--- Returns whether a given sphere intersects this bounding box
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @param radius number The radius of the sphere
--- @return boolean Whether the sphere intersects with this bounding box
function BoundingBox:isIntersectingSphere(centre, radius)
  local closestPoint = self:findClosestPointToSphere(centre)
  return closestPoint:squaredDistance(centre) < radius ^ 2
end

--- @param rotation LuaQuat The new rotation to set for the bounding box
function BoundingBox:changeRotation(rotation)
  self.rotation = quat(rotation)
end

--- @param centre LuaVec3 The new centre to set for the bounding box
function BoundingBox:changeCentre(centre)
  self.centre = vec3(centre)
end

--- @return Array<LuaVec3> The list of all 8 corners of this bounding box in the absolute coordinate system
--- @private
function BoundingBox:getCorners()
  return Array:new(--[[--- @type LuaVec3[] ]] {
    vec3(self.max),
    vec3(self.max.x, self.max.y, self.min.z),
    vec3(self.max.x, self.min.y, self.max.z),
    vec3(self.min.x, self.max.y, self.max.z),
    vec3(self.min.x, self.min.y, self.max.z),
    vec3(self.min.x, self.max.y, self.min.z),
    vec3(self.max.x, self.min.y, self.min.z),
    vec3(self.min),
  }):map(function(corner) return self:transformPoint(corner) end)
end

--- Displays all 8 corners of the bounding box as black spheres
function BoundingBox:debugDisplay()
  self:getCorners():forEach(function(corner)
    worldObjectsHandler.requestTSStatic({ modelName = "sphere", position = corner, instanceColour = "0 0 0 1" })
  end)
end

--- __Constructor__
--- @return BoundingBox The constructed instance
--- @private
function BoundingBox:new()
  local self = classes.setmetatable({}, BoundingBox)
  self.min = vec3(math.huge, math.huge, math.huge)
  self.max = vec3(-math.huge, -math.huge, -math.huge)
  self.centre = vec3()
  self.rotation = quat()
  return self
end

return BoundingBox
