local classes = require("lua/explosionsNG/helpers/classes")
local vehicles = require("lua/explosionsNG/handlers/vehicles")

--- @alias VehicleEffectEvent { position: LuaVec3, vehicle: Vehicle }
--- @alias VehicleEffectProducer fun(id: string, event: VehicleEffectEvent): VehicularEffect @A function that returns a new `VehicleEffect` instance to apply to a vehicle that takes in an `id` of an object that is performing the affection, and additional event context

--- __Abstract__
---
--- A class extending this one will be one which can affect a vehicle.
--- This class comes with two abstract methods `canAffect` and `affect` (see method documentation for type information).
--- When implementing `affect`, for any given `Vehicle` it will be applied to, it is expected that `canAffect` will
--- return `true` when called with the same `Vehicle`.
--- @class VehicleAffectable
--- @field protected id string The unique identifier of the instance
--- @field protected nextId number __Static__ The next available numerical id for an instance (acts as a zero-based counter)
local VehicleAffectable = classes.newClass()
VehicleAffectable.nextId = 0

--- __Abstract__
--- @param vehicle Vehicle The vehicle to check
--- @return boolean Whether the `vehicle` can be affected by this object
--- @protected
function VehicleAffectable:canAffect(vehicle)
  error("canAffect is not implemented")
  return false
end

--- __Abstract__
--- @param vehicle Vehicle The vehicle to affect
--- @return VehicularEffect The effect that can be applied to a vehicle
--- @protected
function VehicleAffectable:affect(vehicle)
  error("affect is not implemented")
  return nil
end

--- __Static__
---
--- Returns the next available id prepended with the given `classId` and increments the `nextId` static field
--- @param classId string The unique identifier for the class generating this id
--- @return string The next available id prepended with `classId`
--- @private
function VehicleAffectable:generateId(classId)
  local nextId = self.nextId
  VehicleAffectable.nextId = nextId + 1
  return classId..nextId
end

--- For the given `vehicle`, this method checks whether `isAffectedBy` on the `vehicle` returns false and that
--- `canAffect` with the `vehicle` returns true. If this is the case, this method calls `affect` and calls `addEffect`
--- on the `vehicle` with the result of the former.
--- @param vehicle Vehicle The vehicle to try and affect
--- @protected
function VehicleAffectable:tryAndAffectVehicle(vehicle)
  if not vehicle:isAffectedBy(self.id) and self:canAffect(vehicle) then
    vehicle:addEffect(self:affect(vehicle))
  end
end

--- For each vehicle retrieved from the vehicles handler, this method calls `tryAndAffectVehicle`
--- @protected
function VehicleAffectable:tryAndAffectVehicles()
  vehicles.getVehicles():forEach(function(vehicle) self:tryAndAffectVehicle(vehicle) end)
end

--- For each vehicle retrieved from the vehicles handler, this method calls `removeEffect` on them with this object's
--- `id`.
--- @protected
function VehicleAffectable:removeAllEffects()
  vehicles.getVehicles():forEach(function(vehicle) vehicle:removeEffect(self.id) end)
end

--- __Constructor__
--- @return VehicleAffectable The constructed instance
function VehicleAffectable:new(classId)
  local self = classes.setmetatable({}, VehicleAffectable)
  self.id = VehicleAffectable:generateId(classId)
  return self
end

return VehicleAffectable
