local classes = require("lua/explosionsNG/helpers/classes")

--- @class CircularPath : Path
--- @field private centre LuaVec3 The centre point of the circular path
--- @field private radius number The radius of the path
--- @field private rotation number The current rotation angle (in radians)
local CircularPath = classes.newClass()

--- Returns a new position along the path that has the given rotation and radius
--- @param radius number The radius
--- @param rotation number The rotation (in radians)
--- @return LuaVec3 The position
--- @private
function CircularPath:getPositionFromRotation(radius, rotation)
  local rCos = radius * math.cos(rotation)
  local rSin = radius * math.sin(rotation)
  return self.centre + vec3(rCos, rSin, 0)
end

--- @return LuaVec3 the next position along this path
function CircularPath:getCurrentPosition()
  return self:getPositionFromRotation(self.radius, self.rotation)
end

--- @return LuaVec3 The position on the of the rotation from the current rotation, plus the given additional rotation
function CircularPath:getPositionAtOffsetRotation(rotation)
  return self:getPositionFromRotation(self.radius, self.rotation + rotation)
end

--- Adds the given increment to the current rotation
--- @param rotationIncrement number The increment to add
function CircularPath:addRotation(rotationIncrement)
  self.rotation = self.rotation + rotationIncrement
end

--- Sets the current radius of the path to the given one
--- @param radius number The new radius to set
function CircularPath:setRadius(radius)
  self.radius = radius
end

--- __Constructor__
--- @param centre LuaVec3 The centre point of the circular path
--- @param radius number The radius of the path
--- @param initialRotation? number The initial rotation along the path. Defaults to `0`
--- @return CircularPath The constructed instance
function CircularPath:new(centre, radius, initialRotation)
  local self = classes.setmetatable({}, CircularPath)
  self.centre = vec3(centre)
  self.radius = radius
  self.rotation = initialRotation or 0
  return self
end

return CircularPath
