local classes = require("lua/explosionsNG/helpers/classes")

--- @class LinearPath : Path
--- @field private startPoint LuaVec3 The starting point on the path
--- @field private endPoint LuaVec3 The ending point on the path
--- @field private distance number The current distance travelled along the path
local LinearPath = classes.newClass()

--- @return number The length of the path
function LinearPath:getLength()
  return (self.endPoint - self.startPoint):length()
end

--- Returns the position that is the given distance along the path from the starting point
--- @param distance number The distance
--- @return LuaVec3 The position that is the given distance along the path from the starting point
--- @private
function LinearPath:getPositionFromDistance(distance)
  local pathDirection = (self.endPoint - self.startPoint):normalized()
  return self.startPoint + pathDirection * distance
end

--- @return LuaVec3 The current position along the path
function LinearPath:getCurrentPosition()
  return self:getPositionFromDistance(self.distance)
end

--- Adds the given increment to the current distance.
--- Note that if adding the increment goes past the end of the path, it will instead be constrained to the end of the
--- path.
--- @param distanceIncrement number The increment to add
function LinearPath:addDistance(distanceIncrement)
  self.distance = math.min(self:getLength(), self.distance + distanceIncrement)
end

--- Resets the distance back to zero again
function LinearPath:reset()
  self.distance = 0
end

--- @return boolean Whether the current point along the path is at the end point
function LinearPath:hasReachedEndPoint()
  return self.distance == self:getLength()
end

--- __Constructor__
--- @param startPoint LuaVec3 The start point of the path
--- @param endPoint LuaVec3 The end point of the path
--- @param initialDistance? number The initial instance travelled on the path. Defaults to `0`
--- @return LinearPath The constructed instance
function LinearPath:new(startPoint, endPoint, initialDistance)
  local self = classes.setmetatable({}, LinearPath)
  self.startPoint = vec3(startPoint)
  self.endPoint = vec3(endPoint)
  self.distance = initialDistance or 0
  return self
end

return LinearPath
