local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local collisions = require("lua/explosionsNG/helpers/collisions")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local Timer = require("lua/explosionsNG/timers/Timer")
local SoundEmitter = require("lua/explosionsNG/worldObjects/SoundEmitter")

--- @alias PayloadFn fun(position: LuaVec3): AnyExplosive

--- @class LoadedProjectile
--- @field velocity LuaVec3 The velocity of the projectile
--- @field position LuaVec3 The position of the projectile
--- @field object SceneObject The in-game world object for this projectile
--- @field soundEmitter SoundEmitter The sound emitter to emit sounds during the launch of the projectile
--- @field timeout Timer A timer that will forcefully treat this projectile as colliding, should it not collide within the `TIMEOUT` period
--- @field collided boolean Whether the projectile has collided with something
--- @field hasPayloadFinished boolean Whether the projectile has detonated and finished detonating its payload
--- @field payload AnyExplosive The `Explosive` payload that this projectile will detonate upon collision
--- @field payloadFn PayloadFn A producer that returns a new `Explosive` from the collision `position`
--- @field COLLISION_TOLERANCE number __Static__ The tolerance at which a projectile must be within to be considered colliding with terrain or another object
--- @field TIMEOUT number __Static__ The expiration time of the projectile (as a safeguard against the projectile never hitting anything and existing forever)
local LoadedProjectile = classes.newClass()
LoadedProjectile.COLLISION_TOLERANCE = 0.5
LoadedProjectile.TIMEOUT = 30

--- Adds the environmental gravity to this projectiles `velocity`. This method will update the `position` with the
--- new velocity, scaled by the `dt` time. This method will then finally update the in-game object's position with the
--- new position.
--- @param dt number The time since the last frame
--- @private
function LoadedProjectile:updatePosition(dt)
  self.velocity = self.velocity + vec3(0, 0, core_environment.getGravity()) * dt
  self.position = self.position + self.velocity * dt

  self.object:setField("position", 0, vectors.toString(self.position))
end

--- Checks and returns whether the projectile is colliding with a TSStatic object. This method will return `true` if
--- there is a static object within the tolerance (see `COLLISION_TOLERANCE`).
--- @return boolean Whether the projectile is colliding with a TSStatic object
--- @private
function LoadedProjectile:isCollidingWithStaticObject()
  -- Important: The final parameter should be greater than the size of the projectile so that the projectile does not get picked up by the ray cast
  return collisions.isCollidingWithStaticObject(self.position, self.velocity:normalized(), LoadedProjectile.COLLISION_TOLERANCE, 1)
end

--- Creates a new `explosive` from the `payloadFn` and sets up the `payload` before activating it. This method will then
--- also release the in-game object to the world objects handler.
--- @private
function LoadedProjectile:onCollision()
  self.payload = self.payloadFn(self.position)
  self.payload:registerOnFinishedHandler(function() self.hasPayloadFinished = true end)
  self.payload:activate()

  self.collided = true

  worldObjectsHandler.releaseTSStatic(self.object)
  self.object = nil
end

--- If the projectile has `collided`, then this method will update the `payload`.
---
--- Otherwise, this method updates the trajectory of the projectile according to the amount of `dt` time elapsed. This
--- method will then check whether the projectile is colliding with either the floor, or a static object. If it is, then
--- this method will then begin the collision process (see `onCollision`).
---
--- The expiry timer will also be updated, if the projectile has not collided still.
--- @param dt number The time since the last frame
--- @private
function LoadedProjectile:update(dt)
  if not self.soundEmitter:isFinished() then
    self.soundEmitter:update(dt)
  end

  if self.hasPayloadFinished then return end

  if self.collided then
    self.payload:update(dt)
    return
  end

  self:updatePosition(dt)

  -- Important: Give precedence to the terrain calculation as it is significantly cheaper than checking for static collisions
  if collisions.isCollidingWithTerrain(self.position) or self:isCollidingWithStaticObject() then
    self:onCollision()
    return
  end

  self.timeout:update(dt)
end

--- @return boolean Whether this projectile has collided and its payload has finished its effects
function LoadedProjectile:isFinished()
  return self.hasPayloadFinished and self.soundEmitter:isFinished()
end

--- Releases all world objects, destroys the payload (if it is active), and sets this instance to be destroyed
function LoadedProjectile:destroy()
  if self.hasPayloadFinished then return end

  self.hasPayloadFinished = true

  worldObjectsHandler.releaseTSStatic(self.object)
  self.object = nil

  self.soundEmitter:destroy()

  if self.payload then
    self.payload:deactivate()
  end
end

--- __Constructor__
--- @param position LuaVec3 The initial position of the projectile
--- @param velocity LuaVec3 The initial velocity of the projectile
--- @param payloadFn PayloadFn A producer that should provide a new `Explosive` from the collision `position`, and that will be detonated upon collision
--- @return LoadedProjectile The constructed instance
function LoadedProjectile:new(position, velocity, payloadFn)
  local self = classes.setmetatable({}, LoadedProjectile)
  self.velocity = vec3(velocity)
  self.position = vec3(position)
  self.payloadFn = payloadFn
  self.object = worldObjectsHandler.requestTSStatic({ modelName = "projectile_sphere", position = self.position, scale = 0.5 })
  self.soundEmitter = SoundEmitter:new({ position = self.position, track = "projectile_launch" })
  self.timeout = Timer:new(LoadedProjectile.TIMEOUT, function() self:onCollision() end, true)
  self.payload = nil
  self.hasPayloadFinished = false
  self.collided = false
  return self
end

return LoadedProjectile
