local classes = require("lua/explosionsNG/helpers/classes")
local maths = require("lua/explosionsNG/helpers/maths")
local Array = require("lua/explosionsNG/misc/Array")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")
local LoadedProjectile = require("lua/explosionsNG/projectiles/LoadedProjectile")

--- Properties:
---  - `ejectionMultiplier`  - the multiplier of the velocity of the projectiles that are thrown out of this group
---  - `numberOfProjectiles` - the number of projectiles that will be created by this group
---  - `interval`            - the time between firing two projectiles in succession
--- @alias ProjectileVolcanoOptions { ejectionMultiplier: number | nil, numberOfProjectiles: number | nil, interval: number | nil }

--- @class ProjectileVolcano
--- @field position LuaVec3 The position of which projectiles will start at
--- @field options ProjectileVolcanoOptions The list of configured options for this group
--- @field payloadProducer PayloadFn A function for returning a new payload for a new projectile
--- @field finishFn fun(): void A callback function to call when this group is finished
--- @field projectiles Array<LoadedProjectile> The list of projectiles in action
--- @field projectileTimer IntervalTimer The interval timer that fires a new projectile at set intervals
--- @field finished boolean Whether the group has finished
--- @field MAX_VELOCITY_XY number The maximum velocity of a projectile in the x or y direction
--- @field VELOCITY_Z number The initial velocity on the z axis of a new projectile
local ProjectileVolcano = classes.newClass()
ProjectileVolcano.MAX_VELOCITY_XY = 10
ProjectileVolcano.VELOCITY_Z = 50

--- __Static__
---
--- Fills in any key not in the given `options` with default values
--- @param options ProjectileVolcanoOptions The list of predefined options
--- @return ProjectileVolcanoOptions The complete set of options with defaults filled in where empty values are given
function ProjectileVolcano:fillOptions(options)
  return {
    ejectionMultiplier = options.ejectionMultiplier or 1,
    numberOfProjectiles = options.numberOfProjectiles or 5,
    interval = options.interval or 0.3
  }
end

--- __Static__
---
--- Note that this assumes `g = -9.81`, the ground is flat and the initial `VELOCITY_Z` is positive
--- @return number The total time a projectile will exist in the air
function ProjectileVolcano:getTotalProjectileAirTime()
  return math.abs(ProjectileVolcano.VELOCITY_Z / -9.81) * 2
end

--- __Static__
--- @return number The maximal range a projectile can travel. Note some assumptions by te `getTotalProjectileAirTime` method
--- @see ProjectileVolcano#getTotalProjectileAirTime
function ProjectileVolcano:getMaximumRange()
  return ProjectileVolcano.MAX_VELOCITY_XY * ProjectileVolcano:getTotalProjectileAirTime()
end

--- __Static__
--- @return number A random XY component of a velocity for the projectile
--- @private
function ProjectileVolcano:getRandomVelocityXY()
  return maths.randomSignedFloat(ProjectileVolcano.MAX_VELOCITY_XY)
end

--- __Static__
---
--- Creates and returns a new `LoadedProjectile` that starts at the given `position` and has a random velocity upwards.
--- @param position LuaVec3 The position at which the projectile start start at
--- @param payloadFn PayloadFn A function that should return a new payload for the created projectile
--- @param ejectionMultiplier number The multiplier to apply to this projectile's initial velocity
--- @return LoadedProjectile The created projectile
function ProjectileVolcano:createProjectile(position, payloadFn, ejectionMultiplier)
  local velocity = vec3(
      ProjectileVolcano:getRandomVelocityXY(),
      ProjectileVolcano:getRandomVelocityXY(),
      ProjectileVolcano.VELOCITY_Z
  ) * ejectionMultiplier

  return LoadedProjectile:new(position, velocity, payloadFn)
end

--- Creates a new `LoadedProjectile` and adds it to the current list of `projectiles`. For details on the projectile
--- itself, see `createProjectile`.
--- @private
function ProjectileVolcano:fireProjectile()
  local projectile = ProjectileVolcano:createProjectile(self.position, self.payloadProducer, self.options.ejectionMultiplier)
  self.projectiles:add(projectile)
end

--- If this group is not finished, then this method will continue to throw out new projectiles until all projectiles
--- have been created. This method will update any existing projectiles and will set the `finished` flag once all
--- projectiles have finished.
--- @param dt number The time since the last frame
function ProjectileVolcano:update(dt)
  if self:isFinished() then return end

  if self.projectiles:length() < self.options.numberOfProjectiles then
    self.projectileTimer:update(dt)
  end

  self.projectiles:forEach(function(projectile) projectile:update(dt) end)

  if self.projectiles:all(function(projectile) return projectile:isFinished() end) then
    self.finished = true
  end
end

--- Starts the interval to create new projectiles and fires the first projectile
function ProjectileVolcano:erupt()
  self.projectileTimer:start()
  self:fireProjectile()
end

--- @return boolean Whether this group has been started and has subsequently finished
function ProjectileVolcano:isFinished()
  return self.projectileTimer:hasStarted() and self.finished
end

--- Destroys all projectiles and any active payloads
function ProjectileVolcano:destroy()
  self.projectiles:forEach(function(projectile) projectile:destroy() end)
end

--- __Constructor__
--- @param position LuaVec3
--- @param options? ProjectileVolcanoOptions | nil
--- @return ProjectileVolcano The constructed instance
function ProjectileVolcano:new(position, payloadProducer, options)
  local self = classes.setmetatable({}, ProjectileVolcano)
  self.options = ProjectileVolcano:fillOptions(options or {})
  self.position = vec3(position)
  self.payloadProducer = payloadProducer
  self.projectiles = Array:new(--[[--- @type LoadedProjectile[] ]] {})
  self.projectileTimer = IntervalTimer:new(self.options.interval, function() self:fireProjectile() end)
  self.finished = false
  return self
end

return ProjectileVolcano
