local tables = require("lua/explosionsNG/helpers/tables")
local Array = require("lua/explosionsNG/misc/Array")

local ShockwaveExplosive = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosive")
local ShockwaveExplosiveStats = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosiveStats")
local AnomalyExplosive = require("lua/explosionsNG/explosives/contained/anomaly/AnomalyExplosive")
local AnomalyExplosiveStats = require("lua/explosionsNG/explosives/contained/anomaly/AnomalyExplosiveStats")
local ClusterExplosive = require("lua/explosionsNG/explosives/contained/cluster/ClusterExplosive")
local ClusterExplosiveStats = require("lua/explosionsNG/explosives/contained/cluster/ClusterExplosiveStats")
local LightningStormExplosive = require("lua/explosionsNG/explosives/contained/lightningStorm/LightningStormExplosive")
local LightningStormExplosiveStats = require("lua/explosionsNG/explosives/contained/lightningStorm/LightningStormExplosiveStats")
local LandMineExplosive = require("lua/explosionsNG/explosives/contained/landMine/LandMineExplosive")
local LandMineExplosiveStats = require("lua/explosionsNG/explosives/contained/landMine/LandMineExplosiveStats")
local SupernovaExplosive = require("lua/explosionsNG/explosives/contained/supernova/SupernovaExplosive")
local SupernovaExplosiveStats = require("lua/explosionsNG/explosives/contained/supernova/SupernovaExplosiveStats")
local ConvergenceExplosive = require("lua/explosionsNG/explosives/contained/convergence/ConvergenceExplosive")
local ConvergenceExplosiveStats = require("lua/explosionsNG/explosives/contained/convergence/ConvergenceExplosiveStats")
local GravityOrbsExplosive = require("lua/explosionsNG/explosives/contained/gravityOrbs/GravityOrbsExplosive")
local GravityOrbsExplosiveStats = require("lua/explosionsNG/explosives/contained/gravityOrbs/GravityOrbsExplosiveStats")
local OrbitalStrikeExplosive = require("lua/explosionsNG/explosives/contained/orbitalStrike/OrbitalStrikeExplosive")
local OrbitalStrikeExplosiveStats = require("lua/explosionsNG/explosives/contained/orbitalStrike/OrbitalStrikeExplosiveStats")
local BlackHoleExplosive = require("lua/explosionsNG/explosives/contained/blackHole/BlackHoleExplosive")
local BlackHoleExplosiveStats = require("lua/explosionsNG/explosives/contained/blackHole/BlackHoleExplosiveStats")

-- Intelligent type checking with all these generics has not been implemented as luanalysis is nowhere near intelligent enough for it :-)

local EXPLOSIVE_DETAILS = --[[--- @type table<string, { explosive: any, stats: any }> ]] {
  shockwave = {
    explosive = ShockwaveExplosive,
    stats = ShockwaveExplosiveStats
  },
  gravityOrbs = {
    explosive = GravityOrbsExplosive,
    stats = GravityOrbsExplosiveStats
  },
  cluster = {
    explosive = ClusterExplosive,
    stats = ClusterExplosiveStats
  },
  lightningStorm = {
    explosive = LightningStormExplosive,
    stats = LightningStormExplosiveStats
  },
  orbitalStrike = {
    explosive = OrbitalStrikeExplosive,
    stats = OrbitalStrikeExplosiveStats
  },
  landMine = {
    explosive = LandMineExplosive,
    stats = LandMineExplosiveStats
  },
  anomaly = {
    explosive = AnomalyExplosive,
    stats = AnomalyExplosiveStats
  },
  blackHole = {
    explosive = BlackHoleExplosive,
    stats = BlackHoleExplosiveStats
  },
  supernova = {
    explosive = SupernovaExplosive,
    stats = SupernovaExplosiveStats
  },
  convergence = {
    explosive = ConvergenceExplosive,
    stats = ConvergenceExplosiveStats
  }
}

--- @return Array<string> The list of explosive ids
local function getExplosives()
  return Array:new(tables.keys(EXPLOSIVE_DETAILS))
end

--- Returns the stats instance for a given preset of an explosive
--- @param explosiveId string The id of the explosive
--- @param presetId string The id (preset) of the preset of the explosive
--- @return ExplosiveStats<any> The stats of the preset for the explosive
local function getExplosiveStatsInstance(explosiveId, presetId)
  local detail = EXPLOSIVE_DETAILS[explosiveId]
  local options = detail.explosive:getPresetOptions(presetId)
  return --[[--- @type ExplosiveStats<any> ]] detail.stats:new(options)
end

--- Returns the stats for a given preset of an explosive
--- @param explosiveId string The id of the explosive
--- @param presetId string The id (preset) of the preset of the explosive
--- @return PresetStats The stats of the preset for the explosive
local function getAbsoluteExplosivePresetStats(explosiveId, presetId)
  local stats = getExplosiveStatsInstance(explosiveId, presetId)
  return stats:getAbsoluteStats()
end

--- Returns the stats for a given preset of an explosive, scaled to values between [0, 1]
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @return PresetStats The stats of the preset for the explosive
local function getExplosivePresetStats(explosiveId, presetId)
  local stats = getExplosiveStatsInstance(explosiveId, presetId)
  return stats:getNormalisedStats()
end

--- Returns the list of presents for an explosive, given its id
--- @param explosiveId string The id of the explosive
--- @return Array<string> The preset ids of the explosive
local function getPresetsForExplosive(explosiveId)
  local detail = EXPLOSIVE_DETAILS[explosiveId]
  return Array:new(--[[--- @type string[] ]] detail.explosive:getPresets())
end

--- Creates and returns the explosive given its corresponding ids and the position to place it at
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @param fuse? number A delay to apply to the return explosive between it being active and it detonating
--- @return AnyExplosive The created explosive
local function createNewExplosive(explosiveId, presetId, position, fuse)
  local detail = EXPLOSIVE_DETAILS[explosiveId]
  local explosive = --[[--- @type AnyExplosive ]] detail.explosive:new(position, { preset = presetId })

  if fuse then
    explosive:setFuse(fuse)
  end

  return explosive
end

--- Returns the bounding area of objects that will be created for a given explosive and preset. This function will
--- return `nil` for explosives that do not create long-lasting objects as part of their explosions.
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @return BoundingArea | nil The bounding area range of the explosive. `nil` if there is no bounding area for the explosive.
local function getBoundingArea(explosiveId, presetId)
  -- TODO Abusive typecasting - not really a good way around it though as this needs to be generic; maybe just checking the ids are sensible?
  if explosiveId == "anomaly"     then return     AnomalyExplosive:getBoundingArea(--[[--- @type AnomalyExplosivePreset     ]] presetId) end
  if explosiveId == "blackHole"   then return   BlackHoleExplosive:getBoundingArea(--[[--- @type BlackHoleExplosivePreset   ]] presetId) end
  if explosiveId == "supernova"   then return   SupernovaExplosive:getBoundingArea(--[[--- @type SupernovaExplosivePreset   ]] presetId) end
  if explosiveId == "convergence" then return ConvergenceExplosive:getBoundingArea(--[[--- @type ConvergenceExplosivePreset ]] presetId) end
  return nil
end

return {
  getExplosives = getExplosives,
  getAbsoluteExplosivePresetStats = getAbsoluteExplosivePresetStats,
  getPresetsForExplosive = getPresetsForExplosive,
  createNewExplosive = createNewExplosive,
  getBoundingArea = getBoundingArea,
  getExplosivePresetStats = getExplosivePresetStats
}