local classes = require("lua/explosionsNG/helpers/classes")

--- @class Timer
--- @field private active boolean Whether the timer is active
--- @field private time number The total time elapsed
--- @field private timeLimit number The time limit of the timer
--- @field private callback fun(): void The callback function to call once the timer is finished
local Timer = classes.newClass()

--- Calls the `callback` function and sets this timer to be inactive
--- @protected
function Timer:onTimerFinished()
  self.active = false
  self.callback()
end

--- Adds the given `time` to the timer's `time`, ensuring that the resulting increment does not increase beyond the
--- `timeLimit`
--- @param time number The time to increment by
--- @private
function Timer:increaseTime(time)
  self.time = math.min(self.timeLimit, self.time + time)
end

--- @return boolean Whether the timer has finished
function Timer:isComplete()
  return self.time >= self.timeLimit
end

--- @return boolean Whether the time is `active`
function Timer:hasStarted()
  return self.active
end

--- Completes the timer
function Timer:complete()
  self.time = self.timeLimit
end

--- If this timer is not `active` or `isComplete`, this method does nothing.
--- Other, this method calls `increaseTime`.
--- If the timer `isComplete` now, this method calls `onTimerFinished`.
--- @param dt number The time since the last frame
function Timer:update(dt)
  if not self.active then return end

  if self:isComplete() then
    -- If the timer has a zero-timer, then it will never pass zero seconds
    if self.timeLimit == 0 then
      self:onTimerFinished()
    end

    return
  end

  self:increaseTime(dt)

  if self:isComplete() then self:onTimerFinished() end
end

--- @return number The ratio of `time` to `timeLimit`, as a number between 0 and 1 where a newly started timer has value `0`. This is the inverse of `timeLeftRatio`
function Timer:elapsedRatio()
  return math.min(1, self.time / self.timeLimit)
end

--- @return number The ratio of `time` to `timeLimit`, as a number between 0 and 1 where a newly started timer has value `1`. This is the inverse of `elapsedRatio`
function Timer:timeLeftRatio()
  return 1 - self:elapsedRatio()
end

--- @return number The time elapsed
function Timer:elapsedTime()
  return self.time
end

--- @return number The time left
function Timer:timeLeft()
  return self.timeLimit - self:elapsedTime()
end

--- Starts the timer
function Timer:start()
  self.active = true
end

--- Stops the timer
function Timer:stop()
  self.active = false
end

-- Sets the `time` back to `0` and starts the timer
function Timer:restart()
  self.time = 0
  self:start()
end

--- __Constructor__
--- @param timeLimit number The time limit of the timer
--- @param callback fun(): void A callback function to call after the timer has finished
--- @param startImmediately? boolean Whether to start the timer immediately
--- @return Timer The constructed instance
function Timer:new(timeLimit, callback, startImmediately)
  local self = classes.setmetatable({}, Timer)
  self.active = startImmediately
  self.time = 0
  self.timeLimit = timeLimit
  self.callback = callback
  return self
end

return Timer
