local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local VehicleNodes = require("lua/explosionsNG/vehicles/VehicleNodes")
local VehicleBounds = require("lua/explosionsNG/vehicles/VehicleBounds")
local VehicleDissolver = require("lua/explosionsNG/vehicles/VehicleDissolver")
local VehiclePlanet = require("lua/explosionsNG/vehicles/VehiclePlanet")

--- @class Vehicle
--- @field private id number The in-game id of the vehicle
--- @field private name string The in-game name of the vehicle
--- @field private object VehicleSceneObject The in-game object representation the vehicle
--- @field private effects Array<VehicularEffect> The list of status effects affecting this vehicle
--- @field private bounds VehicleBounds The bounds of the vehicle. Lazy instantiated, as the vehicle isn't always aligned by TorqueScript straight away
--- @field private dissolver VehicleDissolver A dissolver to dissolve parts of the vehicle
--- @field private planets Array<VehiclePlanet> A list of active beamstate planets on this vehicle
local Vehicle = classes.newClass()

--- @return number The in-game id of the vehicle
function Vehicle:getId()
	return self.id
end

--- @return VehicleSceneObject The vehicle game object
function Vehicle:getGameObject()
	return self.object
end

--- @return LuaVec3 The vehicle's current position
function Vehicle:getPosition()
	return vec3(self.object:getPosition())
end

--- @return LuaQuat The vehicle's current rotation
function Vehicle:getDirection()
	local direction = vec3(self.object:getDirectionVector())
	local upDirection = vec3(self.object:getDirectionVectorUp())
	return quatFromDir(direction, upDirection)
end

--- @return LuaVec3 The velocity vector of the vehicle
function Vehicle:getVelocity()
	return self.object:getVelocity()
end

--- @return number The speed oft he vehicle in ms^-1
function Vehicle:getSpeed()
	return self:getVelocity():length()
end

--- @return number The Id of the reference node of this vehicle
--- @private
function Vehicle:getRefId()
	return self.object:getRefNodeId()
end

--- @return LuaVec3 The position of the reference node of this vehicle
--- @private
function Vehicle:getReferenceNodePosition()
	return vec3(self.object:getNodePosition(self.object:getRefNodeId()))
end

--- @return boolean Whether the vehicle has been removed from the level
function Vehicle:isRemovedFromLevel()
	return not be:getObjectByID(self:getId())
end

--function Vehicle:setPositionRotation()
--	local refNodeId = self:getRefId()
--	--local velocity = self:getVelocity()
--	local position = self:getPosition()
--	local rotation = self:getDirection():inversed()
--	self.object:setClusterPosRelRot(refNodeId, position.x, position.y, position.z, rotation.x, rotation.y, rotation.z, rotation.w)
--	--self.object:applyClusterVelocityScaleAdd(refNodeId, 0, velocity.x, velocity.y, velocity.z)
--end

--- Checks and returns whether the vehicle is within `distance` metres of the iven `position`
--- @param position LuaVec3 The position to check the vehicles distance to
--- @param distance number The maximum distance the vehicle can be within
--- @return boolean Whether the vehicle is within `distance` metres to `position`
function Vehicle:isWithin(position, distance)
	return self:getPosition():distance(position) <= distance
end

--- Queues up the given Lua command to the vehicle `object`
--- @param command string The Lua command to queue up
--- @private
function Vehicle:queueLuaCommand(command)
	self.object:queueLuaCommand(command)
end

--- Queues up the given Lua command as a part of the game engine to the vehicle `object`.
--- This wraps the `command` in a call to `queueGameEngineLua`.
--- @param command string The game engine Lua command to queue up
--- @private
function Vehicle:queueGameEngineLuaCommand(command)
	self:queueLuaCommand("obj:queueGameEngineLua("..command..")")
end

--- If `bounds` is `nil`, this method creates a `VehicleBounds` object and sets the value of `bounds` with it.
--- Otherwise, this method calls `updateFromVehicle` on the `bounds` with the vehicle's current position and direction.
--- @private
function Vehicle:updateBounds()
	local position = self:getPosition()
	local direction = self:getDirection()

	if not self.bounds then
		local nodePositions = VehicleNodes:getAllNodePositions(self.object)
		self.bounds = VehicleBounds:new(nodePositions, position, direction)
		return
	end

	self.bounds:updateFromVehicle(position, direction)
end

--- Returns whether the given `sphere` intersects within this vehicle's `bounds`
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @param radius number The radius of the sphere
--- @return boolean Whether the `sphere` intersects within this vehicle's `bounds`
function Vehicle:withinBounds(centre, radius)
	-- Ensure bounds are up-to-date and defined
	self:updateBounds()
	return self.bounds:isIntersectingSphere(centre, radius)
end

--- Returns the closest point on this bounding box (in the absolute coordinate system) with respect to a sphere's centre
--- position.
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @return LuaVec3 The closest point to the sphere `centre`
function Vehicle:closestPointToBounds(centre)
	-- Ensure bounds are up-to-date and defined
	self:updateBounds()
	return self.bounds:findClosestPointToSphere(centre)
end

--- Sets the planets on the beamstate part of the vehicle, with the list of planets associated with this vehicle
--- @private
function Vehicle:updateBeamstatePlanets()
	local beamstatePlanetsString = VehiclePlanet:arrayToBeamstateString(self.planets)
	self:queueLuaCommand("beamstate.setPlanets("..beamstatePlanetsString..")")
end

--- Adds the given planet to this vehicle to affect this vehicle
--- @param planet VehiclePlanet The planet to add
function Vehicle:addPlanet(planet)
	self.planets:add(planet)
	self:updateBeamstatePlanets()
end

--- Removes the given beamstate planet. If the vehicle is not being affected by that planet, the vehicle will not be
--- affected any differently than before this method was called.
--- @param planet VehiclePlanet The planet to remove
function Vehicle:removePlanet(planet)
	self.planets:remove(planet)
	self:updateBeamstatePlanets()
end

--- Sets the local gravity of this vehicle, to the given value.
--- @param gravity number The local gravity to set
function Vehicle:setLocalGravity(gravity)
	self:queueLuaCommand("obj:setGravity("..gravity..")")
end

--- Sets all the electronics of this vehicle to off, including the ignition level
function Vehicle:disableElectronics()
	self:queueLuaCommand("electrics.set_warn_signal(0)")
	self:queueLuaCommand("electrics.set_lightbar_signal(0)")
	self:queueLuaCommand("electrics.set_fog_lights(0)")
	self:queueLuaCommand("electrics.setLightsState(0)")
	self:queueLuaCommand("electrics.setIgnitionLevel(0)")
end

--- Ignites a random node on the vehicle
function Vehicle:igniteRandomNode()
	self:queueLuaCommand("fire.igniteRandomNode()")
end

--- Deflates a random tire on the vehicle
function Vehicle:deflateRandomTire()
	self:queueLuaCommand("beamstate.deflateRandomTire()")
end

--- Breaks a random "break group" in the vehicle
function Vehicle:breakRandomBreakGroup()
	self:queueLuaCommand([[
		local breakGroups = {}

		for _, beam in pairs(v.data.beams) do
			local breakGroup = beam.breakGroup

			if breakGroup then
				local nestedBreakGroups = type(breakGroup) == "table" and breakGroup or { breakGroup }
				for _, bGroup in ipairs(nestedBreakGroups) do
					table.insert(breakGroups, bGroup)
				end
			end
		end

		local chosenBreakGroupIndex = math.random(1, #breakGroups)
		local chosenBreakGroup = breakGroups[chosenBreakGroupIndex]
		beamstate.breakBreakGroup(chosenBreakGroup)
	]])
end

--- Performs the next increment of dissolving damage to the vehicle
function Vehicle:dissolve()
	self.dissolver:incrementDissolveDamage()
end

--- Calculates certain values and caches them inside the vehicle environment
--- @private
function Vehicle:cacheValues()
	-- Set up ENG_CENTRE_NODE
	self:queueLuaCommand([[
		local centreNode
		local centreNodeDist = math.huge

		for _, node in pairs(v.data.nodes) do
			local nodeDist = node.pos:length()
			if nodeDist < centreNodeDist then
				centreNodeDist = nodeDist
				centreNode = node.cid
			end
		end

		ENG_CENTRE_NODE = centreNode
	]])
end

--- Returns whether the given object id is currently affecting this vehicle (i.e. at least one `VehicleEffect` has the
--- aforementioned id).
--- @param id string The id of the object
--- @return boolean Whether the given object id is currently affecting this vehicle
function Vehicle:isAffectedBy(id)
	return self.effects:some(function(effect) return effect:isSource(id) end)
end

--- Adds the given `effect` to the list of node effects
--- @param effect VehicularEffect The effect to add
function Vehicle:addEffect(effect)
	self.effects:add(effect)
end

--- Removes any effect from this vehicle's `effects` which have the given `sourceId`, and is an "infinite" effect
--- @param sourceId string The id of the source of the effect(s) to remove
function Vehicle:removeEffect(sourceId)
	--- @type Array<VehicularEffect>
	local newEffects = Array:new({})

	self.effects:forEach(function(effect)
		-- TODO This logic will not work with multi-effects as an effect inside of them may be finite but gets treated as infinite
		local isSourceInfiniteEffect = effect:isInfinite() and effect:isSource(sourceId)

		-- Complete the effect and don't add it to the new list
		if isSourceInfiniteEffect then
			effect:completeInfiniteEffect(self)
			return
		end

		newEffects:add(effect)
	end)

	self.effects = newEffects
end

--- If the `effects` are empty, this method does nothing.
--- Otherwise, calls `affect` with this vehicle for all status effects, and removes all effects that aren't active.
--- @param dt number The time since the last frame
--- @private
function Vehicle:updateVehicleEffects(dt)
	self.effects:forEach(function(effect) effect:affect(self, dt) end)
	self.effects = self.effects:filter(function(effect) return effect:isActive() end)
end

--- Updates the vehicle effects for this vehicle
--- @param dt number The time since the last frame
function Vehicle:update(dt)
	self:updateVehicleEffects(dt)
end

--- Resets the vehicle back to its initial state.
function Vehicle:reset()
	self:updateBounds()
	self.effects:clear()
	self.dissolver:reset()

	self.planets:clear()
	self:updateBeamstatePlanets()
end

--- @return boolean Whether the player is controlling this vehicle
function Vehicle:isPlayer()
	return be:getPlayerVehicleID(0) == self.id
end

--- __Constructor__
--- @param id number The in-game id of the vehicle
--- @param object VehicleSceneObject The in-game object representation the vehicle
--- @return Vehicle The constructed instance
function Vehicle:new(id, object)
	local self = classes.setmetatable({}, Vehicle)
	self.id = id
	self.name = object.name
	self.object = object
	self.effects = Array:new(--[[--- @type VehicularEffect[] ]] {})
	self.bounds = nil
	self.dissolver = VehicleDissolver:new(object)
	self.planets = Array:new({})
	self:cacheValues()
	return self
end

return Vehicle
