local classes = require("lua/explosionsNG/helpers/classes")

--- Performs discrete levels of damage to a vehicle, specifically catered towards removing/breaking parts off of the
--- vehicle.
---
--- The level damages are as follows (`N = number of wheels`):
---  - `1` - glass and lights are smashed
---  - `2` - all hinges are broken
---  - `[3, N+2]` - a unique wheel is broken
---  - `N+3` - all breakgroups are destroyed. This is the final, terminal damage level.
--- @class VehicleDissolver
--- @field private vehicle VehicleSceneObject The vehicle that is being dissolved
--- @field private level number The dissolve level. Increments upwards only.
local VehicleDissolver = classes.newClass()

--- Performs the next level of dissolving on the vehicle.
--- Once the final level of dissolving has occurred, this method will no longer affect the vehicle further.
function VehicleDissolver:incrementDissolveDamage()
  self.level = self.level + 1

  -- Break glass/lights
  if self.level == 1 then
    self.vehicle:queueLuaCommand([[
			for _, breakGroup in ipairs(ENG_BREAK_GROUPS.glass) do
				beamstate.breakBreakGroup(breakGroup)
			end
		]])
    return
  end

  -- Break all hinges
  if self.level == 2 then
    self.vehicle:queueLuaCommand("beamstate.breakHinges()")
    return
  end

  -- Attempt to break one wheel at a time; break everything else if all wheels have been broken
  self.vehicle:queueLuaCommand([[
    local index = ]]..self.level..[[ - 2 -- Two non-wheel levels before it
    local numberOfWheels = #ENG_BREAK_GROUPS.wheel

    -- All finished
    if index > numberOfWheels + 1 then return end

    if index > numberOfWheels then
      beamstate.breakAllBreakgroups()
      return
    end

    beamstate.breakBreakGroup(ENG_BREAK_GROUPS.wheel[index])
  ]])
end

--- Sets up an environment constant for the given vehicle so that the instance can perform dissolving actions on the
--- same vehicle in the future.
--- @param vehicle VehicleSceneObject The vehicle to handle dissolving
--- @private
function VehicleDissolver:initBreakgroups(vehicle)
  -- Set up ENG_BREAK_GROUPS
  vehicle:queueLuaCommand([[
		local breakGroups = { glass = {}, wheel = {} }

		local included = function(element, table)
			for _, v in ipairs(table) do
				if v == element then return true end
			end

			return false
		end

		for _, beam in pairs(v.data.beams) do
			local breakGroup = beam.breakGroup

			if breakGroup then
				local nestedBreakGroups = type(breakGroup) == "table" and breakGroup or { breakGroup }

				for _, bGroup in ipairs(nestedBreakGroups) do
					if not included(bGroup, breakGroups.glass) and not included(bGroup, breakGroups.wheel) then
						local formattedName = bGroup:lower()

						local isGlass = formattedName:find("light") or formattedName:find("glass")
						if isGlass then table.insert(breakGroups.glass, bGroup) end

						local isWheel = formattedName:find("wheel")
						if isWheel then table.insert(breakGroups.wheel, bGroup) end
					end
				end
			end
		end

		ENG_BREAK_GROUPS = breakGroups
	]])
end

--- Resets the instance back to its initial state
function VehicleDissolver:reset()
  self.level = 0
end

--- __Constructor__
--- @param vehicle VehicleSceneObject The vehicle to handle dissolving
--- @return VehicleDissolver The constructed instance
function VehicleDissolver:new(vehicle)
  local self = classes.setmetatable({}, VehicleDissolver)
  self.vehicle = vehicle
  self.level = 0
  self:initBreakgroups(vehicle)
  return self
end

return VehicleDissolver
