local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias Node { position: LuaVec3, id: number }
--- @alias MagnitudeFn fun(distance: number): number
--- @alias VehicleParticleType 1 | 29

--- @class VehicleNodes
--- @field private vehicleObject VehicleSceneObject The vehicle object
--- @field private numberOfNodes number The number of nodes inside the vehicle object
--- @field private MAX_DISTANCE_FROM_SOURCE number __Static__ The maximum distance at which zero force will be applied to a node, compared to the source of the force
local VehicleNodes = classes.newClass()
VehicleNodes.MAX_DISTANCE_FROM_SOURCE = 2

--- __Static__
---
--- Returns the position of the reference node for the given in-game vehicle object
--- @param vehicleObject VehicleSceneObject The in-game vehicle object
--- @return LuaVec3 The position of the reference node of this vehicle
--- @private
function VehicleNodes:getReferenceNodePosition(vehicleObject)
  return vec3(vehicleObject:getNodePosition(vehicleObject:getRefNodeId()))
end

--- __Static__
---
--- Returns the node's position of the given index in the absolute coordinate system
--- @param nodeIndex number The index of the node
--- @param vehicleObject VehicleSceneObject The in-game vehicle object to get the node position of
--- @return LuaVec3 The node's position of the given index in the absolute coordinate system
--- @private
function VehicleNodes:getAbsoluteNodePosition(nodeIndex, vehicleObject, vehiclePosition, referenceNodePosition)
  local node = vec3(vehicleObject:getNodePosition(nodeIndex))
  return node - referenceNodePosition + vehiclePosition
end

--- __Static__
---
--- Returns the list of all node positions for the given vehicle game object
--- @param vehicleObject VehicleSceneObject The vehicle game object to get the node positions of
--- @return Array<LuaVec3> The list of all node positions
function VehicleNodes:getAllNodePositions(vehicleObject)
  local vehiclePosition = vec3(vehicleObject:getPosition())
  local referenceNodePosition = VehicleNodes:getReferenceNodePosition(vehicleObject)

  return Array:from(vehicleObject:getNodeCount(), function(i)
    return VehicleNodes:getAbsoluteNodePosition(i, vehicleObject, vehiclePosition, referenceNodePosition)
  end)
end

--- __Static__
---
--- Returns the closest vehicle node's position to the given `position`.
--- This obeys the current vehicle's rotation as well and so the same vehicle placed at the same position but at a
--- different rotation will likely return a different node position. The same applies for damaged vehicles too.
--- In the event a vehicle has no nodes, this method will return `nil`.
--- @param nodes Array<Node> The list of nodes to search
--- @param position LuaVec3 The position to get the closest node to
--- @return Node The node with the closest position to `position`
--- @private
function VehicleNodes:closestNode(nodes, position)
  local closest = nodes
      :map(function(node) return { node = node, distance = position:distance(node.position) } end)
      :findByComparison(function(best, node) return best.distance <= node.distance and best or node end)

  return closest and closest.node or nil
end

--- __Static__
---
--- Creates a new `Node` for every node within the given in-game vehicle object
--- @param vehicleObject VehicleSceneObject The in-game vehicle object
--- @return Array<Node> The nodes within the `vehicleObject`
--- @private
function VehicleNodes:createNodes(vehicleObject)
  return VehicleNodes:getAllNodePositions(vehicleObject):map(function(nodePosition, i)
    return { id = i, position = nodePosition }
  end)
end

--- Affects a node by applying a force. This has a direction vector going from the `source` to the node's position. The
--- magnitude for the node is based on the `magnitudeFn` which is then multiplied by the distance to the `closestNode`.
--- This is linear, and if the node is outside a set distance (see `MAX_DISTANCE_FROM_CLOSEST`) of the `closestNode`,
--- then no force is applied.
--- @param node Node The node to affect
--- @param magnitudeFn MagnitudeFn A function that should return the magnitude of the force (in N) to inflict on this node, given its distance away from the `source`
--- @param source LuaVec3 The source position of the force to inflict on a node
--- @param closestNode LuaVec3 The closest node position to the `source`
--- @private
function VehicleNodes:applyForceToNode(node, magnitudeFn, source, closestNode)
  local nodePosition = node.position
  local distanceToSource = nodePosition:distance(source)
  local distanceToClosestNode = nodePosition:distance(closestNode)

  if distanceToClosestNode > VehicleNodes.MAX_DISTANCE_FROM_SOURCE then return end

  local distanceRatio = (VehicleNodes.MAX_DISTANCE_FROM_SOURCE - distanceToClosestNode) / VehicleNodes.MAX_DISTANCE_FROM_SOURCE
  local magnitude = magnitudeFn(distanceToSource) * math.max(0, distanceRatio)
  local direction = (nodePosition - source):normalized()
  local force = direction * magnitude

  -- Apply a force proportional to the node weight so that all materials experience the same amount of acceleration
  self.vehicleObject:queueLuaCommand([[
    -- Lua is 1-indexed, the vehicle's subsystem is 0-indexed
    local index = ]]..node.id..[[ - 1
    local node = v.data.nodes[index]
    obj:applyForceVector(node.cid, vec3(]]..vectors.toString(force, true)..[[) * node.nodeWeight)
  ]])
end

----- Affects each vehicle node by applying a force with a magnitude derived from the given `magnitudeFn` (see
----- `affectNode` for details).
----- The direction of the force will be away from the given `source` location.
----- @param magnitudeFn MagnitudeFn A function that should return the magnitude of the force (in N) to inflict on a node, given its distance away from the `source`
----- @param source LuaVec3 The source position of the force to inflict on a node
----- @see VehicleNodes#applyForceToNode
--function VehicleNodes:applyForceToNodes(magnitudeFn, source)
--  local nodes = VehicleNodes:createNodes(self.vehicleObject)
--  local closestNode = VehicleNodes:closestNode(nodes, source)
--
--  if not closestNode then return end
--
--  nodes:forEach(function(node) self:applyForceToNode(node, magnitudeFn, source, closestNode.position) end)
--end

--- Applies a particle to a random node on the vehicle
--- @param particle VehicleParticleType The id of the particle to add to a random node
function VehicleNodes:applyParticlesToRandomNode(particle)
  self.vehicleObject:queueLuaCommand([[
	  -- params: node, centreNode, velocity(towards centre node), particleType, scatter, particleCount
		obj:addParticleByNodesRelative(math.random(0, ]]..self.numberOfNodes..[[), ENG_CENTRE_NODE, 0, ]]..particle..[[, 0.01, 10)
	]])
end

--- __Constructor__
--- @param vehicleObject VehicleSceneObject The in-game vehicle object
--- @return VehicleNodes The constructed instance
function VehicleNodes:new(vehicleObject)
  local self = classes.setmetatable({}, VehicleNodes)
  self.vehicleObject = vehicleObject
  self.numberOfNodes = self.vehicleObject:getNodeCount()
  return self
end

return VehicleNodes
