local classes = require("lua/explosionsNG/helpers/classes")

--- @alias VehicleEffectUpdateFrequency "once" | "frame"

--- Properties:
---  - `postAffectFn` - A function to call when the affect has finished
---  - `effectDuration` - The duration of time this effect can affect vehicles
---  - `isInfinite` - Whether this effect is permanent and will never expire. This does NOT mean it will infinitely affect vehicles.
---  - `updateFrequency` - The frequency of updates the effect will apply to a vehicle
--- @alias VehicleEffectOptions { postAffectFn: VehicleAffectFn | nil, effectDuration: number | nil, isInfinite: boolean | nil, updateFrequency: VehicleEffectUpdateFrequency | nil }
--- @alias VehicleAffectFn fun(vehicle: Vehicle): void

--- An effect that can be applied to a `Vehicle`.
---
--- Note that the constructor is protected and instances should be created using the static methods in this class.
---
--- If an infinite effect has been provided with a `postAffectFn` in the options, then the caller should call the
--- `completeInfiniteEffect` method when the caller wishes to close the effect (i.e. the vehicle no longer needs to be
--- under this effect).
--- @class VehicleEffect : VehicularEffect
--- @field private time number | nil The time left of the effect. If `nil` this effect is infinite.
--- @field private options VehicleEffectOptions A list of options to uee
--- @field private affectFn VehicleAffectFn The function to apply when affecting a vehicle
--- @field private sourceId string The id of the object that is the source of the effect
--- @field private finishedAffecting boolean Whether this effect has finished performing any effects
local VehicleEffect = classes.newClass()

--- __Static__
---
--- Fills in any property not set in the given options with default values and returns it
--- @param options VehicleEffectOptions The list of options to fill
--- @return VehicleEffectOptions The filled options
--- @private
function VehicleEffect:fillOptions(options)
  return {
    effectDuration = options.effectDuration or 0, -- TODO This will never finish if the effect `isInfinite` and has `frame` update frequency
    isInfinite = options.isInfinite or false,
    updateFrequency = options.updateFrequency or "frame",
    postAffectFn = options.postAffectFn
  }
end

--- __Static__
---
--- Creates and returns an effect that lasts for an indefinite amount of time and repeatedly calls the affect function
--- every `affect` call.
--- This effect wil never finish (`isActive` will always return true).
--- @param sourceId string The id of the object that is the source of the effect
--- @param affectFn VehicleAffectFn The function to apply when affecting a vehicle
--- @return VehicleEffect The constructed instance
function VehicleEffect:newInfiniteEffect(sourceId, affectFn)
  return VehicleEffect:new(sourceId, affectFn, { isInfinite = true })
end

--- __Static__
---
--- Creates and returns an effect that lasts a particular amount of time and calls the affect function only once at the
--- first `affect` call.
--- This effect finishes (`isActive` will return `false`) after the time elapses.
--- @param sourceId string The id of the object that is the source of the effect
--- @param affectFn VehicleAffectFn The function to apply when affecting a vehicle
--- @param time number The amount of time the effect is in action
--- @param postAffectFn? VehicleAffectFn An option function to call when the affect has finished
--- @return VehicleEffect The constructed instance
function VehicleEffect:newSingleTimedEffect(sourceId, affectFn, time, postAffectFn)
  return VehicleEffect:new(sourceId, affectFn, { updateFrequency = "once", effectDuration = time, postAffectFn = postAffectFn })
end

--- __Static__
---
--- Creates and returns an effect that only calls the affect function once, but never finishes (i.e. `isActive` will
--- always return true)
--- @param sourceId string The id of the object that is the source of the effect
--- @param affectFn VehicleAffectFn The function to apply when affecting a vehicle
--- @param postAffectFn? VehicleAffectFn An option function to call when the affect has finished
--- @return VehicleEffect The constructed instance
function VehicleEffect:newInstantaneousEffect(sourceId, affectFn, postAffectFn)
  return VehicleEffect:new(sourceId, affectFn, { isInfinite = true, updateFrequency = "once", postAffectFn = postAffectFn })
end

--- Updates the effect. If the effect is still affecting vehicles, then the given vehicle will be affected by this
--- effect.
--- @param vehicle Vehicle The vehicle to affect
--- @param dt number The time since the last frame
function VehicleEffect:affect(vehicle, dt)
  if self.time > 0 then self:updateTimer(vehicle, dt) end

  if self.finishedAffecting then return end

  self.affectFn(vehicle)
  -- Only update once if set in the options
  if self.options.updateFrequency == "once" then
    self:onOnceComplete(vehicle)
  end
end

--- Calls the `postAffectFn`, if it is defined
--- @param vehicle Vehicle The vehicle that was affected
--- @private
function VehicleEffect:callPostAffectFn(vehicle)
  local postAffectFn = self.options.postAffectFn
  if postAffectFn then postAffectFn(vehicle) end
end

--- Called when the effect finishes and is in the `"once"` update frequency setting.
--- Sets the instance to have finished affecting and calls the `postAffectFn` (if it is defined and this effect is
--- __not__ an infinite effect) with the given vehicle.
--- @param vehicle Vehicle The vehicle that was affected
--- @private
function VehicleEffect:onOnceComplete(vehicle)
  self.finishedAffecting = true
  if not self:isInfinite() then self:callPostAffectFn(vehicle) end
end

--- Updates the timer of the effect, calling the `postAffectFn` if it is defined and the timer has reached `0`
--- @param vehicle Vehicle The vehicle to affect
--- @param dt number The time since the last frame
--- @private
function VehicleEffect:updateTimer(vehicle, dt)
  self.time = math.max(0, self.time - dt)
  -- If the time is now zero, call the post affect function
  if self.time == 0 then self:callPostAffectFn(vehicle) end
end

--- If this effect is infinite, then this method will call the `postAffectFn` (assuming it is defined).
--- This should be called only if the caller wishes to complete the affect and close it down; note that if the effect is
--- not infinite then this method will do nothing.
--- @param vehicle Vehicle The vehicle to affect
--- @see VehicularEffect#completeInfiniteEffect
function VehicleEffect:completeInfiniteEffect(vehicle)
  if self:isInfinite() then self:callPostAffectFn(vehicle) end
end

--- @return boolean Whether this effect is an infinite effect
function VehicleEffect:isInfinite()
  return self.options.isInfinite
end

--- @return boolean Whether this effect is still active. Active is defined as either an infinite effect or a finite effect that still has time remaining.
function VehicleEffect:isActive()
  return self:isInfinite() or self.time > 0
end

--- Returns whether the given `sourceId` is the source of this effect
--- @param sourceId string The id of the source to check
--- @return boolean Whether `sourceId` is the source of this effect
function VehicleEffect:isSource(sourceId)
  return self.sourceId == sourceId
end

--- __Constructor__
--- @param sourceId string The id of the object that is the source of the effect
--- @param affectFn VehicleAffectFn The function to apply when affecting a vehicle
--- @param options? VehicleEffectOptions An optional list of options to use. Note that any property not given will be defaulted.
--- @return VehicleEffect The constructed instance
--- @protected
function VehicleEffect:new(sourceId, affectFn, options)
  local self = classes.setmetatable({}, VehicleEffect)
  self.sourceId = sourceId
  self.affectFn = affectFn
  self.options = VehicleEffect:fillOptions(options or {})
  self.time = self.options.effectDuration
  self.finishedAffecting = false
  return self
end

return VehicleEffect
