local classes = require("lua/explosionsNG/helpers/classes")
local VehicleNodes = require("lua/explosionsNG/vehicles/VehicleNodes")
local VehicleEffect = require("lua/explosionsNG/vehicles/effects/VehicleEffect")

--- @alias NodeTime { id: number, position: LuaVec3, time: number }

--- A `VehicleEffect` affects individual nodes within any affected vehicle.
---
--- This effect applies a force to each node within the affected vehicle, away from the `source` position. The closest
--- vehicle node to this source position will be labelled the "zero-second" affect node. This node will be affected
--- immediately where all other nodes will calculate their delay before being affected, with the assumption that this is
--- the source of the ripple effect.
---
--- It should be noted that any instance will be both infinite and a single tick effect. In this way, any instance will
--- be an 'instantaneous` effect. See `VehicleEffect#newInstantaneousEffect` for details.
--- @class VehicleForceEffect : VehicleEffect
--- @field private source LuaVec3 The source position of the effect
--- @field private magnitudeFn MagnitudeFn A function that returns a magnitude of force to apply to a node on the vehicle, given the distance away
--- @field private timeElapsed number The amount of time (in seconds) that has elapsed
--- @field private vehicleNodes VehicleNodes The vehicle's node that is being affected
--- @field private sourceVehicleNode Node The vehicle node that is the closest to the source and will be treated as the "zero-second" effect node. See class documentation above.
--- @field private nodeTimes Array<NodeTime> The list of vehicle nodes, plus a time at which they should be affected
--- @field private SPEED number __Static__ The speed (in ms^-1) that the force will ripple through any vehicle
local VehicleForceEffect = classes.newClass(VehicleEffect)
VehicleForceEffect.SPEED = 20

--- __Static__
---
--- @param vehicle Vehicle
--- @param source LuaVec3 The source location of the effect. This determines the direction the force and acceleration will point away from
--- @return Array<NodeTime>
--- @private
function VehicleForceEffect:initVehicleNodes(vehicle, source)
  local vehicleObject = vehicle:getGameObject()
  return VehicleNodes:getAllNodePositions(vehicleObject):map(function(nodePosition, index)
    local distance = nodePosition:distance(source)
    local time = distance / VehicleForceEffect.SPEED
    return --[[--- @type NodeTime ]] { id = index, position = nodePosition, time = time }
  end)
end

--- Updates the effect. If the effect is still affecting vehicles, then the given vehicle will be affected by this
--- effect.
--- @param vehicle Vehicle The vehicle to affect
--- @param dt number The time since the last frame
function VehicleForceEffect:affect(vehicle, dt)
  VehicleEffect.affect(self, vehicle, dt)
  self.timeElapsed = self.timeElapsed + dt
end

--- Updates the node effects to the given vehicle. If all nodes have now been affected, then this method will set the
--- `finishedAffecting` flag to `true`.
--- @param vehicle Vehicle The vehicle to affect
--- @private
function VehicleForceEffect:updateAffect(vehicle)
  if not self.vehicleNodes then self:initNodes(vehicle) end

  --- A predicate for whether a given node can be affect at this time
  --- @param node NodeTime The node to consider
  --- @return boolean Whether the given node can be affect at this time
  local canAffectNode = function(node) return self.timeElapsed >= node.time end

  self.nodeTimes
      :filter(canAffectNode)
      :forEach(function(node)
        self.vehicleNodes:applyForceToNode(node, self.magnitudeFn, self.source, self.sourceVehicleNode.position)
      end)

  -- Remove affected nodes
  self.nodeTimes = self.nodeTimes:removeIf(canAffectNode)
  self.finishedAffecting = self.nodeTimes:isEmpty()
end

--- Sets up various node related fields for this instance, catered towards the given vehicle
--- @param vehicle Vehicle The vehicle to initialise the effect to
--- @private
function VehicleForceEffect:initNodes(vehicle)
  local vehicleObject = vehicle:getGameObject()

  local nodes = VehicleNodes:createNodes(vehicleObject)
  self.sourceVehicleNode = VehicleNodes:closestNode(nodes, self.source)

  if not self.sourceVehicleNode then
    log("E", "explosivesNG", "Could not find a closest node on the vehicle - this means the vehicle has no node!")
    return
  end

  self.vehicleNodes = VehicleNodes:new(vehicleObject)
  self.nodeTimes = VehicleForceEffect:initVehicleNodes(vehicle, self.sourceVehicleNode.position)
end

--- __Constructor__
--- @param sourceId string The id of the object that is the source of the effect
--- @param source LuaVec3 The location of the source of the effect
--- @param magnitudeFn MagnitudeFn A function that should return the magnitude of the force (in N) to inflict on a node, given its distance away from the `source`
--- @return VehicleForceEffect The constructed instance
function VehicleForceEffect:new(sourceId, source, magnitudeFn)
  local self = classes.setmetatable(VehicleEffect:new(sourceId, nil, { isInfinite = true }), VehicleForceEffect)
  -- This is not passed in the super call and set here as `self` is not defined when passing in as a parameter when calling super
  self.affectFn = function(vehicle) self:updateAffect(vehicle) end
  self.source = source
  self.magnitudeFn = magnitudeFn
  self.vehicleNodes = nil
  self.nodeTimes = nil
  self.sourceVehicleNode = nil
  self.timeElapsed = 0
  return self
end

return VehicleForceEffect
