local classes = require("lua/explosionsNG/helpers/classes")

--- A composite class that wraps a list of `VehicleEffect`s and manages/updates each of them as one.
--- Note that the `sourceId` of any instance should be the same in all effects wrapped in this instance.
--- @class VehicleMultiEffect : VehicularEffect
--- @field private sourceId string The id of the object that is the source of the effect
--- @field private effects Array<VehicleEffect> The list of effects to apply
local VehicleMultiEffect = classes.newClass()

--- Applies all effects to the vehicle
--- @param vehicle Vehicle The vehicle to affect
--- @param dt number The time since the last frame
function VehicleMultiEffect:affect(vehicle, dt)
  self.effects:forEach(function(effect) effect:affect(vehicle, dt) end)
end

--- @return boolean Whether at least one effect is still active
function VehicleMultiEffect:isActive()
  return self.effects:some(function(effect) return effect:isActive() end)
end

--- @return boolean Whether at least one effect is an infinite effect
function VehicleMultiEffect:isInfinite()
  return self.effects:some(function(effect) return effect:isInfinite() end)
end

--- @param vehicle Vehicle The vehicle to affect
--- @see VehicularEffect#completeInfiniteEffect
function VehicleMultiEffect:completeInfiniteEffect(vehicle)
  self.effects:forEach(function(effect) return effect:completeInfiniteEffect(vehicle) end)
end

--- Returns whether the given `sourceId` is the source of this effect
--- @param sourceId string The id of the source to check
--- @return boolean Whether `sourceId` is the source of this effect
function VehicleMultiEffect:isSource(sourceId)
  return self.sourceId == sourceId
end

--- __Constructor__
--- @param sourceId string The id of the object that is the source of the effect
--- @param effects Array<VehicleEffect> The list of effects to apply
--- @return VehicleMultiEffect The constructed instance
function VehicleMultiEffect:new(sourceId, effects)
  local self = classes.setmetatable({}, VehicleMultiEffect)
  self.sourceId = sourceId
  self.effects = effects
  return self
end

return VehicleMultiEffect
