local Array = require("lua/explosionsNG/misc/Array")
local maths = require("lua/explosionsNG/helpers/maths")
local Vehicle = require("lua/explosionsNG/vehicles/Vehicle")
local VehicleEffect = require("lua/explosionsNG/vehicles/effects/VehicleEffect")
local VehicleMultiEffect = require("lua/explosionsNG/vehicles/effects/VehicleMultiEffect")
local VehicleForceEffect = require("lua/explosionsNG/vehicles/effects/VehicleForceEffect")
local VehicleSoundEffect = require("lua/explosionsNG/vehicles/effects/VehicleSoundEffect")
local VehicleParticleEffect = require("lua/explosionsNG/vehicles/effects/VehicleParticleEffect")

local BASE_FORCE = 25000

--- Converts a multiplier of force into its maximal force that a shockwave could inflict
--- @param forceMultiplier number The multiplier to return the force for
--- @return number The maximum force (in N) that a shockwave can apply
local function forceMultiplierToMagnitude(forceMultiplier)
  return BASE_FORCE * forceMultiplier
end

--- Returns the force at a distance away from a point. This internally uses the `landingRampDown` maths function (see
--- `/helpers/maths.lua` for visuals). This means a point will inflict the maximal force it can for a distance within
--- 50% of its `maxRadius`. After that, the force drops off at a linear rate to `0` as the distance to the point reaches
--- and exceeds the `maxRadius`.
--- @param distance number The distance away from the point
--- @param maxRadius number The maximum radius of the point
--- @param forceMultiplier number The multiplier of force of the shockwave
--- @param reverse boolean Whether to inverse the force or not
--- @return number The force
local function forceAt(distance, maxRadius, forceMultiplier, reverse)
  local distanceRatio = math.min(1, distance / maxRadius)
  local maxForce = forceMultiplierToMagnitude(forceMultiplier)
  local scaleForced = maxForce * maths.landingRampDown(distanceRatio)
  return reverse and -scaleForced or scaleForced
end

--- Returns a vehicle effect that:
---  - Disables the electronics of the vehicle
---  - Produces particle effects on the vehicle
---  - Plays a sound effect, if the vehicle being affected is a player vehicle
--- @param id string The id of the source of the effect
--- @return VehicleMultiEffect The created effect
local function createEmpEffect(id)
  local disableElectrics = VehicleEffect:newInstantaneousEffect(id, Vehicle.disableElectronics)
  local particleEffect = VehicleParticleEffect:new(id, 2, 1)
  local soundEffect = VehicleSoundEffect:new(id, "electric_sparks")
  return VehicleMultiEffect:new(id, Array:new({ disableElectrics, particleEffect, soundEffect }))
end

--- Returns a vehicle effect that ignites and bursts a random tire on a vehicle
--- @param id string The id of the source of the effect
--- @return VehicleMultiEffect The created effect
local function createFireEffect(id)
  local deflateTire = VehicleEffect:newInstantaneousEffect(id, Vehicle.deflateRandomTire)
  local particleEffect = VehicleParticleEffect:new(id, 3, 29)
  return VehicleMultiEffect:new(id, Array:new({ deflateTire, particleEffect }))
end

--- Returns a vehicle effect that performs the next level of dissolving damage on a vehicle
--- @param id string The id of the source of the effect
--- @param time? number The period of time the effect lasts. If not given, the effect lasts indefinitely
--- @return VehicleEffect The created effect
local function createDissolveEffect(id, time)
  if not time then
    return VehicleEffect:newInstantaneousEffect(id, Vehicle.dissolve)
  end

  return VehicleEffect:newSingleTimedEffect(id, Vehicle.dissolve, time)
end

--- Returns a `VehicleForceEffect` that applies a force to a vehicle. This effect will stay indefinitely on the vehicle
--- until it is removed manually, preventing any further effects from being applied.
--- Note that the force applied is linear and the force applied will become `0` if the position at which the effect is
--- applied away from the source is greater than or equal to the maximum radius.
--- @param id string The id of the source of the effect
--- @param position LuaVec3 The position of the effect at which the collision with a vehicle has occurred
--- @param maxRadius number The maximum radius of the effect
--- @param forceMultiplier? number The multiplier of force to apply. Defaults to `1`
--- @param reverse? boolean Whether to invert the force i.e. apply a force inwards/that attracts vehicles. Defaults to `false`.
--- @return VehicleForceEffect The created instance
local function createForceEffect(id, position, maxRadius, forceMultiplier, reverse)
  --- @type MagnitudeFn
  local magnitudeFn = function(distance) return forceAt(distance, maxRadius, forceMultiplier or 1, reverse or false) end
  return VehicleForceEffect:new(id, position, magnitudeFn)
end

--- Returns a `VehicleEffect` that sets the local gravity for a vehicle. This effect will last for the time given and
--- upon completing, will then reset the local gravity for the vehicle back to the environmental gravity.
--- "Local gravity" refers to the gravity point directly downwards.
--- @param id string The id of the source of the effect
--- @param gravity number The gravity (in ms^-2) to apply to the vehicle from this effect
--- @param time number The period of time the effect lasts
--- @return VehicleEffect The local gravity effect
local function createLocalGravityEffect(id, gravity, time)
  --- @type VehicleAffectFn
  local affectFn = function(vehicle) vehicle:setLocalGravity(gravity) end
  --- @type VehicleAffectFn
  local postAffectFn = function(vehicle) vehicle:setLocalGravity(core_environment.getGravity()) end
  return VehicleEffect:newSingleTimedEffect(id, affectFn, time, postAffectFn)
end

return {
  BASE_FORCE = BASE_FORCE,
  createEmpEffect = createEmpEffect,
  createFireEffect = createFireEffect,
  createDissolveEffect = createDissolveEffect,
  createForceEffect = createForceEffect,
  createLocalGravityEffect = createLocalGravityEffect
}