local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local maths = require("lua/explosionsNG/helpers/maths")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local Timer = require("lua/explosionsNG/timers/Timer")

--- Properties:
---  - `range` - The detection range of the mine
---  - `life`  - The life (in seconds) of the mine before it naturally activates itself
--- @alias MineOptions { range: number | nil, life: number | nil }

--- @class Mine
--- @field position LuaVec3 The centre position of the mine
--- @field object SceneObject The in-game marker object which displays it's range
--- @field detonationFn fun(): void A function to call when the mine is detonated
--- @field options MineOptions The list of options for this mine
--- @field isActivated boolean Whether this mine is activated
--- @field isDetonated boolean Whether this mine has detonated
--- @field light SceneObject The in-game light object
--- @field soundEmitter SceneObject The in-game sound emitter object
--- @field activationTimer Timer The timer that counts down the activation period
--- @field lifeTimer Timer The timer that counts down the mine's life
--- @field OBJECT_POSITION_STRING string Stores a string version of the mine's position. This is a workaround for a bug within BeamNG/TorqueScript: see `updateObjectColours` for details on this usage of this
--- @field DETONATION_DELAY number __Static__ The delay (in seconds) that the mine is activated for before it is considered detonating
--- @field STARTING_COLOUR_HUE number __Static__ The hue colour (between [0, 1]) that the mine starts out with (i.e. is completely unactivated)
--- @field BLINK_FREQUENCY number __Static__ The frequency (in seconds) that the mine will toggle its blink animation, during the activation phase
--- @field MARKER_ALPHA number __Static__ The alpha of the marker area
--- @field MAX_LIGHT_BRIGHTNESS number __Static__ The maximum brightness the light can be
--- @field MIN_LIGHT_BRIGHTNESS number __Static__ The minimum brightness the light can be
--- @field SOUND_EMITTER_MULTIPLIER number __Static__ The multiplier of mine range to use for the sound emitted by any mine
local Mine = classes.newClass()
Mine.DETONATION_DELAY = 1.5
Mine.STARTING_COLOUR_HUE = 60 / 360 -- Equates to a hue of yellow
Mine.BLINK_FREQUENCY = 15
Mine.MARKER_ALPHA = 0.5
Mine.MAX_LIGHT_BRIGHTNESS = 5
Mine.MIN_LIGHT_BRIGHTNESS = 0.25
Mine.SOUND_EMITTER_MULTIPLIER = 5

--- __Static__
---
--- Fills in any key not in the given `options` with default values
--- @param options MineOptions The list of predefined options
--- @return MineOptions The complete set of options with defaults filled in where empty values are given
--- @private
function Mine:fillOptions(options)
  return {
    range = options.range or 10,
    life = options.life or 15
  }
end

--- __Static__
---
--- Returns a boolean value for a given input. A `true` value means the mine should "blink", a `false` value means the
--- mine should not
--- @param x number The input
--- @return boolean Whether the mine should "blink"
--- @private
function Mine:activationFunction(x)
  local xFrequency = x * Mine.BLINK_FREQUENCY
  return maths.decimalPart(xFrequency) > 0.5
end

--- __Static__
---
--- Returns a new sound emitter with an associated track which matches whether the mine this sound emitter is for is
--- activated or not.
--- @param isActivated boolean Whether the mine is activated
--- @param position LuaVec3 The position to place the sound emitter
--- @param options MineOptions The list of options to use
--- @return SceneObject The sound emitter
--- @private
function Mine:createSoundEmitter(isActivated, position, options)
  return worldObjectsHandler.requestSoundEmitter({
    track = isActivated and "mine_activated_ambient" or "mine_ambient",
    isLooping = true,
    position = position,
    maxDistance = options.range * 5
  })
end

--- If this mine is not detonated, this method will do one of the following:
---  - If this mine `isActivated`, then this method updates the activating phase
---  - If this mine has a vehicle within it's range of effect, this method will activate the mine
---  - Otherwise, this method will decrease the mine's life timer and update the various in-game object colours
--- @param dt number The time since the last frame
function Mine:update(dt)
  if self.isDetonated then return end

  if self.isActivated then
    self:updateActivating(dt)
    return
  end

  if vehiclesHandler.isVehicleInRange(self.position, self.options.range) then
    self:activate()
    return
  end

  self.lifeTimer:update(dt)

  local timeLeftRatio = self.lifeTimer:timeLeftRatio()
  self:updateObjectColours(timeLeftRatio)
end

--- Updates the activation timer and updates the in-game light colour with the current elapsed time in this phase.
--- This method will then update the marker colour, flashing between orange and red in quick succession.
--- @param dt number The time since the last frame
--- @private
function Mine:updateActivating(dt)
  local timeRatio = self.activationTimer:timeLeftRatio()
  self:updateLightBlinking(timeRatio)

  local colour = Mine:activationFunction(timeRatio) and 0.5 or 0
  self:updateObjectColours(colour)

  self.activationTimer:update(dt)
end

--- Updates the various in-game objects' colours with the hue value passed. The hue value should be normalised, and
--- where a value of `0` corresponds to the expected hue of `0`deg, and a value of `1` corresponds to a hue of
--- `STARTING_COLOUR_HUE * 360`deg.
--- @param hueRatio number The normalised hue value
--- @private
function Mine:updateObjectColours(hueRatio)
  local r, g, b = HSVtoRGB(hueRatio * Mine.STARTING_COLOUR_HUE, 1, 1)
  self.object:setField("instanceColor", 0, r.." "..g.." "..b.." "..Mine.MARKER_ALPHA)
  self.object:setField("position", 0, self.OBJECT_POSITION_STRING) -- This is to trigger an update as for some reason, setting the instanceColor doesn't seem to update some objects :/
  self.light:setField("color", 0, r.." "..g.." "..b.." 1")
end

--- Updates the in-game light's brightness by flashing between the min and max brightness, see the following static
--- fields respectively: `Mine.MIN_LIGHT_BRIGHTNESS`, `Mine.MAX_LIGHT_BRIGHTNESS `
--- @param timeRatio number The elapsed time as a ratio of time remained to total time
--- @private
function Mine:updateLightBlinking(timeRatio)
  local brightness = Mine:activationFunction(timeRatio) and Mine.MAX_LIGHT_BRIGHTNESS or Mine.MIN_LIGHT_BRIGHTNESS
  self.light:setField("brightness", 0, brightness)
end

--- Begins the mine's activation phase, sets the in-game objects' colours to red and changes the ambient sound emitter
--- @private
function Mine:activate()
  self.isActivated = true
  self.activationTimer:start()
  self:updateObjectColours(0)

  worldObjectsHandler.releaseSoundEmitter(self.soundEmitter)
  self.soundEmitter = Mine:createSoundEmitter(true, self.position, self.options)
end

--- Destroys all in-game objects, set's the internal flag to `true` and calls the `detonationFn` callback function
--- @private
function Mine:onDetonation()
  self:destroy()
  self.isDetonated = true
  self.detonationFn()
end

--- Starts the mine's life timer and proximity checking
function Mine:arm()
  self.lifeTimer:start()
end

--- Releases the in-game objects to the world objects handler
function Mine:destroy()
  -- If this is true then the instance is already destroyed
  if not self.object then return end

  worldObjectsHandler.releaseTSStatic(self.object)
  self.object = nil

  worldObjectsHandler.releasePointLight(self.light)
  self.light = nil

  worldObjectsHandler.releaseSoundEmitter(self.soundEmitter)
  self.soundEmitter = nil
end

--- __Constructor__
--- @field position LuaVec3 The centre position of the mine
--- @field object SceneObject The in-game marker object which displays it's range
--- @field detonationFn fun(): void A function to call when the mine is detonated
--- @field options MineOptions The list of options for this mine
--- @return Mine The constructed instance
function Mine:new(position, object, detonationFn, options)
  local self = classes.setmetatable({}, Mine)
  self.options = Mine:fillOptions(options or {})
  self.position = vec3(position)
  self.object = object
  self.detonationFn = detonationFn
  self.OBJECT_POSITION_STRING = vectors.toString(vec3(object:getPosition()))
  self.isActivated = false
  self.isDetonated = false
  self.light = worldObjectsHandler.requestPointLight({
    colour = "0 0 0",
    applyRadiusToPosition = true,
    position = self.position,
    radius = self.options.range,
    brightness = Mine.MAX_LIGHT_BRIGHTNESS
  })
  self.soundEmitter = Mine:createSoundEmitter(false, self.position, self.options)
  self.activationTimer = Timer:new(Mine.DETONATION_DELAY, function() self:onDetonation() end)
  self.lifeTimer = Timer:new(self.options.life, function() self:activate() end)
  return self
end

return Mine
