local classes = require("lua/explosionsNG/helpers/classes")
local collisions = require("lua/explosionsNG/helpers/collisions")
local vectors = require("lua/explosionsNG/helpers/vectors")
local Timer = require("lua/explosionsNG/timers/Timer")
local VehicleAffectable = require("lua/explosionsNG/misc/VehicleAffectable")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")

--- @alias OrbOptions { colour: string | nil, radius: number | nil, life: number | nil, speed: number | nil }

--- A spherical region that moves in a linear motion, and bounces off of objects (from the centre point of the region).
--- @class Orb : VehicleAffectable
--- @field private position LuaVec3 The current position of the orb
--- @field private direction LuaVec3 The current direction of the orb
--- @field private options OrbOptions The list of options in use
--- @field private object SceneObject The in-game object
--- @field private lifeTimer Timer The life timer
--- @field private finished boolean Whether the orb is finished
--- @field private affectFn VehicleEffectProducer A function that returns a new `VehicleEffect` instance to apply to any vehicle within range of this orb's effect
--- @field private FADE_OUT_PERIOD number __Static__ The ending period of which the orb will begin to fade out
--- @field private MAX_ALPHA number __Static__ The maximum alpha that the orb can be at any one time
local Orb = classes.newClass(VehicleAffectable)
Orb.FADE_OUT_PERIOD = 2
Orb.MAX_ALPHA = 0.2

--- __Static__
---
--- Fills in any option not given in the given list with defaults
--- @param options OrbOptions A partial list of options
--- @return OrbOptions The filled options
function Orb:fillOptions(options)
  return {
    colour = options.colour or "1 1 1",
    radius = options.radius or 5,
    life = options.life or 10,
    speed = options.speed or 1,
  }
end

--- @param vehicle Vehicle The vehicle to check
--- @return boolean Whether the `vehicle` can be affected by this object
--- @see VehicleAffectable#canAffect
function Orb:canAffect(vehicle)
  return vehicle:withinBounds(self.position, self.options.radius)
end

--- @param vehicle Vehicle The vehicle to affect
--- @return VehicularEffect The effect that can be applied to a vehicle
--- @see VehicleAffectable#affect
function Orb:affect(vehicle)
  local collisionPoint = vehicle:closestPointToBounds(self.position)
  return self.affectFn(self.id, { position = collisionPoint, vehicle = vehicle })
end

--- Updates the orbs position, colour and whether it can affect any vehicle(s)
--- @param dt number The time since the last frame
function Orb:update(dt)
  if self.finished then return end

  self:updatePosition(dt)
  self:updateColour(dt)
  self:tryAndAffectVehicles()
end

--- Updates the current position of the orb and updates the in-game object.
--- This method will also check for any collisions.
--- @param dt number The time since the last frame
--- @private
function Orb:updatePosition(dt)
  local positionIncrement = self.direction * (self.options.speed * dt)
  self.position = self.position + positionIncrement
  self.object:setField("position", 0, vectors.toString(self.position))

  self:checkForCollision()
end

--- Updates the colour of the in-game object, specifically its alpha value.
--- This method will also update the life timer.
--- @param dt number The time since the last frame
--- @private
function Orb:updateColour(dt)
  local timeLeft = self.lifeTimer:timeLeft()

  -- Only update alpha if it will be different
  if timeLeft < Orb.FADE_OUT_PERIOD then
    local alpha = Orb.MAX_ALPHA * (timeLeft / Orb.FADE_OUT_PERIOD)
    self.object:setField("instanceColor", 0, self.options.colour.." "..alpha)
  end

  self.lifeTimer:update(dt)
end

--- Checks for any collisions on all axises. If there is a collision, this method will change the direction of the orb
--- to bounce in the opposite direction.
--- @private
function Orb:checkForCollision()
  local collisionDirection = collisions.findDirectionOfCollision(self.position, 0.5, self.options.radius)

  if collisionDirection then
    -- Creates a multiplier vector, where the collision direction remains unchanged, but all `0`s become `1`s
    local multiplier = vectors.defaultIfZero(collisionDirection, 1)
    self.direction = vectors.crossMultiply(self.direction, multiplier)
  end
end

--- Releases its in-game object to the world objects handler, removes the effects on all vehicles, and sets this orb to
--- be destroyed.
function Orb:destroy()
  if self.finished then return end

  self.finished = true

  worldObjectsHandler.releaseTSStatic(self.object)
  self.object = nil

  self:removeAllEffects()
end

--- @return boolean Whether the orb is finished
function Orb:hasFinished()
  return self.finished
end

--- __Constructor__
--- @param position LuaVec3 The initial position of the orb
--- @param direction LuaVec3 The direction the orb travels
--- @param affectFn VehicleEffectProducer A function that returns a new `VehicleEffect` instance to apply to any vehicle within range of this orb's effect
--- @param options? OrbOptions | nil An optional list of options to use
--- @return Orb The constructed instance
function Orb:new(position, direction, affectFn, options)
  local self = classes.setmetatable({}, Orb)
  self.position = position
  self.direction = direction
  self.options = Orb:fillOptions(options or {})
  self.affectFn = affectFn
  self.object = worldObjectsHandler.requestTSStatic({
    modelName = "sphere",
    position = position,
    instanceColour = self.options.colour.." "..Orb.MAX_ALPHA,
    scale = self.options.radius
  })
  self.lifeTimer = Timer:new(self.options.life, function() self:destroy() end, true)
  self.finished = false
  return self
end

return Orb
