local classes = require("lua/explosionsNG/helpers/classes")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local vectors = require("lua/explosionsNG/helpers/vectors")
local Array = require("lua/explosionsNG/misc/Array")

--- A preview of an explosive that contains three in-game objects:
---  - A box object, coloured to the tier of an explosive
---  - A cylindrical marker that extends to the range of an explosive
---  - An optional set of spherical marker(s) that extends to the bounds of an explosive
--- @class Preview
--- @field private id string The id of the preview
--- @field private bounds BoundingArea | nil The bound's range
--- @field private boxObject SceneObject The in-game object representing the explosive
--- @field private rangeObject SceneObject The in-game object representing a range
--- @field private boundsObjects Array<SceneObject> The in-game object representation(s) of the bounds of an object, that exist(s) inside the range
--- @field private RANGE_COLOUR string __Static__ The colour to use for the range in-game object
--- @field private BOUNDS_COLOUR string __Static__ The colour to use for the bounds in-game object
--- @field private nextId number __Static__ The next available id to use
local Preview = classes.newClass()
Preview.RANGE_COLOUR = "0 0.5 1 0.5"
Preview.BOUNDS_COLOUR = "0 1 0 1"
Preview.nextId = 0

--- __Static__
--- @return string The new string identifier for a preview
--- @private
function Preview:createId()
  local id = Preview.nextId
  Preview.nextId = Preview.nextId + 1
  return tostring(id)
end

--- __Static__
---
--- Returns the colour of the given tier in "R G B" format
--- @param tier number The tier of the explosive to get the colour of
--- @return string The tier's colour
--- @private
function Preview:getTierColour(tier)
  if tier == 1 then return "1   1 0" end
  if tier == 2 then return "1 0.5 0" end
  if tier == 3 then return "1   0 0" end
  if tier == 4 then return "1   0 1" end

  log("W", "explosivesNG", "Tier "..tostring(tier).." is not a known tier")
  return "0 0 0"
end

--- __Static__
---
--- Returns a new box object to represent the explosive itself
--- @param position LuaVec3 The centre position to place the object
--- @param tier number The tier of the explosive to represent
--- @return SceneObject The created in-game object
--- @private
function Preview:createBoxObject(position, tier)
  local colour = Preview:getTierColour(tier)
  return worldObjectsHandler.requestTSStatic({
    modelName = "box",
    instanceColour = colour,
    position = position,
    scale = tier
  })
end

--- __Static__
---
--- Returns a new cylindrical marker to represent the given range, centred around the given position
--- @param position LuaVec3 The centre position to place the object
--- @param range number The range (in m) that the range object should reach to
--- @return SceneObject The create in-game object
--- @private
function Preview:createRangeObject(position, range)
  return worldObjectsHandler.requestTSStatic({
    modelName = "cylinder_marker",
    instanceColour = Preview.RANGE_COLOUR,
    -- The cylinder has a radius of 0.5m, so we need to double it on the XY axises
    scale = vec3(range * 2, range * 2, range / 4),
    position = position,
    dynamic = false
  })
end

--- __Static__
---
--- Returns a new spherical marker to represent the given bounds, centred around the given position.
--- @param position LuaVec3 The centre position to place the object
--- @param bounds number The bounds
--- @return SceneObject The create in-game object
--- @private
function Preview:createSingleBoundObject(position, bounds)
  return worldObjectsHandler.requestTSStatic({
    modelName = "sphere_marker",
    instanceColour = Preview.BOUNDS_COLOUR,
    scale = bounds,
    position = position,
    dynamic = false
  })
end

--- __Static__
---
--- Returns a new spherical marker (or markers) to represent the given bounds, centred around the given position.
--- @param position LuaVec3 The centre position to place the object
--- @param bounds BoundingArea The bounds
--- @return Array<SceneObject> The create in-game object(s)
--- @private
function Preview:createBoundsObject(position, bounds)
  if type(bounds) == "number" then
    return Array:new({ Preview:createSingleBoundObject(position, --[[--- @type number ]] bounds) })
  end

  return Array:new(--[[--- @type SphericalBound[] ]] bounds):map(function(sphericalBound)
    local boundPosition = position + sphericalBound.offset
    return Preview:createSingleBoundObject(boundPosition, sphericalBound.radius)
  end)
end

--- Sets the centre of the preview to the given position
--- @param position LuaVec3 The new centre position
function Preview:updatePosition(position)
  local positionString = vectors.toString(position)

  self.boxObject:setField("position", 0, positionString)
  if self.rangeObject then self.rangeObject:setField("position", 0, positionString) end
  self:updateBoundsPosition(position)
end

--- Sets the centres of the bounds to the given position. If a set of bounds are present, this centre position
--- represents the centre of the whole bounding area.
--- If there are no bounds, this method does nothing.
--- @param position LuaVec3 The new centre position
--- @private
function Preview:updateBoundsPosition(position)
  if not self.bounds then return end

  if type(self.bounds) ~= "number" then
    local castedBounds = --[[--- @type SphericalBound[] ]] self.bounds
    self.boundsObjects:forEach(function(bound, index)
      local offset = castedBounds[index].offset
      local boundPosition = position + offset
      bound:setField("position", 0, vectors.toString(boundPosition))
    end)
    return
  end

  if self.boundsObjects:length() ~= 1 then
    log("E", "explosivesNG", "The bounds of this preview is set up to have exactly one bound but there are "..tostring(self.boundsObjects:length()).." bound objects")
    return
  end

  self.boundsObjects:nth(1):setField("position", 0, vectors.toString(position))
end

--- Releases all assets
function Preview:delete()
  worldObjectsHandler.releaseTSStatic(self.boxObject)
  self.boxObject = nil

  if self.rangeObject then
    worldObjectsHandler.releaseTSStatic(self.rangeObject)
    self.rangeObject = nil
  end

  if self.boundsObjects then
    (--[[--- @not nil ]] self.boundsObjects):forEach(function(bound) worldObjectsHandler.releaseTSStatic(bound) end)
    self.boundsObjects = nil
  end
end

--- @return string The id of the preview
function Preview:getId()
  return self.id
end

--- @return BoundingArea | nil The bound's range
function Preview:getBoundsRange()
  return self.bounds
end

--- __Constructor__
--- @param position LuaVec3 The centre position of the preview
--- @param tier number The tier of explosive to create a preview for
--- @param range number The range (in m) that the range part of the preview should reach to
--- @param bounds? BoundingArea | nil An optional `BoundingArea` that the explosive being preview takes up in the world. See type definition for details.
--- @return Preview The constructed instance
function Preview:new(position, tier, range, bounds)
  local self = classes.setmetatable({}, Preview)
  self.bounds = bounds
  self.id = Preview:createId()
  self.boxObject = Preview:createBoxObject(position, tier)
  -- If the explosive has infinite range, do not create a range in-game object for it
  self.rangeObject = range < math.huge and Preview:createRangeObject(position, range)
  self.boundsObjects = bounds and Preview:createBoundsObject(position, bounds) or nil
  return self
end

return Preview
