local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local Array = require("lua/explosionsNG/misc/Array")
local Shockwave = require("lua/explosionsNG/worldObjects/Shockwave")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")

--- Properties:
---  - `numberOfShockwaves` - the number of shockwaves to emit
---  - `interval` - the interval between two shockwaves
---  - `maxRadius` - the maximum radius of each shockwave
---  - `time` - the time each shockwave lasts for
---  - `affectFn` - a function to use when a shockwave affects a vehicle
---  - `onVehicleAffected` a function that will be called when a vehicle is affected by this shockwave
---  - `shockwaveOptions` - a list of options to use for each shockwave
--- @alias PulsarOptions { numberOfShockwaves: number | nil, interval: number | nil, maxRadius: number | nil, time: number | nil, affectFn: VehicleEffectProducer | nil, onVehicleAffected: ShockwaveVehicleAffectedFn | nil, shockwaveOptions: ShockwaveOptions | nil }

--- @class Pulsar
--- @field private position LuaVec3 The centre position of the pulsar
--- @field private options PulsarOptions The list of options
--- @field private interval IntervalTimer The interval which will create a new shockwave at every cycle
--- @field private shockwaves Array<Shockwave> The list of active shockwaves
local Pulsar = classes.newClass()

--- __Static__
---
--- Returns a list of options with any non-given options filled in with defaults.
--- Note that the `affectFn` MUST be given as there is no default value.
--- @param options PulsarOptions The partial list of options
--- @return PulsarOptions
--- @private
function Pulsar:fillOptions(options)
  return {
    numberOfShockwaves = options.numberOfShockwaves or 5,
    interval = options.interval or 2,
    maxRadius = options.maxRadius or 25,
    time = options.time or 2,
    affectFn = options.affectFn,
    onVehicleAffected = options.onVehicleAffected,
    shockwaveOptions = options.shockwaveOptions or {}
  }
end

--- __Static__
---
--- Creates and returns a new shockwave with the given options
--- @param position LuaVec3 The position of the centre of each shockwave
--- @param options PulsarOptions The list of options to use
--- @return Shockwave The created shockwave
--- @private
function Pulsar:createShockwaves(position, options)
  --- @type VehicleEffectProducer
  local affectFn = function(id, event)
    local onVehicleAffected = options.onVehicleAffected
    if onVehicleAffected then onVehicleAffected(event.vehicle) end
    return options.affectFn(id, event)
  end

  local shockwaveOptions = tables.copy(options.shockwaveOptions or {})
  return Shockwave:new(position, options.maxRadius, options.time, affectFn, shockwaveOptions)
end

--- Updates all shockwaves, if this pulsar is not finished
--- @param dt number The time since the last frame
function Pulsar:update(dt)
  if self:isFinished() then return end

  self.shockwaves:forEach(function(shockwave) shockwave:update(dt) end)
  self.interval:update(dt)
end

--- Begins the pulsar and starts the first shockwave
function Pulsar:begin()
  self:addNewShockwave()
  self.interval:start()
end

--- @return boolean Whether the pulsar is finished
function Pulsar:isFinished()
  return self:hasAllShockwaves() and self.shockwaves:all(function(shockwave) return shockwave:isFinished() end)
end

--- @return boolean Whether all shockwaves have been created
--- @private
function Pulsar:hasAllShockwaves()
  return self.shockwaves:length() == self.options.numberOfShockwaves
end

--- Creates a new shockwave and adds it to the internal array.
--- @private
function Pulsar:addNewShockwave()
  local shockwave = Pulsar:createShockwaves(self.position, self.options)
  shockwave:activate()
  self.shockwaves:add(shockwave)
end

--- Called when the interval timer completes a cycle.
--- If all shocker have been created, this method will stop the interval timer. Otherwise, this method will add a new
--- shockwave.
--- @private
function Pulsar:onIntervalFinished()
  if self:hasAllShockwaves() then
    self.interval:stop()
    return
  end

  self:addNewShockwave()
end

--- Destroys all shockwaves and sets the instance to be destroyed
function Pulsar:destroy()
  self.shockwaves:forEach(function(shockwave) shockwave:destroy() end)
  self.interval:stop()
end

--- __Constructor__
--- @param position LuaVec3 The position of the centre of the pulsar
--- @param options PulsarOptions The list of options for all shockwaves
--- @return Pulsar The constructed instance
function Pulsar:new(position, options)
  local self = classes.setmetatable({}, Pulsar)
  self.position = position
  self.options = Pulsar:fillOptions(options)
  self.shockwaves = Array:new({})
  self.interval = IntervalTimer:new(self.options.interval, function() self:onIntervalFinished() end)
  return self
end

return Pulsar
