local classes = require("lua/explosionsNG/helpers/classes")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local VehicleAffectable = require("lua/explosionsNG/misc/VehicleAffectable")
local SoundEmitter = require("lua/explosionsNG/worldObjects/SoundEmitter")
local Timer = require("lua/explosionsNG/timers/Timer")

--- Properties:
---  - `colour` - the colour of the shockwave
---  - `maxAlpha` - the maximum alpha of the shockwave
---  - `audioTrack` - an audio track to play
---  - `modelName` - the name of the model to use for the shockwave in-game object
---  - `expansionFn` - a function to control the expansion from frame to frame
---  - `implode` - whether the shockwave starts off at max size and reduces in size over time. Defaults to false.
--- @alias ShockwaveOptions { colour: string | nil, maxAlpha: number | nil, audioTrack: string | nil, modelName: string | nil, expansionFn: ShockwaveExpansionFn | nil, implode: boolean | nil }
--- @alias ShockwaveExpansionFn fun(time: number): number @A function which should return the new radius of the shockwave, corresponding to the current amount of time elapsed. Note that both the input and the expected output should be normalised.

--- A shockwave that slowly fades out as it increases its range of effect. Note that the shockwave will stay at uniform
--- acceleration throughout its active period. This shockwave can affect vehicles and will call the `affectFn` to apply
--- an effect to vehicles. To start the shockwave, call the `activate` method.
--- @class Shockwave : VehicleAffectable
--- @field private position LuaVec3 The position of the shockwave
--- @field private maxRadius number The maximum radius the shockwave can be
--- @field private radius number The current radius of the shockwave
--- @field private timer Timer The life timer of the shockwave. On completion, the shockwave will be destroyed.
--- @field private active boolean Whether the shockwave is active
--- @field private isDestroyed boolean Whether the instance has been destroyed. This will only be `true` with external influence to the instance.
--- @field private waveObject SceneObject The shockwave in-game object
--- @field private soundEmitter SoundEmitter A SFX emitter to play sound. Will be `nil` if `options.audioTrack` is `nil`.
--- @field private options ShockwaveOptions A list of options
--- @field private affectFn VehicleEffectProducer A function that returns a new `VehicleEffect` instance to apply to any vehicle within range of this shockwave's effect
local Shockwave = classes.newClass(VehicleAffectable)

--- __Static__
---
--- Fills in any key not in the given `options` with default values. Note that this does not set a default for
--- `audioTrack`.
--- @param options ShockwaveOptions The list of predefined options
--- @return ShockwaveOptions The complete set of options with defaults filled in where empty values are given
function Shockwave:fillOptions(options)
  --- @type ShockwaveExpansionFn
  local defaultExpansionFn = function(time) return time end
  return {
    modelName = options.modelName or "sphere",
    colour = options.colour or "1 1 1",
    maxAlpha = options.maxAlpha or 0.125,
    audioTrack = options.audioTrack,
    expansionFn = options.expansionFn or defaultExpansionFn,
    implode = options.implode or false,
  }
end

--- @param vehicle Vehicle The vehicle to check
--- @return boolean Whether the `vehicle` can be affected by this object
--- @see VehicleAffectable#canAffect
function Shockwave:canAffect(vehicle)
  return vehicle:withinBounds(self.position, self.radius)
end

--- @param vehicle Vehicle The vehicle to affect
--- @return VehicularEffect The effect that can be applied to a vehicle
--- @see VehicleAffectable#affect
function Shockwave:affect(vehicle)
  local collisionPoint = vehicle:closestPointToBounds(self.position)
  return self.affectFn(self.id, { position = collisionPoint, vehicle = vehicle })
end

--- If this shockwave is not active or this shockwave `isFinished`, this method does nothing.
--- Calculates a new radius based on the time elapsed and updates the in-game object with this radius.
--- This method will then attempts to affect vehicles, an last updates the in-game sound emitter.
--- @param dt number The time since the last frame
function Shockwave:update(dt)
  if not self.active or self:isFinished() then return end

  if not self.timer:isComplete() then
    self:updateRadius()
    self:updateWaveObject()
    self:tryAndAffectVehicles()
    self.timer:update(dt)
  end

  if self.soundEmitter then self.soundEmitter:update(dt) end
end

--- Calculates a new radius based on the time elapsed and the options, specifically the `expansionFn` and `implode`
--- properties.
--- @private
function Shockwave:updateRadius()
  local timeElapsed = self.timer:elapsedRatio()
  local normalisedRadius = self.options.expansionFn(timeElapsed)
  local explosiveRadius = self.options.implode and 1 - normalisedRadius or normalisedRadius
  self.radius = math.min(self.maxRadius, explosiveRadius * self.maxRadius)
end

--- Sets the `waveObject`'s scale to the current value of `radius`, modified using the `toModelRadius` method.
--- Sets the colour of the `waveObject` using the alpha using the `getRadiusAlpha` method.
--- @private
function Shockwave:updateWaveObject()
  local alpha = self:getRadiusAlpha()
  self.waveObject:setField("scale", 0, self.radius.." "..self.radius.." "..self.radius)
  self.waveObject:setField("instanceColor", 0, self.options.colour.." "..alpha)
end

--- The colour alpha based on the amount of time the shockwave has existed for
--- @return number The colour alpha
--- @private
function Shockwave:getRadiusAlpha()
  return self.options.maxAlpha * self.timer:timeLeftRatio()
end

--- Releases the wave object and removes all effects on vehicles from this shockwave.
--- @private
function Shockwave:cleanupWave()
  if not self.waveObject then return end

  worldObjectsHandler.releaseTSStatic(self.waveObject)
  self.waveObject = nil

  self:removeAllEffects()
end

--- Destroys this instance, including releasing any in-use world objects and removing all effects on vehicles from this
--- shockwave.
function Shockwave:destroy()
  self:cleanupWave()

  if self.soundEmitter then
    self.soundEmitter:destroy()
  end
end

--- @return boolean Whether the explosive has now finished
function Shockwave:isFinished()
  local isSoundFinished = not self.soundEmitter or self.soundEmitter:isFinished()
  return self.timer:isComplete() and isSoundFinished
end

--- @return number The current `radius` of the shockwave
function Shockwave:getCurrentRadius()
  return self.radius
end

--- Sets the shockwave to be active.
--- Additionally, this method creates a sound emitter, if specified in the options.
function Shockwave:activate()
  self.active = true
  self.timer:start()

  local track = self.options.audioTrack
  if track then
    self.soundEmitter = SoundEmitter:new({ track = track, position = self.position, maxDistance = self.maxRadius * 3 })
  end
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param maxRadius number The maximum radius the shockwave can get to
--- @param time number The amount of time the shockwave exists for
--- @param affectFn VehicleEffectProducer A function that returns a new `VehicleEffect` instance to apply to any vehicle within range of this shockwave's effect
--- @param options? ShockwaveOptions | nil A list of options that can be passed. Default values will be used for any empty values
--- @return Shockwave The constructed instance
function Shockwave:new(position, maxRadius, time, affectFn, options)
  local self = classes.setmetatable(VehicleAffectable:new("Shockwave"), Shockwave)
  self.position = vec3(position)
  self.maxRadius = maxRadius
  self.timer = Timer:new(time, function() self:cleanupWave() end)
  self.affectFn = affectFn
  self.options = Shockwave:fillOptions(options or {})
  self.waveObject = worldObjectsHandler.requestTSStatic({
    modelName = self.options.modelName,
    position = position,
    instanceColour = self.options.colour,
    scale = 0
  })
  self.soundEmitter = nil
  self.radius = 0
  self.active = false
  self.isDestroyed = false
  return self
end

return Shockwave
