local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local vectors = require("lua/explosionsNG/helpers/vectors")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local Timer = require("lua/explosionsNG/timers/Timer")

--- A wrapper to a SFXEmitter object. Note that this class should only really be used for non-looping sounds. This class
--- can be used for the aforementioned case, however it will not provide any meaningful abstraction.
--- @class SoundEmitter
--- @field private emitter SceneObject The in-game SFXEmitter object
--- @field private timer Timer The timer to time the sound effect and remove the in-game object on completion
--- @field private finished boolean Whether the sound emitter has finished playing sounds
local SoundEmitter = classes.newClass()

--- __Static__
---
--- Returns the run-time (in seconds) of a particular `track`
--- @param track string The track to get the run-time of
--- @return number The run-time of the track
--- @private
function SoundEmitter:getTrackTime(track)
  if track == "generic_explosion"   then return 10   end
  if track == "lightning_explosion" then return 10   end
  if track == "supernova_explosion" then return  9.3 end
  if track == "lightning1"          then return  9   end
  if track == "lightning2"          then return  7   end
  if track == "lightning3"          then return  5   end
  if track == "emp"                 then return  3.4 end
  if track == "electric_sparks"     then return  4.6 end
  if track == "anomaly_damage"      then return  5.5 end
  if track == "projectile_launch"   then return  2   end
  if track == "star_ambient"        then return 18.6 end
  log("W", "explosivesNG", "Unknown SFX track to play "..tostring(track))
  return 1
end

--- __Static__
---
--- Returns a copy of the given options, with specific properties being overridden, based on the given track
--- @param track string The track to apply overrides of
--- @param options SFXEmitterOptions The copy with overridden options
function SoundEmitter:applyTrackOverrides(track, options)
  local copiedOptions = tables.copy(options)

  -- This track works best when the reference distance matches the maximum distance it reaches
  if track == "supernova_explosion" then
    copiedOptions.referenceDistance = copiedOptions.maxDistance
  end

  return copiedOptions
end

--- Updates the internal timer
--- @param dt number The time since the last frame
function SoundEmitter:update(dt)
  if self.finished then return end
  self.timer:update(dt)
end

--- Sets the sound emitter to the given position
--- @param position LuaVec3 The new position
function SoundEmitter:move(position)
  self.emitter:setField("position", 0, vectors.toString(position))
end

--- Releases the in-game emitter object to the world objects handler and sets this instance to be finished.
--- Note that any sounds will immediately stop playing, if called when the sound emitter is still active.
function SoundEmitter:destroy()
  if self.finished then return end

  self.finished = true
  worldObjectsHandler.releaseSoundEmitter(self.emitter)
  self.emitter = nil
end

--- @return boolean Whether this sound emitter has finished
function SoundEmitter:isFinished()
  return self.finished
end

--- __Constructor__
--- @param options SFXEmitterOptions A list of options to use
--- @return SoundEmitter The constructed instance
function SoundEmitter:new(options)
  local self = classes.setmetatable({}, SoundEmitter)
  self.emitter = worldObjectsHandler.requestSoundEmitter(SoundEmitter:applyTrackOverrides(options.track, options))
  self.timer = Timer:new(SoundEmitter:getTrackTime(options.track), function() self:destroy() end, true)
  self.finished = false
  return self
end

return SoundEmitter
