local classes = require("lua/explosionsNG/helpers/classes")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local Timer = require("lua/explosionsNG/timers/Timer")

--- Properties:
---  - `markerColour` - The colour of the marker object
---  - `areaColour` - The colour of the area object
---  - `markerMaxAlpha` - The maximum alpha of the marker object
---  - `areaMaxAlpha` - The maximum alpha of the area object
---  - `markerScale` - The scale of the marker object
---  - `time` - The time the target is present
---  - `showMarker` - Whether to show the marker or not. If this is `true`, then all other marker options are irrelevant. Defaults to `false`.
---  - `emitSound` - Whether to emit sound. Defaults to `true`.
---  - `reactive` - Whether the colour of the area reacts to whether there is a vehicle inside of it
---  - `onFinish` - A function to call when the target is finished displaying
--- @alias TargetOptions { markerColour: string | nil, areaColour: string | nil, markerMaxAlpha: number | nil, areaMaxAlpha: number | nil, markerScale: number | nil, time: number | nil, showMarker: boolean | nil, emitSound: boolean | nil, onFinish: (fun(): void) | nil, reactive: boolean | nil }

--- @class Target
--- @field private position LuaVec3 The position of the target
--- @field private areaRadius number The radius of the area
--- @field private options TargetOptions The options in use for this target
--- @field private object SceneObject The target in-game object
--- @field private area SceneObject The target in-game object representing a wider area
--- @field private timer Timer The timer that finishes after the time set in the options has elapsed
--- @field private soundEmitter SceneObject The in-game sound emitter
--- @field private isDestroyed boolean Whether the instance is destroyed
--- @field private DEFAULT_COLOUR string __Static__ The default colour for the marker and area colour for any target
--- @field private OCCUPIED_COLOUR string __Static__ The colour for the area colour if there is a vehicle within it
local Target = classes.newClass()
Target.DEFAULT_COLOUR = "1 0.5 0"
Target.OCCUPIED_COLOUR = "1 0 0"

--- __Static__
---
--- Fills in any option not set in the given options
--- @param options TargetOptions The options to fill
--- @return TargetOptions The filed options
--- @private
function Target:fillOptions(options)
  return {
    areaColour = options.areaColour or Target.DEFAULT_COLOUR,
    areaMaxAlpha = options.areaMaxAlpha or 0.35,
    markerColour = options.markerColour or Target.DEFAULT_COLOUR,
    markerMaxAlpha = options.markerMaxAlpha or 1,
    markerScale = options.markerScale or 1,
    time = options.time or 5,
    showMarker = options.showMarker or options.showMarker == nil,
    emitSound = options.emitSound or options.emitSound == nil,
    reactive = options.reactive or false,
    onFinish = options.onFinish
  }
end

--- Updates the colours of the marker and area colours
--- @param dt number The time since the last frame
function Target:update(dt)
  if self.timer:isComplete() then return end

  local ratio = self.timer:timeLeftRatio()

  local areaColour = self:getReactiveAreaColour()
  self.area:setField("instanceColor", 0, areaColour.." "..(self.options.areaMaxAlpha * ratio))

  if self.object then
    self.object:setField("instanceColor", 0, self.options.markerColour.." "..(self.options.markerMaxAlpha * ratio))
  end

  self.timer:update(dt)
end

--- Checks whether this target is react or not. If it isn't, this method will simply return the area colour. If it is,
--- then this method will check whether a vehicle is within the area, and if so, will return a colour dictated by
--- `Target.OCCUPIED_COLOUR`
--- @return string
function Target:getReactiveAreaColour()
  if not self.options.reactive then return self.options.areaColour end

  local vehicles = vehiclesHandler.getVehicles()
  local hasVehicleInside = vehicles:some(function(vehicle) return vehicle:isWithin(self.position, self.areaRadius) end)
  return hasVehicleInside and Target.OCCUPIED_COLOUR or self.options.areaColour
end

--- Removes all in-game objects and releases them to the world objects handler. If an `onFinish` handler was set in the
--- options, then this method will call this function as well.
function Target:destroy()
  if self.isDestroyed then return end

  if self.object then
    worldObjectsHandler.releaseTSStatic(self.object)
    self.object = nil
  end

  worldObjectsHandler.releaseTSStatic(self.area)
  self.area = nil

  if self.soundEmitter then
    worldObjectsHandler.releaseSoundEmitter(self.soundEmitter)
    self.soundEmitter = nil
  end

  local onFinish = self.options.onFinish
  if onFinish then onFinish() end

  self.isDestroyed = true
end

--- __Constructor__
--- @param position LuaVec3 The position to place the target at
--- @param areaRadius number The radius of the area target. It is expected that this should be greater than the `objectScale`.
--- @param options? TargetOptions | nil The list of options to use
--- @return Target The constructed instance
function Target:new(position, areaRadius, options)
  local self = classes.setmetatable({}, Target)
  self.position = position
  self.areaRadius = areaRadius
  self.options = Target:fillOptions(options or {})
  self.object = self.options.showMarker
      and worldObjectsHandler.requestTSStatic({
        dynamic = false,
        modelName = "sphere_marker",
        instanceColour = self.options.markerColour.." "..self.options.markerMaxAlpha,
        position = position,
        scale = self.options.markerScale
      })
  self.area = worldObjectsHandler.requestTSStatic({
    modelName = "sphere",
    instanceColour = self.options.areaColour.." "..self.options.areaMaxAlpha,
    position = position,
    scale = areaRadius,
  })
  self.soundEmitter = self.options.emitSound and worldObjectsHandler.requestSoundEmitter({ track = "ping", position = position, maxDistance = 300 })
  self.timer = Timer:new(self.options.time, function() self:destroy() end, true)
  self.isDestroyed = false
  return self
end

return Target
