local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local Array = require("lua/explosionsNG/misc/Array")
local Events = require("lua/explosionsNG/misc/Events")
local Timer = require("lua/explosionsNG/timers/Timer")
local anomalyAnimationStateMachine = require("lua/explosionsNG/worldObjects/anomalies/animations")
local AnomalyLayer = require("lua/explosionsNG/worldObjects/anomalies/AnomalyLayer")
local SoundEmitter = require("lua/explosionsNG/worldObjects/SoundEmitter")

--- Properties:
---   - `scale`          - the scale of the anomaly. A scale of `1` will equate to a radius of `UNIT_RADIUS` metres.
---                        Defaults to `1.5`.
---   - `life`           - the lifetime (in seconds) this anomaly will stay open for. Defaults to `30`
---   - `numberOfLayers` - how many layers make up the anomaly. Defaults to `10`
--- @alias AnomalyOptions { scale: number | nil, life: number | nil, numberOfLayers: number | nil, colour: string | nil }
--- @alias AnomalyQueueableState "fluctuating" | "discharging" | "locking" | "unlocking" | "damaging" | "shimmering"
--- @alias AnomalyState "ready" | "opening" | "idle" | "locked" | "closing" | "violentlyClosing" | "exploding" | "closed" | "destroyed" | AnomalyQueueableState
--- @alias AnomalyEvent "fluctuating" | "discharging" | "opened" | "closed" | "locked" | "unlocked" | "passive" | "damaged" | "shimmered" | "violentlyClosed" @A list of events that can be registered to
--- @alias AnomalyReleaseFn fun(anomaly: Anomaly): void A function that passes the cleaned up anomaly back to the caller

--- A spherical object made up of triangular "shards" in multiple layers.
---
--- It should be noted that the documentation for this class will refer to anomalies being "open". An anomaly is
--- considered "open" if its state is in one of the `AnomalyQueueableState`s or "idle". See the next section on how to
--- access the state diagram of this mod to better illustrate this.
---
--- __States and Lifecycle__
---
--- (_It should be noted that a lifecycle diagram of an anomaly is depicted in the README (or accessible directly) in
--- the root folder of this mod_)
---
--- An anomaly can be in a single state at a time (see `AnomalyState`). The typical progression of an anomaly will be
--- `ready -> opening -> idle -> closing -> closed -> destroyed`. This state sequence will happen when the anomaly is
--- created by a caller and then not interacted with, other than by use of the `update` method. There are additional
--- states that the anomaly can be in once it has become `"idle"`. These states are considered "queueable states" (see
--- `AnomalyQueueableState`). By calling the `queueState` function, the respective states are added to the current queue
--- of states. If the anomaly is currently `idle`, then the anomaly will change state to the next state in the queue. It
--- should be noted that these states will decrease the lifetime of the anomaly, and the anomaly can begin closing
--- before the queue is empty. Furthermore, an anomaly which is not yet "idle", or has begun closing will not attempt to
--- change to one of theses queueable states; in the case of the latter, any queued up states will be ignored
--- permanently as an anomaly can never open again once it has begun to close/has closed. In the state diagram linked
--- above, it can be seen that once the anomaly has entered either the "closing" or "violentlyClosing" states, the
--- anomaly can never become idle again and so the queue of states will forever be ignored.
--- @class Anomaly : EventedDynamicObject<AnomalyEvent>
--- @field private state AnimationStateMachine<AnomalyState, Anomaly> The animation state machine of the anomaly
--- @field private layers Array<AnomalyLayer> The list of layers
--- @field private options AnomalyOptions The options of the anomaly
--- @field private light SceneObject The in-game point light object
--- @field private ambientSoundEmitter SceneObject The in-game SFX emitter object for ambient sounds
--- @field private terminalSoundEmitter SceneObject The in-game SFX emitter object for opening and closing sounds
--- @field private damageSoundEmitter SoundEmitter The sound emitter for when the anomaly is damaged
--- @field private endingSoundTimer Timer A timer that lasts for the duration of the `CLOSING_SOUND_PERIOD` and that will cleanup all game objects when completed
--- @field private position LuaVec3 The position of the centre of the anomaly
--- @field private rotation number The rotation of the first layer of the anomaly about the z-axis (subsequent layers will have a value based on this one)
--- @field private queuedStates Array<{ state: AnomalyQueueableState, context: any }> A list of queued anomaly states that the anomaly can switch to when it is open
--- @field private events Events<AnomalyEvent> A list of event handlers per state
--- @field private halfwayQueuedState boolean A flag to store whether the anomaly, when in an open state, has called its callback function halfway into the animation
--- @field private BASE_ROTATION_MULTIPLIER number __Static__ The rotation multiplier that all rotations of this anomaly will be based on
--- @field private LIGHT_BRIGHTNESS number __Static__ The brightness of the light at the centre of the anomaly
--- @field private AFTER_CLOSING_PERIOD number __Static__ The total time (in seconds) after the anomaly has closed to wait for the `anomaly_closing` sound file to finish
--- @field private UNIT_RADIUS number __Static__ The in-game radius of the layer model at a scale of `1`
--- @field public FLUCTUATION_PERIOD number __Static__ The time an anomaly will spend during the `"fluctuating"` state
--- @field public DISCHARGING_PERIOD number __Static__ The time an anomaly will spend during the `"discharging"` state
--- @field public LOCKING_SWITCH_PERIOD number __Static__ The time an anomaly will spend during the `"locking"` or `"unlocking"` state
--- @field private DAMAGING_PERIOD number __Static__ The time an anomaly will spend during the `"damaging"` state
--- @field private SHIMMERING_PERIOD number __Static__ The time the anomaly will spend during the `"shimmering"` state
--- @field private EXPLODING_PERIOD number __Static__ The time the anomaly will spend during the `"exploding"` state
--- @field private MAX_NON_EXPLODING_SCALE_DEVIATION number __Static__ The maximum deviation (from 1) the scale of the anomaly can change by during regular state animations
--- @field private MAX_EXPLODING_SCALE_DEVIATION number __Static__ The maximum deviation (from 1) the scale of the anomaly can change by during the exploding state animations
local Anomaly = classes.newClass()
Anomaly.BASE_ROTATION_MULTIPLIER = 0.5
Anomaly.LIGHT_BRIGHTNESS = 10
Anomaly.UNIT_RADIUS = 12.195
Anomaly.AFTER_CLOSING_PERIOD = 5
Anomaly.FLUCTUATION_PERIOD = 1.5
Anomaly.DISCHARGING_PERIOD = 2.5
Anomaly.LOCKING_SWITCH_PERIOD = 0.75
Anomaly.DAMAGING_PERIOD = 2.5
Anomaly.SHIMMERING_PERIOD = 3
Anomaly.EXPLODING_PERIOD = 10
Anomaly.MAX_NON_EXPLODING_SCALE_DEVIATION = 0.375
Anomaly.MAX_EXPLODING_SCALE_DEVIATION = 0.475

--- __Static__
---
--- Sets and returns a given options object with the default values
--- @param options AnomalyOptions A set of partially filled options
--- @return AnomalyOptions The `options` object with defaults set for undefined keys
--- @private
function Anomaly:fillOptions(options)
  return {
    scale = options.scale or 1.5,
    life = options.life or 30,
    numberOfLayers = options.numberOfLayers or 10,
    colour = options.colour or "1 0.5 0 1"
  }
end

--- __Static__
---
--- Returns the light radius for a given `scale`
--- @param scale number The anomaly scale
--- @return number The scale of a light for the given `scale`
--- @private
function Anomaly:scaleToLightRadius(scale)
  return scale * Anomaly.UNIT_RADIUS * 5
end

--- __Static__
---
--- Returns the SFX emitter reference distance for a given `scale`
--- @param scale number The anomaly scale
--- @return number The reference distance of a SFX emitter for the given `scale`
--- @private
function Anomaly:scaleToSfxRefDistance(scale)
  return scale * Anomaly.UNIT_RADIUS * 5
end

--- __Static__
---
--- Returns the SFX emitter maximum distance for a given `scale`
--- @param scale number The anomaly scale
--- @return number The maximum distance of a SFX emitter for the given `scale`
--- @private
function Anomaly:scaleToSfxMaxDistance(scale)
  return scale * Anomaly.UNIT_RADIUS * 12
end

--- __Static__
---
--- Returns the radius of the outer-most layer from the perspective of the anomaly being fully open and unchanging
--- @param scale number The scale of the anomaly
--- @return number The radius of the outer-most layer from the perspective of the anomaly being fully open and unchanging
function Anomaly:getStaticRadiusOf(scale)
  return Anomaly.UNIT_RADIUS * scale
end

--- __Static__
---
--- Returns the radius of the outer-most layer from the perspective of the anomaly being as expanded as possible. The
--- `couldExplode` parameter determines whether the returned radius should consider the state where the anomaly is
--- exploding. In this case, the return radius will be greater than if this is not the case.
--- @param scale number The scale of the anomaly
--- @param couldExplode? boolean Whether the anomaly could explode
--- @return number The radius of the outer-most layer from the perspective of the anomaly being as expanded as possible
function Anomaly:getMaxRadiusOf(scale, couldExplode)
  local deviation = couldExplode and Anomaly.MAX_EXPLODING_SCALE_DEVIATION or Anomaly.MAX_NON_EXPLODING_SCALE_DEVIATION
  local unitScaleWithDeviation = 1 + deviation
  return Anomaly.UNIT_RADIUS * unitScaleWithDeviation * scale
end

--- Creates and returns a new layer of an anomaly in-game at the current anomaly position.
--- The scale is initially set to `0`.
--- @param index number The index of the layer to create for
--- @return AnomalyLayer The create layer
--- @private
function Anomaly:createLayer(index)
  local layerRatio = index / self.options.numberOfLayers
  -- Value between [0, 2PI]
  local rotationalPhase = layerRatio * 2 * math.pi
  return AnomalyLayer:new(self.position, index, layerRatio, rotationalPhase)
end

--- Creates all necessary objects in-game for this anomaly.
--- @private
function Anomaly:createGameObjects()
  self.layers = Array:from(self.options.numberOfLayers, function(i) return self:createLayer(i) end)

  self.light = worldObjectsHandler.requestPointLight({
    position = self.position,
    radius = 0,
    colour = self.options.colour,
    brightness = Anomaly.LIGHT_BRIGHTNESS
  })

  self.ambientSoundEmitter = worldObjectsHandler.requestSoundEmitter(self:getSfxOptions("anomaly_ambient", true))
  self.terminalSoundEmitter = worldObjectsHandler.requestSoundEmitter(self:getSfxOptions("anomaly_opening", false))
end

--- Creates and returns a list of options to apply to an SFX emitter for this anomaly
--- @param isAmbient boolean Whether to return options for an ambient emitter
--- @return SFXEmitterOptions A list of options to use for SFX emitter in-game objects for this anomaly
--- @private
function Anomaly:getSfxOptions(track, isAmbient)
  local scale = self.options.scale
  return {
    track = track,
    position = self.position,
    referenceDistance = Anomaly:scaleToSfxRefDistance(scale),
    maxDistance = Anomaly:scaleToSfxMaxDistance(scale),
    isLooping = isAmbient
  }
end

--- Updates the state of the anomaly, including its current animation
--- @param dt number The time since the last frame
function Anomaly:update(dt)
  self.state:update(dt)

  if self.damageSoundEmitter then
    self.damageSoundEmitter:update(dt)
  end
end

--- If the `animationTime` is half of the `statePeriod`, this method calls the callback function once
--- @param callbackFn fun(): void A callback function to call when the animation is halfway to the state period
--- @param animationTime number The time into the animation
--- @param statePeriod number The total period of time the anomaly will be in this state
--- @private
function Anomaly:checkHalfwayQueuedStatePeriod(callbackFn, animationTime, statePeriod)
  if not self.halfwayQueuedState and animationTime >= statePeriod / 2 then
    callbackFn()
    self.halfwayQueuedState = true
  end
end

--- Decrease the `life` of this anomaly by the `dt` time
--- @param dt number The time since the last frame
--- @private
function Anomaly:decreaseLife(dt)
  self.life = math.max(0, self.life - dt)
end

--- Increases the `rotation` by the increment given.
--- For each layer, this method sets the rotation of that layer based on the anomaly's total `rotation`.
--- See `rotateLayer` for details.
--- @param increment number The amount to add to the rotation
--- @private
function Anomaly:rotateLayers(increment)
  self.rotation = self.rotation + increment * Anomaly.BASE_ROTATION_MULTIPLIER
  self.layers:forEach(function(layer) layer:rotate(increment) end)
end

--- Sets the scale of all in-game objects to the given scale
--- @param scale number The new scale to set
--- @private
function Anomaly:applyScale(scale)
  self.layers:forEach(function(layer) layer:scale(scale) end)
  self.light:setField("radius", 0, Anomaly:scaleToLightRadius(scale))
end

--- Sets the scale of all in-game objects to the given scale
--- @param scale number The new scale to set
--- @private
function Anomaly:applyLayerScale(scale)
  self.layers:forEach(function(layer) layer:scale(scale) end)
end

--- Scales the layers of the anomaly, given a function that produces a scale for each layer
--- @param scaleFn fun(index: number, numberOfLayers: number, baseScale: number): number A function that should return the scale for a particular layer, given its index, the number of layers in this anomaly and the base scale of the anomaly itself
--- @private
function Anomaly:applyNonUniformScale(scaleFn)
  local numberOfLayers = self.options.numberOfLayers
  local baseScale = self.options.scale

  self.layers:forEach(function(layer, index)
    local scale = scaleFn(index, numberOfLayers, baseScale)
    layer:scale(scale)
  end)

  local lastLayerScale = scaleFn(numberOfLayers, numberOfLayers, baseScale)
  self.light:setField("radius", 0, Anomaly:scaleToLightRadius(lastLayerScale))
end

--- Sets the colour of the light of the anomaly to the given colour
--- @param colour string The new colour to set the anomaly light to
--- @private
function Anomaly:changeColour(colour)
  self.light:setField("color", 0, colour)
end

--- Sets the alpha of the layers of the anomaly to the given value
--- @param alpha number The new alpha to set the anomaly layers to
--- @private
function Anomaly:setAlpha(alpha)
  self.layers:forEach(function(layer) layer:setAlpha(alpha) end)
end

--- Opens the anomaly and creates the necessary in-game objects, including the sound emitter object.
--- Note that this method should only be called on an anomaly whose state is `"ready"`.
function Anomaly:open()
  if self.state:getState() ~= "ready" then
    log("W", "explosivesNG", "An anomaly whose state is not 'ready' cannot be opened")
    return
  end

  self.state:switch("opening")
  self:createGameObjects()
end

--- Switches the internal state to the given one and calls any event handler that is registered to the given event
--- @param newState AnomalyState The new state to switch to
--- @param event AnomalyEvent The event type that is causing the change
--- @private
function Anomaly:onStateChange(newState, event)
  self.state:switch(newState)
  self.events:emit(event)
end

--- Switches the anomaly's state to be closing.
--- Stops playing the ambient sound effect and begins playing the relevant closing sound effect.
--- @param violentClosure? boolean Whether to close the anomaly violently, culmination in an explosion
function Anomaly:beginClosing(violentClosure)
  local newState = violentClosure and "violentlyClosing" or "closing"
  self.state:switch(newState)

  worldObjectsHandler.releaseSoundEmitter(self.ambientSoundEmitter)
  self.ambientSoundEmitter = nil

  worldObjectsHandler.releaseSoundEmitter(self.terminalSoundEmitter)
  local terminalSound = violentClosure and "anomaly_explosion" or "anomaly_closing"
  self.terminalSoundEmitter = worldObjectsHandler.requestSoundEmitter(self:getSfxOptions(terminalSound, false))

  self.endingSoundTimer:start()
end

--- Sets the state of the anomaly to be closed, and sets the scale of all layers to `0`.
--- @private
function Anomaly:close()
  self.state:switch("closed")
  self.layers:forEach(function(layer) layer:scale(0) end)
end

--- Releases the in-game objects to the world objects handler and sets the `state` to `"destroyed"`.
--- This method will then emit the `"closed"` event.
function Anomaly:destroy()
  local state = self.state:getState()

  if state == "destroyed" then return end

  self.state:switch("destroyed")

  if state == "ready" then return end

  self.layers:forEach(AnomalyLayer.destroy)

  worldObjectsHandler.releasePointLight(self.light)
  self.light = nil
  worldObjectsHandler.releaseSoundEmitter(self.terminalSoundEmitter)
  self.terminalSoundEmitter = nil

  -- If this method is called while the anomaly is idle, this will still be present
  if self.ambientSoundEmitter then
    worldObjectsHandler.releaseSoundEmitter(self.ambientSoundEmitter)
  end

  self.events:emit("closed")
end

--- @return boolean Whether the anomaly has been closed and all in-game objects have been destroyed
function Anomaly:isFullyClosed()
  return self.state:getState() == "destroyed"
end

--- @return number The radius of the outer-most layer from the perspective of the anomaly being fully open and unchanging
function Anomaly:getStaticRadius()
  return Anomaly:getStaticRadiusOf(self.options.scale)
end

--- @return number The radius of the outer-most layer from the perspective of the anomaly being as expanded as possible
function Anomaly:getMaxRadius()
  return Anomaly:getMaxRadiusOf(self.options.scale)
end

--- @return LuaVec3 The current position of the anomaly
function Anomaly:getPosition()
  return vec3(self.position)
end

--- Sets the centre of the anomaly to the given centre, and updates the in-game objects
--- @param newPosition LuaVec3 The new position to set
function Anomaly:setPosition(newPosition)
  self.position = vec3(newPosition)
  self.layers:forEach(function(layer) layer:setPosition(newPosition) end)

  local positionString = vectors.toString(newPosition)
  self.light:setField("position", 0, positionString)
  self.ambientSoundEmitter:setField("position", 0, positionString)

  if self.terminalSoundEmitter then
    self.terminalSoundEmitter:setField("position", 0, positionString)
  end

  if self.damageSoundEmitter and not self.damageSoundEmitter:isFinished() then
    self.damageSoundEmitter:move(newPosition)
  end
end

--- Creates a sound emitter for the damaging sound of the anomaly
--- @private
function Anomaly:createDamageSound()
  if self.damageSoundEmitter and not self.damageSoundEmitter:isFinished() then
    log("W", "explosivesNG", "Damage sound emitter is already active - sound will not be overwritten.")
    return
  end

  self.damageSoundEmitter = SoundEmitter:new(self:getSfxOptions("anomaly_damage", false))
end

--- Queues up the given state for this anomaly. See class documentation for details on queued states.
--- The context should be given if the state being passed is `"shimmering"`, a should include the desired colour as the
--- context parameter.
--- @param state AnomalyQueueableState The state to queue
--- @param context? any Any context parameters to use for the animation. These should be given dependent on the state being passed.
function Anomaly:queueState(state, context)
  if state ~= "unlocking" then
    self.queuedStates:add({ state = state, context = context })
    return
  end

  -- TODO Could implement a state machine to define allowed state transitions?
  if self.state:getState() ~= "locked" then
    log("W", "explosivesNG", "Cannot unlock an anomaly which is not locked")
    return
  end

  self.state:switch("unlocking")
end

--- @return Events<AnomalyEvent> A registrable set of events that can be registered to or unregistered from
function Anomaly:getEvents()
  return self.events
end

--- __Constructor__
--- @param position LuaVec3 The centre position of the anomaly
--- @param options? AnomalyOptions | nil An optional list of options to use. See `fillOptions` method for defaults and an explanation of each property.
--- @return Anomaly The constructed instance
function Anomaly:new(position, options)
  local self = classes.setmetatable({}, Anomaly)
  self.layers = Array:new(--[[--- @type AnomalyLayer[] ]] {})
  self.options = Anomaly:fillOptions(options or {})
  self.light = nil
  self.ambientSoundEmitter = nil
  self.terminalSoundEmitter = nil
  self.damageSoundEmitter = nil
  self.endingSoundTimer = Timer:new(Anomaly.AFTER_CLOSING_PERIOD, function() self:destroy() end)
  self.position = vec3(position)
  self.rotation = 0
  self.life = self.options.life
  self.queuedStates = Array:new(--[[--- @type AnomalyQueueableState[] ]] {})
  self.state = anomalyAnimationStateMachine(self)
  self.events = Events:new()
  self.halfwayQueuedState = false
  return self
end

return Anomaly
