local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")

--- @class AnomalyLayer : DynamicObject
--- @field private rotation number The rotation angle of the layer
--- @field private rotationDirection number The direction the layer is rotating in. `1` if rotating clockwise, `-1` if rotating anticlockwise
--- @field private SCALE_MULTIPLIER number The multiplier to apply to when the scale of this layer is changed
--- @field private LAYER_ANGLE number The angle to add onto the current rotation, and will be used to set the X and Y axis rotation angles to. See `rotate` method.
--- @field private object SceneObject The in-game object of the layer
local AnomalyLayer = classes.newClass()

--- Sets the scale to apply to this layer. The true scale will be multiplied by this layer's `SCALE_MULTIPLIER`.
--- @param baseScale number The base scale to apply to this layer
function AnomalyLayer:scale(baseScale)
  local scale = baseScale * self.SCALE_MULTIPLIER
  self.object:setField("scale", 0, scale .." ".. scale .." ".. scale)
end

--- Rotates the layer about the Z axis by the increment given
--- @param increment number The amount to add to the rotation
function AnomalyLayer:rotate(increment)
  self.rotation = self.rotation + increment

  local layerRotation = vec3(self.LAYER_ANGLE, self.LAYER_ANGLE, self.LAYER_ANGLE + self.rotation)
  local directionalRotation = layerRotation * self.rotationDirection

  self.object:setField("rotationEuler", 0, vectors.toString(directionalRotation))
end

--- Places the in-game layer to the given position
--- @param newPosition LuaVec3 The new position to set
function AnomalyLayer:setPosition(newPosition)
  self.object:setField("position", 0, vectors.toString(newPosition))
end

--- Sets the alpha of the layer to the given value
--- @param alpha number The new alpha to set the layer to
function AnomalyLayer:setAlpha(alpha)
  self.object:setField("instanceColor", 0, "1 1 1 "..alpha)
end

--- Releases the in-game layer object to the world objects handler
function AnomalyLayer:destroy()
  if not self.object then return end

  worldObjectsHandler.releaseTSStatic(self.object)
  self.object = nil
end

--- __Constructor__
--- @param position LuaVec3 The position of the layer
--- @param layerNumber number The position (zero-based) of this layer in the associated anomaly
--- @param layerRatio number The ratio of this layer's position in the associated anomaly, with respect to the innermost layer of the anomaly. `0` if this is the first layer, `1` if this is the last layer, linear between.
--- @param layerAngle number The angle to add onto the current rotation, and will be used to set the X and Y axis rotation angles to. See `rotate` method.
--- @return AnomalyLayer The constructed instance
function AnomalyLayer:new(position, layerNumber, layerRatio, layerAngle)
  local self = classes.setmetatable({}, AnomalyLayer)
  self.rotation = 0
  -- Odd indexed layers rotate the other way
  self.rotationDirection = layerNumber % 2 == 0 and 1 or -1
  self.SCALE_MULTIPLIER = layerRatio
  self.LAYER_ANGLE = layerAngle
  self.object = worldObjectsHandler.requestTSStatic({ modelName = "anomaly_layer", position = position, scale = 0 })
  return self
end

return AnomalyLayer