local classes = require("lua/explosionsNG/helpers/classes")
local LinearPath = require("lua/explosionsNG/pathing/LinearPath")
local Target = require("lua/explosionsNG/worldObjects/Target")
local Anomaly = require("lua/explosionsNG/worldObjects/anomalies/Anomaly")

--- @alias MovingAnomalyOptions { pathTime: number | nil, target: TargetOptions | nil }

--- An `Anomaly` object that starts at a given point and moves linearly to another point. The anomaly will lock while at
--- the starting point and, once locked, will transition to the ending point, where it will then unlock and the
--- `onComplete` callback will fire.
--- @class MovingAnomaly
--- @field private path LinearPath The path the anomaly will take
--- @field private anomaly Anomaly The anomaly that is moving
--- @field private isMoving boolean Whether the anomaly is currently moving
--- @field private target Target The representation of where the anomaly will move to next
--- @field private onComplete fun(): void A callback function to call when the movement is complete
--- @field private MOVEMENT_INCREMENT number The increment the anomaly will move along in (ms^-1)
local MovingAnomaly = classes.newClass()

--- @param options MovingAnomalyOptions
--- @return MovingAnomalyOptions
function MovingAnomaly:fillOptions(options)
  return {
    pathTime = options.pathTime or 3,
    target = options.target or {}
  }
end

--- Updates the anomaly and the path it is moving on
--- @param dt number The time since the last frame
function MovingAnomaly:update(dt)
  if not self.anomaly then return end
  if self.isMoving then self:updateMoving(dt) end
  self.target:update(dt)
  self.anomaly:update(dt)
end

--- Updates the anomaly and the path it is moving on. If the anomaly has now reached the end of its path, it is unlocked
--- @param dt number The time since the last frame
--- @private
function MovingAnomaly:updateMoving(dt)
  self:moveAnomalyAlongPath(dt)

  if self.path:hasReachedEndPoint() then
    self.anomaly:queueState("unlocking")
    self.isMoving = false
  end
end

--- Updates the path the anomaly is moving along and sets the anomaly's position to the new position
--- @param dt number The time since the last frame
--- @private
function MovingAnomaly:moveAnomalyAlongPath(dt)
  self.path:addDistance(dt * self.MOVEMENT_INCREMENT)
  local newPosition = self.path:getCurrentPosition()
  self.anomaly:setPosition(newPosition)
end

--- Sets up all event handlers on the anomaly and begins locking the anomaly
--- @private
function MovingAnomaly:initAnomaly()
  local anomalyEvents = self.anomaly:getEvents()
  local lockedFn = function() self.isMoving = true end

  --- Cleans up the anomaly event handlers, before finally calling the `onComplete` callback.
  --- The function is defined like this so it can refer to itself.
  local unlockedFn
  unlockedFn = function()
    anomalyEvents:unregisterEventHandler("locked", lockedFn)
    anomalyEvents:unregisterEventHandler("unlocked", unlockedFn)
    self.onComplete()
  end

  anomalyEvents:registerEventHandler("locked", lockedFn)
  anomalyEvents:registerEventHandler("unlocked", unlockedFn)
  self.anomaly:queueState("locking")
end

--- Destroys this instance, including the target. this will __NOT__ destroy the underlying anomaly instance.
function MovingAnomaly:destroy()
  self.target:destroy()
end

--- __Constructor__
--- @param endPoint LuaVec3 The target position for the anomaly to move to
--- @param anomaly Anomaly The anomaly that is moving
--- @param areaCoverageRadius number The area that the anomaly can cover (for example, for an attack range)
--- @param onComplete fun(): void A callback function to call when the movement is complete
--- @param options? MovingAnomalyOptions | nil The list of options to use. Note that the `time` and `markerScale` for the `target` will be overridden
--- @return MovingAnomaly The constructed instance
function MovingAnomaly:new(endPoint, anomaly, areaCoverageRadius, onComplete, options)
  local filledOptions = MovingAnomaly:fillOptions(options or {})
  local targetOptions = --[[--- @not nil ]] filledOptions.target
  -- Add the locking time so that the target will disappear after the anomaly has moved (in its locked state) to the target location
  targetOptions.time = filledOptions.pathTime + Anomaly.LOCKING_SWITCH_PERIOD
  targetOptions.markerScale = anomaly:getStaticRadius()

  local self = classes.setmetatable({}, MovingAnomaly)
  self.anomaly = anomaly
  self.onComplete = onComplete
  self.path = LinearPath:new(anomaly:getPosition(), endPoint)
  self.isMoving = false
  self.target = Target:new(endPoint, areaCoverageRadius, targetOptions)
  self.MOVEMENT_INCREMENT = self.path:getLength() / filledOptions.pathTime
  self:initAnomaly()
  return self
end

return MovingAnomaly
