local maths = require("lua/explosionsNG/helpers/maths")
local strings = require("lua/explosionsNG/helpers/strings")
local AnimationStateMachine = require("lua/explosionsNG/animations/AnimationStateMachine")

------------------------------------------------------------------------------------------------------------------------
--- Constants
------------------------------------------------------------------------------------------------------------------------

-- The maximum speed at which the anomaly can spin (in rads^-1) while in the `opening` state
local MAX_OPENING_SPIN_SPEED = 10
-- The maximum speed at which the anomaly can spin (in rads^-1) while in the `discharging` state
local MAX_DISCHARGING_SPIN_SPEED = 3
-- The time it takes for the anomaly to fully contract, and also the same time it takes to fully expand to the original scale, while in the `discharging` state
local DISCHARGING_DEVIATION_TIME = 0.75
-- The maximum deviation of scale that the anomaly can contract by while in the `fluctuation` state
local MAX_FLUCTUATION_DEVIATION = 0.2
-- The time (in seconds) in the `anomaly_closing` sound file that the anomaly audibly closes
local CLOSING_SOUND_CLIMAX_PERIOD = 5.2
-- The time (in seconds) in the `anomaly_opening` sound file that the anomaly audibly opens
local OPENING_SOUND_CLIMAX_PERIOD = 5.2
-- The time the anomaly will spend from the last expansion to the final contraction, when in the `closing` or `exploding` state
local CLOSING_TIME = 0.75
-- The time the anomaly will spend expanding and contracting, when in the `closing` state
local CLOSING_INSTABILITY_TIME = CLOSING_SOUND_CLIMAX_PERIOD - CLOSING_TIME
-- The minimum scale of the anomaly when in the `locked`
local LOCK_SCALE = 0.35
-- The maximum speed at which the anomaly can spin (in rads^-1) while in the `locked` or `locking` state
local MAX_LOCK_SPIN_SPEED = 2
-- The minimum scale of the anomaly when in the `damaging` state
local DAMAGING_SCALE = 0.9
-- The time it takes to fully change from one colour to another
local COLOUR_CHANGE_TIME = 0.25
-- The time (in seconds) in the `anomaly_exploding` sound file the anomaly audibly closes
local EXPLODING_SOUND_CLOSED = 7.4
-- The time (in seconds) in the `anomaly_exploding` sound file the anomaly audibly explodes
local EXPLODING_SOUND_EXPLODE = 7.6
-- The time (in seconds) between the closing part and the exploding part in the `anomaly_exploding` sound file that there is a gap in audio
local EXPLODING_WAIT_TIME = EXPLODING_SOUND_EXPLODE - EXPLODING_SOUND_CLOSED
-- The time the anomaly will spend expanding and contracting, when in the `violentlyClosing` state
local EXPLODING_INSTABILITY_TIME = EXPLODING_SOUND_CLOSED - CLOSING_TIME
-- The multiplier to apply to the normalised scale calculation when in the `exploding` state
local EXPLODING_SCALE_MULTIPLIER = 50
-- The multiplier to apply to the normalised spin speed calculation when in the `exploding` state
local EXPLODING_SPIN_SPEED_MULTIPLIER = 0.3

------------------------------------------------------------------------------------------------------------------------
--- Helper Functions
------------------------------------------------------------------------------------------------------------------------

--- Applies the following to the given anomaly:
---  - A non-uniform scale from the given ratio
---  - A rotation from the given ratio
---  - A decrease in life
--- @param anomaly Anomaly The anomaly to animate
--- @param dt number The time since the last frame
--- @param ratio number The amount to contract/expand by
--- @param targetScale number The scale at which when the `ratio` is `1`, all layers of the anomaly are at
local function contractExpandLocking(anomaly, dt, ratio, targetScale)
  local linearNorm = 1 - (1 - targetScale) * maths.smoothRampUp(ratio)

  --- Each layer gets a scale factor based on the layer number.
  --- For instance, in contraction:
  ---  - Initially: ratio = 0, scale = default scale
  ---  - Finally: ratio = 1, scale = LOCK_SCALE for all layers
  ---  - Each layer will therefore get a slightly different gradient so that they diverge when ratio = 0 and converge at
  ---    ratio = 1 (vice versa for expansion)
  --- @param index number The index of the layer to get the scale of
  --- @param numberOfLayers number The number of layers in the anomaly
  --- @param baseScale number The base scale of the anomaly
  --- @return number The scale for the given layer
  local scaleFn = function(index, numberOfLayers, baseScale)
    local gradient = (numberOfLayers / index) - 1
    local layerScaleFactor = gradient * ratio + 1
    return baseScale * linearNorm * layerScaleFactor
  end

  anomaly:applyNonUniformScale(scaleFn)
  anomaly:rotateLayers(dt * (1 + ratio * MAX_LOCK_SPIN_SPEED))
  anomaly:decreaseLife(dt)
end

--- Changes the state to `"idle"` with the `"passive"` event source trigger
--- @param anomaly Anomaly The anomaly changing state
local function onQueuedStateChange(anomaly)
  anomaly:onStateChange("idle", "passive")
  anomaly.halfwayQueuedState = false
end

--- Changes the colour of the given anomaly which follows a smooth, plateau curve whereby the anomaly changes colour to
--- the given colour at the middle points of the animation, before returning back to the original anomaly colour.
--- @param anomaly Anomaly The anomaly to change the colour of
--- @param colourRatio number The normalised ratio along the animation to change the colour by
--- @param targetColour string The target colour of the anomaly when fully coloured
local function smoothChangeColour(anomaly, colourRatio, targetColour)
  local mainColour = --[[--- @not nil ]] anomaly.options.colour
  -- This function is using x (colourRatio) as a normalised value so the last parameter is `1` to match this
  local smoothedColourRatio = maths.smoothPlateau(colourRatio, COLOUR_CHANGE_TIME, 1)

  local mainColourRGB = strings.split(mainColour)
  local targetColourRGB = strings.split(targetColour)

  --- Returns a new colour for a channel which follows a linear path with respect to the `smoothedColourRatio` value
  --- @param default number The anomaly's default colour for this channel
  --- @param target number The target colour to change to
  local function linearChannelColour(default, target)
    local gradient = target - default
    return gradient * (smoothedColourRatio - 1) + target
  end

  local red   = linearChannelColour(tonumber(mainColourRGB[1]), tonumber(targetColourRGB[1]))
  local green = linearChannelColour(tonumber(mainColourRGB[2]), tonumber(targetColourRGB[2]))
  local blue  = linearChannelColour(tonumber(mainColourRGB[3]), tonumber(targetColourRGB[3]))

  anomaly:changeColour(red .." ".. green .." ".. blue .." 1")
end

--- Returns a function that will return a scale, governed by a mathematical formula that allows for ever-increasing
--- sinusoidal 'wiggling' (see https://www.desmos.com/calculator/frufey8abf for the formula).
--- @param wigglingPeriod number
--- @param closePeriod number
--- @param scaleDeviation number
local function wigglingFn(wigglingPeriod, closePeriod, scaleDeviation)
  local f = function(x) return ((x / wigglingPeriod) * scaleDeviation) * math.sin((9 * math.pi * x) / (2 * wigglingPeriod)) + 1 end
  local g = function(x) return math.max(0, f(wigglingPeriod) * (1 - ((x - wigglingPeriod) / closePeriod) ^ 3)) end
  return function(x) if x <= wigglingPeriod then return f(x) end return g(x) end
end

------------------------------------------------------------------------------------------------------------------------
--- Animation Functions
------------------------------------------------------------------------------------------------------------------------

--- Increases the scale by a factor amount of the given dt`
--- @type AnimationTimeFn<Anomaly>
local opening = function (anomaly, dt, timeContext)
  local scale = timeContext.elapsedRatio * anomaly.options.scale
  anomaly:applyScale(scale)

  local rotation = dt + dt * (MAX_OPENING_SPIN_SPEED - 1) * timeContext.timeLeftRatio
  anomaly:rotateLayers(rotation)
end

-- TODO Animations should not be responsible for this
--- This method first checks if the lifetime is now `0`, and calls `beginClosing` if so.
--- Otherwise, this method checks whether there is a queued state in `queuedStates`. If so, this method changes the
--- `state` to the next queued state.
--- In neither of these conditions is true, this method:
---   - Reduces the lifetime of the anomaly by the `dt` time.
---   - Rotates the layers with the `dt` time.
--- @type AnimationFn<Anomaly>
local idle = function(anomaly, dt)
  if anomaly.life == 0 then
    anomaly:beginClosing()
    return
  end

  -- Check if there is a queued state
  if anomaly.queuedStates:isPopulated() then
    local nextState = anomaly.queuedStates:pop()
    anomaly.state:switch(nextState.state, nextState.context)

    if nextState.state == "damaging" then
      anomaly:createDamageSound()
    end

    return
  end

  anomaly:decreaseLife(dt)
  anomaly:rotateLayers(dt)
end

--- Fluctuates the scale of the anomaly such that it increases in scale and then decreases until it reaches the original
--- scale.
--- This method rotates the layers with the dt time and decreases the life of this anomaly.
--- @type AnimationTimeFn<Anomaly>
local fluctuating = function(anomaly, dt, timeContext)
  local smoothedTime = maths.repeatingHill(timeContext.elapsedRatio)
  local scale = (1 + MAX_FLUCTUATION_DEVIATION * smoothedTime) * anomaly.options.scale

  anomaly:applyScale(scale)
  anomaly:rotateLayers(dt)
  anomaly:decreaseLife(dt)

  local callback = function() anomaly.events:emit("fluctuating") end
  anomaly:checkHalfwayQueuedStatePeriod(callback, timeContext.elapsed, anomaly.FLUCTUATION_PERIOD)
end

--- Sets the scale to an amount governed by a mathematical formula that contracts the anomaly for a period of time
--- before expanding back to the original scale (see `maths.smoothPlateau` for details).
---
--- This method also rotates the layers of the anomaly by the same contraction/expansion of the scale and decreases the
--- life of this anomaly.
--- @type AnimationTimeFn<Anomaly>
local discharging = function(anomaly, dt, timeContext)
  local discharge = maths.smoothPlateau(timeContext.elapsed, DISCHARGING_DEVIATION_TIME, anomaly.DISCHARGING_PERIOD)
  local scale = (1 - discharge * anomaly.MAX_NON_EXPLODING_SCALE_DEVIATION) * anomaly.options.scale

  anomaly:applyScale(scale)
  anomaly:rotateLayers(dt * (1 + discharge * MAX_DISCHARGING_SPIN_SPEED))
  anomaly:decreaseLife(dt)

  local callback = function() anomaly.events:emit("discharging") end
  anomaly:checkHalfwayQueuedStatePeriod(callback, timeContext.elapsed, anomaly.DISCHARGING_PERIOD)
end

--- Changes the colour of the anomaly from what the default is to a fully red hue, and expands all inner layers to match
--- the outermost layer's scale (similar to the `locking` or `unlocking` state).
--- Both animations are smoothened and return back to the original state of the anomaly after finishing.
--- @type AnimationTimeFn<Anomaly>
local damaging = function(anomaly, dt, timeContext)
  local elapsedRatio = timeContext.elapsedRatio
  smoothChangeColour(anomaly, elapsedRatio, "1 0 0")
  contractExpandLocking(anomaly, dt, maths.repeatingHill(elapsedRatio), DAMAGING_SCALE)
end

--- Changes the colour of the anomaly from what the default is to the given colour from the animation context.
--- The animation is smoothened and returns back to the original state of the anomaly after finishing.
--- @type AnimationTimeFn<Anomaly>
local shimmering = function(anomaly, dt, timeContext, context)
  if not context or type(context) ~= "string" then
    log("E", "explosivesNG", "Anomaly is in the 'shimmering' state but no colour was given")
    return
  end

  local elapsedRatio = timeContext.elapsedRatio
  smoothChangeColour(anomaly, elapsedRatio, context)
  anomaly:rotateLayers(dt)
end

--- Sets the scale to an amount governed by the `wigglingFn` function.
--- This method also rotates the layers of the anomaly by the dt time, rotating faster and faster as time elapses.
--- @type AnimationTimeFn<Anomaly>
local violentlyClosing = function(anomaly, dt, timeContext)
  local timeElapsed = timeContext.elapsed

  local scaleFn = wigglingFn(EXPLODING_INSTABILITY_TIME, CLOSING_TIME, anomaly.MAX_EXPLODING_SCALE_DEVIATION)
  local scale = scaleFn(timeElapsed) * anomaly.options.scale

  anomaly:applyScale(scale)
  anomaly:rotateLayers((1 + timeElapsed) * dt)
end

--- Waits for the exploding to commence, before scaling the anomaly's layers outwards linearly, and gradually phasing
--- the alpha of the layers to `0`. The anomaly rotates at a slow pace before also phasing out.
--- @type AnimationTimeFn<Anomaly>
local exploding = function(anomaly, dt, timeContext)
  if timeContext.elapsed < EXPLODING_WAIT_TIME then return end

  local scale = (1 - math.exp(-1.5 * timeContext.elapsedRatio)) * EXPLODING_SCALE_MULTIPLIER * anomaly.options.scale
  anomaly:applyLayerScale(scale)
  anomaly:setAlpha(timeContext.timeLeftRatio)
  anomaly:rotateLayers(dt * timeContext.timeLeftRatio * EXPLODING_SPIN_SPEED_MULTIPLIER)
end

--- Sets the scale to an amount governed by the `wigglingFn` function.
--- This method also rotates the layers of the anomaly by the dt time.
--- @type AnimationTimeFn<Anomaly>
local closing = function(anomaly, dt, timeContext)
  local scaleFn = wigglingFn(CLOSING_INSTABILITY_TIME, CLOSING_TIME, anomaly.MAX_NON_EXPLODING_SCALE_DEVIATION)
  local scale = scaleFn(timeContext.elapsed) * anomaly.options.scale
  anomaly:applyScale(scale)
  anomaly:rotateLayers(dt)
end

--- Updates the ending sound of the anomaly
--- @type AnimationTimeFn<Anomaly>
local closed = function(anomaly, dt)
  anomaly.endingSoundTimer:update(dt)
end

--- Decreases the scale and increases the rotation of the anomaly
--- @type AnimationTimeFn<Anomaly>
local lock = function(anomaly, dt, timeContext)
  contractExpandLocking(anomaly, dt, timeContext.elapsedRatio, LOCK_SCALE)
end

--- Updates the rotation of the anomaly
--- @type AnimationFn<Anomaly>
local locked = function(anomaly, dt)
  anomaly:rotateLayers(dt * MAX_LOCK_SPIN_SPEED)
end

--- Increases the scale and decreases the rotation of the anomaly back to normal
--- @type AnimationTimeFn<Anomaly>
local unlock = function(anomaly, dt, timeContext)
  contractExpandLocking(anomaly, dt, timeContext.timeLeftRatio, LOCK_SCALE)
end

--- Returns a new set of animations to animate the given anomaly for every state it can be in
--- @param anomaly Anomaly The star to animate
--- @return AnimationStateMachine<AnomalyState, Anomaly> The star animations for the given star
return function(anomaly)
  local animations = --[[--- @type AnimationStateMachine<AnomalyState, Anomaly> ]] AnimationStateMachine:new(anomaly)
  animations:add("opening", opening, function() anomaly:onStateChange("idle", "opened") end, OPENING_SOUND_CLIMAX_PERIOD)
  animations:add("idle", idle)
  animations:add("locking", lock, function() anomaly:onStateChange("locked", "locked") end, anomaly.LOCKING_SWITCH_PERIOD)
  animations:add("locked", locked)
  animations:add("unlocking", unlock, function() anomaly:onStateChange("idle", "unlocked") end, anomaly.LOCKING_SWITCH_PERIOD)
  animations:add("fluctuating", fluctuating, function() onQueuedStateChange(anomaly) end, anomaly.FLUCTUATION_PERIOD)
  animations:add("discharging", discharging, function() onQueuedStateChange(anomaly) end, anomaly.DISCHARGING_PERIOD)
  animations:add("damaging", damaging, function() anomaly:onStateChange("idle", "damaged") end, anomaly.DAMAGING_PERIOD)
  animations:add("shimmering", shimmering, function() anomaly:onStateChange("idle", "shimmered") end, anomaly.SHIMMERING_PERIOD)
  animations:add("violentlyClosing", violentlyClosing, function() anomaly:onStateChange("exploding", "violentlyClosed") end, EXPLODING_SOUND_CLOSED)
  animations:add("exploding", exploding, function() anomaly:close() end, anomaly.EXPLODING_PERIOD)
  animations:add("closing", closing, function() anomaly:close() end, CLOSING_SOUND_CLIMAX_PERIOD)
  animations:add("closed", closed)
  animations:addNoop("ready")
  animations:addNoop("destroyed")

  -- Set initial state
  animations:switch("ready")

  return animations
end