local classes = require("lua/explosionsNG/helpers/classes")
local tables = require("lua/explosionsNG/helpers/tables")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local ExplosiveArea = require("lua/explosionsNG/explosives/containers/ExplosiveArea")
local Events = require("lua/explosionsNG/misc/Events")
local LightningExplosive = require("lua/explosionsNG/explosives/composed/lightning/LightningExplosive")
local ShockwaveExplosive = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosive")
local OrbitalStrikeExplosive = require("lua/explosionsNG/explosives/contained/orbitalStrike/OrbitalStrikeExplosive")

--- @alias AreaAttackOnStartFn fun(attackType: AreaAttackType): void

--- @alias AreaAttackType "lightning" | "implosion" | "orbitalStrike"
--- @alias AreaAttack { explosiveProducer: (fun(position: LuaVec3): AnyExplosive), soundEffects: ExplosiveAreaSoundEffectOptions | nil, areaOptions: ExplosiveAreaOptions }

--- Properties
---  - `attackType` - The type of attack to perform
---  - `numberOfExplosives` - The number of explosives in the attack
---  - `targetDelay` - The delay between a target appearing for an explosive and the explosive detonating
---  - `areaRadius` - The radius of the area the attack can occur in
---  - `effectRadius` - The radius of the effect area (i.e. the shockwave area)
---  - `placement` - The placement to use when placing the explosives
---  - `interval` - The interval between two sequential explosives appearing in the attack. If not given, explosives occur simultaneously. This should only be given if the `attackType` is not `"lightning"`.
---  - `lightningOptions` - A list of options for a lightning attack bolt. This will only be used if the `attackType` is `"lightning"`.
---  - `shockwaveOptions` - A list of options for an implosion attack shockwave. This will only be used if the `attackType` is `"implosion"`.
---  - `pulsarOptions` - A list of options for an orbital strike attack pulsar. This will only be used if the `attackType` is `"orbitalStrike"`.
--- @alias AreaAttackPatternSetting { attackType: AreaAttackType, numberOfExplosives: number, targetDelay: number, areaRadius: number | nil, effectRadius: number | nil, placement: ExplosiveAreaPlacement | nil, interval: number | nil, lightningOptions: LightningOptions | nil, shockwaveOptions: ShockwaveOptions | nil, pulsarOptions: PulsarOptions | nil }

--- @class AreaAttackPattern : AttackPattern
--- @field private anomaly Anomaly The anomaly performing the attacks
--- @field private attackSettings Array<AreaAttackPatternSetting> The list of attacks to perform
--- @field private explosiveArea ExplosiveArea<AnyExplosive> The explosive area
--- @field private awaitingPlayerVehicle boolean Whether this pattern is waiting for the player to be in a vehicle
--- @field private FORCE_MULTIPLIER number __Static__ The force multiplier of all shockwave-based attacks
--- @field private MAX_TARGET_AREA_ALPHA number __Static__ The maximum alpha to use for all targets created by attacks
local AreaAttackPattern = classes.newClass()
AreaAttackPattern.FORCE_MULTIPLIER = 5
AreaAttackPattern.MAX_TARGET_AREA_ALPHA = 0.09

--- __Static__
---
--- Combines a partial list of `ExplosiveAreaOptions` with a list of `AreaAttackPatternOptions`. This method will also
--- use the given sound effects and persist those in the returned options.
--- @param areaOptions ExplosiveAreaOptions The area options
--- @param soundEffects ExplosiveAreaSoundEffectOptions | nil A set of sound effects to play at various events
--- @param patternOptions AreaAttackPatternSetting The pattern options
--- @return ExplosiveAreaOptions The combined options into one area options object
--- @private
function AreaAttackPattern:combineOptions(areaOptions, soundEffects, patternOptions)
  local interval = (patternOptions.attackType == "implosion" or patternOptions.attackType == "orbitalStrike")
      and patternOptions.interval
      or 0

  return {
    soundEffects = soundEffects,
    interval = interval,
    areaRadius = patternOptions.areaRadius,
    numberOfExplosives = patternOptions.numberOfExplosives,
    placement = patternOptions.placement,
    explosiveRadius = areaOptions.explosiveRadius,
    targetOptions = areaOptions.targetOptions
  }
end

--- __Static__
---
--- Returns a new lightning attack. Note that sound will not be emitted from the lightning object, even if it is set to
--- `true` in the given options.
--- @param attackSetting AreaAttackPatternSetting The attack setting
--- @param onVehicleAffected ShockwaveVehicleAffectedFn A function to when an explosive affects a vehicle
--- @return AreaAttack The lightning attack
--- @private
function AreaAttackPattern:getLightningAttack(attackSetting, onVehicleAffected)
  return {
    explosiveProducer = function(position)
      return --[[--- @type AnyExplosive ]] LightningExplosive:new(position, {
        options = {
          shockwave = {
            time = 1,
            maxRadius = attackSetting.effectRadius,
            forceMultiplier = AreaAttackPattern.FORCE_MULTIPLIER,
            onVehicleAffected = onVehicleAffected
          },
          lightning = tables.merge(--[[--- @type LightningOptions ]] attackSetting.lightningOptions or {}, {
            emitSound = false
          })
        }
      })
    end,
    soundEffects = {
      firstDetonation = "lightning_explosion",
      firstTarget = "ping"
    },
    areaOptions = {
      explosiveRadius = attackSetting.effectRadius,
      targetOptions = {
        areaMaxAlpha = AreaAttackPattern.MAX_TARGET_AREA_ALPHA,
        areaColour = "0 0.5 1",
        time = attackSetting.targetDelay,
        emitSound = false,
        reactive = true
      }
    }
  }
end

--- __Static__
---
--- Returns a new implosion attack. Note that the  shockwave will always implode inwards, even if it is set to `false`
--- in the given options.
--- @param attackSetting AreaAttackPatternSetting The attack setting
--- @param onVehicleAffected ShockwaveVehicleAffectedFn A function to when an explosive affects a vehicle
--- @return AreaAttack The implosion attack
--- @private
function AreaAttackPattern:getImplosionAttack(attackSetting, onVehicleAffected)
  return {
    explosiveProducer = function(position)
      return --[[--- @type AnyExplosive ]] ShockwaveExplosive:new(position, {
        options = {
          time = 1,
          maxRadius = attackSetting.effectRadius,
          forceMultiplier = AreaAttackPattern.FORCE_MULTIPLIER,
          onVehicleAffected = onVehicleAffected,
          shockwaveOptions = tables.merge(--[[--- @type ShockwaveOptions ]] attackSetting.shockwaveOptions or {}, {
            implode = true
          })
        }
      })
    end,
    areaOptions = {
      explosiveRadius = attackSetting.effectRadius,
      targetOptions = {
        areaMaxAlpha = AreaAttackPattern.MAX_TARGET_AREA_ALPHA,
        areaColour = "1 1 0",
        time = attackSetting.targetDelay,
        reactive = true
      }
    }
  }
end

--- __Static__
---
--- Returns a new orbital strike attack. Note that the shockwave will always implode inwards, even if it is set to
--- `false` in the given options.
--- @param attackSetting AreaAttackPatternSetting The attack setting
--- @param onVehicleAffected ShockwaveVehicleAffectedFn A function to when an explosive affects a vehicle
--- @return AreaAttack The orbital strike attack
--- @private
function AreaAttackPattern:getOrbitalStrikeAttack(attackSetting, onVehicleAffected)
  return {
    explosiveProducer = function(position)
      return --[[--- @type AnyExplosive ]] OrbitalStrikeExplosive:new(position, {
        options = {
          forceMultiplier = AreaAttackPattern.FORCE_MULTIPLIER,
          pulsarOptions = tables.merge(--[[--- @type PulsarOptions ]] attackSetting.pulsarOptions or {}, {
            maxRadius = attackSetting.effectRadius,
            onVehicleAffected = onVehicleAffected
          }),
          laserOptions = {
            colour = "0 1 0",
            radius = attackSetting.effectRadius
          }
        }
      })
    end,
    areaOptions = {
      explosiveRadius = attackSetting.effectRadius,
      targetOptions = {
        areaMaxAlpha = AreaAttackPattern.MAX_TARGET_AREA_ALPHA,
        areaColour = "0 1 0",
        time = attackSetting.targetDelay,
        reactive = true
      }
    }
  }
end

--- __Static__
---
--- Returns a new `AreaAttack`, given it attack settings
--- @param attackSetting AreaAttackPatternSetting The attack setting
--- @param onVehicleAffected ShockwaveVehicleAffectedFn A function to when an explosive affects a vehicle
--- @return AreaAttack The attack that matches the given attack setting
--- @private
function AreaAttackPattern:getAttack(attackSetting, onVehicleAffected)
  if attackSetting.attackType == "lightning" then
    return AreaAttackPattern:getLightningAttack(attackSetting, onVehicleAffected)
  end

  if attackSetting.attackType == "implosion" then
    return AreaAttackPattern:getImplosionAttack( attackSetting, onVehicleAffected)
  end

  if attackSetting.attackType == "orbitalStrike" then
    return AreaAttackPattern:getOrbitalStrikeAttack(attackSetting, onVehicleAffected)
  end
end

--- __Static__
---
--- Returns a new `ExplosiveArea` for the given attack. This method will activate this area before returning it.
--- @param attackSetting AreaAttackPatternSetting The attack setting
--- @param centre LuaVec3 The centre position to set for the area
--- @param onVehicleAffected ShockwaveVehicleAffectedFn A function to when an explosive affects a vehicle
--- @return ExplosiveArea<AnyExplosive> A targeted area for the attack type given
--- @private
function AreaAttackPattern:createExplosiveArea(attackSetting, centre, onVehicleAffected)
  local attack = AreaAttackPattern:getAttack(attackSetting, onVehicleAffected)
  local attackOptions = AreaAttackPattern:combineOptions(attack.areaOptions, attack.soundEffects, attackSetting)

  local explosiveArea = ExplosiveArea:new(centre, attack.explosiveProducer, { options = attackOptions })
  explosiveArea:activate()
  return explosiveArea
end

--- If the explosive area is not yet finished, this method updates the area and anomaly, before checking whether the
--- area is now finished.
--- @param dt number The time since the last frame
function AreaAttackPattern:update(dt)
  if self.awaitingPlayerVehicle then
    self.anomaly:update(dt)
    return
  end

  if self.explosiveArea:isDetonationFinished() then return end

  self.anomaly:update(dt)
  self.explosiveArea:update(dt)

  if self.explosiveArea:isDetonationFinished() then
    self:postAttack()
  end
end

--- Begins the first attack
function AreaAttackPattern:begin()
  self:nextAttack()
end

--- Performs the next attack in the setting array.
--- If this array is empty, this method will call the `onFinish` function.
--- @private
function AreaAttackPattern:postAttack()
  if self.attackSettings:isEmpty() then
    self.events:emit("finished")
    return
  end

  self:nextAttack()
end

--- Creates a new area attack from the next setting and centres it around the player.
--- If the player vehicle is not present, this method will wait for the player vehicle to be spawned. See the
--- `waitForPlayer` method for more details.
--- @private
function AreaAttackPattern:nextAttack()
  local player = vehiclesHandler.getPlayerVehicle()

  if not player then
    self:waitForPlayer()
    return
  end

  self:createAndBeginAreaAttack(player)
end

--- Creates a new attack from the first element inside the settings array
--- @param player Vehicle The player vehicle to centre the attack area around
--- @private
function AreaAttackPattern:createAndBeginAreaAttack(player)
  local setting = self.attackSettings:pop()
  local centre = player:getPosition()
  local onVehicleAffected = function(vehicle) self.events:emit("vehicleAffected", vehicle) end
  self.explosiveArea = AreaAttackPattern:createExplosiveArea(setting, centre, onVehicleAffected)

  self.events:emit("started", setting.attackType)
end

--- Registers a callback with the vehicles' handler, which will fire when the player vehicle is spawned. This callback
--- will then perform the tasks specified by the `createAndBeginAreaAttack` method.
--- @private
function AreaAttackPattern:waitForPlayer()
  self.awaitingPlayerVehicle = true

  vehiclesHandler.registerHandler("playerSpawned", function(playerVehicle)
    self:createAndBeginAreaAttack(playerVehicle)
    self.awaitingPlayerVehicle = false
  end)
end

--- @return Events<AttackPatternEvent> A registrable set of events that can be registered to or unregistered from
function AreaAttackPattern:getEvents()
  return self.events
end

--- Destroys all attacks being performed. This will __NOT__ destroy the underlying anomaly object.
function AreaAttackPattern:destroy()
  self.explosiveArea:deactivate()
end

--- __Constructor__
--- @param anomaly Anomaly The anomaly performing the attacks
--- @param attackSettings Array<AreaAttackPatternSetting> The list of attacks to perform
--- @return AreaAttackPattern The constructed instance
function AreaAttackPattern:new(anomaly, attackSettings)
  local self = classes.setmetatable({}, AreaAttackPattern)
  self.anomaly = anomaly
  self.attackSettings = attackSettings
  self.events = Events:new()
  self.explosiveArea = nil
  self.awaitingPlayerVehicle = false
  return self
end

return AreaAttackPattern
