local classes = require("lua/explosionsNG/helpers/classes")
local positioning = require("lua/explosionsNG/helpers/positioning")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local Timer = require("lua/explosionsNG/timers/Timer")
local Events = require("lua/explosionsNG/misc/Events")
local Anomaly = require("lua/explosionsNG/worldObjects/anomalies/Anomaly")
local AttackingAnomaly = require("lua/explosionsNG/worldObjects/anomalies/attacks/AttackingAnomaly")
local MovingAnomaly = require("lua/explosionsNG/worldObjects/anomalies/MovingAnomaly")

--- Properties:
---  - `radius` - The radius of the attack
---  - `time` - The total time the attack will last for
---  - `type` - The type of attack to perform
---  - `numberOfAttacks` - The total number of attacks this attack pattern will do
---  - `cooldown` - The cooldown (in seconds) between one attack and the anomaly moving
---  - `movementTime` - The time (in seconds) it takes for the anomaly to move from one position to the next
---  - `accuracyMultiplier` - A multiple to increase the accuracy of the anomaly's movement. A lower value will mean less accuracy.
--- @alias AssaultAttackPatternOptions { radius: number, time: number, type: AttackingAnomalyAttack, numberOfAttacks: number, cooldown: number, movementTime: number, accuracyMultiplier: number }

--- An `AttackPattern` with a single anomaly that moves around, following the player and attacking them
--- @class AssaultAttackPattern : AttackPattern
--- @field private anomaly Anomaly The anomaly performing the attacks
--- @field private attackOptions AssaultAttackPatternOptions The list of options being used
--- @field private attackingAnomaly AttackingAnomaly The `anomaly`, wrapped in an attacking object
--- @field private attackTimer Timer The timer that countdowns the time until the next attack
--- @field private attacksMade number The number of attacks currently made so far
--- @field private awaitingPlayerVehicle boolean Whether this pattern is waiting for the player to be in a vehicle
--- @field private isDestroyed boolean Whether the instance has been destroyed. This will only be `true` with external influence to the instance.
--- @field private movementPhase MovingAnomaly | nil The phase movement between two attacks
--- @field protected BASE_MOVEMENT_ACCURACY number __Static__ How fuzzy/precise the anomaly movement will be (in m) i.e. if the anomaly wants to move to a point, it will actually move to a point within Xm of that point.
local AssaultAttackPattern = classes.newClass()
AssaultAttackPattern.BASE_MOVEMENT_ACCURACY = 50

--- Updates the movement of the anomaly (if it is moving) or the attacking state of the anomaly.
--- If the attack pattern has finished then this method will emit the `"finished" event, and cease to do anything on
--- subsequent method calls.
--- In the event this pattern is waiting for a player vehicle to be present, this method will simply update the anomaly.
--- @param dt number The time since the last frame
function AssaultAttackPattern:update(dt)
  if self.awaitingPlayerVehicle then
    self.attackingAnomaly:update(dt)
    return
  end

  if self:isFinished() then return end

  if self.movementPhase then
    (--[[--- @not nil ]] self.movementPhase):update(dt)
    -- Update any ongoing attacks
    self.attackingAnomaly:updateAttacks(dt)
    return
  end

  self.attackingAnomaly:update(dt)
  self.attackTimer:update(dt)

  if self:isFinished() then
    self.attackingAnomaly:destroy()
    self.attackingAnomaly = nil
    self.events:emit("finished")
  end
end

--- Generates a random position within a set range of the player, governed by the `MOVEMENT_ACCURACY` value.
--- @param player Vehicle The player vehicle
--- @return LuaVec3 The generated position
--- @protected
function AssaultAttackPattern:generateNewPosition(player)
  -- The total time includes the movement time, plus the locking/unlocking time at each end for the anomaly
  local time = self.attackOptions.movementTime + 2 * Anomaly.LOCKING_SWITCH_PERIOD
  local fuzziness = AssaultAttackPattern.BASE_MOVEMENT_ACCURACY / self.attackOptions.accuracyMultiplier

  local predictedPosition = positioning.predictPosition(player, time, fuzziness)

  -- Account for the anomaly radius
  local anomalyRadius = self.anomaly:getStaticRadius()
  local finalAltitude = predictedPosition.z + anomalyRadius

  return vec3(predictedPosition.x, predictedPosition.y, finalAltitude)
end

--- Moves the anomaly to a random position within a set range of the player, and sets up a new movement phase
--- @param player Vehicle The player vehicle
--- @private
function AssaultAttackPattern:moveAnomalyToPlayer(player)
  local targetPosition = self:generateNewPosition(player)
  local onComplete = function() self:attack() end
  --[[--- @type MovingAnomalyOptions ]] local options = { pathTime = self.attackOptions.movementTime }
  self.movementPhase = MovingAnomaly:new(targetPosition, self.anomaly, self.attackOptions.radius, onComplete, options)
end

--- Increments the number of attacks made by this anomaly. If there are still more attacks to be made, the timer is
--- restarted.
--- @private
function AssaultAttackPattern:postAttack()
  self.attacksMade = self.attacksMade + 1

  if not self:hasMadeAllAttacks() then
    self.attackTimer:restart()
  end
end

--- Moves the anomaly to the next position, based on the position of the player.
--- If the player vehicle is not present, this method will wait for the player vehicle to be spawned. See the
--- `waitForPlayer` method for more details.
--- @private
function AssaultAttackPattern:preAttack()
  local player = vehiclesHandler.getPlayerVehicle()

  if not player then
    self:waitForPlayer()
    return
  end

  self:moveAnomalyToPlayer(player)
end

--- Removes the movement phase add queues up another attack by the anomaly
--- @private
function AssaultAttackPattern:attack()
  self.movementPhase = nil

  self.attackingAnomaly:attack({
    animation = "discharging",
    time = self.attackOptions.time,
    maxRadius = self.attackOptions.radius,
    attack = self.attackOptions.type,
    onAttackFn = function() self:postAttack() end
  })
end

--- Registers a callback with the vehicles' handler, which will fire when the player vehicle is spawned. This callback
--- will then perform the tasks specified by the `moveAnomalyToPlayer` method.
--- @private
function AssaultAttackPattern:waitForPlayer()
  self.awaitingPlayerVehicle = true

  vehiclesHandler.registerHandler("playerSpawned", function(playerVehicle)
    self:moveAnomalyToPlayer(playerVehicle)
    self.awaitingPlayerVehicle = false
  end)
end

--- @return boolean Whether all attacks have been performed by the anomaly and it has finished attacking completely. If the instance is destroyed, this method will return `true`.
--- @private
function AssaultAttackPattern:isFinished()
  return self.isDestroyed or self:hasMadeAllAttacks() and (not self.attackingAnomaly or not self.attackingAnomaly:isAttacking())
end

--- @return boolean Whether all attacks have been performed by the anomaly
--- @private
function AssaultAttackPattern:hasMadeAllAttacks()
  return self.attacksMade == self.attackOptions.numberOfAttacks
end

--- Creates the attacking anomaly instance and registers to the vehicle affecting event
--- @param anomaly Anomaly The anomaly performing the attacks
--- @return AttackingAnomaly The created attacking anomaly
--- @private
function AssaultAttackPattern:createAttackingAnomaly(anomaly)
  local attackingAnomaly = AttackingAnomaly:new(anomaly)
  attackingAnomaly:registerOnVehicleAffected(function(vehicle) self.events:emit("vehicleAffected", vehicle) end)
  return attackingAnomaly
end

--- @return Events<AttackPatternEvent> A registrable set of events that can be registered to or unregistered from
function AssaultAttackPattern:getEvents()
  return self.events
end

--- Destroys all attacks being performed. This will __NOT__ destroy the underlying anomaly object.
function AssaultAttackPattern:destroy()
  if self.isDestroyed then return end

  if self.movementPhase then
    (--[[--- @not nil ]] self.movementPhase):destroy()
  end

  self.attackingAnomaly:destroy()
  self.attackTimer:stop()
  self.isDestroyed = true
end

--- __Constructor__
--- @param anomaly Anomaly The anomaly performing the attacks
--- @param attackOptions AssaultAttackPatternOptions The list of options to use
--- @return AssaultAttackPattern The constructed instance
function AssaultAttackPattern:new(anomaly, attackOptions)
  local self = classes.setmetatable({}, AssaultAttackPattern)
  self.anomaly = anomaly
  self.attackOptions = attackOptions
  self.attackingAnomaly = self:createAttackingAnomaly(anomaly)
  self.movementPhase = nil
  self.events = Events:new()
  self.attackTimer = Timer:new(attackOptions.cooldown, function() self:preAttack() end)
  self.attacksMade = 0
  self.awaitingPlayerVehicle = false
  self.isDestroyed = false
  self:preAttack() -- Begin the first attack immediately
  return self
end

return AssaultAttackPattern
