local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local Shockwave = require("lua/explosionsNG/worldObjects/Shockwave")
local effectFactory = require("lua/explosionsNG/vehicles/effects/factory")

--- @alias AttackingAnomalyAnimationState "fluctuating" | "discharging"
--- @alias AttackingAnomalyAttack "force" | "emp" | "fire" | "dissolve"
--- @alias AttackingAnomalyQueuedAttack { attack: AttackingAnomalyAttack, animation: AttackingAnomalyAnimationState, maxRadius: number, time: number, onAttackFn: (fun(): void) | nil }

--- A wrapper to an `Anomaly` that can be set to perform different attacks. This class manages those attacks and works
--- on a queue based system with FIFO rules.
---
--- __Important: Callers MUST explicitly call the `destroy` method when finished with an instance to properly release
--- the underlying anomaly object for further use outside of this instance. If this is not done, the event handling with
--- that anomaly will be blocked or perform incorrectly.__
--- @class AttackingAnomaly
--- @field private anomaly Anomaly The anomaly that will perform the attacks
--- @field private currentAttack AttackingAnomalyQueuedAttack The current attack of the anomaly
--- @field private queuedAttacks Array<AttackingAnomalyQueuedAttack> The list of currently queue attacks
--- @field private shockwaves Array<Shockwave> The list of active shockwaves from attacks
--- @field private onVehicleAffectFn VehicleAffectFn A function to call whenever a vehicle is affected
--- @field private unregisterFn fun(): void A function that cleanups all registered events with the anomaly
--- @field private ATTACK_FORCE_MULTIPLIER number __Static__ The multiplier of force that the `"force"` applies to a vehicle
local AttackingAnomaly = classes.newClass()
AttackingAnomaly.ATTACK_FORCE_MULTIPLIER = 12

--- __Static__
---
--- Returns the shockwave options for a particular attack
--- @param attack AttackingAnomalyAttack The attack to get the shockwave options of
--- @return ShockwaveOptions The shockwave options for the attack
--- @private
function AttackingAnomaly:getAttackOptions(attack)
  if attack == "force"    then return { colour = "1 1 1", audioTrack = "emp" } end
  if attack == "emp"      then return { colour = "0 1 0", audioTrack = "emp" } end
  if attack == "fire"     then return { colour = "1 0 0", audioTrack = "emp" } end
  if attack == "dissolve" then return { colour = "1 1 0", audioTrack = "emp" } end
end

--- __Static__
---
--- Returns a `VehicleEffectProducer` for the given attack. Note that the optional parameter is only currently necessary for a
--- `"force"` affect.
--- @param attack AttackingAnomalyAttack The attack to get the vehicle affect function for
--- @param maxRadius? number The maximum radius an effect can have. Note this only applies to the `"force"` affect
--- @param onVehicleAffectFn? VehicleAffectFn A function to call when a vehicle is affected by this function
--- @return VehicleEffectProducer The affect function
--- @private
function AttackingAnomaly:getAffectFn(attack, maxRadius, onVehicleAffectFn)
  return function(id, event)
    if onVehicleAffectFn then onVehicleAffectFn(event.vehicle) end

    if attack == "force"    then return effectFactory.createForceEffect(id, event.position, maxRadius, AttackingAnomaly.ATTACK_FORCE_MULTIPLIER) end
    if attack == "emp"      then return effectFactory.createEmpEffect(id) end
    if attack == "fire"     then return effectFactory.createFireEffect(id) end
    if attack == "dissolve" then return effectFactory.createDissolveEffect(id) end
  end
end

--- Updates the state of the anomaly and any active shockwaves from attacks. This method will then attempt to attack
--- again.
--- @param dt number The time since the last frame
function AttackingAnomaly:update(dt)
  self.anomaly:update(dt)
  self:updateAttacks(dt)
  self:tryAndAttack()
end

--- Updates any active shockwaves from attacks. Note that unlike the `update` method, this method will not update the
--- anomaly in any way, including trying to attack again.
--- @param dt number The time since the last frame
function AttackingAnomaly:updateAttacks(dt)
  self.shockwaves:forEach(function(shockwave) shockwave:update(dt) end)
  self.shockwaves = self.shockwaves:removeIf(Shockwave.isFinished)
end

--- Checks whether a current attack is underway. If not, and there is at least one queued attack, this method takes the
--- first queued attack from the queue and sets the anomaly to begin the attack.
--- @private
function AttackingAnomaly:tryAndAttack()
  if not self.currentAttack and self.queuedAttacks:isPopulated() then
    self.currentAttack = self.queuedAttacks:pop()
    self.anomaly:queueState(self.currentAttack.animation)
  end
end

--- Queues a new attack on the anomaly. This method will try to attack straight away and will do, unless there is
--- another attack already underway.
--- @param attack AttackingAnomalyQueuedAttack The new attack to queue
function AttackingAnomaly:attack(attack)
  self.queuedAttacks:add(attack)
  self:tryAndAttack()
end

--- Performs the current attack from the anomaly. This method creates a new shockwave from the attack that is centred on
--- the current anomaly position.
--- @private
function AttackingAnomaly:onAttack()
  local options = AttackingAnomaly:getAttackOptions(self.currentAttack.attack)
  local affectFn = AttackingAnomaly:getAffectFn(self.currentAttack.attack, self.currentAttack.maxRadius, self.onVehicleAffectFn)
  local position = self.anomaly:getPosition()

  local shockwave = Shockwave:new(position, self.currentAttack.maxRadius, self.currentAttack.time, affectFn, options)
  shockwave:activate()
  self.shockwaves:add(shockwave)
end

--- Calls the attack function on the current attack, if it exists. This method will then reset the current attack to
--- `nil` and try to attack again.
--- @private
function AttackingAnomaly:postAttack()
  local onAttackFn = self.currentAttack.onAttackFn
  if onAttackFn then onAttackFn() end

  self.currentAttack = nil
  self:tryAndAttack()
end

--- Registers an event handler for all anomaly attacks used by this class and internally sets up the unregister function
--- @private
function AttackingAnomaly:initAnomaly()
  local attackFn = function() self:onAttack() end
  local passiveFn = function() self:postAttack() end

  local anomalyEvents = self.anomaly:getEvents()

  anomalyEvents:registerEventHandler("fluctuating", attackFn)
  anomalyEvents:registerEventHandler("discharging", attackFn)
  anomalyEvents:registerEventHandler("passive", passiveFn)

  self.unregisterFn = function()
    anomalyEvents:unregisterEventHandler("fluctuating", attackFn)
    anomalyEvents:unregisterEventHandler("discharging", attackFn)
    anomalyEvents:unregisterEventHandler("passive", passiveFn)
  end
end

--- @return boolean Whether the anomaly is still attacking
function AttackingAnomaly:isAttacking()
  return self.currentAttack ~= nil or self.shockwaves:isPopulated()
end

--- @return boolean Whether the anomaly has now fully closed
--- @see Anomaly#isFullyClosed
function AttackingAnomaly:isFullyClosed()
  return self.anomaly:isFullyClosed()
end

--- Opens the anomaly
--- @see Anomaly#open
function AttackingAnomaly:open()
  self.anomaly:open()
end

--- Registers the given callback function to be called whenever the anomaly attacks a vehicle
--- @param onVehicleAffectFn VehicleAffectFn The function to call when a vehicle is affected by this anomaly
function AttackingAnomaly:registerOnVehicleAffected(onVehicleAffectFn)
  if self.onVehicleAffectFn then
    log("W", "explosivesNG", "This attacking anomaly already has a vehicle effect event handler")
    return
  end

  self.onVehicleAffectFn = onVehicleAffectFn
end

--- Cleans up the underlying anomaly to be used elsewhere. This will __NOT__ destroy or close the underlying anomaly.
function AttackingAnomaly:destroy()
  self.unregisterFn()
  self.onVehicleAffectFn = nil
  self.shockwaves:forEach(function(shockwave) shockwave:destroy() end)
end

--- __Constructor__
--- @return AttackingAnomaly The constructed instance
function AttackingAnomaly:new(anomaly)
  local self = classes.setmetatable({}, AttackingAnomaly)
  self.anomaly = anomaly
  self.currentAttack = nil
  self.queuedAttacks = Array:new(--[[--- @type AttackingAnomalyQueuedAttack[] ]] {})
  self.shockwaves = Array:new(--[[--- @type Shockwave[] ]] {})
  self.onVehicleAffectFn = nil
  self.unregisterFn = nil
  self:initAnomaly()
  return self
end

return AttackingAnomaly