local classes = require("lua/explosionsNG/helpers/classes")
local Events = require("lua/explosionsNG/misc/Events")
local AssaultAttackPattern = require("lua/explosionsNG/worldObjects/anomalies/attacks/AssaultAttackPattern")

--- @alias MultiAssaultAttackSetting { anomaly: Anomaly, attackOptions: AssaultAttackPatternOptions }

--- Groups and maintains multiple `AssaultAttackPattern` as one wrapper class. When the last attack pattern hsa
--- finished, the `onFinish` event will be emitted.
--- @class MultiAssaultAttackPattern : AttackPattern
--- @field private attackPattens Array<AssaultAttackPattern> The list of attack patterns
local MultiAssaultAttackPattern = classes.newClass()

--- Updates each pattern
--- @param dt number The time since the last frame
function MultiAssaultAttackPattern:update(dt)
  self.attackPattens:forEach(function(attack) attack:update(dt) end)
end

--- Called when a pattern finishes. Updates the number of patterns finished and if this is the last pattern to finish,
--- then this method will emit the `"finished"` event.
--- @private
function MultiAssaultAttackPattern:onPatternFinished()
  self.patternsFinished = self.patternsFinished + 1

  if self.patternsFinished == self.attackPattens:length() then
    self.events:emit("finished")
  end
end

--- Creates and returns a new attack pattern with the given setting's anomaly and options. On finish, the attack pattern
--- will call the `onPatternFinished` method.
--- @param setting MultiAssaultAttackSetting The attack setting
--- @return AssaultAttackPattern The created pattern
--- @private
function MultiAssaultAttackPattern:initPattern(setting)
  local pattern = AssaultAttackPattern:new(setting.anomaly, setting.attackOptions)
  local events = pattern:getEvents()
  events:registerEventHandler("finished", function() self:onPatternFinished() end)
  events:registerEventHandler("vehicleAffected", function(context) self.events:emit("vehicleAffected", context) end)
  return pattern
end

--- @return Events<AttackPatternEvent> A registrable set of events that can be registered to or unregistered from
function MultiAssaultAttackPattern:getEvents()
  return self.events
end

--- Destroys all attack patterns. This will __NOT__ destroy any anomalies.
function MultiAssaultAttackPattern:destroy()
  self.attackPattens:forEach(function(attack) attack:destroy() end)
end

--- __Constructor__
--- @param attackSettings Array<MultiAssaultAttackSetting> The list of attacks with their settings
--- @return MultiAssaultAttackPattern The constructed instance
function MultiAssaultAttackPattern:new(attackSettings)
  local self = classes.setmetatable({}, MultiAssaultAttackPattern)
  self.attackPattens = attackSettings:map(function(setting) return self:initPattern(setting) end)
  self.events = Events:new()
  self.patternsFinished = 0
  return self
end

return MultiAssaultAttackPattern
