local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local SkyEffect = require("lua/explosionsNG/worldObjects/environment/SkyEffect")
local Events = require("lua/explosionsNG/misc/Events")
local CircularPath = require("lua/explosionsNG/pathing/CircularPath")
local Anomaly = require("lua/explosionsNG/worldObjects/anomalies/Anomaly")
local ConvergencePhaseOne = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhaseOne")
local ConvergencePhaseTwo = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhaseTwo")
local ConvergencePhaseThree = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhaseThree")
local ConvergencePhaseFour = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhaseFour")

--- Properties:
---  - `phaseOptions` - The list of options to use for each phase
--- @alias ConvergenceOptions { phaseOptions: ConvergencePhaseOptions | nil }
--- @alias ConvergenceEvent "phaseUpdated" | "completed" | "playerAffected"

--- @class Convergence : Evented<ConvergenceEvent>
--- @field private options ConvergenceOptions The list of options in use
--- @field private mainAnomaly Anomaly The main, big anomaly
--- @field private orbitalAnomalies Array<Anomaly> The anomalies that orbit the main anomaly and are used as part of a subphase
--- @field private orbitPath CircularPath The path that the orbiting anomalies follow
--- @field private phase number The active phase number
--- @field private currentPhase ConvergencePhase The current active phase
--- @field private events Events<ConvergenceEvent> The list of registrable events
--- @field private progress Array<ConvergencePhaseProgress> The current progress the player is through this boss
--- @field private skyEffect SkyEffect The sky brightness effect for when the main entity is spawned and killed
--- @field private NUMBER_OF_ORBITAL_ANOMALIES number __Static__ The number of anomalies that orbit the convergence anomaly
--- @field private ORBITAL_SPEED number __Static__ The speed (in rads^-1) the orbiting anomalies rotate at
--- @field private ORBITAL_RADIUS number __Static__ The radius (in m) the orbiting anomalies rotate at
--- @field private SKY_EFFECT_PERIOD number __Static__ The time (in seconds) the sky brightness effect should take to complete
local Convergence = classes.newClass()
Convergence.NUMBER_OF_ORBITAL_ANOMALIES = 4
Convergence.ORBITAL_SPEED = 0.25
Convergence.ORBITAL_RADIUS = 300
Convergence.SKY_EFFECT_PERIOD = 7.5

--- __Static__
---
--- Fills in any option not given in the given options with default values
--- @param options ConvergenceOptions The partial list of options
--- @return ConvergenceOptions The filled options
--- @private
function Convergence:fillOptions(options)
  return {
    phaseOptions = options.phaseOptions or {}
  }
end

--- __Static__
--- @return Array<ConvergencePhaseProgress> A new progress list for a completely brand new instance
--- @private
function Convergence:initProgress()
  local phaseFour = ConvergencePhaseFour:getUnenteredPhaseProgress()
  -- The final phase starts off active
  phaseFour.active = true

  return Array:new({
    ConvergencePhaseOne:getUnenteredPhaseProgress(),
    ConvergencePhaseTwo:getUnenteredPhaseProgress(),
    ConvergencePhaseThree:getUnenteredPhaseProgress(),
    phaseFour
  })
end

--- __Static__
---
--- Returns the position of an orbiting anomaly, given its index, around a circular path.
--- This assumes equidistant points around the orbit (i.e. all other anomalies will stay in the same place if one moves
--- out of the array).
--- @param path CircularPath The circular path
--- @param index number The index of the orbiting anomaly
--- @return LuaVec3 The offset position for an orbital anomaly
--- @private
function Convergence:getOrbitalAnomalyPosition(path, index)
  local rotationalOffset = (2 * math.pi / Convergence.NUMBER_OF_ORBITAL_ANOMALIES) * index
  return path:getPositionAtOffsetRotation(rotationalOffset)
end

--- __Static__
--- @return AnomalyOptions The list of options for an orbiting anomaly
--- @private
function Convergence:getOrbitalAnomalyOptions()
  return { scale = 2.5, numberOfLayers = 9, life = math.huge }
end

--- __Static__
--- @return AnomalyOptions The list of options for the convergence anomaly
--- @private
function Convergence:getMainAnomalyOptions()
  return { scale = 10, numberOfLayers = 18, colour = "1 1 1", life = math.huge }
end

--- __Static__
---
--- Creates and returns a list of anomalies that will orbit the convergence anomaly
--- @param position LuaVec3 The initial position to place the anomaly at
--- @return Array<Anomaly> The list of created anomalies
--- @private
function Convergence:createOrbitalAnomalies(position)
  return Array:from(Convergence.NUMBER_OF_ORBITAL_ANOMALIES, function()
    local options = Convergence:getOrbitalAnomalyOptions()
    return Anomaly:new(position, options)
  end)
end

--- __Static__
--- @return SunSkyMap The effect options to use when the convergence is active
--- @private
function Convergence:getSkyEffectOptions()
  return --[[--- @type SunSkyMap ]] { skyBrightness = 0.3, brightness = 0.15 }
end

--- __Static__
--- @return SphericalBound[] The area of the objects that make up this entity
function Convergence:getBoundingArea()
  local circularPath = CircularPath:new(vec3(), Convergence.ORBITAL_RADIUS)

  local mainOptions = Convergence:getMainAnomalyOptions()
  local mainAnomalyRadius = Anomaly:getMaxRadiusOf(mainOptions.scale, true)
  local mainAnomalyBound = --[[--- @type SphericalBound ]] { offset = vec3(), radius = mainAnomalyRadius }

  local orbitalAnomalyOptions = Convergence:getOrbitalAnomalyOptions()
  local orbitalAnomalyRadius = Anomaly:getMaxRadiusOf(orbitalAnomalyOptions.scale)
  --- @type Array<SphericalBound>
  local orbitalAnomalyBounds = Array:from(Convergence.NUMBER_OF_ORBITAL_ANOMALIES, function(index)
    local offset = Convergence:getOrbitalAnomalyPosition(circularPath, index)
    return { offset = offset, radius = orbitalAnomalyRadius }
  end)

  return orbitalAnomalyBounds:add(mainAnomalyBound):toPrimitive()
end

--- Updates the main convergence anomaly, the orbiting anomalies and the current phase, if it exists
--- @param dt number The time since the last frame
function Convergence:update(dt)
  -- The main anomaly is used elsewhere in this phase
  if self.phase ~= 4 then
    self.mainAnomaly:update(dt)
  end

  self.skyEffect:update(dt)
  self:updateOrbitalAnomalies(dt)

  if self.currentPhase then
    self.currentPhase:update(dt)
  end
end

--- Updates the orbit path for the orbiting anomalies, and then updates those anomalies with their new position and
--- current animation.
--- @param dt number The time since the last frame
--- @private
function Convergence:updateOrbitalAnomalies(dt)
  if self.orbitalAnomalies:isEmpty() then return end

  self.orbitPath:addRotation(dt * Convergence.ORBITAL_SPEED)

  self.orbitalAnomalies:forEach(function(anomaly, index)
    local position = Convergence:getOrbitalAnomalyPosition(self.orbitPath, index)
    anomaly:setPosition(position)
    anomaly:update(dt)
  end)
end

--- Opens the main convergence anomaly and all orbiting anomalies
function Convergence:open()
  self.mainAnomaly:open()
  self.orbitalAnomalies:forEach(Anomaly.open)
  self.skyEffect:start()
end

--- @return Anomaly The next orbital anomaly
--- @private
function Convergence:takeNextOrbitalAnomaly()
  return self.orbitalAnomalies:pop(true)
end

--- @return ConvergencePhase The next phase, based on the current phase number. If there is no next phase, `nil` will be returned
--- @private
function Convergence:getNextPhase()
  local phaseOptions = self.options.phaseOptions

  if self.phase == 1 then
    local anomaly = self:takeNextOrbitalAnomaly()
    return ConvergencePhaseOne:new(anomaly, phaseOptions)
  end

  if self.phase == 2 then
    local anomaly = self:takeNextOrbitalAnomaly()
    return ConvergencePhaseTwo:new(anomaly, phaseOptions)
  end

  if self.phase == 3 then
    local anomalies = Array:new({ self:takeNextOrbitalAnomaly(), self:takeNextOrbitalAnomaly() })
    return ConvergencePhaseThree:new(anomalies, phaseOptions)
  end

  if self.phase == 4 then
    return ConvergencePhaseFour:new(self.mainAnomaly, phaseOptions)
  end
end

--- Called when a phase is completed. This method creates the next phase and begins it immediately. If there is no next
--- phase, then this method will emit the `"completed"` event.
--- @private
function Convergence:onPhaseChange()
  self.phase = self.phase + 1
  local phase = self:getNextPhase()

  if not phase then
    self.events:emit("completed")
    return
  end

  -- Set initial progress (in most cases, this will just be setting the "active" flag to true)
  local progress = phase:getProgress()
  self:updatePhaseProgress(progress)

  local phaseEvents = phase:getEvents()
  phaseEvents:registerEventHandler("subphaseCompleted", function() self:onSubphaseComplete(phase) end)
  phaseEvents:registerEventHandler("phaseCompleted", function() self:onPhaseCompleted(phase) end)
  phaseEvents:registerEventHandler("playerAffected", function(wasPlayerAffected) self:onPlayerAffected(wasPlayerAffected) end)

  phase:begin()
  self.currentPhase = phase
end

--- Called when a subphase within a phase is completed.
--- This method updates the current progress by the player, and if the given phase has now complete all subphases and is
--- the final phase, this method will revert the sky brightness effect in operation.
--- @param phase ConvergencePhase The phase whose subphase has completed
--- @private
function Convergence:onSubphaseComplete(phase)
  local progress = phase:getProgress()
  self:updatePhaseProgress(progress)

  -- If this is the final subphase of the final phase, revert the sky effect
  if self.phase == 4 and phase:hasFinishedAllSubphases() then
    self.skyEffect:revert()
  end
end

--- Called when a phase is completed.
--- This method updates the current progress by the player and changes to the next phase.
--- @param phase ConvergencePhase The completed phase
--- @private
function Convergence:onPhaseCompleted(phase)
  local finishedProgress = phase:getCompletePhaseProgress()
  self:updatePhaseProgress(finishedProgress)
  self:onPhaseChange()
end

--- Called when this entity affects (or does not affect) the player vehicle.
--- Emits the event via the `"playerAffected"` event.
--- @param wasPlayerAffected boolean Whether the player vehicles was affected
--- @private
function Convergence:onPlayerAffected(wasPlayerAffected)
  self.events:emit("playerAffected", wasPlayerAffected)
end

--- Updates the phase progress of the current phase and emits the `"phaseUpdated"` event
--- @param progress ConvergencePhaseProgress The updated progress
--- @private
function Convergence:updatePhaseProgress(progress)
  self.progress:replace(self.phase, progress)
  self.events:emit("phaseUpdated")
end

--- Creates a new, main convergence anomaly and returns it
--- @param position LuaVec3 The position to place the main anomaly
--- @return Anomaly The created main anomaly
--- @private
function Convergence:createMainAnomaly(position)
  local anomaly = Anomaly:new(position, Convergence:getMainAnomalyOptions())
  local anomalyEvents = anomaly:getEvents()

  anomalyEvents:registerEventHandler("opened", function()
    self:onPhaseChange()
    self:updatePhaseProgress(self.currentPhase:getProgress())
  end)

  return anomaly
end

--- @return ConvergencePhaseProgress[] The current progress
function Convergence:getCurrentProgress()
  return self.progress:toPrimitive()
end

--- @return boolean Whether the player has failed the current phase
function Convergence:hasFailedCurrentPhase()
  return self.currentPhase and self.currentPhase:hasFailedSubphase()
end

--- Destroys the instance and releases any world objects
function Convergence:destroy()
  self.mainAnomaly:destroy()
  self.orbitalAnomalies:forEach(function(anomaly) anomaly:destroy() end)

  if self.currentPhase then
    self.currentPhase:destroy()
  end

  self.skyEffect:revertInstantly()
end

--- @return Registrable<ConvergenceEvent> A registrable set of events that can be registered to or unregistered from
function Convergence:getEvents()
  return self.events
end

--- __Constructor__
--- @param position LuaVec3 The position of the main entities of the convergence
--- @param options? ConvergenceOptions | nil A list of options to use
--- @return Convergence The constructed instance
function Convergence:new(position, options)
  local self = classes.setmetatable({}, Convergence)
  self.options = Convergence:fillOptions(options or {})
  self.mainAnomaly = self:createMainAnomaly(position)
  self.orbitalAnomalies = Convergence:createOrbitalAnomalies(position)
  self.orbitPath = CircularPath:new(position, Convergence.ORBITAL_RADIUS)
  self.phase = 0
  self.currentPhase = nil
  self.progress = Convergence:initProgress()
  self.events = --[[--- @type Events<ConvergenceEvent> ]] Events:new()
  self.skyEffect = SkyEffect:new(Convergence.SKY_EFFECT_PERIOD, Convergence:getSkyEffectOptions())
  return self
end

return Convergence
