local classes = require("lua/explosionsNG/helpers/classes")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local Events = require("lua/explosionsNG/misc/Events")

--- Properties:
---  - `timeMultiplier` - The multiplier of time to apply to all time-related periods during the convergence phases.
---                       A value `< 1` decreases attack cooldowns, a value `> 1` increases attack cooldowns.
--- @alias ConvergencePhaseOptions { timeMultiplier: number | nil }
--- @alias ConvergencePhaseEvent "subphaseCompleted" | "phaseCompleted" | "playerAffected"
--- @alias ConvergencePhaseProgress { segments: number, segmentsCompleted: number, dead: boolean, active: boolean }

--- The base phase class that all convergence phases should inherit from
--- @class ConvergencePhase : Evented<ConvergencePhaseEvent>
--- @field protected options ConvergencePhaseOptions The list of options in use
--- @field protected anomalies Array<Anomaly> The anomalies involved in the phase
--- @field protected numberOfSubphases number The number of subphases in this phase
--- @field protected attackPattern AttackPattern | nil The currently active attack pattern of the anomaly
--- @field protected subphase number The current subphase
--- @field protected hasAffectedPlayer boolean Whether the player has been affected by the current subphase
--- @field protected markedForDestruction boolean Whether the instance is in the process of being destroyed
--- @field protected events Events<ConvergencePhaseEvent> A list of events and event handlers to emit
local ConvergencePhase = classes.newClass()

--- __Static__
---
--- Fills in any option not given in the given options with default values
--- @param options ConvergencePhaseOptions The partial list of options
--- @return ConvergencePhaseOptions The filled options
--- @private
function ConvergencePhase:fillOptions(options)
  return {
    timeMultiplier = options.timeMultiplier or 1
  }
end

--- __Static__
--- @return ConvergencePhaseProgress A progress object that is not entered / is fully incomplete for this phase
function ConvergencePhase:getUnenteredPhaseProgress()
  local numberOfSubphases = self:getNumberOfSubphases() -- Using self so that the correct subclass method is called
  return {
    segments = numberOfSubphases,
    segmentsCompleted = 0,
    dead = false,
    active = false
  }
end

-- TODO This method doesn't really need to be static after all...
--- __Static__
---
--- @return ConvergencePhaseProgress A progress object that is fully complete for this phase
function ConvergencePhase:getCompletePhaseProgress()
  local numberOfSubphases = self:getNumberOfSubphases() -- Using self so that the correct subclass method is called
  return {
    segments = numberOfSubphases,
    segmentsCompleted = numberOfSubphases,
    dead = true,
    active = false
  }
end

--- __Abstract__
---
--- An implementation of this method __must__ remain static in nature (i.e. not access any instance fields or methods)
--- in order for this method to be called correctly in the constructor. This method is defined as an instance method so
--- that this base class can access the properties in the above static methods from subclasses (which is a limitation of
--- Lua OOP).
--- @return number The number of subphases
--- @protected
function ConvergencePhase:getNumberOfSubphases()
  error("getNumberOfSubphases is not implemented")
  return nil
end

--- __Abstract__
--- @return AttackPattern A new attack pattern for a new phase
--- @protected
function ConvergencePhase:createAttackPattern()
  error("createSubphase is not implemented")
  return nil
end

--- Damages each anomaly in this phase
--- @protected
function ConvergencePhase:onDamage()
  self.anomalies:forEach(function(anomaly) anomaly:queueState("damaging") end)
end

--- Called when the last subphase is completed and thus the phase has been formally completed.
--- This method will close all anomalies in the phase.
--- @protected
function ConvergencePhase:onLastSubphaseComplete()
  self.anomalies:forEach(function(anomaly) anomaly:beginClosing() end)
end

--- Updates the anomalies. This will be called when the attack pattern has finished and so it is indicative that the
--- anomalies are being updated to close.
--- @param dt number The time since the last frame
--- @private
function ConvergencePhase:updateNonAttack(dt)
  self.anomalies:forEach(function(anomaly) anomaly:update(dt) end)
end

--- Checks whether all subphases have completed. If they have, then this method will begin the completion of the phase,
--- otherwise it will create a new active attack pattern.
--- @private
function ConvergencePhase:postDamage()
  if self:hasFinishedAllSubphases() then
    self:postLastSubphaseComplete()
    return
  end

  self:onCreateNewAttackPattern()
end

--- Called when an attack pattern is to be created.
--- This method will create a new attack pattern using `createAttackPattern`, and resets the `hasAffectedPlayer` flag
--- back to `false`.
--- @private
function ConvergencePhase:onCreateNewAttackPattern()
  local pattern = self:createAttackPattern()

  local events = pattern:getEvents()
  events:registerEventHandler("finished", function() self:onSubphaseCompleted() end)
  events:registerEventHandler("vehicleAffected", function(vehicle) self:onVehicleAffected(vehicle) end)

  self.attackPattern = pattern
  self.hasAffectedPlayer = false
end

--- Called when an attack pattern has affected a vehicle.
--- This method will set the the `hasAffectedPlayer` flag to whether the given vehicle is the player vehicle. Note that
--- the flag will not be overwritten if it is already `true`.
--- @param vehicle Vehicle The affected vehicle
--- @private
function ConvergencePhase:onVehicleAffected(vehicle)
  -- Don't overwrite
  if self.hasAffectedPlayer then return end

  local player = vehiclesHandler.getPlayerVehicle()
  if not player then return end

  self.hasAffectedPlayer = vehicle:getId() == player:getId()

  if self.hasAffectedPlayer then
    self.events:emit("playerAffected", true)
  end
end

--- Checks whether the player was attacked and if so, recreates another attack pattern for the same subphase. If not,
--- then this method will create a new attack pattern for the next subphase.
--- In the event all subphases have now be created and finished, this method will clear the current attack pattern and
--- leave it empty.
--- If the phase is currently paused, then no new attack pattern is created.
--- @private
function ConvergencePhase:onSubphaseCompleted()
  if self.hasAffectedPlayer then
    self.events:emit("playerAffected", false)
    self:onCreateNewAttackPattern()
    return
  end
  
  self.subphase = self.subphase + 1
  self.events:emit("subphaseCompleted")

  self.attackPattern = nil
  self:onDamage()
end

--- Removes the current attack pattern. This method will then call the `onLastSubphaseComplete` method.
--- @private
function ConvergencePhase:postLastSubphaseComplete()
  self.attackPattern = nil
  self:onLastSubphaseComplete()
end

--- @return boolean Whether all subphases have finished
function ConvergencePhase:hasFinishedAllSubphases()
  return self.subphase > self.numberOfSubphases
end

--- Updates the current phase
--- @param dt number The time since the last frame
function ConvergencePhase:update(dt)
  if self.attackPattern then
    (--[[--- @not nil ]] self.attackPattern):update(dt)
    return
  end

  self:updateNonAttack(dt)
end

--- Creates and begins a new attack pattern
function ConvergencePhase:begin()
  self:onCreateNewAttackPattern()
end

--- @return ConvergencePhaseProgress The current progress of this phase so far
function ConvergencePhase:getProgress()
  local subphasesComplete = self.subphase - 1
  return {
    segments = self.numberOfSubphases,
    segmentsCompleted = subphasesComplete,
    dead = subphasesComplete == self.numberOfSubphases,
    active = true
  }
end

--- @return boolean Whether the player has failed the current subphase
function ConvergencePhase:hasFailedSubphase()
  return self.hasAffectedPlayer
end

--- @return Registrable<ConvergencePhaseEvent> A registrable set of events that can be registered to or unregistered from
function ConvergencePhase:getEvents()
  return self.events
end

--- Sets up the given anomalies to this phase instance before returning it for convenience. Note that only the first
--- anomaly will have events handled.
--- @param anomalies Array<Anomaly> The anomalies to initialise
--- @return Array<Anomaly> The given anomalies for use elsewhere
--- @private
function ConvergencePhase:initAnomalies(anomalies)
  -- Only need to register on one
  local anomalyEvents = anomalies:nth(1):getEvents()

  local onDamaged = function() self:postDamage() end
  local onClosed
  onClosed = function()
    anomalyEvents:unregisterEventHandler("closed", onClosed)
    anomalyEvents:unregisterEventHandler("damaged", onDamaged)

    -- Only emit this event if this instance is not being forcibly destroyed
    if not self.markedForDestruction then
      self.events:emit("phaseCompleted")
    end
  end

  anomalyEvents:registerEventHandler("closed", onClosed)
  anomalyEvents:registerEventHandler("damaged", onDamaged)
  return anomalies
end

--- Destroys all anomalies in this phase and any attack pattern
function ConvergencePhase:destroy()
  self.markedForDestruction = true

  self.anomalies:forEach(function(anomaly) anomaly:destroy() end)

  if self.attackPattern then
    (--[[--- @not nil ]] self.attackPattern):destroy()
  end
end

--- __Constructor__
--- @param anomalies Array<Anomaly> The anomalies involved in the phase
--- @param Subclass ConvergencePhase The subclass extending this base class
--- @param options? ConvergencePhaseOptions | nil A list of options to use
--- @return ConvergencePhase The constructed instance
function ConvergencePhase:new(anomalies, Subclass, options)
  local self = classes.setmetatable({}, ConvergencePhase)
  self.options = ConvergencePhase:fillOptions(options or {})
  self.anomalies = self:initAnomalies(anomalies)
  self.numberOfSubphases = Subclass:getNumberOfSubphases()
  self.subphase = 1
  self.attackPattern = nil
  self.hasAffectedPlayer = false
  self.markedForDestruction = false
  self.events = --[[--- @type Events<ConvergencePhaseEvent> ]] Events:new()
  return self
end

return ConvergencePhase
