local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local ConvergencePhase = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhase")
local AreaAttackPattern = require("lua/explosionsNG/worldObjects/anomalies/attacks/AreaAttackPattern")

--- @class ConvergencePhaseFour : ConvergencePhase
local ConvergencePhaseFour = classes.newClass(ConvergencePhase)

--- __Static__
---
--- Returns the RGB colour an anomaly should be, if it is performing given attack type
--- @param attackType AreaAttackType The attack type to get the colour of
--- @return string The associated anomaly colour of the attack
function ConvergencePhaseFour:getAnomalyColourFromAttack(attackType)
  if attackType == "lightning"     then return "0 0.5 1" end -- TODO Repeated constants in AreaAttackPattern
  if attackType == "implosion"     then return "1 1   0" end
  if attackType == "orbitalStrike" then return "0 1   0" end
end

--- __Static__
---
--- Returns a list of attack settings for the given subphase
--- @param subphase number The subphase to get the attack settings for
--- @param timeMultiplier number The multiplier to use for any time-based properties within the setting
--- @return AreaAttackPatternSetting[] The list of attack settings for the subphase
--- @private
function ConvergencePhaseFour:getSubphaseAttackSetting(subphase, timeMultiplier)
  if subphase == 1 then
    return {
      {
        attackType = "lightning",
        areaRadius = 75,
        effectRadius = 40,
        targetDelay = 3 * timeMultiplier,
        numberOfExplosives = 5
      },
      {
        attackType = "lightning",
        areaRadius = 125,
        effectRadius = 40,
        targetDelay = 3 * timeMultiplier,
        numberOfExplosives = 7
      }
    }
  end

  if subphase == 2 then
    return {
      {
        attackType = "implosion",
        interval = 1.5 * timeMultiplier,
        targetDelay = 3 * timeMultiplier,
        effectRadius = 35,
        numberOfExplosives = 5,
        placement = "followPlayer"
      },
      {
        attackType = "implosion",
        interval = 0.9 * timeMultiplier,
        targetDelay = 3 * timeMultiplier,
        effectRadius = 35,
        numberOfExplosives = 8,
        placement = "followPlayer"
      },
    }
  end

  if subphase == 3 then
    return {
      {
        attackType = "orbitalStrike",
        areaRadius = 200,
        interval = 0.5 * timeMultiplier,
        targetDelay = 4 * timeMultiplier,
        effectRadius = 75,
        numberOfExplosives = 8,
        placement = "ring",
        pulsarOptions = {
          numberOfShockwaves = 3
        }
      },
      {
        attackType = "orbitalStrike",
        areaRadius = 300,
        interval = 0.5 * timeMultiplier,
        targetDelay = 2 * timeMultiplier,
        effectRadius = 100,
        numberOfExplosives = 12,
        placement = "ring",
        pulsarOptions = {
          interval = timeMultiplier,
          numberOfShockwaves = 3
        }
      },
    }
  end

  if subphase == 4 then
    return {
      {
        attackType = "lightning",
        areaRadius = 150,
        effectRadius = 40,
        targetDelay = 3 * timeMultiplier,
        numberOfExplosives = 15,
        lightningOptions = {
          life = 2 * timeMultiplier
        }
      },
      {
        attackType = "implosion",
        interval = 0.5 * timeMultiplier,
        targetDelay = 5 * timeMultiplier,
        effectRadius = 40,
        numberOfExplosives = 10,
        placement = "followPlayer"
      },
      {
        attackType = "orbitalStrike",
        areaRadius = 200,
        interval = 0.25 * timeMultiplier,
        targetDelay = 2.5 * timeMultiplier,
        effectRadius = 75,
        numberOfExplosives = 8,
        placement = "ring",
        pulsarOptions = {
          interval = 1.5 * timeMultiplier,
          numberOfShockwaves = 3
        }
      },
    }
  end

  if subphase == 5 then
    return {
      {
        attackType = "implosion",
        interval = 0.7 * timeMultiplier,
        targetDelay = 3 * timeMultiplier,
        effectRadius = 30,
        numberOfExplosives = 6,
        placement = "followPlayer"
      },
      {
        attackType = "lightning",
        areaRadius = 300,
        effectRadius = 40,
        targetDelay = 3 * timeMultiplier,
        numberOfExplosives = 25
      },
      {
        attackType = "lightning",
        areaRadius = 300,
        effectRadius = 40,
        targetDelay = 3 * timeMultiplier,
        numberOfExplosives = 25
      },
      {
        attackType = "orbitalStrike",
        areaRadius = 500,
        interval = 0.5 * timeMultiplier,
        targetDelay = 5 * timeMultiplier,
        effectRadius = 450,
        numberOfExplosives = 8,
        placement = "ring",
        pulsarOptions = {
          time = 4 * timeMultiplier,
          numberOfShockwaves = 2
        }
      },
    }
  end

  if subphase == 6 then
    return {
      {
        attackType = "implosion",
        interval = 0.4 * timeMultiplier,
        targetDelay = 4 * timeMultiplier,
        effectRadius = 35,
        numberOfExplosives = 10,
        placement = "followPlayer"
      },
      {
        attackType = "lightning",
        areaRadius = 250,
        targetDelay = 4 * timeMultiplier,
        effectRadius = 75,
        numberOfExplosives = 25,
        lightningOptions = {
          life = 2 * timeMultiplier
        }
      },
      {
        attackType = "orbitalStrike",
        areaRadius = 600,
        interval = 0.25 * timeMultiplier,
        targetDelay = 2 * timeMultiplier,
        effectRadius = 225,
        numberOfExplosives = 12,
        placement = "ring",
        pulsarOptions = {
          interval = 1 * timeMultiplier,
          time = 3 * timeMultiplier,
          numberOfShockwaves = 5
        }
      },
      {
        attackType = "lightning",
        areaRadius = 200,
        targetDelay = 4 * timeMultiplier,
        effectRadius = 75,
        numberOfExplosives = 25,
        lightningOptions = {
          life = 2 * timeMultiplier
        }
      },
      {
        attackType = "orbitalStrike",
        areaRadius = 200,
        interval = 0.25 * timeMultiplier,
        targetDelay = 2 * timeMultiplier,
        effectRadius = 170,
        numberOfExplosives = 8,
        placement = "ring",
        pulsarOptions = {
          interval = timeMultiplier,
          numberOfShockwaves = 3
        }
      },
      {
        attackType = "lightning",
        areaRadius = 500,
        targetDelay = 6 * timeMultiplier,
        effectRadius = 100,
        numberOfExplosives = 50
      },
    }
  end
end

--- @return number The number of subphases
--- @see ConvergencePhase#getNumberOfSubphases
function ConvergencePhaseFour:getNumberOfSubphases()
  return 6
end

--- @return AttackPattern A new attack pattern for a new phase
--- @see ConvergencePhase#createAttackPattern
function ConvergencePhaseFour:createAttackPattern()
  --- Shimmers the anomaly with the colour of the attack type
  --- @type AreaAttackOnStartFn
  local onAttackStarted = function(attackType)
    local colour = ConvergencePhaseFour:getAnomalyColourFromAttack(attackType)
    self.anomalies:nth(1):queueState("shimmering", colour)
  end

  -- Type is incompatible and I have no idea why. It's a generic type - why are you upset?! :(
  local options = Array:new(--[[--- @type any[] ]] ConvergencePhaseFour:getSubphaseAttackSetting(self.subphase, self.options.timeMultiplier))
  local pattern = AreaAttackPattern:new(self.anomalies:nth(1), options)

  local events = pattern:getEvents()
  events:registerEventHandler("started", onAttackStarted)

  pattern:begin()
  return pattern
end

--- @see ConvergencePhase#onLastSubphaseComplete
function ConvergencePhaseFour:onLastSubphaseComplete()
  self.anomalies:nth(1):beginClosing(true)
end

--- __Constructor__
--- @param anomaly Anomaly The anomaly that will be used in this phase
--- @param options? ConvergencePhaseOptions | nil A list of options to use
--- @return ConvergencePhaseFour The constructed instance
function ConvergencePhaseFour:new(anomaly, options)
  local self = classes.setmetatable(ConvergencePhase:new(Array:new({ anomaly }), ConvergencePhaseFour, options), ConvergencePhaseFour)
  return self
end

return ConvergencePhaseFour
