local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local Events = require("lua/explosionsNG/misc/Events")
local Star = require("lua/explosionsNG/worldObjects/celestials/star/Star")
local CircularPath = require("lua/explosionsNG/pathing/CircularPath")
local Timer = require("lua/explosionsNG/timers/Timer")

--- @alias BinaryStarSystemEvent "formed" | "collided"

--- @class BinaryStarSystem : Celestial<BinaryStarSystemEvent>
--- @field private stars Array<Star> The stars in the system
--- @field private path CircularPath The circular path of which the stars are moving along
--- @field private events Events<BinaryStarSystemEvent> The list of registrable events that can be emitted
--- @field private INITIAL_SYSTEM_RADIUS number The initial radius of the system (centre point at which both stars rotate around each other)
--- @field private INITIAL_STAR_RADIUS number The initial radius of the stars before any collapsing/compression occurs
--- @field private mergeTimer Timer The timer to time the merging phase
--- @field private isCollapsing boolean Whether the system is current collapsing in on itself
--- @field private TIME_UNTIL_MERGE number __Static__ The time (in s) that it takes for the system to go from collapsing to both stars merged
local BinaryStarSystem = classes.newClass()
BinaryStarSystem.TIME_UNTIL_MERGE = 10

--- Updates the star system, if the star system is not dead
--- @param dt number The time since the last frame
function BinaryStarSystem:update(dt)
  if self:isDead() then return end

  local elapsedTime = self.mergeTimer:elapsedTime()
  local timeRatio = self.mergeTimer:timeLeftRatio()

  if self.isCollapsing then self:updateCollapsingPath(timeRatio) end

  local rotationIncrement = (1 + (elapsedTime / 3) ^ 3) * dt
  self.path:addRotation(rotationIncrement)

  self:updateFirstStar(dt)
  self:updateSecondStar(dt)

  self.mergeTimer:update(dt)
end

--- Updates the first star's state, including its position from its circular path
--- @param dt number The time since the last frame
--- @private
function BinaryStarSystem:updateFirstStar(dt)
  local newPosition = self.path:getCurrentPosition()
  self:updateStar(1, newPosition, dt)
end

--- Updates the second star's state, including its position from its circular path
--- @param dt number The time since the last frame
--- @private
function BinaryStarSystem:updateSecondStar(dt)
  local newPosition = self.path:getPositionAtOffsetRotation(math.pi)
  self:updateStar(2, newPosition, dt)
end

--- Updates the star's position and calls its update method, given its index
--- @param index number The index of the star to update
--- @param position LuaVec3 The new position to set the star
--- @param dt number The time since the last frame
--- @private
function BinaryStarSystem:updateStar(index, position, dt)
  local star = self.stars:nth(index)
  star:setPosition(position);
  star:update(dt);
end

--- Sets the radius of the path that the stars are on, decreasing it dependent on the ratio of time that has passed.
--- Note that the new radius of the path will be at a minimum of the `INITIAL_STAR_RADIUS`.
--- @param timeRatio number The ratio of time elapsed to time left
--- @private
function BinaryStarSystem:updateCollapsingPath(timeRatio)
  -- Minimum radius is the width of the star - i.e. both stars will be 0.5r inside each other
  local newSystemRadius = math.max(0, ((self.INITIAL_SYSTEM_RADIUS - self.INITIAL_STAR_RADIUS) * timeRatio) + self.INITIAL_STAR_RADIUS)
  self.path:setRadius(newSystemRadius)
end

--- @return boolean Whether both stars in the star system are dead
function BinaryStarSystem:isDead()
  return self.stars:all(Star.isDead)
end

--- Begins compressing both stars
--- @private
function BinaryStarSystem:compress()
  self.stars:forEach(function(star) star:compress() end)
end

--- Begins the collapsing of this star system
function BinaryStarSystem:collapse()
  self.isCollapsing = true
end

--- @return Events<BinaryStarSystemEvent> A registrable set of events that can be registered to or unregistered from
function BinaryStarSystem:getEvents()
  return self.events
end

--- Registers to various events from the given stars.
--- One star will be set up to be listened to be this instance, and the other will be set up to be listened to be
--- outside sources and indirectly exposed as part of the `getEvents` method.
--- @param stars Array<Star> The list of stars
--- @private
function BinaryStarSystem:setUpStarEvents(stars)
  -- The order does not matter, but one needs to be handled by this instance and the other should accept external handlers

  local exposedEvents = stars:nth(1):getEvents()
  exposedEvents:registerEventHandler("formed", function() self.events:emit("formed") end)
  exposedEvents:registerEventHandler("compressed", function() self.events:emit("collided") end)

  local internalEvents = stars:nth(2):getEvents()
  internalEvents:registerEventHandler("formed", function() self.mergeTimer:start() end)
end

--- Destroys all celestials within this system
--- @see Celestial#destroy
function BinaryStarSystem:destroy()
  self.stars:forEach(function(star) star:destroy() end)
end

--- __Constructor__
--- @param position LuaVec3 The centre position of the star system. This will be the point at which both stars orbit around.
--- @param systemRadius number The radius of the star system. This will be the radius of the orbits of both stars.
--- @param starRadius number The radius of each star in the system
--- @param starColour string The colour of each star in the system
--- @return BinaryStarSystem The constructed instance
function BinaryStarSystem:new(position, systemRadius, starRadius, starColour)
  local self = classes.setmetatable({}, BinaryStarSystem)
  self.INITIAL_SYSTEM_RADIUS = systemRadius
  self.INITIAL_STAR_RADIUS = starRadius
  self.stars = Array:new({ Star:new(position, starRadius, starColour), Star:new(position, starRadius, starColour) })
  self.path = CircularPath:new(position, systemRadius)
  self.mergeTimer = Timer:new(BinaryStarSystem.TIME_UNTIL_MERGE, function() self:compress() end)
  self.isCollapsing = false
  self.events = Events:new()

  self:setUpStarEvents(self.stars)

  return self
end

return BinaryStarSystem
