local classes = require("lua/explosionsNG/helpers/classes")
local Events = require("lua/explosionsNG/misc/Events")
local blackHoleAnimationStateMachine = require("lua/explosionsNG/worldObjects/celestials/blackHole/animations")
local worldObjectHandler = require("lua/explosionsNG/handlers/worldObjects/handler")

--- @alias BlackHoleState "forming" | "idle" | "evaporating" | "dead"
--- @alias BlackHoleEvent "formed" | "evaporated"

--- @class BlackHole : Celestial<BlackHoleEvent>
--- @field private accretionDisc SceneObject The in-game representation of the accretion disc
--- @field private eventHorizon SceneObject The in-game representation of the event horizon
--- @field private lightObject SceneObject The in-game representation of the light emitted
--- @field private position LuaVec3 The position of the object
--- @field private radius number The radius (in m) of the whole object
--- @field private state AnimationStateMachine<BlackHoleState, BlackHole> The state machine with animations of the black hole
--- @field private events Events<BlackHoleEvent> The list of registrable events that can be emitted
--- @field private LIGHT_RADIUS_MULTIPLIER number __Static__ The multiplier of radius of the light emitter
--- @field private EVENT_HORIZON_SCALE_MULTIPLIER number __Static__ The multiplier of the radius of the event horizon
local BlackHole = classes.newClass()
BlackHole.LIGHT_RADIUS_MULTIPLIER = 2
BlackHole.EVENT_HORIZON_SCALE_MULTIPLIER = 0.2

--- __Static__
---
--- Returns a new PointLight object with the given position and radius
--- @param position LuaVec3 The position to create the light at
--- @param radius number The radius of the light
--- @return SceneObject The created light
--- @private
function BlackHole:createLight(position, radius)
  return worldObjectHandler.requestPointLight({
    position = position,
    radius = BlackHole:radiusToLightRadius(radius),
    brightness = 10,
    colour = "1 0.5 0",
    applyRadiusToPosition = true
  })
end

--- __Static__
---
--- Converts a black hole radius into it corresponding light radius
--- @param radius number The radius of the black hole
--- @return number The radius of the light
--- @private
function BlackHole:radiusToLightRadius(radius)
  return radius * BlackHole.LIGHT_RADIUS_MULTIPLIER
end

--- __Static__
---
--- Converts a black hole radius into it corresponding event horizon radius
--- @param radius number The radius of the black hole
--- @return number The radius of the event horizon
--- @private
function BlackHole:radiusToEventHorizonRadius(radius)
  -- /2 as the model works with a 1m diameter
  return (radius / 2) * BlackHole.EVENT_HORIZON_SCALE_MULTIPLIER
end

--- Updates the black hole, based on the current state. If the black hole is dead, this method does nothing.
--- @param dt number The time since the last frame
--- @see Celestial#update
function BlackHole:update(dt)
  if self:isDead() then return end
  self.state:update(dt)
end

--- Scales the event horizon in-game object
--- @param ratio number The scale to apply as a normalised ratio of the radius
--- @private
function BlackHole:scaleEventHorizon(ratio)
  local scale = ratio * BlackHole:radiusToEventHorizonRadius(self.radius)
  self.eventHorizon:setField("scale", 0, scale.." "..scale.." "..scale)
end

--- Scales the light in-game object
--- @param ratio number The scale to apply as a normalised ratio of the radius
--- @private
function BlackHole:scaleLight(ratio)
  local radius = BlackHole:radiusToLightRadius(self.radius) * ratio
  self.lightObject:setField("radius", 0, radius)
end

--- Rotates the in-game accretion disc object
--- @param angle number The angle to add
--- @private
function BlackHole:rotateAccretionDisc(angle)
  self.rotation = self.rotation + angle
  self.accretionDisc:setField("rotationEuler", 0, "0 0 "..self.rotation)
end

--- Sets the alpha of the in-game accretion disc object
--- @param alpha number The alpha to set
--- @private
function BlackHole:setAccretionDiscAlpha(alpha)
  self.accretionDisc:setField("instanceColor", 0, "1 1 1 "..alpha)
end

--- Deletes all assets created by the black hole and emits the `"evaporated"` event.
--- @private
function BlackHole:setDead()
  self:destroy()
  self.events:emit("evaporated")
end

--- Sets the black hole to be in its idle state and emits the `"formed"` event
--- @private
function BlackHole:idle()
  self.state:switch("idle")
  self.events:emit("formed")
end

--- Beings the evaporation of the black hole
function BlackHole:evaporate()
  self.state:switch("evaporating")
end

--- Deletes all assets created by the black hole.
--- The black hole will be considered functionally dead from now on.
--- @see Celestial#destroy
function BlackHole:destroy()
  if self.state:getState() == "dead" then return end

  worldObjectHandler.releaseTSStatic(self.accretionDisc)
  self.accretionDisc = nil

  worldObjectHandler.releaseTSStatic(self.eventHorizon)
  self.eventHorizon = nil

  worldObjectHandler.releasePointLight(self.lightObject)
  self.lightObject = nil

  self.state:switch("dead")
end

--- @return boolean Whether the black hole is dead
--- @see Celestial#isDead
function BlackHole:isDead()
  return self.state:getState() == "dead"
end

--- @return Events<BlackHoleEvent> A registrable set of events that can be registered to or unregistered from
--- @see Celestial#getEvents
function BlackHole:getEvents()
  return self.events
end

--- __Constructor__
--- @param position LuaVec3 The position of the black hole
--- @param radius number The total radius of the black hole (i.e. of the accretion disc)
--- @return BlackHole The constructed instance
function BlackHole:new(position, radius)
  local self = classes.setmetatable({}, BlackHole)
  self.position = position
  self.radius = radius
  self.accretionDisc = worldObjectHandler.requestTSStatic({ modelName = "accretion_disc", position = position, scale = radius })
  self.eventHorizon = worldObjectHandler.requestTSStatic({ modelName = "sphere_opaque", position = position, scale = 0, instanceColour = "0 0 0 1" })
  self.lightObject = BlackHole:createLight(position, radius)
  self.rotation = 0
  self.events = Events:new()
  self.state = blackHoleAnimationStateMachine(self)

  -- To calculate the adjustment to the radius of the light, the radius has to be initially set to the full value in the
  -- `createLight` method. However, we want the light to start off with no radius, so we set it here instead.
  self.lightObject:setField("radius", 0, "0")

  return self
end

return BlackHole
