local maths = require("lua/explosionsNG/helpers/maths")
local AnimationStateMachine = require("lua/explosionsNG/animations/AnimationStateMachine")

-- The maximum rotation the black hole can rotate at in any state
local MAXIMUM_ROTATION = 2
-- The time the black hole takes to go from the `"forming"` state to the `"idle"` state
local FORMING_PERIOD = 3
-- The time the black hole takes to go from the `"evaporating"` state to the `"dead"` state
local EVAPORATING_PERIOD = 5

--- Animates the given black hole as an opening transition. If the ratio decreases in subsequent calls, then the
--- transition will appear backwards i.e. closing.
--- @param blackHole BlackHole The black hole to animate
--- @param dt number The time since the last frame
--- @param ratio number The ratio
local function openTransition(blackHole, dt, ratio)
  blackHole:rotateAccretionDisc(dt * MAXIMUM_ROTATION)
  blackHole:setAccretionDiscAlpha(ratio)
  blackHole:scaleEventHorizon(ratio)
  blackHole:scaleLight(maths.landingRampUp(ratio, 0.4))
end

--- Rotates the accretion disc and increases the alpha from nothing to full alpha, and scales the event horizon to its
--- radius.
--- @type AnimationTimeFn<BlackHole>
local forming = function(blackHole, dt, timeContext)
  openTransition(blackHole, dt, timeContext.elapsedRatio)
end

--- Rotates the black hole at a constant rate
--- @type AnimationTimeFn<BlackHole>
local idle = function(blackHole, dt)
  blackHole:rotateAccretionDisc(dt * MAXIMUM_ROTATION)
end

--- Rotates the accretion disc and decreases the alpha from nothing to full alpha, and scales the event horizon to zero
--- @type AnimationTimeFn<BlackHole>
local evaporating = function(blackHole, dt, timeContext)
  openTransition(blackHole, dt, timeContext.timeLeftRatio)
end

--- Returns a new set of animations to animate the given black hole for every state it can be in
--- @param blackHole BlackHole The star to animate
--- @return AnimationStateMachine<BlackHoleState, BlackHole> The animations for the given black hole
return function(blackHole)
  local animations = --[[--- @type AnimationStateMachine<BlackHoleState, BlackHole> ]] AnimationStateMachine:new(blackHole)
  animations:add("forming", forming, function() blackHole:idle() end, FORMING_PERIOD)
  animations:add("idle", idle)
  animations:add("evaporating", evaporating, function() blackHole:setDead() end, EVAPORATING_PERIOD)
  animations:addNoop("dead")

  -- Set initial state
  animations:switch("forming")

  return animations
end