local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local Events = require("lua/explosionsNG/misc/Events")
local starAnimationStateMachine = require("lua/explosionsNG/worldObjects/celestials/star/animations")
local worldObjectHandler = require("lua/explosionsNG/handlers/worldObjects/handler")

--- @alias StarState "forming" | "idle" | "collapsing" | "compressing" | "dead"
--- @alias StarEvent "formed" | "compressed"

--- @class Star : Celestial<StarEvent>
--- @field private position LuaVec3 The centre position
--- @field private radius number The radius (in m)
--- @field private rotation number The rotation (in radians)
--- @field private colour string The colour of the star
--- @field private starObject SceneObject The in-game representation of the star
--- @field private lightObject SceneObject The in-game representation of the light emitted
--- @field private soundEmitter SceneObject The in-game representation of the sound emitted
--- @field private state AnimationStateMachine<StarState, Star> The state machine with animations of the star
--- @field private events Events<StarEvent> The list of registrable events that can be emitted
--- @field private MODEL_RADIUS number __Static__ The radius of the model with unit scale applied to it
--- @field private COLOUR string __Static__ The colour of the star and the light it emits
--- @field private SOUND_RADIUS_MULTIPLIER number __Static__ The multiplier of star radius to sound emitter radius
--- @field private LIGHT_RADIUS_MULTIPLIER number __Static__ The multiplier of star radius to light radius
--- @field private LIGHT_BRIGHTNESS number __Static__ The brightness of the light emitted
--- @field private AXIS_TILT number __Static__ The tilt of the star when it rotates (in radians)
local Star = classes.newClass()
Star.MODEL_RADIUS = 0.03
Star.SOUND_RADIUS_MULTIPLIER = 10
Star.LIGHT_RADIUS_MULTIPLIER = 25
Star.LIGHT_BRIGHTNESS = 5
Star.AXIS_TILT = 0.6

--- __Static__
---
--- Converts a star radius into it corresponding model scale radius
--- @param radius number The radius of the star
--- @return number The model radius
--- @private
function Star:radiusToModelRadius(radius)
  return radius / Star.MODEL_RADIUS
end

--- __Static__
---
--- Converts a star radius into it corresponding light radius
--- @param radius number The radius of the star
--- @return number The light radius
--- @private
function Star:radiusToLightRadius(radius)
  return radius * Star.LIGHT_RADIUS_MULTIPLIER
end

--- __Static__
---
--- Converts a star radius into it corresponding sound distance
--- @param radius number The radius of the star
--- @return number The sound distance
--- @private
function Star:radiusToSoundDistance(radius)
  return radius * Star.LIGHT_RADIUS_MULTIPLIER
end

--- __Static__
---
--- Creates a star object that is placed at the given position
--- @param position LuaVec3 The position to place the star at
--- @param colour string The colour of the star
--- @return SceneObject The static object
--- @private
function Star:createStarObject(position, colour)
  return worldObjectHandler.requestTSStatic({
    modelName = "star",
    position = position,
    scale = 0,
    instanceColour = colour
  })
end

--- __Static__
---
--- Creates a new light object that is placed at the given position.
--- @param position LuaVec3 The position to place the light at
--- @param colour string The colour of the light
--- @return SceneObject The light object
--- @private
function Star:createLightObject(position, colour)
  return worldObjectHandler.requestPointLight({
    position = position,
    radius = 0,
    brightness = Star.LIGHT_BRIGHTNESS,
    colour = colour
  })
end

--- __Static__
---
--- Creates a new sound emitter object. This sound emitter will be placed at the given position and has a radius
--- calculated using the `radiusToSoundDistance` method with the given star's radius.
--- @param position LuaVec3 The position to place the star at
--- @param starRadius number The radius of the star
--- @return SceneObject The sound emitter object
--- @private
function Star:createSoundEmitterObject(position, starRadius)
  local distance = Star:radiusToSoundDistance(starRadius)
  return worldObjectHandler.requestSoundEmitter({
    track = "star_ambient",
    isLooping = true,
    maxDistance = distance,
    position = position
  })
end

--- Updates the star, based on the current state. If the star is dead, this method does nothing.
--- @param dt number The time since the last frame
function Star:update(dt)
  if self:isDead() then return end
  self.state:update(dt)
end

--- Updates the scale of the star and its light by a ratio amount
--- @param ratio number The scale ratio. A number between [0, 1]
--- @private
function Star:applyScale(ratio)
  local scale = ratio * Star:radiusToModelRadius(self.radius)
  self.starObject:setField("scale", 0, scale.." "..scale.." "..scale)

  local lightRadius = ratio * Star:radiusToLightRadius(self.radius)
  self.lightObject:setField("radius", 0, lightRadius)
end

--- Moves the star's position by random amount in a random direction. The `maxJitter` controls the maximum amount of
--- movement from the star's position is allowed.
--- @param maxJitter number The maximum amount of jitter that can be applied
--- @private
function Star:applyJitter(maxJitter)
  local jitter = self.position + vectors.random(maxJitter)
  self.starObject:setField("position", 0, vectors.toString(jitter))
end

--- Updates the rotation of the star with the given time, including the axis tilt
--- @param dt number The time since the last frame
--- @private
function Star:addRotation(dt)
  self.rotation = self.rotation + dt
  local rotationVector = vec3(0, Star.AXIS_TILT, self.rotation)
  self.starObject:setField("rotationEuler", 0, vectors.toString(rotationVector))
end

--- Deletes all assets created by the star.
--- The star will be considered functionally dead from now on.
--- @see Celestial#destroy
function Star:destroy()
  if self.state:getState() == "dead" then return end

  worldObjectHandler.releaseTSStatic(self.starObject)
  self.starObject = nil

  worldObjectHandler.releasePointLight(self.lightObject)
  self.lightObject = nil

  worldObjectHandler.releaseSoundEmitter(self.soundEmitter)
  self.soundEmitter = nil

  self.state:switch("dead")
end

--- Deletes all assets created by the star and emits the `"compressed"` event.
--- @private
function Star:setDead()
  self:destroy()
  self.events:emit("compressed")
end

--- Sets the star to be in its idle state and emits the `"formed"` event
--- @private
function Star:idle()
  self.state:switch("idle")
  self.events:emit("formed")
end

--- Begins compressing the star
function Star:compress()
  self.state:switch("compressing")
end

--- Begins collapsing the star
function Star:collapse()
  self.state:switch("collapsing")
end

--- Sets the centre of the star to the given position and updates the in-game objects
--- @param position LuaVec3 The position to set the star to
function Star:setPosition(position)
  self.position = vec3(position)
  self:updatePositionInGame()
end

--- Updates the positions of the in-game objects, from the current position of the star
--- @private
function Star:updatePositionInGame()
  local positionString = vectors.toString(self.position)
  self.starObject:setField("position", 0, positionString)
  self.lightObject:setField("position", 0, positionString)
end

--- @return boolean Whether the star is dead
--- @see Celestial#isDead
function Star:isDead()
  return self.state:getState() == "dead"
end

--- @return Events<StarEvent> A registrable set of events that can be registered to or unregistered from
--- @see Celestial#getEvents
function Star:getEvents()
  return self.events
end

--- __Constructor__
--- @param position LuaVec3 The position to place the star at
--- @param radius number The radius of the star
--- @param colour string The colour of the star
--- @return Star The constructed instance
function Star:new(position, radius, colour)
  local self = classes.setmetatable({}, Star)
  self.position = vec3(position)
  self.radius = radius
  self.starObject = Star:createStarObject(position, colour)
  self.lightObject = Star:createLightObject(position, colour)
  self.soundEmitter = Star:createSoundEmitterObject(position, radius)
  self.rotation = 0
  self.events = Events:new()
  self.state = starAnimationStateMachine(self)
  return self
end

return Star
