local maths = require("lua/explosionsNG/helpers/maths")
local AnimationStateMachine = require("lua/explosionsNG/animations/AnimationStateMachine")

-- The maximum rotation the star can rotate at in any state
local MAX_ROTATION = 50
-- The jitter multiplier to apply when the star is jittering during the `"collapsing"` state
local COLLAPSE_JITTER_MULTIPLIER = 2
-- The time the star takes to go from the `"forming"` state to the `"idle"` state
local FORMING_PERIOD = 2
-- The time the star takes to go from the `"collapsing"` state to the `"compressing"` state
local COLLAPSING_PERIOD = 10
-- The time the star takes to compress
local COMPRESSING_PERIOD = 0.07

--- Updates the forming state of the star, increasing the scale of star and rotating the star at, initially, the
--- `MAX_ROTATION`, with subsequent calls rotating at an ever decreasing amount,
--- @type AnimationTimeFn<Star>
local forming = function(star, dt, timeContext)
  local ratio = maths.smoothRampUp(timeContext.elapsedRatio)
  star:applyScale(ratio)
  star:addRotation(dt + dt * (1 - ratio) * (MAX_ROTATION - 1))
end

--- Updates the collapsing state of the star, jittering the positioning of the star by an ever increasing amount and
--- rotating the star.
--- @type AnimationTimeFn<Star>
local collapsing = function(star, dt, timeContext)
  local jitter = math.exp(timeContext.elapsed - COLLAPSING_PERIOD) * COLLAPSE_JITTER_MULTIPLIER
  star:applyJitter(jitter)
  star:addRotation(dt)
end

--- Updates the compressing state of the star. This function decreases the scale of the star, jitters the positioning of
--- the star and rotates it at an ever increasing amount.
--- @type AnimationTimeFn<Star>
local compressing = function(star, dt, timeContext)
  local rotationMultiplier = math.max(1 / timeContext.timeLeft, MAX_ROTATION)
  star:applyScale(timeContext.timeLeft)
  star:addRotation(dt * rotationMultiplier)
  star:applyJitter(COLLAPSE_JITTER_MULTIPLIER)
end

--- Updates the idle state of the star. This function rotates the star at a constant rate
--- @type AnimationTimeFn<Star>
local idle = function(star, dt)
  star:addRotation(dt)
end

--- Returns a new set of animations to animate the given star for every state it can be in
--- @param star Star The star to animate
--- @return AnimationStateMachine<StarState, Star> The star animations for the given star
return function(star)
  local animations = --[[--- @type AnimationStateMachine<StarState, Star> ]] AnimationStateMachine:new(star)
  animations:add("forming", forming, function() star:idle() end, FORMING_PERIOD)
  animations:add("idle", idle)
  animations:add("collapsing", collapsing, function() star:compress() end, COLLAPSING_PERIOD)
  animations:add("compressing", compressing, function() star:setDead() end, COMPRESSING_PERIOD)
  animations:addNoop("dead")

  -- Set initial state
  animations:switch("forming")

  return animations
end