local classes = require("lua/explosionsNG/helpers/classes")
local vectors = require("lua/explosionsNG/helpers/vectors")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local simObjects = require("lua/explosionsNG/handlers/worldObjects/simObjects")
local Timer = require("lua/explosionsNG/timers/Timer")

--- @class FiniteLight : DynamicObject
--- @field private options PointLightOptions
--- @field private lightObject SceneObject The in-game PointLight object
--- @field private timer Timer The timer to time the light and remove the in-game object on completion
--- @field private finished boolean Whether the point light has finished emitting light
local FiniteLight = classes.newClass()

--- Updates the brightness of the light and internal timer
--- @param dt number The time since the last frame
function FiniteLight:update(dt)
  if self.finished then return end

  local timeLeftRatio = self.timer:timeLeftRatio()
  self:updateLight(timeLeftRatio)

  self.timer:update(dt)
end

--- Places the light at the given position, provide this light is not finished.
--- @param newPosition LuaVec3 The new position to place the light
function FiniteLight:setPosition(newPosition)
  if self.finished then return end
  self.lightObject:setField("position", 0, vectors.toString(newPosition))
end

--- Updates the brightness of the light, based on the time left and the initial brightness of the light
--- @param timeLeft number The normalised time left
--- @private
function FiniteLight:updateLight(timeLeft)
  self.lightObject:setField("brightness", 0, timeLeft * self.options.brightness)
end

--- Releases the in-game light object to the world objects handler and sets this instance to be finished
function FiniteLight:destroy()
  if self.finished then return end

  self.finished = true
  worldObjectsHandler.releasePointLight(self.lightObject)
  self.lightObject = nil
end

--- @return boolean Whether this light has finished
function FiniteLight:isFinished()
  return self.finished
end

-- TODO Repeated structure and logic with SoundEmitter
--- __Constructor__
--- @param time number
--- @param options PointLightOptions
--- @return FiniteLight The constructed instance
function FiniteLight:new(time, options)
  local self = classes.setmetatable({}, FiniteLight)
  self.options = simObjects.fillPointLightOptions(options)
  self.timer = Timer:new(time, function() self:destroy() end, true)
  self.lightObject = worldObjectsHandler.requestPointLight(self.options)
  self.finished = false
  return self
end

return FiniteLight
