local classes = require("lua/explosionsNG/helpers/classes")
local maths = require("lua/explosionsNG/helpers/maths")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local Timer = require("lua/explosionsNG/timers/Timer")
local SoundEmitter = require("lua/explosionsNG/worldObjects/SoundEmitter")

--- Properties:
---  - `colour` - The colour of the bolt and accompanying lighting
---  - `life` - The amount of time the lightning bolt is active. Note that the actual life of the lightning will be at minimum the length of the sound effect played, even if this value is less.
---  - `emitSound` - Whether to emit sound or not
--- @alias LightningOptions { colour: string | nil, life: number | nil, emitSound: boolean | nil }

--- @class Lightning
--- @field options LightningOptions The options to use
--- @field bolt SceneObject The in-game object for the bolt of lightning
--- @field light SceneObject The in-game object for the light
--- @field soundEmitter SoundEmitter The sound emitter for playing a sound effect
--- @field isDead boolean Whether the lightning bolt has finished its visual effects
--- @field lifeTimer Timer The timer to time the visual effects of the lightning and remove in-game objects on completion
--- @field LIGHT_BRIGHTNESS number __Static__ The brightness of the light in-game object
--- @field LIGHT_RADIUS number __Static__ The radius of the light in-game object
--- @field BOLT_SCALE LuaVec3 __Static__ The scale of the bolt in-game object
local Lightning = classes.newClass()
Lightning.BOLT_SCALE = vec3(100, 1, 1000)
Lightning.LIGHT_BRIGHTNESS = 10
Lightning.LIGHT_RADIUS = 10

--- __Static__
---
--- Fills in any property not given in the given options object
--- @param options LightningOptions The options to fill
--- @return LightningOptions The filled options
function Lightning:fillOptions(options)
  local emitSound = options.emitSound
  if emitSound == nil then emitSound = true end

  return {
    colour = options.colour or "0 0.2 1",
    life = options.life or 5,
    emitSound = emitSound
  }
end

--- __Static__
---
--- Creates and returns the point light for a lightning bolt
--- @param position LuaVec3 The position to place the point light at
--- @param options LightningOptions The list of options of the lightning instance
--- @return SceneObject The create in-game object
function Lightning:createPointLight(position, options)
  return worldObjectsHandler.requestPointLight({
    position = position,
    radius = Lightning.LIGHT_RADIUS,
    colour = options.colour,
    applyRadiusToPosition = true
  })
end

--- __Static__
---
--- Creates and returns the bolt for a lightning bolt
--- @param position LuaVec3 The position to place the bolt at
--- @return SceneObject The create in-game object
function Lightning:createBolt(position)
  return worldObjectsHandler.requestTSStatic({
    modelName = "lightning"..math.random(1, 3),
    position = position,
    scale = Lightning.BOLT_SCALE,
    rotationEuler = vec3(0, 0, maths.randomSignedFloat(360)) -- Give a random direction to the bolt
  })
end

--- __Static__
---
--- Creates and returns the sound emitter for a lightning bolt
--- @param position LuaVec3 The position to place the sound emitter at
--- @return SoundEmitter The create sound emitter
function Lightning:createSoundEmitter(position)
  local trackName = "lightning"..math.random(1, 3)
  return SoundEmitter:new({ track = trackName, position = position, maxDistance = 1000 })
end

--- Updates the sound effects (if present) and also the in-game bolt and light objects. If this instance `isFinished`,
--- then this method will do nothing.
--- @param dt number The time since the last frame
function Lightning:update(dt)
  if self:isFinished() then return end

  if self.soundEmitter then
    self.soundEmitter:update(dt)
  end

  if not self.isDead then
    local lifeLeftRatio = self.lifeTimer:timeLeftRatio()
    self:updateBolt(lifeLeftRatio)
    self:updateLight(lifeLeftRatio)
    self.lifeTimer:update(dt)
  end
end

--- Updates the `bolts` in-game colour
--- @param lifeLeftRatio number The ratio of the time left to maximum life
--- @private
function Lightning:updateBolt(lifeLeftRatio)
  self.bolt:setField("instanceColor", 0, self.options.colour.." "..lifeLeftRatio)
end

--- Updates the `light` in-game brightness
--- @param lifeLeftRatio number The ratio of the time left to maximum life
--- @private
function Lightning:updateLight(lifeLeftRatio)
  local brightness = lifeLeftRatio * Lightning.LIGHT_BRIGHTNESS
  self.light:setField("brightness", 0, brightness)
end

--- Sets this instance to be dead and releases the `light` and `bolt` in-game objects to the world objects handler.
--- Note that this method does not destroy the `soundEmitter`, as this will clean itself up
--- @private
function Lightning:markDead()
  self.isDead = true

  worldObjectsHandler.releasePointLight(self.light)
  worldObjectsHandler.releaseTSStatic(self.bolt)
  self.light = nil
  self.bolt = nil
end

--- @return boolean Whether the lightning has finished and played its sound effect
function Lightning:isFinished()
  return self.isDead and (not self.soundEmitter or self.soundEmitter:isFinished())
end

--- Releases all world objects and destroys this instance
function Lightning:destroy()
  if self.isDead then return end

  self:markDead()
  self.soundEmitter:destroy()
end

--- __Constructor__
--- @param position LuaVec3 The position to place the lightning bolt
--- @param options? LightningOptions | nil The options to use
--- @return Lightning The constructed instance
function Lightning:new(position, options)
  local self = classes.setmetatable({}, Lightning)
  self.options = Lightning:fillOptions(options or {})
  self.bolt = Lightning:createBolt(position)
  self.light = Lightning:createPointLight(position, self.options)
  self.soundEmitter = self.options.emitSound and Lightning:createSoundEmitter(position)
  self.isDead = false
  self.lifeTimer = Timer:new(self.options.life, function() self:markDead() end)
  self.lifeTimer:start()
  return self
end

return Lightning
