local classes = require("lua/explosionsNG/helpers/classes")
local Timer = require("lua/explosionsNG/timers/Timer")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias SunSkyMap { [SunSkyField]: number } @A mapping of `SunSkyField` to a numeric value

--- Performs an effect on the `"sunsky"` in-game object, over a period of time. This class allows for reverting the
--- changes to allow complete cleanup of the object and returning it back to its original state.
---
--- In the reverting case, this class uses a quartic function when transitioning. This is because any brightness
--- properties of the sunsky object have a dramatic effect at low values. For example, brightness 0.1 = little
--- brightness, 1 = decent brightness, 10 = not much more brightness. In other words, it is not linear for the
--- brightness to be put into effect when the lighting engine kicks in.
--- @class SkyEffect
--- @field private period number The total time the sky effect is active for, per direction
--- @field private sunsky SunSkySceneObject The in-game sunsky object
--- @field private initialSunSky SunSkyMap The sunsky settings before this effect began
--- @field private transitionTo SunSkyMap The desired settings that this effect should transition to
--- @field private timer Timer The timer to transition between settings
--- @field private reverting boolean Whether the effect is going backwards (i.e. desired settings to initial settings)
--- @field private KEYS Array<SunSkyField> __Static__ A list of keys that are transitionable
local SkyEffect = classes.newClass()
SkyEffect.KEYS = Array:new({ "brightness", "skyBrightness" })

--- Updates the sky effect, if it is active
--- @param dt number The time since the last frame
function SkyEffect:update(dt)
  if self.timer:isComplete() then return end

  self:updateSunSkyObject()
  self.timer:update(dt)
end

--- Updates the in-game sunsky object.
--- If this effect is happening in the normal direction, the effect is linear.
--- If this effect is happening in the reverse direction, the effect is instead of a quartic nature. For details on
--- this, see the class documentation.
--- @private
function SkyEffect:updateSunSkyObject()
  local time = self.timer:elapsedTime()

  --- @param key SunSkyField
  local linear = function(key)
    local gradient = (self.transitionTo[key] - self.initialSunSky[key]) / self.period
    return gradient * time + self.initialSunSky[key]
  end
  --- @param key SunSkyField
  local quartic = function(key)
    local gradient = self.initialSunSky[key] - self.transitionTo[key]
    return gradient * (time / self.period) ^ 4 + self.transitionTo[key]
  end

  SkyEffect.KEYS:forEach(function(key)
    local newValue = self.reverting and quartic(key) or linear(key)
    self.sunsky:setField(key, 0, newValue)
    self.sunsky:postApply()
  end)
end

--- Begins the sky effect
function SkyEffect:start()
  self.timer:start()
end

--- Begins the sky effect in the reverse direction
function SkyEffect:revert()
  self.timer:restart()
  self.reverting = true
end

--- Resets the sky effects back to the defaults instantaneously
function SkyEffect:revertInstantly()
  SkyEffect.KEYS:forEach(function(key)
    local initialValue = self.initialSunSky[key]
    self.sunsky:setField(key, 0, initialValue)
    self.sunsky:postApply()
  end)
end

--- __Constructor__
--- @param period number The period the effect should last for
--- @param transitionTo SunSkyMap The desired sky effect that should be applied when this effect is complete
--- @return SkyEffect The constructed instance
function SkyEffect:new(period, transitionTo)
  local self = classes.setmetatable({}, SkyEffect)
  self.sunsky = scenetree.findObject("sunsky")
  self.period = period
  self.initialSunSky = { skyBrightness = self.sunsky.skyBrightness, brightness = self.sunsky.brightness }
  self.transitionTo = transitionTo
  self.timer = Timer:new(period, function() end) -- Callback not needed
  self.reverting = false
  return self
end

return SkyEffect
