local classes = require("lua/explosionsNG/helpers/classes")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local IntervalTimer = require("lua/explosionsNG/timers/IntervalTimer")
local Events = require("lua/explosionsNG/misc/Events")
local laserAnimationStateMachine = require("lua/explosionsNG/worldObjects/laser/animations")

--- Properties:
---  - `radius` - the radius of the laser
---  - `pulseTime` - the period between two "pulses" of the laser
---  - `colour` - the colour of the laser
--- @alias LaserOptions { radius: number | nil, pulseTime: number | nil, colour: string | nil }
--- @alias LaserState "opening" | "idle" | "closing" | "dead"
--- @alias LaserEvent "opened" | "closed"

--- @class Laser : Evented<LaserEvent>
--- @field private position LuaVec3 The position of the laser
--- @field private options LaserOptions The list of options to use
--- @field private laserObject SceneObject The in-game laser object
--- @field private light SceneObject The in-game light
--- @field private pulseTimer IntervalTimer The timer for the pulses
--- @field private state AnimationStateMachine<LaserState, Laser>
--- @field private events Events<LaserEvent> The list of registrable events that can be emitted
--- @field private LIGHT_RADIUS_MULTIPLIER number __Static__ The multiplier to use when converting a radius of the laser to the light it emits
--- @field private SOUND_EMITTER_RADIUS_MULTIPLIER number __Static__ The multiplier to use when converting a radius of the laser to the sound it emits
local Laser = classes.newClass()
Laser.LIGHT_RADIUS_MULTIPLIER = 5
Laser.SOUND_EMITTER_RADIUS_MULTIPLIER = 10

--- __Static__
---
--- Fills in any option not given in the given options with default values
--- @param options LaserOptions The partial list of options
--- @return LaserOptions The list of partial options with defaulted options
--- @private
function Laser:fillOptions(options)
  return {
    radius = options.radius or 1,
    pulseTime = options.pulseTime or 5,
    colour = options.colour or "1 1 1"
  }
end

--- __Static__
---
--- Creates and returns a laser in-game object
--- @param position LuaVec3 The position to set the object to
--- @param options LaserOptions The list of options to use
--- @return SceneObject The created object
--- @private
function Laser:createLaserObject(position, options)
  return worldObjectsHandler.requestTSStatic({
    modelName = "cylinder_electric",
    dynamic = false,
    position = position,
    scale = options.radius,
    instanceColour = options.colour
  })
end

--- __Static__
---
--- Creates and returns a point light in-game object for a laser
--- @param position LuaVec3 The position to set the light to
--- @param options LaserOptions The list of options to use
--- @return SceneObject The created light
--- @private
function Laser:createPointLight(position, options)
  return worldObjectsHandler.requestPointLight({
    applyRadiusToPosition = true,
    position = position,
    radius = options.radius * Laser.LIGHT_RADIUS_MULTIPLIER,
    colour = options.colour,
    brightness = 0
  })
end

--- Creates and returns a list of options to apply to an SFX emitter for this anomaly
--- @param isAmbient boolean Whether to return options for an ambient emitter
--- @return SFXEmitterOptions A list of options to use for SFX emitter in-game objects for this anomaly
--- @private
function Laser:getSfxOptions(track, isAmbient)
  local radius = self.options.radius
  return {
    track = track,
    position = self.position,
    maxDistance = radius * Laser.SOUND_EMITTER_RADIUS_MULTIPLIER,
    referenceDistance = radius * Laser.SOUND_EMITTER_RADIUS_MULTIPLIER / 2,
    isLooping = isAmbient
  }
end

--- Updates the state of this laser
--- @param dt number The time since the last frame
function Laser:update(dt)
  self.state:update(dt)
end

--- Updates the alpha fade of the laser in-game object
--- @param alpha number The alpha to set
--- @private
function Laser:updateScale(alpha)
  self.laserObject:setField("instanceColor", 0, self.options.colour.." "..alpha)
  self.laserObject:postApply()
end

--- Changes the state to the `"idle"` state and emits the `"opened"` event.
--- This method will then begin the pulse timer.
--- @private
function Laser:idle()
  self.state:switch("idle")
  self.events:emit("opened")
  self.pulseTimer:start()
end

--- Begins to close the laser
function Laser:close()
  if self.state:getState() ~= "idle" then
    log("W", "explosivesNG", "Cannot close a laser which is not idle")
    return
  end

  self.state:switch("closing")

  worldObjectsHandler.releaseSoundEmitter(self.ambientSoundEmitter)
  self.ambientSoundEmitter = nil

  worldObjectsHandler.releaseSoundEmitter(self.terminalSoundEmitter)
  self.terminalSoundEmitter = worldObjectsHandler.requestSoundEmitter(self:getSfxOptions("laser_closing", false))

  worldObjectsHandler.releasePointLight(self.light)
  self.light = nil
end

--- Destroys all in-game objects and emits the `"closed"` event
function Laser:destroy()
  if self.state:getState() == "dead" then return end

  self.state:switch("dead")

  if self.laserObject then
    worldObjectsHandler.releaseTSStatic(self.laserObject)
    self.laserObject = nil
  end

  if self.ambientSoundEmitter then
    worldObjectsHandler.releaseSoundEmitter(self.ambientSoundEmitter)
    self.ambientSoundEmitter = nil
  end

  if self.terminalSoundEmitter then
    worldObjectsHandler.releaseSoundEmitter(self.terminalSoundEmitter)
    self.terminalSoundEmitter = nil
  end

  if self.light then
    worldObjectsHandler.releasePointLight(self.light)
    self.light = nil
  end

  self.events:emit("closed")
end

--- @return Events<LaserEvent> A registrable set of events that can be registered to or unregistered from
function Laser:getEvents()
  return self.events
end

--- __Constructor__
--- @param position LuaVec3 The position to place the laser
--- @param options? LaserOptions | nil An optional list of options to use
--- @return Laser The constructed instance
function Laser:new(position, options)
  local self = classes.setmetatable({}, Laser)
  self.options = Laser:fillOptions(options or {})
  self.position = position
  self.laserObject = Laser:createLaserObject(position, self.options)
  self.light = Laser:createPointLight(position, self.options)
  self.pulseTimer = IntervalTimer:new(self.options.pulseTime, function() end)
  self.terminalSoundEmitter = worldObjectsHandler.requestSoundEmitter(self:getSfxOptions("laser_opening", false))
  self.ambientSoundEmitter = worldObjectsHandler.requestSoundEmitter(self:getSfxOptions("laser_ambient", true))
  self.state = laserAnimationStateMachine(self)
  self.events = Events:new()
  return self
end

return Laser
