local collisions = require("lua/explosionsNG/helpers/collisions")
local levels = require("lua/explosionsNG/helpers/levels")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local explosivesHandler = require("lua/explosionsNG/handlers/explosives")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local previewsHandler = require("lua/explosionsNG/handlers/previews")
local chaosModeHandler = require("lua/explosionsNG/handlers/chaosMode")
local explosivesFacade = require("lua/explosionsNG/stores/explosives/facade")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias RequestParams { explosiveId: string, presetId: string, uiId: string, position: Vector, fuse: number }
--- @alias ModState { explosives: ExplosiveIdentifier[], bossData: BossData, isPlayerInAVehicle: boolean, needsReset: boolean, areMaterialsSetUp: boolean, chaosMode: ChaosModeConfigurationDto | nil }

--- A flag that stores whether the UI needs resetting.
--- For example: a new freeroam level was loaded and the UI configuration should not persist.
local uiNeedsResetting = false

--- A flag that stores whether the texture materials for this mod have been auto-generated or not, BEFORE the map was
--- loaded.
local areMaterialsSetUp = false

--- Wraps the given hook in a function that will not call the hook on calling, should the player not be in freeroam
--- @param hook function The hook to wrap
--- @return function The wrapped hook
local freeroamHookWrapper = function(hook)
  return function(...)
    if not extensions.freeroam_freeroam.state.freeroamActive then return end
    hook(...)
  end
end

--- @param dt number The time since the last frame
local function onPreRender(dt)
  if simTimeAuthority.getPause() then return end

  dt = dt * bullettime.get()

  explosivesHandler.update(dt)
  chaosModeHandler.update(dt)
  vehiclesHandler.update(dt)
end

--- Cleans up all handlers, and sets up the necessary flags for this extension
local function onFreeroamLoaded()
  worldObjectsHandler.initialise()
  vehiclesHandler.clear()
  explosivesHandler.clear()
  previewsHandler.clear()
  chaosModeHandler.stop()
  collisions.onMapLoaded()

  uiNeedsResetting = true
  areMaterialsSetUp = levels.checkAndCopyMaterialForCurrentLevel()
end

--- Creates and places the explosives that matches the given list of explosive and preset ids as well as the position to
--- place each explosive.
--- @param requests RequestParams[] The request parameters containing the details for each explosive
local function placeExplosives(requests)
  Array:new(requests):forEach(function(request)
    local explosive = explosivesFacade.createNewExplosive(request.explosiveId, request.presetId, vec3(request.position), request.fuse)
    explosivesHandler.addExplosive(request.uiId, request.explosiveId, request.presetId, explosive)
  end)
end

--- Sets the chaos mode configuration before starting the mode
--- @param configuration ChaosModeConfigurationDto The configuration to set
local function startChaosMode(configuration)
  chaosModeHandler.setConfigurationByDto(configuration)
  chaosModeHandler.start()
end

--- Collates and returns a snapshot of the current state of the mod
--- @return ModState
local function getCurrentState()
  local doesUiNeedResetting = uiNeedsResetting
  uiNeedsResetting = false -- Reset to false so this is only returned once for the UI

  return {
    explosives = explosivesHandler.getIds():toPrimitive(),
    bossData = explosivesHandler.getBossData(),
    isPlayerInAVehicle = vehiclesHandler.isPlayerInAVehicle(),
    needsReset = doesUiNeedResetting,
    areMaterialsSetUp = areMaterialsSetUp,
    chaosMode = chaosModeHandler.getActiveConfigurationDto()
  }
end

return {
  -- BeamNG interface
  onPreRender = freeroamHookWrapper(onPreRender),
  onVehicleSpawned = freeroamHookWrapper(vehiclesHandler.onVehicleSpawned),
  onVehicleResetted = freeroamHookWrapper(vehiclesHandler.onVehicleResetted),
  onVehicleDestroyed = freeroamHookWrapper(vehiclesHandler.onVehicleDestroyed),
  onVehicleSwitched = freeroamHookWrapper(vehiclesHandler.onVehicleSwitched),
  onFreeroamLoaded = onFreeroamLoaded,

  -- UI Interface
  getCurrentState = getCurrentState,
  placeExplosives = placeExplosives,
  getListOfExplosives = explosivesFacade.getListOfExplosives,
  createPreviewByOption = previewsHandler.createPreviewByOption,
  createPreviewByPosition = previewsHandler.createPreviewByPosition,
  updatePreviewPosition = previewsHandler.updatePreviewPosition,
  updatePreviewPositionByType = previewsHandler.updatePreviewPositionByType,
  deletePreview = previewsHandler.deletePreview,
  destroyAll = explosivesHandler.deactivateAll,
  destroyExplosive = explosivesHandler.deactivateExplosive,
  startChaosMode = startChaosMode,
  stopChaosMode = chaosModeHandler.stop
}
