angular.module("beamng.apps")
    /**
     * @ngdoc directive
     * @name beamng.apps:engBossBarContainer
     * @description A container component that displays a multiphase boss bar, currently specifically tailored towards
     * the "The Convergence" explosive.
     * @param {boolean} showAnimations Whether to show animations on the boss bar or not
     * @requires engBossesFactory
     * @requires engMiscFactory
     */
    .directive("engBossBarContainer", ["engBossesFactory", "engMiscFactory", (engBossesFactory, engMiscFactory) => {
      // noinspection JSUnusedGlobalSymbols
      return {
        templateUrl: "../../modules/apps/ExplosionController/containers/boss-bar-container/boss-bar-container.html",
        restrict: "E",
        scope: {
          showAnimations: "<"
        },
        controller: $scope => {
          /** @type ConvergencePhaseProgress */
          $scope.mainPhase = undefined;
          /** @type ConvergencePhaseProgress[] */
          $scope.phases = [];
          $scope.failedSegment = false;
          $scope.failedMainPhaseSegment = false;

          /**
           * Sets the phases list and main phases to the given progress. If the progress is `null`, then the phases list
           * is set to an empty array and the main phase is set to `undefined`.
           * @param {ConvergencePhaseProgress[] | null} progress The next progress list
           */
          const bossBattleProgressSubscription = progress => {
            if (!engMiscFactory.isDefined(progress)) {
              $scope.phases = [];
              $scope.mainPhase = undefined;
              return;
            }

            const lastElementIndex = progress.length - 1;
            $scope.phases = progress.slice(0, lastElementIndex);
            $scope.mainPhase = progress[lastElementIndex];
          }

          /**
           * Sets the current segments has been failed to the given flag.
           * This function will also set whether the main phase segment was failed, which will become `true` if all over
           * phases are dead and the given flag is itself `true`.
           * @param {boolean} failed The next failed flag value
           */
          const setHasFailedSegmentSubscription = failed => {
            $scope.failedSegment = failed;
            $scope.failedMainPhaseSegment = $scope.phases.every(({ dead }) => dead) && failed;
          };

          /**
           * Updates the {@link engBossesFactory} with the new progress data
           * @param {ConvergencePhaseProgress[] | undefined} data The progress data
           */
          function onUpdate(data) {
            const newData = data !== undefined ? data : null;
            engBossesFactory.setBossBattleProgress(newData);
          }

          /**
           * Updates whether the player has failed the current segment of the boss on the {@link engBossesFactory}
           * @param {boolean} failed Whether the player has failed the current segment of the boss
           */
          function onFailedUpdate(failed) {
            engBossesFactory.setHasFailedSegment(failed);
          }

          /**
           * Initialises the component.
           * Subscribes to various subscriptions on the {@link engBossesFactory}.
           */
          function init() {
            engBossesFactory.bossBattleProgress.subscribe(bossBattleProgressSubscription);
            engBossesFactory.hasFailedSegment.subscribe(setHasFailedSegmentSubscription);
          }

          $scope.$on("engBossBarUpdate", (_event, data) => onUpdate(/** @type {ConvergencePhaseProgress[] | undefined} */ data));

          $scope.$on("engBossBarFailed", (_event, data) => onFailedUpdate(/** @type boolean */ data));

          /**
           * Unsubscribes from any subscribed observables
           */
          $scope.$on("$destroy", () => {
            engBossesFactory.bossBattleProgress.unsubscribe(bossBattleProgressSubscription);
            engBossesFactory.hasFailedSegment.unsubscribe(setHasFailedSegmentSubscription);
          });

          init();

          // setTimeout(() => {
          //   onUpdate([
          //     { active: true, dead: false, segments: 3, segmentsCompleted: 0 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 2 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 0 },
          //     { active: true, dead: false, segments: 6, segmentsCompleted: 3 }
          //   ]);
          //   onFailedUpdate(true);
          // }, 1000);
          //
          // setTimeout(() => {
          //   onUpdate([
          //     { active: true, dead: true, segments: 3, segmentsCompleted: 3 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 2 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 0 },
          //     { active: true, dead: false, segments: 6, segmentsCompleted: 3 }
          //   ]);
          //   onFailedUpdate(true);
          // }, 2500);
        }
      }
    }]);