angular.module("beamng.apps")
    /**
     * @ngdoc directive
     * @name beamng.apps:engChaosModeContainer
     * @description A container component that manages the chaos mode of the mod
     * @requires engUiFactory
     * @requires engChaosModeFactory
     * @requires engEndpointsFactory
     */
    .directive("engChaosModeContainer", ["engUiFactory", "engChaosModeFactory", "engEndpointsFactory", (
        engUiFactory,
        engChaosModeFactory,
        engEndpointsFactory
    ) => {
      // noinspection JSUnusedGlobalSymbols
      return {
        templateUrl: "../../modules/apps/ExplosionController/containers/chaos-mode-container/chaos-mode-container.html",
        restrict: "E",
        controller: $scope => {
          /** @type ConfiguredExplosive[] */
          $scope.explosiveOptions = [];
          /** @type ChaosModeConfiguration */
          $scope.configuration = undefined;
          $scope.isActive = false;
          $scope.selectedExplosivesErrors = [];

          /**
           * Sets the current configuration to the new given configuration.
           * Additionally, this function will enforce validation and set the selected explosives errors with any found.
           * @param {ChaosModeConfiguration} configuration The new configuration
           */
          const configurationSubscription = configuration => {
            $scope.configuration = configuration;
            $scope.selectedExplosivesErrors = configuration.allowedExplosives.length === 0
                ? ["Select at least one explosive"]
                : [];
          };

          /**
           * Sets the list of valid explosive options to the given ones.
           * @param {ConfiguredExplosive[]} explosiveOptions The new explosive options
           */
          const selectedExplosiveSubscription = explosiveOptions => { $scope.explosiveOptions = explosiveOptions; };

          /**
           * Sets whether the mode is active or not
           * @param {ConfiguredExplosive[]} isActive Whether the mode is currently active
           */
          const isActiveSubscription = isActive => { $scope.isActive = isActive; };

          /**
           * Sets the view back to the `"overview"` view on the {@link engUiFactory}
           */
          function onBackClick() {
            engUiFactory.setView("overview");
          }

          /**
           * Toggles the chaos mode on the {@link engEndpointsFactory} and the {@link engChaosModeFactory}, before
           * setting the view back to the `"overview"` view on the {@link engUiFactory}.
           */
          function onToggleClick() {
            const isNowActive = !$scope.isActive;
            engChaosModeFactory.setActive(isNowActive);

            if (!isNowActive) {
              engEndpointsFactory.stopChaosMode();
              return;
            }

            engEndpointsFactory.startChaosMode($scope.configuration);
            engUiFactory.setView("overview");
          }

          /**
           * Updates the configuration on the {@link engChaosModeFactory} with the current configuration, replacing the
           * list of [allowedExplosives]{@link ChaosModeConfiguration.allowedExplosives} with the given options.
           * @param {string[]} options The newly selected options
           */
          function onSelectedOptionsChange(options) {
            /** @type ChaosModeConfiguration */
            const newConfiguration = { ...$scope.configuration, allowedExplosives: options };
            engChaosModeFactory.setConfiguration(newConfiguration);
          }

          /**
           * Initialises the component.
           * Initialises the {@link engChaosModeFactory} and subscribes to all necessary subscriptions on all used
           * factories.
           */
          function init() {
            engChaosModeFactory.configuration.subscribe(configurationSubscription);
            engChaosModeFactory.explosiveOptions.subscribe(selectedExplosiveSubscription);
            engChaosModeFactory.isActive.subscribe(isActiveSubscription);
          }

          /**
           * Unsubscribes from any subscribed observables
           */
          $scope.$on("$destroy", () => {
            engChaosModeFactory.configuration.unsubscribe(configurationSubscription);
            engChaosModeFactory.explosiveOptions.unsubscribe(selectedExplosiveSubscription);
            engChaosModeFactory.isActive.unsubscribe(isActiveSubscription);
          });

          $scope.onBackClick = onBackClick;
          $scope.onToggleClick = onToggleClick;
          $scope.onSelectedOptionsChange = onSelectedOptionsChange;

          init();
        }
      }
    }]);