angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engExplosiveContainer
   * @description A container component that allows configuration of an explosive by the user
   * @requires engExplosivesFactory
   * @requires engMiscFactory
   * @requires engEndpointsFactory
   * @requires engPreviewsFactory
   * @requires engUiFactory
   */
  .directive("engExplosiveContainer", ["engExplosivesFactory", "engMiscFactory", "engEndpointsFactory", "engPreviewsFactory", "engUiFactory", (
      engExplosivesFactory,
      engMiscFactory,
      engEndpointsFactory,
      engPreviewsFactory,
      engUiFactory
  ) => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/containers/explosive-container/explosive-container.html",
      restrict: "E",
      scope: false,
      controller: $scope => {
        /** @type ExplosiveDto[] */
        $scope.explosives = [];
        /** @type {ConfiguredExplosive | undefined} */
        $scope.selectedExplosive = undefined;

        /** @type ButtonControls */
        $scope.cycleExplosives = { disablePrevious: true, disableNext: true };
        /** @type ButtonControls */
        $scope.cyclePresets = { disablePrevious: true, disableNext: true };
        /** @type {{ explosive: HorizontalDirection, preset: HorizontalDirection }} */
        $scope.animationDirection = { explosive: "left", preset: "left" };

        $scope.lockTier4s = false;
        $scope.isPlayerInAVehicle = false;
        $scope.isPositioning = false;
        /** @type ConfiguredExplosive */
        $scope.explosiveShowingInformation = undefined;
        /** @type CoordinateInputType */
        $scope.selectedCoordinateOption = undefined;
        /** @type string */
        $scope.previewId = undefined;

        /**
         * A predicate that returns `true` if the given {@link ExplosiveDto} has a matching id to the
         * {@link selectedExplosive}.
         * @param {ExplosiveDto["id"]} id The id to check
         * @returns {boolean} Whether the given explosive's id matches
         */
        const equalsSelected = ({ id }) => id === $scope.selectedExplosive.id;

        /**
         * A predicate that returns `true` if the given {@link ExplosivePreset} has a matching id to the one on the
         * {@link selectedExplosive}.
         * @param {ExplosivePreset["id"]} id The id to check
         * @returns {boolean} Whether the given explosive preset's id matches
         */
        const equalsSelectedPreset = ({ id }) => id === $scope.selectedExplosive.preset.id;

        /**
         * Sets the currently selected explosive to the new given explosive
         * @param {ConfiguredExplosive} explosive The new selected explosive
         */
        const selectedExplosiveSubscription = explosive => { $scope.selectedExplosive = explosive; };

        /**
         * Sets whether tier 4 explosives are currently locked from being selected
         * @param {boolean} locked The new value
         */
        const lockTier4sSubscription = locked => { $scope.lockTier4s = locked; };

        /**
         * Sets whether the player is currently in a vehicle or not to the given value
         * @param {boolean} isPlayerInAVehicle The new value
         */
        const isPlayerInAVehicleSubscription = isPlayerInAVehicle => {
          $scope.isPlayerInAVehicle = isPlayerInAVehicle;
        };

        /**
         * Sets whether to show boss information
         * @param {boolean} showingInformation Whether to show boss information
         */
        const isShowingBossInformationSubscription = showingInformation => {
          $scope.isShowingBossInformation = showingInformation;
        };

        /**
         * Sets the preview ID to the first element in the given array
         * @param {Preview[]} previews The new previews
         */
        const previewSubscription = previews => {
          if (previews.length > 1) {
            console.warn("More than one preview is active. This is not expected.");
          }

          $scope.previewId = previews.length > 0 ? previews[0].previewId : undefined;
        };

        /**
         * Sets whether an explosive is currently being positioned on the UI to the given value
         * @param {boolean} isPositioning The new value
         */
        const isPositioningSubscription = isPositioning => { $scope.isPositioning = isPositioning; };

        /**
         * Returns a new {@link ButtonControls} object based on an array with a given length and the selected index of
         * the array the controls would be cycling through.
         * More formally, the [disablePrevious]{@link ButtonControls.disablePrevious} will be set to whether the given
         * {@link index} is equal to `0`, and the [disableNext]{@link ButtonControls.disableNext} will be set to whether
         * the given {@link index} is 1 less than the given {@link arrayLength}.
         * @param {number} index The index
         * @param {number} arrayLength The length of the array. This should be greater than `0`.
         * @returns {ButtonControls} The constructed controls
         * @example
         * newButtonControlsByArray(0, 10) => { disablePrevious: true, disableNext: false }
         * newButtonControlsByArray(5, 6) => { disablePrevious: false, disableNext: true }
         */
        function newButtonControlsByArray(index, arrayLength) {
          return { disablePrevious: index === 0, disableNext: index === arrayLength - 1 };
        }

        /**
         * Selects the next explosive in the list, relative to the currently selected explosive.
         * The newly selected explosive will default to the first preset in the explosive's preset list and will
         * maintain the currently configured {@link Position} of the {@link selectedExplosive}.
         * If {@link selectNext} is `false`, then this function selects the previous explosive in the list.
         * @param {boolean} selectNext Whether to select the next explosive (`true`), or the previous (`false`)
         */
        function updateExplosive(selectNext) {
          const index = $scope.explosives.findIndex(equalsSelected);
          const newIndex = index + (selectNext ? 1 : -1);
          const dto = $scope.explosives[newIndex];

          const newPresetIndex = 0;

          const nextExplosive = /** @type ConfiguredExplosive */ {
            ...engExplosivesFactory.dtoToExplosive(dto, newPresetIndex),
            position: $scope.selectedExplosive.position // Persist location
          };

          engExplosivesFactory.setConfiguredExplosiveByIndex(nextExplosive);
          $scope.cycleExplosives = newButtonControlsByArray(newIndex, $scope.explosives.length);
          $scope.cyclePresets = newButtonControlsByArray(newPresetIndex, dto.presets.length);
        }

        /**
         * Selects the next explosive preset in the currently selected explosive's presets list.
         * If {@link selectNext} is `false`, then this function selects the previous preset.
         * @param {boolean} selectNext Whether to select the next preset (`true`), or the previous (`false`)
         */
        function updatePreset(selectNext) {
          const { presets } = $scope.explosives.find(equalsSelected);
          const index = presets.findIndex(equalsSelectedPreset);
          const newIndex = index + (selectNext ? 1 : -1);

          engExplosivesFactory.setSelectedPresetByIndex(newIndex);
          $scope.cyclePresets = newButtonControlsByArray(newIndex, presets.length);
        }

        /**
         * Sets the {@link animationDirection} with the given list of partial updates
         * @param {{ explosive: HorizontalDirection | undefined, preset: HorizontalDirection | undefined }} animationDirection
         * A partial object to update keys inside {@link animationDirection} with
         */
        function updateAnimation(animationDirection) {
          $scope.animationDirection = { ...$scope.animationDirection, ...animationDirection };
        }

        /**
         * Sets the selected explosive's position to the given one
         * @param {Position} position The position to use
         */
        function setExplosivePosition(position) {
          const { x, y, z } = position;
          $scope.selectedExplosive.position = {
            x: engMiscFactory.roundTo(x, 3),
            y: engMiscFactory.roundTo(y, 3),
            z: engMiscFactory.roundTo(z, 3)
          };
        }

        /**
         * Creates the in-world preview by calling various related endpoints on the {@link engEndpointsFactory}:
         *  - If the given {@link option} is `"manual"`, then the {@link position} will be used to place this preview
         *  - In all other cases, the {@link option} will be sent along with the request. and the {@link position}
         *    ignored
         * @param {CoordinateInputType} option The option to use for getting the position to place the preview at
         * @param {Position} [position] The position to place the preview at. Should be given if {@link option} is
         * `"manual"`. In all other cases, this parameter is ignored even if it is given
         */
        function createPreview(option, position) {
          const { uiId, id: explosiveId } = $scope.selectedExplosive;
          const presetId = $scope.selectedExplosive.preset.id;

          /** @param {string} id The id of the preview to add */
          const addPreview = id => engPreviewsFactory.addPreview({ previewId: id, explosiveUiId: uiId });

          if (option !== "manual") {
            engEndpointsFactory.createPreviewByOption(option, explosiveId, presetId, ({ id, position: serverPosition }) => {
              addPreview(id);
              // Update from the position returned
              setExplosivePosition(serverPosition);
            });
            return;
          }

          if (!engMiscFactory.isDefined(position)) {
            console.error("position is mandatory if the input type is 'manual'");
            return;
          }

          engEndpointsFactory.createPreviewByPosition(position, explosiveId, presetId, addPreview);
        }

        /**
         * Updates the in-game preview using the currently {@link selectedCoordinateOption} and
         * {@link selectedExplosive}'s position.
         */
        function updatePreviewForCurrentlySelected() {
          if ($scope.selectedCoordinateOption === "manual") {
            const position = $scope.selectedExplosive.position;
            // If the coordinates are not fully defined, then it probably means the user has not finished typing.
            // In any case, do not send these coordinates to the server.
            if (!engMiscFactory.isDefined(position.x) || !engMiscFactory.isDefined(position.y) || !engMiscFactory.isDefined(position.z)) {
              return;
            }

            engEndpointsFactory.updatePreviewPosition($scope.previewId, position);
            return;
          }

          engEndpointsFactory.updatePreviewPositionByType(
              $scope.previewId,
              $scope.selectedCoordinateOption,
              setExplosivePosition
          );
        }

        /**
         * Sets the {@link selectedCoordinateOption} to `"manual"` if the {@link selectedExplosive}'s position is
         * defined (see ...). Otherwise, sets this option to the {@link defaultOption}.
         * If the position has been created by the player (i.e. it was persisted), then we don't want to overwrite them.
         * @param {CoordinateInputType} defaultOption
         */
        function updateCoordinateOption(defaultOption) {
          $scope.selectedCoordinateOption = engMiscFactory.isPositionDefined($scope.selectedExplosive.position)
              ? "manual"
              : defaultOption;
        }

        /**
         * Closes the positioning window and deletes the in-game preview.
         * This function will also default any `undefined` coordinate to `0`, and enforce the validation on the fuse
         * property.
         */
        function closePositioningWindow() {
          engEndpointsFactory.deletePreview($scope.previewId);
          engPreviewsFactory.setPreviews([]);
          engUiFactory.setIsPositioningExplosive(false);

          // Default any coordinate to `0` if not given when the window is closed
          const { position, fuse } = $scope.selectedExplosive;
          position.x = engMiscFactory.withDefaultValue(position.x, 0);
          position.y = engMiscFactory.withDefaultValue(position.y, 0);
          position.z = engMiscFactory.withDefaultValue(position.z, 0);

          // Enforce fuse validation
          $scope.selectedExplosive.fuse = fuse >= 0 && fuse <= 999 ? fuse : 0;
        }

        /**
         * Displays the positioning layer and shows the in-game preview, requesting it to be created if it does not
         * exist yet.
         */
        function onSelectClicked() {
          // If the animation is playing, ignore any further clicks on this button
          if ($scope.isPositioning) return;

          engUiFactory.setIsPositioningExplosive(true);
          updateCoordinateOption($scope.selectedCoordinateOption);

          if ($scope.previewId !== undefined) {
            console.warn("previewId should not be defined at this stage");
          }

          createPreview($scope.selectedCoordinateOption, $scope.selectedExplosive.position);
        }

        /**
         * Updates the selected explosive to the previous in the array of DTOs.
         * Updates the animations for both explosive and preset to `"right"`.
         */
        function onPreviousExplosiveClicked() {
          updateExplosive(false);
          // Overwrite preset animation as well to match the explosive
          updateAnimation({ explosive: "right", preset: "right" });
        }

        /**
         * Sets the selected explosive to the next in the array of DTOs.
         * Sets the animations for both explosive and preset to `"left"`.
         */
        function onNextExplosiveClicked() {
          updateExplosive(true);
          // Overwrite preset animation as well to match the explosive
          updateAnimation({ explosive: "left", preset: "left" });
        }

        /**
         * Selects the previous preset in the currently selected explosive's preset list.
         * Sets the preset animation to be `"left"`.
         */
        function onPreviousPresetClicked() {
          updatePreset(false);
          updateAnimation({ preset: "left" });
        }

        /**
         * Selects the next preset in the currently selected explosive's preset list.
         * Sets the preset animation to be `"right"`.
         */
        function onNextPresetClicked() {
          updatePreset(true);
          updateAnimation({ preset: "right" });
        }

        /**
         * Closes the positioning window
         */
        function onOverlayClicked() {
          closePositioningWindow();
        }

        /**
         * If the selected explosive is the `"convergence"` explosive, this function will set the information flag to
         * `true` on the {@link engUiFactory}.
         * Otherwise, this function hides the positioning window and preview and sets the shown view to the `"overview"`
         * page.
         */
        function onPositionPlaceClicked() {
          // This will be true if the place button has been clicked twice in quick succession.
          // We are navigating away from the position screen, but the button can still be clicked while the animation
          // is being displayed.
          if (!engMiscFactory.isDefined($scope.selectedExplosive)) return;

          if ($scope.selectedExplosive.id === "convergence") {
            engUiFactory.setIsShowingInformation(true);
            return;
          }

          closePositioningWindow();
          engUiFactory.setView("overview");
        }

        /**
         * Updates the preview for the currently {@link selectedExplosive}'s new position
         */
        function onPositionChanged() {
          updatePreviewForCurrentlySelected();
        }

        /**
         * Sets the fuse for the current explosive with the given value
         * @param {number} fuse The new fuse to set
         */
        function onFuseChanged(fuse) {
          $scope.selectedExplosive.fuse = fuse;
        }

        /**
         * Updates the {@link selectedCoordinateOption} with the given {@link option}.
         * This function will then {@link updatePreviewForCurrentlySelected} option.
         * @param {CoordinateInputType} option The new coordinate option
         */
        function onCoordinateOptionUpdated(option) {
          $scope.selectedCoordinateOption = option;
          updatePreviewForCurrentlySelected(option);
        }

        /**
         * Hides the showing of the information box on the {@link engUiFactory}.
         * This function will then place closing the position window and change to the `"overview"` view on the
         * {@link engUiFactory}. It is important to note that this will happen after a delay, which coincides with the
         * animation delay of the information box closing. This is internally using a simple timeout, and has been
         * implemented like this as the ngAnimate directive (when animating ng-if directives) does not allow hooking on
         * to animation events, from that element.
         */
        function onInformationClosed() {
          engUiFactory.setIsShowingInformation(false);

          setTimeout(() => {
            closePositioningWindow();
            engUiFactory.setView("overview");
          }, 500);
        }

        /**
         * Initialises the component.
         * Initialises the {@link engExplosivesFactory} and subscribes to all necessary subscriptions on all used
         * factories.
         *
         * This function will set the list of {@link explosives} from this factory as well as then set up
         * the {@link cycleExplosives} and {@link cyclePresets} once the list is retrieved.
         *
         * This function will additionally set the {@link selectedCoordinateOption} to `"manual"` if the selected
         * explosive has a defined position (see [isPositionDefined]{@link engMiscFactory.isPositionDefined}), otherwise
         * it is set to `"player"` if a player vehicle exists, otherwise `"camera"` as a default.
         */
        function init() {
          engPreviewsFactory.previews.subscribe(previewSubscription);
          engUiFactory.isPositioningExplosive.subscribe(isPositioningSubscription);
          engUiFactory.isPlayerInAVehicle.subscribe(isPlayerInAVehicleSubscription);
          engUiFactory.isShowingInformation.subscribe(isShowingBossInformationSubscription);

          engExplosivesFactory.init(() => {
            engExplosivesFactory.selectedExplosive.subscribe(selectedExplosiveSubscription);
            engExplosivesFactory.lockTier4s.subscribe(lockTier4sSubscription);
            $scope.explosives = engExplosivesFactory.getExplosiveDTOs();

            const index = $scope.explosives.findIndex(equalsSelected);
            const dto = $scope.explosives[index];
            const presetIndex = dto.presets.findIndex(equalsSelectedPreset);
            $scope.cycleExplosives = newButtonControlsByArray(index, $scope.explosives.length);
            $scope.cyclePresets = newButtonControlsByArray(presetIndex, dto.presets.length);

            const defaultOption = $scope.isPlayerInAVehicle ? "player" : "camera";
            updateCoordinateOption(defaultOption);

            // If the previewId already exists, don't try to create a new preview
            if ($scope.isPositioning && $scope.previewId === undefined) {
              createPreview($scope.selectedCoordinateOption, $scope.selectedExplosive.position);
            }
          });
        }

        $scope.onSelectClicked = onSelectClicked;
        $scope.onPreviousExplosiveClicked = onPreviousExplosiveClicked;
        $scope.onNextExplosiveClicked = onNextExplosiveClicked;
        $scope.onPreviousPresetClicked = onPreviousPresetClicked;
        $scope.onNextPresetClicked = onNextPresetClicked;
        $scope.onOverlayClicked = onOverlayClicked;
        $scope.onPositionPlaceClicked = onPositionPlaceClicked;
        $scope.onPositionChanged = onPositionChanged;
        $scope.onFuseChanged = onFuseChanged;
        $scope.onCoordinateOptionUpdated = onCoordinateOptionUpdated;
        $scope.onInformationClosed = onInformationClosed;

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engExplosivesFactory.selectedExplosive.unsubscribe(selectedExplosiveSubscription);
          engExplosivesFactory.lockTier4s.unsubscribe(lockTier4sSubscription);
          engPreviewsFactory.previews.unsubscribe(previewSubscription);
          engUiFactory.isPlayerInAVehicle.unsubscribe(isPlayerInAVehicleSubscription);
          engUiFactory.isPositioningExplosive.unsubscribe(isPositioningSubscription);
          engUiFactory.isShowingInformation.unsubscribe(isShowingBossInformationSubscription);
        });

        init();
      }
    }
  }])