angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engController
   * @description The app container that is created by BeamNG when the app is added to the game. Acts as the root
   * element that is responsible for what view is shown but does not control the behaviour of these views. This acts as
   * an AngularJS controller, but is defined as a directive (as it must in order to work correctly as a BeamNG app).
   * @requires engExplosivesFactory
   * @requires engNavigationFactory
   * @requires engChaosModeFactory
   * @requires engEndpointsFactory
   * @requires engPreviewsFactory
   * @requires engMiscFactory
   * @requires engUiFactory
   */
  .directive("engController", ["engExplosivesFactory", "engNavigationFactory", "engChaosModeFactory", "engEndpointsFactory", "engPreviewsFactory", "engBossesFactory", "engMiscFactory", "engUiFactory", (
      engExplosivesFactory,
      engNavigationFactory,
      engChaosModeFactory,
      engEndpointsFactory,
      engPreviewsFactory,
      engBossesFactory,
      engMiscFactory,
      engUiFactory,
  ) => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/containers/main-container/main-container.html",
      restrict: "E",
      scope: false,
      controller: $scope => {
        /** @type {View | undefined} */
        $scope.view = undefined;
        /** @type ConfiguredExplosive[] */
        $scope.configuredExplosives = [];
        $scope.showAnimations = true;
        $scope.areMaterialsSetUp = true;

        /**
         * Sets the current view to the new given view.
         * @param {View} view The new view
         */
        const viewSubscription = view => { $scope.view = view; };

        /**
         * Sets the currently configured explosives to the new given list
         * @param {ConfiguredExplosive[]} explosives The new explosives
         */
        const configuredExplosivesSubscription = explosives => { $scope.configuredExplosives = explosives; };

        /**
         * Removes the explosive withe given uiId from the {@link engExplosivesFactory}
         * @param {string} explosiveUiId The uiId of the explosive
         */
        function onExplosiveFinished(explosiveUiId) {
          const explosiveToDelete = $scope.configuredExplosives.find(({ uiId }) => uiId === explosiveUiId);

          if (!engMiscFactory.isDefined(explosiveToDelete)) {
            console.warn(`Explosive with id ${explosiveUiId} has finished but is not present in the configuration`);
            return;
          }

          engExplosivesFactory.removeExplosive(explosiveUiId);
        }

        /**
         * Updates the {@link engBossesFactory}'s state with the given boss data
         * @param {BossData | undefined} bossData The data from the currently active boss
         */
        function updateBossFactoryState(bossData) {
          const hasActiveBoss = engMiscFactory.isDefined(bossData);
          engBossesFactory.setBossBattleProgress(hasActiveBoss ? bossData.progress : null);
          engBossesFactory.setHasFailedSegment(hasActiveBoss && bossData.failed);
        }

        /**
         * Sets whether the chaos mode is active, and the current state on the {@link engChaosModeFactory}
         * @param {ChaosModeConfiguration | undefined} configuration The new chaos mode configuration
         */
        function updateChaosModeState(configuration) {
          const isChaosModeActive = engMiscFactory.isDefined(configuration);
          engChaosModeFactory.setActive(isChaosModeActive);
          engChaosModeFactory.setTimer(isChaosModeActive ? configuration.timer : null);
          if (isChaosModeActive) engChaosModeFactory.setConfiguration(configuration);
        }

        /**
         * Merges the state's various factories with the given data from the server. This function will set whether the
         * player is in a vehicle on the {@link engUiFactory}, the boss data on the {@link engBossesFactory}, the chaos
         * mode state on the {@link engChaosModeFactory}, and combine the configured explosives on the
         * {@link engExplosivesFactory}.
         * @param {ExplosiveIdentifier[]} explosives The list of explosive identifiers on the server
         * @param {BossData | undefined} bossData The data from the currently active boss
         * @param {boolean} isPlayerInAVehicle Whether the player is currently in a vehicle
         * @param {ChaosModeConfiguration | undefined} chaosMode The current chaos mode configuration
         */
        function updateStateFromServer(explosives, bossData, isPlayerInAVehicle, chaosMode) {
          engUiFactory.setPlayerInAVehicle(isPlayerInAVehicle);
          engExplosivesFactory.softSyncWithServer(explosives);
          updateBossFactoryState(bossData);
          updateChaosModeState(chaosMode);
        }

        /**
         * Resets and sets the states of various factories with the given data from the server. This function will set
         * whether the player is in a vehicle on the {@link engUiFactory}, the boss data on the {@link engBossesFactory}
         * , the chaos mode state on the {@link engChaosModeFactory}, and set the configured explosives on the
         * {@link engExplosivesFactory}.
         *
         * The {@link callback} will be called once the states have been successfully updated.
         * @param {ExplosiveIdentifier[]} explosives The list of explosive identifiers on the server
         * @param {BossData | undefined} bossData The data from the currently active boss
         * @param {boolean} isPlayerInAVehicle Whether the player is currently in a vehicle
         * @param {ChaosModeConfiguration | undefined} chaosMode The current chaos mode configuration
         * @param {() => void} callback A callback function to call once the factories have been set
         */
        function resetStateToServer(explosives, bossData, isPlayerInAVehicle, chaosMode, callback) {
          engUiFactory.reset();
          engBossesFactory.reset();
          engPreviewsFactory.reset();
          engChaosModeFactory.reset();

          engUiFactory.setPlayerInAVehicle(isPlayerInAVehicle);
          updateBossFactoryState(bossData);
          updateChaosModeState(chaosMode);

          engExplosivesFactory.hardSyncWithServer(explosives, callback);
        }

        /**
         * Subscribe to all observables used by this component
         */
        function initSubscriptions() {
          engUiFactory.view.subscribe(viewSubscription);
          engExplosivesFactory.init(() => {
            engExplosivesFactory.configuredExplosives.subscribe(configuredExplosivesSubscription);
          });
        }

        /**
         * Initialises the component.
         * Calls the server to get the current state of the mod. If the UI needs a full reset, then this function will
         * reset the UI's state to that server's state (see {@link resetStateToServer}). Otherwise, this function will
         * combine the UI's current state with this server state (see {@link updateStateFromServer}).
         * Both functions will initialise the subscriptions.
         * @see resetStateToServer
         * @see updateStateFromServer
         */
        function init() {
          // engNavigationFactory.init();

          engEndpointsFactory.loadState(({
            explosives,
            bossData,
            isPlayerInAVehicle,
            needsReset,
            areMaterialsSetUp,
            chaosMode
          }) => {
            $scope.areMaterialsSetUp = areMaterialsSetUp;

            if (needsReset) {
              resetStateToServer(explosives, bossData, isPlayerInAVehicle, chaosMode, initSubscriptions);
              $scope.showAnimations = true;
              return;
            }

            updateStateFromServer(explosives, bossData, isPlayerInAVehicle, chaosMode);
            initSubscriptions();

            $scope.showAnimations = false;
            /**
             * We want to "flicker" the showAnimations so that it does not show animations initially, but does
             * subsequently after.
             * A little bit hackish, but works really nicely and simply.
             */
            setTimeout(() => $scope.showAnimations = true, 300);
          });
        }

        /**
         * An explosive has finished on the server. Removes the explosive from the {@link engExplosivesFactory}.
         */
        $scope.$on("engExplosiveFinished", (_event, id) => onExplosiveFinished(/** @type string */ id));

        /**
         * An explosive was added on the server. Adds the explosive to the {@link engExplosivesFactory}.
         */
        $scope.$on("engExplosiveAdded", (_event, id) => engExplosivesFactory.addConfiguredExplosiveByIdentifier(/** @type ExplosiveIdentifier */ id));

        /**
         * The player vehicle has been added/removed/changed. Update whether the player is currently in a vehicle on the {@link engUiFactory}.
         */
        $scope.$on("engIsPlayerInAVehicle", (_event, isPlayerInAVehicle) => engUiFactory.setPlayerInAVehicle( /** @type boolean */ isPlayerInAVehicle));

        /**
         * The timer has elapsed further on the server. Update the stored value on the {@link engChaosModeFactory}
         */
        $scope.$on("engChaosModeTimer", (_event, timeLeftRatio) => {
          const timer = engMiscFactory.isDefined(timeLeftRatio) ? /** @type number */ timeLeftRatio : null;
          engChaosModeFactory.setTimer(timer);
        });

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engUiFactory.view.unsubscribe(viewSubscription);
          engExplosivesFactory.configuredExplosives.unsubscribe(configuredExplosivesSubscription);
        });

        // $scope.$on("UINavigation", (_event, action, value) => {
          // console.log("UINavigation", action, value);
        // });

        init();
      }
    }
  }]);