angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engOverviewContainer
   * @description A container component that displays an overview for every configured explosive by the user
   * @requires engExplosivesFactory
   * @requires engNavigationFactory
   * @requires engEndpointsFactory
   * @requires engChaosModeFactory
   * @requires engPreviewsFactory
   * @requires engMiscFactory
   * @requires engUiFactory
   */
  .directive("engOverviewContainer", ["engExplosivesFactory", "engNavigationFactory", "engEndpointsFactory", "engChaosModeFactory", "engPreviewsFactory", "engMiscFactory", "engUiFactory", "$timeout", (
      engExplosivesFactory,
      engNavigationFactory,
      engEndpointsFactory,
      engChaosModeFactory,
      engPreviewsFactory,
      engMiscFactory,
      engUiFactory,
      $timeout
  ) => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/containers/overview-container/overview-container.html",
      restrict: "E",
      scope: false,
      controller: $scope => {
        /** @type ConfiguredExplosive[] */
        $scope.configuredExplosives = [];
        /** @type Preview[] */
        $scope.previews = [];
        $scope.isDetonateDisabled = true;
        $scope.isMinimised = false;

        $scope.chaosModeTimer = null;

        /**
         * This is to get around the classic funkiness of AngularJS and the state inside here not necessary updating
         * the template HTML.
         *
         * This will be set from within a {@link $timeout} call so that the explosives are reflected in the DOM.
         *
         * This approach has been taken, rather than just doing this on the {@link configuredExplosives} themselves,
         * as a lot of functions inside here rely on the synchronous nature of the `Observable` class and this would be
         * broken if it were to be wrapped in a {@link $timeout} call.
         *
         * This list will be maintained and will store the exactly same reference to the {@link configuredExplosives}.
         * @type ConfiguredExplosive[]
         **/
        $scope.templateExplosives = $scope.configuredExplosives;

        /**
         * Sets the current list of previews to the given previews
         * @param {Preview[]} previews The next list of previews
         */
        const previewsSubscription = previews => { $scope.previews = previews; };

        /**
         * Sets whether the app is currently minimise to the new given value
         * @param {boolean} isMinimised The new state
         */
        const isMinimisedSubscription = isMinimised => { $scope.isMinimised = isMinimised; };

        /**
         * Sets the chaos mode timer to the new given value
         * @param {number | null} timer The new timer value
         */
        const chaosModeTimerSubscription = timer => { $scope.chaosModeTimer = timer; };

        /**
         * Sets the current configured explosives to the given list of {@link explosives}.
         * Updates whether the detonation button is disabled.
         * @param {ConfiguredExplosive[]} explosives The new value of configured explosives
         * @see isDetonationAllowed
         */
        const configuredExplosivesSubscription = explosives => {
          $scope.configuredExplosives = explosives;
          $scope.isDetonateDisabled = !isDetonationAllowed(explosives);

          $timeout(() => {
            $scope.templateExplosives = explosives;
          });
        };

        /**
         * Sets the currently selected index to the given one before going to the explosive's configuration view
         * @param {number} index The index of the explosive clicked
         */
        function goToExplosiveViewByIndex(index) {
          engExplosivesFactory.setSelectedExplosiveByIndex(index);
          engUiFactory.setView("explosive");
        }

        /**
         * Deletes an in-game preview using the {@link engEndpointsFactory}, given the preview, and additionally removes
         * it from the {@link engPreviewsFactory}.
         * @param {Preview} preview The preview to remove
         */
        function requestPreviewDeletion(preview) {
          engEndpointsFactory.deletePreview(preview.previewId);
          engPreviewsFactory.removePreview(preview);
        }

        /**
         * Deletes an in-game preview using the {@link engEndpointsFactory} for a given explosive
         * @param {ConfiguredExplosive} explosive The explosive to delete the preview for
         */
        function deletePreview(explosive) {
          const preview = $scope.previews.find(({ explosiveUiId }) => explosiveUiId === explosive.uiId);

          if (!engMiscFactory.isDefined(preview)) {
            console.error("Cannot delete preview for this explosive as it does not exist");
            return;
          }

          requestPreviewDeletion(preview);
        }

        /**
         * Converts a {@link ConfiguredExplosive} to a {@link PlaceExplosiveRequest} object by extracting out the
         * various ids and properties.
         * @param {ConfiguredExplosive} explosive The explosive
         * @returns {PlaceExplosiveRequest} The request from the explosive
         */
        function explosiveToRequest(explosive) {
          const { uiId, id, preset, position, fuse } = explosive;
          return { uiId, position, fuse, explosiveId: id, presetId: preset.id };
        }

        /**
         * Returns a new explosive from the given explosive where the [showing]{@link ConfiguredExplosive#showing} flag
         * is set to `false`, and the [inProgress]{@link ConfiguredExplosive#inProgress} flag is set to `true`.
         * @param {ConfiguredExplosive} explosive The explosive to set
         * @returns {ConfiguredExplosive} The updated explosive
         */
        function setExplosiveDetonationState(explosive) {
          return explosive.inProgress ? explosive : { ...explosive, showing: false, inProgress: true };
        }

        /**
         * Returns whether the given list of explosives is populated with at least one explosive, and that there is at
         * least one explosive that is not yet in progress.
         * @param {ConfiguredExplosive[]} explosives The list of explosives to check
         * @returns {boolean} Whether there is at least one explosive and that at least one is not yet in progress
         */
        function isDetonationAllowed(explosives) {
          return explosives.length > 0 && explosives.some(({ inProgress }) => !inProgress);
        }

        /**
         * Checks whether a preview exists for the configured explosive at the given index. If it does, this function
         * will then request its deletion and update the explosive on the {@link engExplosivesFactory} with it's
         * `showing` flag set to `false`.
         * @param {number} index The index of the explosive to attempt to delete a preview of
         */
        function removeExistingPreview(index) {
          const explosive = $scope.configuredExplosives[index];
          const preview = $scope.previews.find(({ explosiveUiId }) => explosiveUiId === explosive.uiId);

          if (engMiscFactory.isDefined(preview)) {
            requestPreviewDeletion(preview);
            engExplosivesFactory.setConfiguredExplosiveByIndex({ ...explosive, showing: false }, index);
          }
        }

        /**
         * Deletes the preview from each explosive, if they are being previewed.
         * Note that this will only update the internal list of configured explosives, and the
         * {@link engExplosivesFactory}, if the {@link updateState} flag is set to `true`.
         * @param {boolean} [updateState] Whether to update the state as well. Defaults to `false`.
         */
        function removeAllPreviewsFromExplosives(updateState = false) {
          $scope.configuredExplosives
              .filter(({ showing }) => showing)
              .forEach(deletePreview);

          if ($scope.previews.length > 0) {
            console.warn("There are still previews shown but no explosives are being shown currently. This is a de-sync warning.");
          }

          if (updateState) {
            const updatedExplosives = $scope.configuredExplosives.map(explosive => /** @type ConfiguredExplosive */ ({ ...explosive, showing: false }));
            engExplosivesFactory.setConfiguredExplosives(updatedExplosives);
          }
        }

        /**
         * Toggles whether the app is minimised on the {@link engUiFactory}
         */
        function onAppMinimise() {
          const isNowMinimised = !$scope.isMinimised;
          engUiFactory.setIsMinimised(isNowMinimised);
        }

        /**
         * Sets the view to the `"chaos"` view on the {@link engUiFactory}
         */
        function onChaosModeClicked() {
          engUiFactory.setView("chaos");
        }

        /**
         * Sets the currently selected index to the given one before going to the explosive's configuration positioning
         * view.
         *
         * If the explosive is currently in progress, this function will do nothing.
         * @param {number} index The index of the explosive clicked
         */
        function onOpenClick(index) {
          if ($scope.configuredExplosives[index].inProgress) return;

          removeAllPreviewsFromExplosives(true);
          goToExplosiveViewByIndex(index);
          engUiFactory.setIsPositioningExplosive(true);
        }

        /**
         * Adds a new explosive to the list of configured explosives, before selecting and going to this newly created
         * explosive's configuration view.
         */
        function onAddNewExplosiveClick() {
          engExplosivesFactory.addNewConfiguredExplosive();
          removeAllPreviewsFromExplosives(true);
          goToExplosiveViewByIndex($scope.configuredExplosives.length - 1);
        }

        /**
         * Removes an explosive given its index. If the explosive is current in progress, then this function will
         * request its destruction on the {@link engEndpointsFactory}. If not, and a preview is being shown for this
         * explosive, this function will remove the in-game preview.
         * @param {number} index The index of the explosive to remove/destroy
         */
        function onDeleteClick(index) {
          const { inProgress, uiId } = $scope.configuredExplosives[index];

          if (inProgress) {
            engEndpointsFactory.destroyExplosive(uiId);
            // Note that we do not need to remove it from the engExplosivesFactory - the server will alert the UI when
            // the above end point is called and the server destroys the explosive on its end.
            return;
          }

          removeExistingPreview(index);
          engExplosivesFactory.removeExplosive(index);
        }

        /**
         * Places the explosives using the {@link engEndpointsFactory} endpoint with the configured explosives.
         * This function will then hide and delete all previews on the {@link engPreviewsFactory}, before also updating
         * the configured explosives on the {@link engExplosivesFactory}.
         */
        function onDetonate() {
          // Place all non-detonated explosives
          const explosiveRequests = $scope.configuredExplosives
              .filter(({ inProgress }) => !inProgress)
              .map(explosiveToRequest);
          engEndpointsFactory.placeExplosives(explosiveRequests);

          // Delete all previews, but do not update state as more modifications need to be made to the explosives
          removeAllPreviewsFromExplosives();

          // Update the factory with the newly detonated explosives
          const updatedExplosives = $scope.configuredExplosives.map(setExplosiveDetonationState);
          engExplosivesFactory.setConfiguredExplosives(updatedExplosives);
        }

        /**
         * Toggles whether the explosive at the given {@link index} is showing or not. If the explosive is no longer
         * being shown then this function will delete the in-game preview. Otherwise, this function will create it.
         * @param {number} index The index of the explosive to show/hide preview
         */
        function onPreviewClick(index) {
          const explosive = $scope.configuredExplosives[index];
          const newExplosive = { ...explosive, showing: !explosive.showing };

          engExplosivesFactory.setConfiguredExplosiveByIndex(newExplosive, index);

          if (!newExplosive.showing) {
            deletePreview(newExplosive);
            return;
          }

          const { id, uiId, position, preset } = newExplosive;
          engEndpointsFactory.createPreviewByPosition(position, id, preset.id, previewId => {
            engPreviewsFactory.addPreview({ previewId, explosiveUiId: uiId });
          });
        }

        /**
         * Initialises the component.
         * Initialises the {@link engExplosivesFactory}, and subscribes to various observables on various factories.
         * Additionally, this function will reset the focusable elements on the {@link engNavigationFactory}.
         */
        function init() {
          engNavigationFactory.resetFocusableElements();

          engExplosivesFactory.init(() => {
            engExplosivesFactory.configuredExplosives.subscribe(configuredExplosivesSubscription);
          });

          engUiFactory.isMinimised.subscribe(isMinimisedSubscription);
          engPreviewsFactory.previews.subscribe(previewsSubscription);
          engChaosModeFactory.timer.subscribe(chaosModeTimerSubscription);
        }

        $scope.onAppMinimise = onAppMinimise;
        $scope.onChaosModeClicked = onChaosModeClicked;
        $scope.onOpenClick = onOpenClick;
        $scope.onAddNewExplosiveClick = onAddNewExplosiveClick;
        $scope.onDeleteClick = onDeleteClick;
        $scope.onDetonate = onDetonate;
        $scope.onPreviewClick = onPreviewClick;

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engExplosivesFactory.configuredExplosives.unsubscribe(configuredExplosivesSubscription);
          engPreviewsFactory.previews.unsubscribe(previewsSubscription);
          engUiFactory.isMinimised.unsubscribe(isMinimisedSubscription);
          engChaosModeFactory.timer.unsubscribe(chaosModeTimerSubscription);
        });

        init();
      }
    }
  }])