angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engChaosModeFactory
   * @description A store that manages the state of the chaos mode
   * @requies engObservablesFactory
   */
  .factory("engChaosModeFactory", ["engObservablesFactory", "engExplosivesFactory", (engObservablesFactory, engExplosivesFactory) => {
    /** @type {Observable<ExplosiveDto[]>} */
    const explosiveOptions = engObservablesFactory.newObservable([]);
    /** @type {Observable<ChaosModeConfiguration>} */
    const configuration = engObservablesFactory.newObservable(getDefaultConfiguration());
    /** @type {Observable<boolean>} */
    const isActive = engObservablesFactory.newObservable(false);
    /**
     * Note that this is the normalised timer value
     * @type {Observable<number | null>}
     */
    const timer = engObservablesFactory.newObservable(null);

    /**
     * @return {ChaosModeConfiguration} The default configuration
     */
    function getDefaultConfiguration() {
      return {
        maximumConcurrentExplosives: 5,
        interval: 15,
        placementRandomness: 50,
        allowedExplosives: explosiveOptions.value.map(({ id }) => id)
      };
    }

    /**
     * Sets whether the mode is active or not
     * @param {boolean} newValue The new active whether
     */
    function setActive(newValue) {
      isActive.next(newValue);
    }

    /**
     * Sets the timer
     * @param {number | null} newValue The new value to set
     */
    function setTimer(newValue) {
      timer.next(newValue);
    }

    /**
     * Sets the current configuration
     * @param {ChaosModeConfiguration} newConfiguration The configuration to set
     */
    function setConfiguration(newConfiguration) {
      configuration.next(newConfiguration);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      configuration.next(getDefaultConfiguration());
      isActive.next(false);
    }

    /**
     * Initialises the {@link engExplosivesFactory} and populates the {@link explosiveOptions} with valid explosives (
     * all except tier 4 explosives).
     * This function then updates the {@link configuration} with the default configuration.
     */
    function init() {
      engExplosivesFactory.init(() => {
        // Disallow tier 4 explosives
        const allowedOptions = engExplosivesFactory.getExplosiveDTOs().filter(({ tier }) => tier < 4);
        explosiveOptions.next(allowedOptions);
        // Update from above changes
        configuration.next(getDefaultConfiguration());
      });
    }

    init();

    return {
      configuration: /** @type {Subscribable<ChaosModeConfiguration>} */ configuration,
      explosiveOptions: /** @type {Subscribable<ExplosiveDto[]>} */ explosiveOptions,
      isActive: /** @type {Subscribable<boolean>} */ isActive,
      timer: /** @type {Subscribable<number | null>} */ timer,
      setActive,
      setTimer,
      setConfiguration,
      reset
    };
  }]);