angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engEndpointsFactory
   * @description Provides an API to make requests to the ExplosiveNG server extension, both retrieving data from and
   * sending data to this server.
   */
  .factory("engEndpointsFactory", () => {
    // Location of the server
    const EXTENSION = "extensions.explosionsNG";

    /**
     * The `bngApi` incorrectly returns an empty object instead of an empty array when returned from the server.
     * This function corrects the type.
     * @template T
     * @param {{} | T[]} arrayOrObject The array or object returned from the `bngApi`
     * @returns {T[]} The fixed array
     */
    function fixBngApiArray(arrayOrObject) {
      return Array.isArray(arrayOrObject) ? arrayOrObject : [];
    }

    /**
     * Makes a request to the server extension to get the current state on the server side
     * @param {(response: StateResponse) => void} callback The function to run with the server response. The server will
     * return a {@link StateResponse}.
     */
    function loadState(callback) {
      bngApi.engineLua(`${EXTENSION}.getCurrentState()`, state => {
        callback(/** @type StateResponse */ { ...state, explosives: fixBngApiArray(state.explosives) });
      });
    }

    /**
     * Makes a request to the server extension with the given combined list of explosives to create and place
     * @param {PlaceExplosiveRequest[]} explosives The explosives to create and place on the server response
     */
    function placeExplosives(explosives) {
      bngApi.engineLua(`${EXTENSION}.placeExplosives(${bngApi.serializeToLua(explosives)})`);
    }

    /**
     * Makes a request to the server extension for the list of explosives available. The given {@link callback} will be
     * called on server response.
     * @param {(explosives: ExplosiveDto[]) => void} callback The function to run with the server response. The server
     * will return the list of all explosives that can be created by this mod.
     */
    function getListOfExplosives(callback) {
      bngApi.engineLua(`${EXTENSION}.getListOfExplosives()`, response => callback(fixBngApiArray(response)));
    }

    /**
     * Makes a request to the server extension to create a new preview at the given {@link position}. The given
     * {@link callback} will be called on server response.
     * @param {string} explosiveId The explosive id
     * @param {string} presetId The preset's id of the explosive
     * @param {Position} position The position to use
     * @param {(id: string) => void} callback The function to run with the server response. The server will return the
     * id of the created preview
     */
    function createPreviewByPosition(position, explosiveId, presetId, callback) {
      bngApi.engineLua(
          `${EXTENSION}.createPreviewByPosition(${bngApi.serializeToLua(position)}, "${explosiveId}", "${presetId}")`,
          callback
      );
    }

    /**
     * Makes a request to the server extension to create a new preview with the {@link option} to set its position. The
     * given {@link callback} will be called on server response.
     * @param {AutoCoordinateInputType} option The coordinate type to use as the position
     * @param {string} explosiveId The explosive id
     * @param {string} presetId The preset's id of the explosive
     * @param {(response: { id: string, position: Position }) => void} callback The function to run with the server
     * response. The server will return the id of the created preview and its position it was placed at.
     */
    function createPreviewByOption(option, explosiveId, presetId, callback) {
      bngApi.engineLua(`${EXTENSION}.createPreviewByOption("${option}", "${explosiveId}", "${presetId}")`, callback);
    }

    /**
     * Makes a request to the server extension with the given {@link position} to update the preview position
     * @param {string} id The id of the preview to update
     * @param {Position} position The position to use
     */
    function updatePreviewPosition(id, position) {
      bngApi.engineLua(`${EXTENSION}.updatePreviewPosition("${id}", ${bngApi.serializeToLua(position)})`);
    }

    /**
     * Makes a request to the server extension with the given {@link option} to update the preview position. The given
     * {@link callback} will be called on server response.
     * @param {string} id The id of the preview to update
     * @param {AutoCoordinateInputType} option The coordinate type to use as the position
     * @param {(response: { id: string, position: Position }) => void} callback The function to run with the server
     * response. The server will return the position at which the preview was placed at
     */
    function updatePreviewPositionByType(id, option, callback) {
      bngApi.engineLua(`${EXTENSION}.updatePreviewPositionByType("${id}", "${option}")`, callback);
    }

    /**
     * Makes a request to the server extension to delete the displayed preview for the preview that matches the given
     * {@link id}.
     * @param {string} id The id of the preview to delete
     */
    function deletePreview(id) {
      bngApi.engineLua(`${EXTENSION}.deletePreview("${id}")`);
    }

    /**
     * Makes a request to the server extension to destroy all active explosives
     */
    function destroyExplosives() {
      bngApi.engineLua(`${EXTENSION}.destroyAll()`);
    }

    /**
     * Makes a request to the server extension to destroy all active explosives
     */
    function destroyExplosive(uiId) {
      bngApi.engineLua(`${EXTENSION}.destroyExplosive("${uiId}")`);
    }

    /**
     * Makes a request to the server extension to start chaos mode with the given configuration
     * @param {ChaosModeConfiguration} configuration The configuration
     */
    function startChaosMode(configuration) {
      bngApi.engineLua(`${EXTENSION}.startChaosMode(${bngApi.serializeToLua(configuration)})`)
    }

    /**
     * Makes a request to the server extension to stop chaos mode
     */
    function stopChaosMode() {
      bngApi.engineLua(`${EXTENSION}.stopChaosMode()`);
    }

    return {
      loadState,
      placeExplosives,
      getListOfExplosives,
      createPreviewByPosition,
      createPreviewByOption,
      updatePreviewPosition,
      updatePreviewPositionByType,
      deletePreview,
      destroyExplosives,
      destroyExplosive,
      startChaosMode,
      stopChaosMode
    };
  });