angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engExplosivesFactory
   * @description A store that manages the state of all things for this mod. This includes configured explosives by the
   * user, a list of all possible explosive that can be selected and the current view that is being shown to the user.
   *
   * For components using this factory, the {@link init} method should be called as part of the component's init. This
   * will initiate all async requests and set the initial state of the factory, referred to as "synchronised". If two or
   * more components require this factory, then once the factory is synchronised, the next component will not affect
   * this factory and instead immediately return (see {@link init} for more details).
   *
   * This factory uses observables to store and manage information in this factory. To retrieve data, a component will
   * need to subscribe to the individual data by setting up a subscription (see {@link Subscribable#subscribe}).
   * Again, the {@link init} method is set up for this purpose. It is important that the component that sets up these
   * subscriptions also unsubscribes from the same data when it is destroyed, or not needed anyway (see
   * {@link Subscribable#unsubscribe}).
   * @requires engObservablesFactory
   * @requires engMiscFactory
   * @requires engEndpointsFactory
   */
  .factory("engExplosivesFactory", ["engObservablesFactory", "engMiscFactory", "engEndpointsFactory", (
      engObservablesFactory,
      engMiscFactory,
      engEndpointsFactory,
  ) => {
    let synchronised = false;
    /** @type {(() => void)[]} */
    let onSyncSubscribers = [];

    let nextId = 0;

    /** @type ExplosiveDto[] */
    let explosiveDTOs = [];
    /** @type Observable<ConfiguredExplosive[]> */
    const configuredExplosives = engObservablesFactory.newObservable([]);
    /** @type {Observable<ConfiguredExplosive | undefined>} */
    const selectedExplosive = engObservablesFactory.newObservable(undefined);
    /** @type {Observable<boolean>} */
    const lockTier4s = engObservablesFactory.newObservable(false);

    // Set up a subscription to update whether tier 4 locking should come into effect
    configuredExplosives.subscribe(checkTier4Locking);

    /**
     * Converts an {@link ExplosiveDto} object into an {@link ConfiguredExplosive} object. The returned object will have
     * its [preset]{@link ExplosivePreset} set to the corresponding preset of the given index from the {@link presets}
     * list given. The position will be set to the origin.
     * @param {ExplosiveDto} explosive The explosive DTO
     * @param {number} selectedPresetIndex The index of the presets within {@link presets} to use as the return
     * preset
     * @returns {ConfiguredExplosive} The converted object
     */
    function dtoToExplosive(explosive, selectedPresetIndex) {
      const { id, name, tier, description, properties, presets } = explosive;
      return {
        uiId: (nextId++).toString(),
        id,
        name,
        tier,
        description,
        properties,
        preset: presets[selectedPresetIndex],
        position: { x: 0, y: 0, z: 0 },
        showing: false,
        inProgress: false,
        fuse: 0
      };
    }

    /**
     * Maps an {@link ExplosiveIdentifier} to a {@link ConfiguredExplosive} object
     * @param {ExplosiveIdentifier} explosiveIdentifier The identifier to map
     * @returns {ConfiguredExplosive} The mapped configured explosive
     */
    function identifierToConfiguredExplosive(explosiveIdentifier) {
      const { uiId, explosiveId, presetId, fuse } = explosiveIdentifier;

      const { id, name, tier, description, properties, presets } = explosiveDTOs.find(({ id }) => id === explosiveId);
      /** @type ExplosivePreset */ // Not sure why type-cast is needed by Intellij?
      const preset = presets.find(({ id }) => id === presetId);
      const position = { x: 0, y: 0, z: 0 }; // This is unknown; it will never be used in the UI so shouldn't matter

      return { uiId, id, name, tier, description, properties, preset, position, fuse, showing: false, inProgress: true };
    }

    /**
     * Checks whether more than one tier 4 explosive has been configured, and sets the result to the {@link lockTier4s}
     * observable.
     * @param {ConfiguredExplosive[]} explosives The new list of configured explosive
     */
    function checkTier4Locking(explosives) {
      const shouldLock = explosives.filter(({ tier }) => tier === 4).length > 1;
      lockTier4s.next(shouldLock);
    }

    /**
     * @returns {ConfiguredExplosive} The first {@link ExplosiveDto} object in the factory, converted to a
     * {@link ConfiguredExplosive} object with the first available preset.
     */
    function createDefaultExplosive() {
      const dto = explosiveDTOs[0];
      return dtoToExplosive(dto, 0);
    }

    /**
     * Requests the list of explosives from the {@link engEndpointsFactory}.
     * This function will additionally set this factory to be synchronised.
     * @see synchronise
     */
    function initExplosives() {
      engEndpointsFactory.getListOfExplosives(explosives => {
        explosiveDTOs = explosives;

        // /** @type ConfiguredExplosive[] */
        // const createdExplosives = explosives.flatMap((explosive, explosiveIndex) => explosive.presets.map((_, index) => {
        //   if (explosiveIndex > 5) return;
        //   const position = { x: Math.random() * 100, y: Math.random() * 100, z: 0 };
        //   return { ...dtoToExplosive(explosive, index), position };
        // })).filter(engMiscFactory.isDefined);
        // const createdExplosives = Array(50).fill(null).map(() => {
        //   const position = { x: Math.random() * 100, y: Math.random() * 100, z: 0 };
        //   const explosive = explosives[Math.floor(Math.random() * (explosives.length - 1))];
        //   const presetIndex = Math.floor(Math.random() * (explosive.presets.length - 1));
        //   return { ...dtoToExplosive(explosive, presetIndex), position };
        // })
        // configuredExplosives.next(createdExplosives);

        synchronise();
      });
    }

    /**
     * Creates a new explosive with default settings and adds it to the list of configured explosives
     * @see createDefaultExplosive
     */
    function addNewConfiguredExplosive() {
      configuredExplosives.next([...configuredExplosives.value, createDefaultExplosive()]);
    }

    /**
     * Creates a {@link ConfiguredExplosive} from the given identifier and adds it to the list of configured explosives
     * @param {ExplosiveIdentifier} explosiveIdentifier The identifier to create and add an explosive for
     */
    function addConfiguredExplosiveByIdentifier(explosiveIdentifier) {
      const newConfiguredExplosive = identifierToConfiguredExplosive(explosiveIdentifier);
      configuredExplosives.next([...configuredExplosives.value, newConfiguredExplosive]);
    }

    /**
     * Removes an explosive from the list of configured explosives given its index or uiId
     * @param {number | string} property Either the index of the explosive, or the explosive uiId to remove
     */
    function removeExplosive(property) {
      if (typeof property === "number") {
        removeByIndex(property);
        return;
      }

      removeById(property);
    }

    /**
     * Removes an explosive from the list of configured explosives given its index
     * @param {number} index The index of the explosive to remove
     */
    function removeByIndex(index) {
      const explosives = configuredExplosives.value;
      const length = explosives.length;

      if (index < 0 || index >= length) {
        console.error(`Index ${index} must be positive and less than the array length (${length})`);
        return;
      }

      const explosiveToRemove = explosives[index];
      const selected = selectedExplosive.value;

      if (engMiscFactory.isDefined(selected) && selected === explosiveToRemove) {
        console.warn("Attempted to remove the currently selected explosive. Removal has been prevented");
        return;
      }

      configuredExplosives.next(explosives.filter((_, i) => i !== index));
    }

    /**
     * Removes an explosive from the list of configured explosives given its uiId
     * @param {string} id The uiId of the explosive to remove
     */
    function removeById(id) {
      const explosives = configuredExplosives.value;
      const explosiveToRemove = explosives.find(({ uiId }) => id === uiId);

      if (!engMiscFactory.isDefined(explosiveToRemove)) {
        console.warn(`Explosive with id ${id} is not present in the configuration`);
        return;
      }

      const filteredExplosives = explosives.filter(explosive => explosive !== explosiveToRemove);
      configuredExplosives.next(filteredExplosives);
    }

    /**
     * Overwrites the explosive at the given index to the given explosive. If the {@link index} is not given, then the
     * selected explosive index will be used instead.
     * @param {ConfiguredExplosive} explosive
     * @param {number} [index] The index of the explosive to set in the configuration array. Optional and if not given,
     * the selected explosive's index is used instead
     */
    function setConfiguredExplosiveByIndex(explosive, index) {
      const selectedExplosiveIndex = getSelectedExplosiveIndex();
      const explosiveIndex = engMiscFactory.withDefaultValue(index, selectedExplosiveIndex);
      const newExplosives = [...configuredExplosives.value];

      newExplosives[explosiveIndex] = explosive;
      configuredExplosives.next(newExplosives);

      // Update selected explosive if it has changed
      if (explosiveIndex === selectedExplosiveIndex) {
        selectedExplosive.next(explosive);
      }
    }

    /**
     * Sets the currently configured explosives to the given list. This will update the currently selected explosive
     * with either the updated explosive, or if it does not exist in the list, to `undefined`.
     * @param {ConfiguredExplosive[]} explosives The list of explosives to set
     */
    function setConfiguredExplosives(explosives) {
      configuredExplosives.next(explosives);

      const updatedSelectedExplosive = engMiscFactory.isDefined(selectedExplosive.value)
          ? explosives.find(({ uiId }) => uiId === selectedExplosive.value.uiId)
          : null;

      // Update selected explosive if it has changed
      if (engMiscFactory.isDefined(updatedSelectedExplosive)) {
        selectedExplosive.next(updatedSelectedExplosive);
        return
      }

      // Unset the selected explosive if it no longer exists
      selectedExplosive.next(undefined);
    }

    /**
     * Overwrites the preset of the currently selected explosive given the index of the new preset to set
     * @param {number} presetIndex The index of the preset to set
     */
    function setSelectedPresetByIndex(presetIndex) {
      const presets = getSelectedExplosiveDto().presets;
      const preset = presets[presetIndex];
      const updatedExplosive = /** @type ConfiguredExplosive */ { ...selectedExplosive.value, preset };

      setConfiguredExplosiveByIndex(updatedExplosive);
      selectedExplosive.next(updatedExplosive);
    }
    /**
     * Sets the currently selected explosive to the explosive that exists in the configured explosives with the given
     * index
     * @param {number} index The index of the list of currently configured explosives to get the new selected explosive
     */
    function setSelectedExplosiveByIndex(index) {
      const newSelectedExplosive = configuredExplosives.value[index];
      selectedExplosive.next(newSelectedExplosive);
    }

    /**
     * Resets the currently selected explosive to `undefined`
     */
    function resetSelectedExplosive() {
      selectedExplosive.next(undefined);
    }

    /**
     * @returns {number} The index of the currently selected explosive in the list of configured explosives
     */
    function getSelectedExplosiveIndex() {
      return configuredExplosives.value.findIndex(explosive => explosive === selectedExplosive.value);
    }

    /**
     * @returns {ExplosiveDto} The DTO of the currently selected explosive;
     */
    function getSelectedExplosiveDto() {
      const configurationId = selectedExplosive.value.id;
      return explosiveDTOs.find(({ id }) => id === configurationId);
    }

    /**
     * @returns {ExplosiveDto[]} The list of explosives available, in DTO format
     */
    function getExplosiveDTOs() {
      return [...explosiveDTOs];
    }

    /**
     * Sets this factory to {@link synchronised}. This function will also call each of the {@link onSyncSubscribers},
     * before emptying the array.
     */
    function synchronise() {
      synchronised = true;
      onSyncSubscribers.forEach(subscriber => subscriber());
      onSyncSubscribers = [];
    }

    /**
     * Merges the current list of configured explosives with the given list of explosive identifiers from the server.
     * The list of configured explosives is updated in the following ways:
     *  - If the explosive is not marked as detonating/in progress in this factory, then it remains in the list
     *  - If the explosive is marked as detonating/in progress in this factory, and it exists in the server list
     *  too, then it also remains in the list
     *  - If the explosive is marked as detonating/in progress in this factory, but does not exist in the server list,
     *  then it is removed
     *  - If the explosive exists **only** on the server, then it is mapped to a {@link ConfiguredExplosive} and added
     *  to the list
     * @param {ExplosiveIdentifier[]} serverExplosives The list of explosive identifiers from the server
     */
    function softSyncWithServer(serverExplosives) {
      const serverUiIds = serverExplosives.map(({ uiId }) => uiId);

      // Keep any explosive on the factory that is either not in progress, or also exists on the server
      const existingExplosives = configuredExplosives.value.filter(({ uiId, inProgress }) => !inProgress || serverUiIds.includes(uiId));

      const serverOnlyExplosives = serverExplosives
          .filter(({ uiId }) => !configuredExplosives.value.find(e => e.uiId === uiId))
          .map(identifierToConfiguredExplosive);

      // Keep the order from the factory, and add new explosives to the end of the list
      const currentExplosives = [...existingExplosives, ...serverOnlyExplosives];
      configuredExplosives.next(currentExplosives);
    }

    /**
     * Creates and sets the list of configured explosives, mapped from a list of explosive identifiers from the server.
     * This function will overwrite all currently configured explsovies, included the currently selected explosive.
     * Note that this function is not necessarily synchronous; the given callback function will be called once this
     * function has completed.
     * @param {ExplosiveIdentifier[]} serverExplosives The list of explosive identifiers from the server
     * @param {() => void} callback The callback function to call once done
     */
    function hardSyncWithServer(serverExplosives, callback) {
      // Ensure that the DTOs are available for use
      init(() => {
        const serverConfiguredExplosives = serverExplosives.map(identifierToConfiguredExplosive);

        selectedExplosive.next(undefined);
        configuredExplosives.next(serverConfiguredExplosives);

        callback();
      });
    }

    /**
     * Initialises this factory by performing all async requests it needs. Once these have completed, the given callback
     * function will be called. This should typically be used to set up subscriptions.
     * If this factory has already performed its async requests, then this method will immediately call the callback.
     * @param {() => void} onSync The callback to call after synchronisation has completed
     */
    function init(onSync) {
      if (synchronised) {
        onSync();
        return;
      }

      initExplosives();
      onSyncSubscribers.push(onSync);
    }

    return {
      configuredExplosives: /** @type Subscribable<ConfiguredExplosive[]> */ configuredExplosives,
      selectedExplosive: /** @type Subscribable<ConfiguredExplosive> */ selectedExplosive,
      lockTier4s: /** @type Subscribable<boolean> */ lockTier4s,
      dtoToExplosive,
      addNewConfiguredExplosive,
      addConfiguredExplosiveByIdentifier,
      removeExplosive,
      setConfiguredExplosiveByIndex,
      setConfiguredExplosives,
      setSelectedPresetByIndex,
      setSelectedExplosiveByIndex,
      resetSelectedExplosive,
      getExplosiveDTOs,
      softSyncWithServer,
      hardSyncWithServer,
      init
    }
  }
]);