angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engMiscFactory
   * @description Provides miscellaneous functions
   */
  .factory("engMiscFactory", () => {
    /**
     * Rounds a value to a given number of decimal places
     * @param {number} value The value to round
     * @param {number} decimalPlaces The number of decimal places to round to
     * @returns {number} The rounded value
     */
    function roundTo(value, decimalPlaces) {
      const power10 = Math.pow(10, decimalPlaces);
      return Math.round(value * power10) / power10;
    }

    /**
     * Checks and returns whether the given {@link value} is defined
     * @template T
     * @param {T | null | undefined} value The value to check
     * @returns {boolean} Whether the value is defined
     */
    function isDefined(value) {
      return value !== undefined && value !== null;
    }

    /**
     * Returns the {@link value} if it's defined, else the {@link defaultValue}
     * @template T
     * @param {T | null | undefined} value The value
     * @param {T} defaultValue The default value
     * @returns {T} The {@link value} if it's defined, else the {@link defaultValue}
     */
    function withDefaultValue(value, defaultValue) {
      return isDefined(value) ? value : defaultValue;
    }

    /**
     * Checks whether the given {@link position} is fully defined, which is defined as being anywhere in the coordinate
     * space that is __not__ the origin.
     * @param {Position} position The position to check
     * @returns {boolean} Whether the position is fully defined
     */
    function isPositionDefined(position) {
      return !(position.x === 0 && position.y === 0 && position.z === 0);
    }

    /**
     * Partitions an array into two subsets. The first subset contains all elements that pass the given predicate, with
     * the second subset for all other elements.
     * @template T
     * @param {T[]} array The array to partition
     * @param {(element: T) => boolean} predicate The predicate to divide the array into partitions
     * @returns {[T[], T[]]} The two partitions of the original array
     */
    function binaryPartition(array, predicate) {
      const subset1 = [];
      const subset2 = [];

      array.forEach(element => {
        const subset = predicate(element) ? subset1 : subset2;
        subset.push(element);
      })

      return [subset1, subset2];
    }

    return {
      roundTo,
      isDefined,
      withDefaultValue,
      isPositionDefined,
      binaryPartition
    };
  });