angular.module("beamng.apps")
    /**
     * @ngdoc factory
     * @name beamng.apps:engObservablesFactory
     * @description Creates an Observable class that stores a current value that can be changed whereby any subscribers to
     * that value will be notified of the change in value. This factory exposes a single method to create new observables.
     */
    .factory("engObservablesFactory", () => {
      /**
       * An object that stores a current value and broadcasts changes in its value to subscribers. It should be noted
       * that unlike most implementations of observables, this class is synchronous. Note the example below:
       *
       * @example
       * const value = 0;
       * const observable = new Observable(0);
       * observable.subscribe(x => value = x);
       *
       * console.log(value); // Will print 0
       * observable.next(1);
       * console.log(value); // Will print 1 as `Observable.next` call is synchronous
       *
       * @template T
       */
      class Observable {
        /**
         * @constructor
         * @param {T} initialValue The value to set as the initial value of this observable
         */
        constructor(initialValue) {
          this.value = initialValue;
          /** @type {SubscriptionFn[]} */
          this.subscriptions = [];
        }

        /**
         * Sets the value of the source to the given one and emits to all subscribers
         * @param {T} newValue The value to set
         */
        next(newValue) {
          this.value = newValue;
          this.subscriptions.forEach(subscriber => subscriber(this.value));
        }

        /**
         * Adds the emitter function to this observable. Whenever this observable has a new value, the passed subscription
         * will be called with this new value.
         * This method will also call the subscription with the current value when this method is called.
         * @param {SubscriptionFn} subscription The subscription function to call when changes happen
         */
        subscribe(subscription) {
          this.subscriptions.push(subscription);
          subscription(this.value);
        }

        /**
         * Removes the emitter function from this observable
         * @param {SubscriptionFn} subscription The subscription to remove
         */
        unsubscribe(subscription) {
          this.subscriptions = this.subscriptions.filter(s => s !== subscription);
        }
      }

      /**
       * Creates a new {@link Observable} with the given initial value
       * @template T
       * @param {T} initialValue The value to set as the initial value of the observable
       * @return {Observable<T>} The created observable
       */
      function newObservable(initialValue) {
        return new Observable(initialValue);
      }

      return { newObservable };
    });