angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engUiFactory
   * @description A store that manages states that affect how components render options and subcomponents, across
   * components. This includes things like the current view as well as extra panels that may be open.
   * @requires engObservablesFactory
   * @requires engExplosivesFactory
   */
  .factory("engUiFactory", ["engObservablesFactory", "engExplosivesFactory", (
      engObservablesFactory,
      engExplosivesFactory
  ) => {
    /** @type Observable<View> */
    const view = engObservablesFactory.newObservable("overview");
    /** @type Observable<boolean> */
    const isPositioningExplosive = engObservablesFactory.newObservable(false);
    /** @type Observable<boolean> */
    const isPlayerInAVehicle = engObservablesFactory.newObservable(false);
    /** @type Observable<boolean> */
    const isMinimised = engObservablesFactory.newObservable(false);
    /** @type {Observable<boolean>} */
    const isShowingInformation = engObservablesFactory.newObservable(false);

    /**
     * Sets the current view to the given one.
     * This function will reset the selected explosive on the {@link engExplosivesFactory} if the view has changed from
     * `"explosive"` to another view.
     * @param {View} newView The view to set
     */
    function setView(newView) {
      // Unset selected explosive if we are moving away from the configure explosive view
      if (view.value === "explosive" && newView !== "explosive") {
        engExplosivesFactory.resetSelectedExplosive();
      }

      view.next(newView);
    }

    /**
     * Sets whether an explosive is being positioned
     * @param {boolean} value The value to set
     */
    function setIsPositioningExplosive(value) {
      isPositioningExplosive.next(value);
    }

    /**
     * Sets the flag of whether the player is in a vehicle
     * @param {boolean} isInAVehicle Whether the player is now in a vehicle
     */
    function setPlayerInAVehicle(isInAVehicle) {
      isPlayerInAVehicle.next(isInAVehicle);
    }

    /**
     * Sets the flag of whether the app is minimised
     * @param {boolean} minimised Whether the app is minimised
     */
    function setIsMinimised(minimised) {
      isMinimised.next(minimised);
    }

    /**
     * Updates whether the information box is visible
     * @param {boolean} showingInformation Whether the information related to a boss is being displayed
     */
    function setIsShowingInformation(showingInformation) {
      isShowingInformation.next(showingInformation);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      view.next("overview");
      isPositioningExplosive.next(false);
      isPlayerInAVehicle.next(false);
      isShowingInformation.next(false);
    }

    return {
      view: /** @type Subscribable<View> */ view,
      isPositioningExplosive: /** @type Subscribable<boolean> */ isPositioningExplosive,
      isPlayerInAVehicle: /** @type Subscribable<boolean> */ isPlayerInAVehicle,
      isMinimised: /** @type Subscribable<boolean> */ isMinimised,
      isShowingInformation: /** @type {Subscribable<boolean>} */ isShowingInformation,
      setView,
      setIsPositioningExplosive,
      setPlayerInAVehicle,
      setIsMinimised,
      setIsShowingInformation,
      reset
    };
  }]);